import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt, log10


from tklib.tkobject import tkObject
from tklib.tkutils import pint, pfloat
from tklib.tkutils import pconv
from tklib.tkfile import tkFile
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me, Ea_Arrhenius, kayserToHz
from tklib.tksci.tkmatrix import make_matrix2


def read_input(path, print_level = 0):
    fin = tkFile(path, IsPrint = False)
    if not fin.fp:
        return None

    inf = {}
    while True:
        line = fin.readline()
        if not line: break

        line = line.strip()
        if line[0:5] == "&cell":
            inf["scale"] = fin.readline().strip()
            a0 = [pfloat(s) for s in fin.readline().split()]
            if len(a0) == 9:
                inf["latt"] = None
                inf["aij"] = [a0[0:3], a0[3:6], a0[6:9]]
            elif len(a0) == 6:
                inf["latt"] = a0
                inf["aij"] = None
            else:
                a1 = [pfloat(s) for s in fin.readline().split()]
                a2 = [pfloat(s) for s in fin.readline().split()]
                inf["latt"] = None
                inf["aij"] = [a0, a1, a2]
            
            while True:
                line = fin.readline()
                if not line or line[0] == '/': break

        elif line[0:7] == "&kpoint":
            inf["nk"] = pint(fin.readline())
            inf["k"] = [pfloat(s) for s in fin.readline().split()]
            while True:
                line = fin.readline()
                if not line or line[0] == '/': break

        texts = line.split(';')
        for text in texts:
            if '=' in text:
                _aa = text.split('=', 1)
                key = _aa[0].strip()
                val = _aa[1].strip()
                inf[key] = pconv(val, strict = True, defval = val)

    fin.close()
    
    return inf

def read_taudata(path, target = '', kmin = 0.0, iklist = None, print_level = 0):
    fin = tkFile(path, IsPrint = False)
    if not fin.fp:
        return None, None

    line = fin.skip_to(r"^#\s+ik,\s+is,\s+Frequency")
#    print("line=", line)

    keys = ["ik", "is", "f(cm-1)", "f(THz)", "tau", "v", "MFP", "kxx", "kxy", "kxz", "kyx", "kyy", "kyz", "kzx", "kzy", "kzz"]
    if target in keys:
        itarget = keys.index(target)
    else:
        itarget = None

    tau_list = []
    while 1:
        line = fin.readline()
        if not line:
            break
            
        aa = line.split()
        if len(aa) < 16:
            break

#        print("aa=", *aa)
        ik = pint(aa[0])
        if iklist and ik not in iklist:
            continue
        
        _is = pint(aa[1])
        fk   = pfloat(aa[2])
        fTHz = kayserToHz(pfloat(aa[2])) * 1.0e-12
        tau = pfloat(aa[3])
        v = pfloat(aa[4])
        MFP = pfloat(aa[5])
        mult = pint(aa[6])
        kappa_list = [1.0e6 * pfloat(s) for s in aa[7:]]
        vals = [ik, _is, fk, fTHz, tau, v, MFP, *kappa_list]

#        print("target=", itarget, target, vals[itarget], kmin)
        if itarget is not None and vals[itarget] < kmin:
#            print(f"  target=#{itarget}[{target}]: val={vals[itarget]} < {kmin}: skip")
            continue


        tau_list.append({ "ik": ik, "is": _is, "f(cm-1)": fk, 
                          "f": fTHz, "tau": tau, "v": v, 
                          "l": MFP, "mult": mult, "kappa": kappa_list })

    fin.close()
    
    return tau_list

def read_log(path, print_level = 0):
    inf = {}

    fin = tkFile(path, IsPrint = False)
    if not fin.fp:
        return None, None, None

    line = fin.skip_to(r"^\s*Input variables:")
    line = fin.skip_to(r"^\s*General:")
    for line in fin.fp:
        if line[:3] == ' --':
            break
        
        _aa = line.split(";")
        for s in _aa:
            if '=' not in s:
                continue

            _a2 = s.split("=", 1)
            key = _a2[0].strip()
            val = _a2[1].strip()
            inf[key] = pconv(val, val)

    line = fin.skip_to(r"^\s*Crystal structure")
    line = fin.skip_to(r"^\s*Lattice Vectors:")
    line = fin.skip_to(r"^\s*\*\s*Supercell")
    fin.readline()
    aij = make_matrix2(3, 3)
    for i in range(3):
        line = fin.readline()
        _aa = line.split()
        for j in range(3):
            aij[i][j] = pfloat(_aa[j])

    fin.readline()
    raij = make_matrix2(3, 3)
    for i in range(3):
        line = fin.readline()
        _aa = line.split()
        for j in range(3):
            raij[i][j] = pfloat(_aa[j])

    line = fin.skip_to(r"^\s*\*\s*Primitive cell")
    fin.readline()
    paij = make_matrix2(3, 3)
    for i in range(3):
        line = fin.readline()
        _aa = line.split()
        for j in range(3):
            paij[i][j] = pfloat(_aa[j])

    fin.readline()
    praij = make_matrix2(3, 3)
    for i in range(3):
        line = fin.readline()
        _aa = line.split()
        for j in range(3):
            praij[i][j] = pfloat(_aa[j])

    inf["aij"]   = aij
    inf["raij"]  = raij
    inf["paij"]  = paij
    inf["praij"] = praij

    line = fin.skip_to(r"Volume of the primitive cell\s*:\s*(\S+)")
    if fin.last_match:
        g = fin.last_match.groups()
        if g:
            inf["pVolume(a.u.)"] = pfloat(g[0])
#    print("inf=", inf)

    given_FC2XML = inf.get("FC2XML", None)
#    print("given_FC2XML=", given_FC2XML)
    if inf.get("FC2XML", None) is not None:
        line = fin.skip_to(r"^\s*FC2XML is given:")
        line = fin.skip_to(r"^\s*\*\s*Supercell")
        fin.readline()
        haij = make_matrix2(3, 3)
        for i in range(3):
            line = fin.readline()
            _aa = line.split()
            for j in range(3):
                haij[i][j] = pfloat(_aa[j])

        fin.readline()
        hraij = make_matrix2(3, 3)
        for i in range(3):
            line = fin.readline()
            _aa = line.split()
            for j in range(3):
                hraij[i][j] = pfloat(_aa[j])
        
        inf["haij"]   = haij
        inf["hraij"]  = hraij

    line = fin.skip_to(r"^\s*List of irreducible k points")
    kp_list = []
    while 1:
        line = fin.readline()
        if not line:
            break
            
        a1 = line.split(':')
        if len(a1) < 2:
            break
        a2 = a1[1].split()
        if len(a2) < 4:
            break

#        print("aa=", *aa)
        kp_list.append({ "ik": pint(a1[0]), "kx": pfloat(a2[0]), "ky": pfloat(a2[1]), "kz": pfloat(a2[2]), "w": pfloat(a2[3]) })

    fin.close()

    kp_list_ik = [None] * (len(kp_list) + 1)
    for kd in kp_list:
        ik = kd["ik"]
        for kd2 in kp_list:
            if kd2["ik"] == ik:
                kp_list_ik[ik] = kd2

    return inf, kp_list, kp_list_ik

def read_kpoints(path, print_level = 0):
    inf, kp_list, kp_list_ik = read_log(path, print_level = print_level)
    return kp_list, kp_list_ik


def read_properties(kplog_file, taudat_file, target = '', kmin = 0.0, iklist = None, print_level = 1):
    if print_level:
        print("")
        print(f"Read k points from {kplog_file}")
    inf_log, kp_list, kp_list_ik = read_log(kplog_file, print_level = print_level)
    if inf_log is None:
        app.terminate(f"Error: Can not read [{cparams.kplog_file}].", pause = True)
    
    print("  # of k points: ", len(kp_list))
    print("  K points:")
    for i in range(1, len(kp_list_ik)):
        l = kp_list_ik[i]
        ik = l["ik"]
        kx = l["kx"]
        ky = l["ky"]
        kz = l["kz"]
        w  = l["w"]
        if print_level:
            print(f"    {i:2d}: ik={ik:2d}: k=({kx:12.6f}, {ky:12.6f}, {kz:12.6f}) w={w:12.6f}")

    if print_level:
        print("")
        print(f"Read tau data from {taudat_file}")
    prop_list = read_taudata(taudat_file, target = target, kmin = kmin, iklist = iklist, print_level = print_level)
    if prop_list is None:
        app.terminate(f"Error: Can not read [{cparams.taudat_file}].", pause = True)
    if print_level:
        print("tau data:")
    sn_list = []
    ik_list = []
    is_list = []
    kx_list = []
    ky_list = []
    kz_list = []
    w_list  = []
    fk_list   = []
    fTHz_list = []
    _tau_list = []
    v_list   = []
    MFP_list = []
    kxx_list = []
    kxy_list = []
    kxz_list = []
    kyx_list = []
    kyy_list = []
    kyz_list = []
    kzx_list = []
    kzy_list = []
    kzz_list = []
    for i in range(len(prop_list)):
        l   = prop_list[i]
        ik  = l["ik"]
        _is = l["is"]

        ik0 = kp_list_ik[ik]["kx"]
        kx  = kp_list_ik[ik]["kx"]
        ky  = kp_list_ik[ik]["ky"]
        kz  = kp_list_ik[ik]["kz"]
        w   = kp_list_ik[ik]["w"]

        fk    = l["f(cm-1)"] # cm-1
        fTHz  = l["f"]       # THz
        tau   = l["tau"]
        v     = l["v"]
        MFP   = l["l"]
        kappa = l["kappa"]
        strs = [f'{kappa[i]:10.4g}' for i in range(len(l["kappa"]))]
        str = ''
        for s in strs:
            str += s + ' '
        
        if print_level:
            print(f' {i:05d}: ik={ik:03d} ({kx:8.3f},{ky:8.3f},{kz:8.3f}) is={_is:03d}: '
                + f'f={fTHz:10.4f} tau={tau:10.4f} v={v:10.4f} MFP={MFP:10.4f} kappa={str}')

        sn_list.append(i)
        ik_list.append(ik)
        is_list.append(_is)
        kx_list.append(kx)
        ky_list.append(ky)
        kz_list.append(kz)
        w_list.append(w)
        fk_list.append(fk)
        fTHz_list.append(fTHz)
        _tau_list.append(tau)
        v_list.append(v)
        MFP_list.append(MFP)
        kxx_list.append(kappa[0])
        kxy_list.append(kappa[1])
        kxz_list.append(kappa[2])
        kyx_list.append(kappa[3])
        kyy_list.append(kappa[4])
        kyz_list.append(kappa[5])
        kzx_list.append(kappa[6])
        kzy_list.append(kappa[7])
        kzz_list.append(kappa[8])

    return sn_list, ik_list, is_list, kx_list, ky_list, kz_list, w_list, fk_list, fTHz_list, _tau_list, v_list, MFP_list, \
                 kxx_list, kxy_list, kxz_list, kyx_list, kyy_list, kyz_list, kzx_list, kzy_list, kzz_list, inf_log

def cal_contrib(prop_list, prop_label, w_list, nprop_mesh, print_level = 1):
    pmax = max(prop_list) * 1.2
    pmin = pmax * 1.0e-6
    logpmin = log10(pmin)
    logpmax = log10(pmax)
    logpstep = (logpmax - logpmin) / (nprop_mesh - 1)

    xlogp_list = [logpmin + i * logpstep for i in range(nprop_mesh)]
    xp_list = pow(10.0, np.array(xlogp_list))
    nhit_list = np.zeros(nprop_mesh)
    wprop_list = np.zeros(nprop_mesh)
    for ip, p in enumerate(prop_list):
        if p <= 0.0:
            continue

        idx = int((log10(p) - logpmin) / logpstep + 1.0e-5)
        if idx < 0:
            continue

        nhit_list[idx] += w_list[ip]
        wprop_list[idx] += w_list[ip] * p

    contrib_list = np.empty(nprop_mesh)
    contrib_list[0] = 0.0
    for idx in range(1, nprop_mesh):
        contrib_list[idx] = contrib_list[idx - 1] + wprop_list[idx]
#        contrib_list[idx] = contrib_list[idx - 1] + nhit_list[idx] * xp_list[idx]

    p_tot = sum([p * w for p, w in zip(prop_list, w_list)])
    p_max = contrib_list[nprop_mesh - 1]
    contrib_list /= p_max

    if print_level:
        print(f"{prop_label}_tot: {p_tot}")
        print(f"{prop_label}_max: {p_max}")

    return xp_list, xlogp_list, nhit_list, wprop_list, contrib_list

def convert_k(praij, kx, ky, kz):
    return praij[0][0] * kx + praij[0][1] * ky + praij[0][2] * kz, \
           praij[1][0] * kx + praij[1][1] * ky + praij[1][2] * kz, \
           praij[2][0] * kx + praij[2][1] * ky + praij[2][2] * kz

def convert_k_list(praij, kx_list, ky_list, kz_list):
    kxc_list = []
    kyc_list = []
    kzc_list = []
    for i in range(len(kx_list)):
        kxc, kyc, kzc = convert_k(praij, kx_list[i], ky_list[i], kz_list[i])
        kxc_list.append(kxc)
        kyc_list.append(kyc)
        kzc_list.append(kzc)
    
    return kxc_list, kyc_list, kzc_list

    
    