from numpy import exp, log, sin, cos, tan, sqrt
import numpy as np
from numpy import linalg as la
from pprint import pprint

from tklib.tkobject import tkObject
import tklib.tkre as tkre
from tklib.tkre import DelQuote
from tklib.tkutils import SplitFilePath
from tklib.tksci.tksci import Round
from tklib.tksci.tksci import a0, NA, pi, pi2, torad, todeg, Round, Factorize, IsFactor, Factors
from tklib.tksci.tksci import degcos, degsin, degtan, acos, asin, atan, degacos, degasin, degatan
from tklib.tkcrystal.tkspacegroup import tkSpaceGroup
from tklib.tkcrystal.tkatomtypeobject import SplitAtomName, Charge2f
from tklib.tkcrystal.tkatomsiteobject import RoundParameter, Round01
from tklib.tkcrystal.tkatomtype import tkAtomType
from tklib.tkcrystal.tkatomsite import tkAtomSite


"""
Root Package to manage crystal structure
 class tkCrystalObject
"""


class tkCrystalObject(tkObject):
    def __init__(self, **args):
        self.InitializeRoot()
 
    def __del__(self):
        pass

    def __str__(self):
        return self.ClassPath()


    def InitializeRoot(self):
        self.__CrystalName = None
        self.__SampleName  = None
        self.__chemicalcomposition = ''
        self.__sumchemicalComposition = ''
 
        self.__aij   = None
        self.__gij   = None
        self.__Latt  = None
        self.__Raij  = None
        self.__Rgij  = None
        self.__RLatt = None
        self.__SpaceGroup = None

        self.ClearSpaceGroup()
        self.ClearAtomTypeList()
        self.ClearAtomSiteList()
        self.SetTemperature(25.0)

    def ClearAtomSiteList(self):
        self.__AtomSiteList = []
        return self.__AtomSiteList

    def ClearAtomTypeList(self):
        self.__AtomTypeList = []
        return self.__AtomTypeList

    def ClearSpaceGroup(self):
        self.__SpaceGroup = tkSpaceGroup()
        self.ConventionalSPGName = ''
        self.__SpaceGroup.SetP1()
        return self.__SpaceGroup


    def RoundLatticeParameters(self, val, eps = 1.0e-8):
        if abs(val - 90.0) < eps:
            return 90.0
        elif abs(val - 60.0) < eps:
            return 60.0
        elif abs(val - 120.0) < eps:
            return 120.0
        else:
            RoundParameter(val, eps)
        return val

    def SetSampleName(self, name = None, MakeFromPath = 0):
        if name is not None:
            self.__SampleName = name
            return
        if MakeFromPath:
            dir, filename, filebody, ext = SplitFilePath(self.path)
            self.__SampleName = filebody

    def SampleName(self, Create = 0):
        if self.__SampleName is not None:
            return self.__SampleName
        if Create:
            try:
                dir, filename, filebody, ext = SplitFilePath(self.path)
                return filebody
            except:
                return None

    def SetCrystalName(self, name = None, MakeFromPath = 0):
        if name is not None:
            self.__CrystalName = name
            return
        if MakeFromPath:
            dir, filename, filebody, ext = SplitFilePath(self.path)
            self.__CrystalName = filebody

    def CrystalName(self, Create = 0):
        if self.__CrystalName is not None:
            return self.__CrystalName
        if Create:
            comp = self.ChemicalComposition()
            if comp != '':
                return comp
            dir, filename, filebody, ext = SplitFilePath(self.path)
            return filebody

    def SetTemperature(self, T):
        self.__temperature = T

    def Temperature(self,):
        return self.__temperature


    def nAtomType(self):
        return len(self.__AtomTypeList)

    def AtomTypeList(self):
        return self.__AtomTypeList

    def natom_site(self):
        return len(self.__AtomSiteList)

    def nAtomSite(self):
        return self.natom_site()

    def atom_site_list(self, mode = 0):
        if mode == 0:
            return self.__AtomSiteList
        else:
            return self.__ExpandedAtomSiteList

    def AtomSiteList(self, mode = 0):
        return self.atom_site_list(mode = mode)

    def nexpanded_atom_site(self):
        return len(self.__ExpandedAtomSiteList)

    def nExpandedAtomSiteList(self):
        return self.nexpanded_atom_site()

    def expanded_atom_site_list(self):
        return self.__ExpandedAtomSiteList

    def ExpandedAtomSiteList(self):
        return self.expanded_atom_site_list()
    
    def ChemicalComposition(self):
        return self.__chemicalcomposition
    
    def SetChemicalComposition(self, composition):
        self.__chemicalcomposition = composition

    def SumChemicalComposition(self):
        try:
            return self.__sumchemicalcomposition
        except:
            return ''

#mode: all|asym|type
    def get_atom_name_list(self, mode = 'all', NameOnly = True):
        if mode == 'aym':
            sites = self.AtomSiteList()
        elif mode == 'all':
            sites = self.ExpandedAtomSiteList()
        elif mode == 'type':
            sites = self.AtomTypeList()
        else:
            print("\nError in tkcrystal.get_atom_name_list: Invalide mode [{}]\n".format(mode))
            exit()

        namelist = []
        for i in range(len(sites)):
            site = sites[i]
            if NameOnly:
                namelist.append(sites[i].AtomName())
            else:
                namelist.append(sites[i].AtomNameOnly())

        return namelist

#mode: all|asym
    def get_position_list(self, mode = 'all', IsReduce01 = False):
        if mode == 'aym':
            sites = self.AtomSiteList()
        elif mode == 'all':
            sites = self.ExpandedAtomSiteList()
        else:
            print("\nError in tkcrystal.get_atom_name_list: Invalide mode [{}]\n".format(mode))
            exit()

        poslist = []
        for i in range(len(sites)):
            site = sites[i]
            poslist.append(sites[i].Position(IsReduce01))

        return poslist

    def SetSumChemicalComposition(self, composition):
        self.__sumchemicalcomposition = composition

    def FormulaUnit(self):
        try:
            return self.__formulaunit
        except:
            return ''    

    def SetFormulaUnit(self, Z):
        self.__formulaunit = Z

    def IsAnionic(self, name):
        if tkre.Match(r'(O|S|Se|Te|F|Cl|I|N|As|Sb)$', name, 'i'):
            return 1
        return 0

    def IsCationic(self, name):
        return not self.IsAnionic(name)

    def AnalyzeChemicalComposition(self):
        ExpandedAtomSites  = self.ExpandedAtomSiteList()
        nExpandedAtomSite  = self.nExpandedAtomSiteList()
        AtomCount = {}
        for i in range(nExpandedAtomSite):
            site = ExpandedAtomSites[i]
            name = site.AtomNameOnly(1);
            if name not in AtomCount.keys():
                AtomCount[name] = 0.0
            AtomCount[name] += site.Occupancy()

        Names = list(AtomCount.keys())
        for key in Names:
            AtomCount[key] = RoundParameter(AtomCount[key], 1.0e-4)
#        print("AtomCount=", AtomCount)

        nk = len(Names)
        MinN = 1000000
        C1 = ''
        C2 = ''
        for i in range(nk):
            name = Names[i]
            n = AtomCount[name]
            MinN = min([n, MinN])
            if n == 1:
                n = ''
            else:
                n = str(n)
            n = tkre.Sub('0+$', '', n)    
            n = tkre.Sub(r'\.$', '', n)    
            if self.IsCationic(name):
                C1 += name + n
            else:
                C2 += name + n
        Composition = C1 + C2
        self.SetSumChemicalComposition(Composition)
#        print("Composition=", Composition)

        elements = Factors(MinN);
#        elements = Factorize(MinN);
        fac = 1
        ne = len(elements)
        for i in range(ne):
            num = elements[i]

        IsCommonFactor = 1
        for j in range(nk):
            name = Names[j];
            n = AtomCount[name]
            if IsFactor(n, num*fac) == 0:
                IsCommonFactor = 0
        if IsCommonFactor:
            fac *= num 
        self.SetFormulaUnit(fac)

        C1 = ''
        C2 = ''
        for i in range(nk):
            name = Names[i]
            n = AtomCount[name] / fac
            if n == 1:
                n = ''
            else:
                n = str(n)
            n = tkre.Sub('0+$', '', n)    
            n = tkre.Sub(r'\.$', '', n)    
            if self.IsAnionic(name) == 0:
                C1 += name + n
            else:
                C2 += name + n
        Composition = C1 + C2
        self.SetChemicalComposition(Composition)

    def UnitcellWeight(self):
        try:
            return self.__UnitcellWeight
        except:
            return 0.0
            
    def SetUnitcellWeight(self, MW):
        self.__UnitcellWeight = MW

    def Density(self):
        try:
            return self.__density
        except:
            return 0.0
   
    def AtomDensity(self):
         AtomSites = self.ExpandedAtomSiteList()
         v = self.Volume()
         self.__AtomDensity = len(AtomSites) / (v * 1.0e-24)
         return self.__AtomDensity

    def SetDensity(self, density):
        self.__density = density

    def SetAtomDensity(self, density):
        self.__AtomDensity = density

    def CalculateDensity(self):
        AtomTypes = self.AtomTypeList()
        AtomSites = self.AtomSiteList()
        nAtomSite = self.nAtomSite()
        MW = 0.0;
        self.dprint(1, "tkcrystal.tkcrystal.CalculateDensity: nS=", nAtomSite)
        for i in range(nAtomSite):
            AtomSite  = AtomSites[i]
            AtomName  = AtomSite.AtomNameOnly(0)
            mult      = AtomSite.Multiplicity()
            iAtomType = self.FindiAtomType(AtomName)
            if iAtomType < 0:
                print("Error in tkcrystal.CalculateDensity: Invalid iAtomType={}".format(iAtomType))
                exit()

            AtomType   = AtomTypes[iAtomType]
            occ        = AtomSite.Occupancy()
            AtomicMass = AtomType.AtomicMass()

            if occ is None:
                occ = 1.0
            if mult is None:
                mult = 1
            MW += AtomicMass * mult * occ
            self.dprint(1, "tkcrystal.tkcrystal.CalculateDensity: " 
                    + "iAtomSite={}: M={:6.4f} m={:6.4f} occ={:6.4f} MW={:6.4f}" 
                    .format(i, AtomicMass, mult, occ, MW))

        self.SetUnitcellWeight(MW)
        vol = self.Volume()
        self.dprint(1, "tkcrystal.tkcrystal.CalculateDensity:MW=", MW)
        self.dprint(1, "tkcrystal.tkcrystal.CalculateDensity:vol=", vol)
        if vol == 0.0:
            self.__density = 0.0
        else:
            d = MW / NA / vol * 1.0e24
            self.__density = Round(d, 4)
        return self.__density


#iAtomTypeは0から始まる
    def FindAtomType(self, atomtype):
        for i in range(len(self.__AtomTypeList)):
            if atomtype == self.__AtomTypeList[i].AtomType():
                return i
        else:
            return -1

#iAtomSiteは0から始まる
    def FindAtomSite(self, atomname, mode = 0):
        if mode == 0:
            AtomSites = self.AtomSiteList(mode)
        else:
            pass
        for i in range(len(AtomSites)):
            atom = AtomSites[i]
            name = atom.AtomNameOnly()
            if atomname.upper() == name.upper():
                return atom

#iAtomTypeは0から始まる
    def FindiAtomType(self, atomname):
        at = tkAtomType()
        name, nametype, type, charge = SplitAtomName(atomname)
#        name, nametype, type, charge = at.SplitAtomName(atomname)
        name = name.upper()
        AtomTypeList = self.AtomTypeList()
        nAtomType = self.nAtomType()
        for i in range(nAtomType):
            name1 = AtomTypeList[i].AtomTypeOnly().upper()
            if name == name1:
                return i
        return -1

    def find_nearest_site(self, pos0, irange = [1, 1, 1]):
        AtomSites = self.AtomSiteList(1)
        minidx = -1
        mindis = 1.0e10;
        for i in range(len(AtomSites)):
            atom = AtomSites[i]
            pos1 = atom.Position(1)
            dis = self.get_nearest_interatomic_distance(pos0, pos1, irange = irange)
            if dis < mindis:
                mindis = dis
                minidx = i
        return minidx

    def FindNearestSite(self, pos0, irange = [1, 1, 1]):
        return self.find_nearest_site(pos0, irange = [1, 1, 1])

    def BurstToP1(self):
#    my @ExpandedAtomSiteList = $this->GetCExpandedAtomSiteList();
#    SPG = $this->GetCSpaceGroup(); 
        self.SpaceGroupp().SetP1()
#        self.SetSpaceGroup("P 1", 1, 1)
#        self.SetLatticeSystem('triclinic', 0)
        self.SetAtomSiteList(self.ExpandedAtomSiteList())
        self.ExpandCoordinates();

    def GetBravaisLattice(self):
        SPG = self.SpaceGroup()
        return SPG.BravaisLattice()

    def expand_coordinates(self, DoTranslation = 1):
#初期化
        self.__nTotalExpandedAtomSite  = 0;
        self.__Density  = 0.0;

#同一原子の判断半径を決める
        latt = self.LatticeParameters()
        OverlapCriteriaRadius = 0.001 * min([latt[0], latt[1], latt[2]])

#非対称構造中のパラメータ
        AtomTypes = self.AtomTypeList()
        AtomSites = self.AtomSiteList()
        nAtomType = len(AtomTypes)
        nAtomSite = len(AtomSites)
        SPG = self.SpaceGroup()
        if DoTranslation == 1:
            nTranslation = 1 
        else:
            nTranslation = SPG.nTranslation()
        nSymmetryOperation = SPG.nSymmetryOperation()
        nTotalSymmetry = nTranslation * nSymmetryOperation

#座標展開を開始
        ExpandedAtomSiteList = []
        nExpandedAtomSiteList = []
        count = 0;
        for ia in range(nAtomSite):
            nExpandedAtomSiteList.append(0)
            atom         = AtomSites[ia]
            label        = atom.Label()
            AtomName     = atom.AtomName()
            Charge       = atom.Charge()
            iAtomType    = self.FindiAtomType(AtomName)
            AtomType     = AtomTypes[iAtomType]
            pos   = atom.Position(1)
            pos[0] = self.RoundSymmetricPosition(pos[0])
            pos[1] = self.RoundSymmetricPosition(pos[1])
            pos[2] = self.RoundSymmetricPosition(pos[2])
            Occupancy = atom.Occupancy()
            force     = atom.Force()
            v         = atom.Velocity()
            atom.SetId(ia)
#            print("ia={}: {} ({}, {}, {})".format(ia, label, pos[0], pos[1], pos[2]))

            ExpandedAtomSiteList.append(tkAtomSite())
            ExpandedAtomSiteList[count].SetIdAsymmetricAtomSite(ia)
            ExpandedAtomSiteList[count].SetiAtomType(iAtomType)
            ExpandedAtomSiteList[count].SetAtomType(AtomType)
            ExpandedAtomSiteList[count].SetLabel(label)
            ExpandedAtomSiteList[count].SetAtomName(AtomName)
            ExpandedAtomSiteList[count].SetCharge(Charge)
            ExpandedAtomSiteList[count].SetPosition(pos)
            if force is not None:
                ExpandedAtomSiteList[count].SetForce(force) 
            ExpandedAtomSiteList[count].SetOccupancy(Occupancy)
            if v is not None:
                ExpandedAtomSiteList[count].SetVelocity(v)

            nExpandedAtomSiteList[ia] += 1
            count += 1
#print "($ia) x,y,z: $x,$y,$z<br>\n";
            for iss in range(nTotalSymmetry):
                symop  = SPG.SymmetryOperation(iss)
                poss   = SPG.DoSymmetryOperation(iss, pos, 1)
                forces = SPG.DoVelocitySymmetryOperation(iss, force, 1)
                vs     = SPG.DoVelocitySymmetryOperation(iss, v, 1)
#                print(" iss={}: [{}] to ({}, {}, {})".format(iss, symop, poss[0], poss[1], poss[2]))

                IsPassed = 1
                for it in range(count):
                    id1 = ExpandedAtomSiteList[it].Id()
#                    print("  id1={}".format(id1))
                    if ia != id1:
                        continue
#                    iat = ExpandedAtomSiteList[it].iAtomType()
#                    if iat != iAtomType:
#                        continue

                    pos0 = ExpandedAtomSiteList[it].Position(1)
#                    print("  id1={}: ({}, {}, {})".format(id1,  pos0[0], pos0[1], pos0[2]))

                    dis = self.GetNearestInterAtomicDistance(pos0, poss)
#                    print("   dis=", dis, " < {}?".format(OverlapCriteriaRadius))
                    if dis < OverlapCriteriaRadius:
                        IsPassed = 0
                        break

                if IsPassed == 0:
                    continue

                ExpandedAtomSiteList.append(tkAtomSite())
                ExpandedAtomSiteList[count].SetIdAsymmetricAtomSite(ia)
                ExpandedAtomSiteList[count].SetiAtomType(iAtomType)
                ExpandedAtomSiteList[count].SetAtomType(AtomType)
                ExpandedAtomSiteList[count].SetLabel(label)
                ExpandedAtomSiteList[count].SetAtomName(AtomName)
                ExpandedAtomSiteList[count].SetCharge(Charge)
                ExpandedAtomSiteList[count].SetPosition(poss)
                ExpandedAtomSiteList[count].SetForce(forces)
                ExpandedAtomSiteList[count].SetOccupancy(Occupancy)
                ExpandedAtomSiteList[count].SetVelocity(vs)
#Idは1からはじまる
                ExpandedAtomSiteList[count].SetId(ia)
#print "v: ($vx,$vy,$vz) ($vxs, $vys, $vzs)<br>\n";
                nExpandedAtomSiteList[ia] += 1
                count += 1

        self.__ExpandedAtomSiteList   = ExpandedAtomSiteList
        self.__nTotalExpandedAtomSite = count
        self.__nExpandedAtomSite      = nExpandedAtomSiteList

#多重度を、非対称原子に設定
        for ia in range(nAtomSite):
            atom = AtomSites[ia]
            atom.SetMultiplicity(nExpandedAtomSiteList[ia])
#多重度を、全原子に設定
        self.__MultiplitictyList = []
        for i in range(len(ExpandedAtomSiteList)):
            atom = ExpandedAtomSiteList[i]
            ia   = atom.Id()
#            print("i=", i, " ia=", ia, "m=", nExpandedAtomSiteList[ia])
            self.__MultiplitictyList.append(nExpandedAtomSiteList[ia])
            atom.SetMultiplicity(nExpandedAtomSiteList[ia])

        self.CalculateDensity()
        self.AnalyzeChemicalComposition()

    def ExpandCoordinates(self, DoTranslation = 1):
        return self.expand_coordinates(DoTranslation = DoTranslation)

    def GetimaxByRmax(self, Rmax):
        latt = self.LatticeParameters()
        return int(Rmax/latt[0])+1, int(Rmax/latt[1])+1, int(Rmax/latt[2])+1

    def GetReciprocalDistanceFromK(self, k0, k1):
        dk = np.empty(3)
        dk[0] = k1[0] - k0[0]
        dk[1] = k1[1] - k0[1]
        dk[2] = k1[2] - k0[2]

        gij, Rgij = self.Metrics()
        dk2 = Rgij[0][0] * dk[0]*dk[0] + Rgij[1][1] * dk[1]*dk[1] + Rgij[2][2] * dk[2]*dk[2] \
            + 2.0 * (Rgij[0][1] * dk[0]*dk[1] + Rgij[0][2] * dk[0]*dk[2] + Rgij[1][2] * dk[1]*dk[2])
        return sqrt(dk2)


    def distance(self, pos0, pos1, **kwargs):
        dx = np.empty(3)
        dx[0] = pos1[0] - pos0[0]
        dx[1] = pos1[1] - pos0[1]
        dx[2] = pos1[2] - pos0[2]

        gij, Rgij = self.Metrics()
        r2 = gij[0][0] * dx[0]*dx[0] + gij[1][1] * dx[1]*dx[1] + gij[2][2] * dx[2]*dx[2] \
           + 2.0 * (gij[0][1] * dx[0]*dx[1] + gij[0][2] * dx[0]*dx[2] + gij[1][2] * dx[1]*dx[2])

#        print(f"pos1=({pos1[0]:7.4f} {pos1[1]:7.4f} {pos1[2]:7.4f})-({pos0[0]:7.4f} {pos0[1]:7.4f} {pos0[2]:7.4f} r={r2}")

        return sqrt(r2)

    def GetInterAtomicDistance(self, pos0, pos1, **kwargs):
        return self.distance(pos0, pos1, **kwargs)

    def FindNearestEquivalentFractionCoordinate(self, x2, x):
        d = x2 - x + 0.5
        if d < 0:
            d = int(d - 1)
        else:
            d = int(d)
        xret = x2 - d
#        print("x,x2,xret=", x, x2, xret)
        return xret

        """
        xret = x2
        mind = abs(x2 - x)
        while 1:
            if mind > abs(x2 + 1.0 - x):
                xret = x2 + 1.0
                mind = abs(x2 + 1.0 - x)
            elif mind > abs(x2 - 1.0 - x):
                xret = x2 - 1.0
                mind = abs(x2 - 1.0 - x)
            else:
                break
        return xret
        """

    def get_nearest_interatomic_distance_with_offset(self, pos0, pos1, AllowZero = True, irange = [1, 1, 1], **kwargs):
        pos00 = pos0.copy()
        pos00[0] = self.FindNearestEquivalentFractionCoordinate(pos0[0], pos1[0])
        pos00[1] = self.FindNearestEquivalentFractionCoordinate(pos0[1], pos1[1])
        pos00[2] = self.FindNearestEquivalentFractionCoordinate(pos0[2], pos1[2])
#        print("pos change: ({:8.4g}, {:8.4g}, {:8.4g})=>({:8.4g}, {:8.4g}, {:8.4g})".format(*pos0, *pos00))

        r  = self.GetInterAtomicDistance(pos00, pos1);
        if AllowZero and r < 1.0e-5:
#            print("off:", pos00[0] - pos0[0], pos00[1] - pos0[1], pos00[2] - pos0[2])
            return r, pos00[0] - pos0[0], pos00[1] - pos0[1], pos00[2] - pos0[2]

        gij, Rgij = self.Metrics()

        offset = None
        if r > 1.0e-5:
            r = 1.0e10
            for ix in range(-irange[0], irange[0]+1):
                for iy in range(-irange[1], irange[1]+1):
                    for iz in range(-irange[2], irange[2]+1):
                        r2 = self.GetInterAtomicDistance([pos00[0] + ix, pos00[1] + iy, pos00[2] + iz], pos1, **kwargs)
#                        print(f"for idx={v}, ixyz={ix} {iy} {iz} pos=({pos00[0]:7.4f} {pos00[1]:7.4f} {pos00[2]:7.4f})-({pos1[0]:7.4f} {pos1[1]:7.4f} {pos1[2]:7.4f}) r={r2:7.4f}")
#                        print("   r2=", ix, iy, iz, r2)
                        if r2 > 1.0e-5 and r2 < r:
                            r = r2
                            offset = [pos00[0] + ix - pos0[0], pos00[1] + ix - pos0[1], pos00[2] + ix - pos0[2]]

        if offset is None: return None, None, None, None

        return r, offset[0], offset[1], offset[2]

    def GetNearestInterAtomicDistanceWithOffset(self, pos0, pos1, AllowZero = True, irange = [1, 1, 1], **kwargs):
        return self.get_nearest_interatomic_distance_with_offset(pos0, pos1, AllowZero = True, irange = [1, 1, 1], **kwargs)

    def get_nearest_interatomic_aistance(self, pos0, pos1, AllowZero = True, irange = [1, 1, 1], **kwargs):
        r, xoff, yoff, zoff = self.GetNearestInterAtomicDistanceWithOffset(pos0, pos1, AllowZero, irange = irange, **kwargs)
        return r

    def GetNearestInterAtomicDistance(self, pos0, pos1, AllowZero = True, irange = [1, 1, 1], **kwargs):
        return self.get_nearest_interatomic_aistance(pos0, pos1, AllowZero = True, irange = [1, 1, 1], **kwargs)

    def GetNearestInterAtomicDistances(self, pos0, pos1, CoordRMax = 10.0, AllowZero = True, irange = None):
        if irange is None:
            irange = self.GetimaxByRmax(CoordRMax);
        list = []
        c = 0
        for ix in range(-irange[0], irange[0]+1):
            for iy in range(-irange[1], irange[1]+1):
                for iz in range(-irange[2], irange[2]+1):
                    r = self.GetInterAtomicDistance(pos0, [pos1[0] + ix, pos1[1] + iy, pos1[2] + iz])
                    if r < CoordRMax:
                        if not AllowZero and r < 1.0e-5: continue
                        list[c] = {'piCell': [ix, iy, iz], 'r': r}
                        c += 1
        return list

    def GetInterAtomicAngle(self, pos0, pos1, pos2):
        dis01 = self.GetInterAtomicDistance(pos0, pos1)
        if dis01 == 0.0:
            return 0.0
        dis02 = self.GetInterAtomicDistance(pos0, pos2)
        if dis02 == 0.0:
            return 0.0

        dx01 = pos1 - pos0
        dx02 = pos2 - pos0
        gij, Rgij = self.Metrics()
        ip  = 0.0
        for i in range(3):
            for j in range(3):
                ip += gij[i][j] * dx01[i] * dx02[j]

        cosa = ip / dis01 / dis02
        angle = degacos(cosa)
        if angle > 180.0:
            angle = 360.0 - angle 

        return angle;

    def AddAtomType(self, atomtype = None, charge = None):
        at = tkAtomType()

        name, nametype, type, charge1 = SplitAtomName(atomtype)
#        name, nametype, type, charge1 = at.SplitAtomName(atomtype)
#        print("CO: AddAtomType:", atomtype, name, nametype, type, charge1)
        if charge is None:
            charge = charge1
        if charge is None:
            charge = 0.0

        ret = self.FindAtomType(name)
        if ret >= 0:
            return ret, self.__AtomTypeList[ret]

        t = tkAtomType(nametype, charge)
        self.__AtomTypeList.append(t)
        idx = len(self.__AtomTypeList)

        return idx-1, self.__AtomTypeList[idx-1]

    def SetAtomSite(self, iatom, iatomtype = None, atomtype = None,
                        label = None, atomname = None, m = None, ws = None,
                         pos = [], force = None, occ = 1.0, hydrogen = None):
        n = len(self.__AtomSiteList)
        if n <= iatom:
            return None
        at = self.__AtomSiteList[iatom]
        if iatomtype is not None:
            at.SetiAtomType(iAtomType)
        if atomtype is not None:
            at.SetAtomType(atomtype)
        if label is not None:
            at.SetLabel(label)
        if atomname is not None:
            at.SetAtomName(atomname)
        if pos is not None:
            at.SetPosition(pos)
        if occ is not None:
            at.SetOccupancy(occ)
        if force is not None:
            at.SetForce(force)
        at.__ws = ws
        at.__hydrogen = hydrogen

    def AddAtomSite(self, label = None, name = None, charge = None, m = None, ws = None,
                     pos = None, occ = 1.0, force = None, velocity = None, hydrogen = None):
        iAtomType, AtomType = self.AddAtomType(name)
        atomlist = self.AtomSiteList()
#        print("{} {} {} na={}".format(name, iAtomType, AtomType, len(atomlist)))

        if label is None:
            n = self.count_by_type(name, 'short', 'asymmetric')
            label = "{}{}".format(name, n + 1)

#        print("label=", label, "  iatomtype=", iAtomType, AtomType)
        at = tkAtomSite()
        at.SetAtomSite(atomtype = AtomType, iatomtype = iAtomType, label = label, name = name, charge = charge,
                m = m, ws = ws, pos = pos, occ = occ, hydrogen = hydrogen)
#        at.SetiAtomType(iAtomType)
#        at.SetAtomType(AtomType)
#        print("     iatomtype=", at.iAtomType(), at.AtomType())

        atomlist.append(at)
        at.SetId(len(atomlist))
        

    def DoSymmetryOperation(self, pos, i):
        return self.SpaceGroup().DoSymmetryOperation(pos, i)

    def nSymmetryOperation(self):
        return self.SpaceGroup().nSymmetryOperation()

    def SymmetryOperation(self, i):
        SPG = self.SpaceGroup()
        return SPG.SymmetryOperation(i)
    
    def AddSymmetry(self, sym):
        sym = DelQuote(sym)
        self.SpaceGroup().AddSymmetryOperation(sym)

    def SpaceGroup(self):
        if self.__SpaceGroup is None:
            return self.ClearSpaceGroup()
        return self.__SpaceGroup

    def LatticeAxis(self):
        return self.lattice_axis()

    def lattice_axis(self):
        SPG = self.SpaceGroup()
        return SPG.LatticeAxis().lower()

    def lattice_system(self):
        SPG = self.SpaceGroup()
        ls = SPG.LatticeSystem()
        return ls.lower()

    def LatticeSystem(self):
        return self.lattice_system()

    def SetLatticeSystem(self, latticesystem, SearchByLatticeParameter, tollat, tolangle):
        SPG = self.SpaceGroup()
        SPG.SetLatticeSystem(latticesystem, SearchByLatticeParameter, tollat, tolangle)

    def ReciprocalLatticeParameters(self, UseAtomicUnit = 0):
        rlatt = list(self.__RLatt)
        if UseAtomicUnit:
            return a0 * 1.0e10 * rlatt
        return rlatt

    def ReciprocalLatticeVectors(self, UseAtomicUnit = 0):
        if UseAtomicUnit:
            return a0 * 1.0e10 * self.__Raij
        else:
            return self.__Raij

    def DiffractionAngle(self, wl, h, k, l): # wl in angstrom
        dhkl = self.CalculateHKLInterplanarSpacing(h, k, l)  # angstrom
        sB = 1.0 / 2.0 / dhkl        # A^-1
        Q2 = 2.0 * arcsin(wl * sB)
        return dhkl, sB, Q2

    def Fhkl(self, sB, h, k, l):
        sites = self.ExpandedAtomSiteList()
        Fhkl = 0.0
        for i in range(len(sites)):
            atom     = sites[i]
            x, y, z  = atom.Position()
            occ      = atom.Occupancy()
            atomtype = atom.AtomType()

            phase = h * x + k * y + l * z
            Fhkl += occ * atomtype.asf(sB) * exp(1.0j * pi2 * phase)
        return Fhkl

    def SetVolume(self, vol):
        if vol is not None:
            self.__Volume = vol

    def Volume(self, UseAtomicUnit = 0):
        if UseAtomicUnit:
            k = 1.0 / (a0 * 1.0e10)
            return self.__Volume * k * k * k
        return self.__Volume

    def ReciprocalVolume(self, UseAtomicUnit = 0):
        if UseAtomicUnit:
            k = a0 * 1.0e10
            return self.__RVolume * k * k * k
        return self.__RVolume

    def CalculateVolume(self):
        latt = self.__Latt
        cosa = degcos(latt[3])
        cosb = degcos(latt[4])
        cosg = degcos(latt[5])
#        print("latt=", latt[0], latt[1], latt[2])
#        print("cos a,b,g=", cosa, cosb, cosg)
        vol = 1.0 - cosa*cosa - cosb*cosb - cosg*cosg + 2.0 * cosa*cosb*cosg
        vol = latt[0] * latt[1] * latt[2] * sqrt(vol)
#        print("vol=", vol)
#        self.__Volume = Round(vol, 4)
        self.__Volume = vol
        return self.__Volume

    def CalculateReciprocalVolume(self):
        Rlatt = self.__RLatt
        cosa = degcos(Rlatt[3])
        cosb = degcos(Rlatt[4])
        cosg = degcos(Rlatt[5])
        Rvol = 1.0 - cosa*cosa - cosb*cosb - cosg*cosg + 2.0 * cosa*cosb*cosg
        Rvol = Rlatt[0]*Rlatt[1]*Rlatt[2] * sqrt(Rvol);
#        self.__RVolume = Round(Rvol, 4);
        self.__RVolume = Rvol
        return self.__RVolume

    def fractional2cartesian(self, x, y, z):
        xc, yc, zc = self.__aij.transpose() @ [x, y, z]
        
        return [xc, yc, zc]

    def FractionalToCartesian(self, x, y, z):
        return self.fractional2cartesian(x, y, z)

    def CartesianToFractional(self, xc, yc, zc):
        xyz  = la.solve(self.__aij.transpose(), [xc, yc, zc])
        return xyz

    def CalculateHKLInterplanarSpacing(self, h, k, l):
        Rg00, Rg01, Rg02 = self.__Rgij[0][0], self.__Rgij[0][1], self.__Rgij[0][2]
        Rg10, Rg11, Rg12 = self.__Rgij[1][0], self.__Rgij[1][1], self.__Rgij[1][2]
        Rg20, Rg21, Rg22 = self.__Rgij[2][0], self.__Rgij[2][1], self.__Rgij[2][2]
        
        d = Rg00 * h * h + Rg11 * k * k + Rg22 * l * l \
          + 2.0 * (Rg01 * h * k + Rg12 * k * l + Rg02 * l * h)
        if d == 0.0:
            return None
        else:
            d = sqrt(1.0 / d)
            return d

    def SetSpaceGroup(self, name = None, number = None, set = 1):
        self.SpaceGroup().SetSpaceGroup(name, number, set)
        SPG = self.SpaceGroup()
        SPG.SetSPGName(name)
        SPG.SetiSPG(number)
        SPG.SetLatticeSystem()
        SPG.SetLatticeAxis()
        if set is not None:
            SPG.SetiSet(set)

    def GetSpaceGroupInf(self):
        sg = self.SpaceGroup()
        return sg.SPGName(), sg.iSPG(), sg.iSet()

    def GetLatticeRange(self, Rmax):
        a, b, c, alpha, beta, gamma = self.LatticeParameters()
        sinA = abs(sin(alpha * torad));
        sinB = abs(sin(beta  * torad));
        sinG = abs(sin(gamma * torad));
        if sinB < sinA:
            sinA = sinB 
        if sinG < sinA:
            sinA = sinG 
        nx = int(Rmax / a / sinA)
        ny = int(Rmax / b / sinA)
        nz = int(Rmax / c / sinA)
        return nx+1, ny+1, nz+1

    """
    def SplitAtomName(self, atomname):
        list = tkre.Match(r'([A-Z][a-z]?(\.[a-zA-Z]*)?)(\[.*\])?([\d+\-\.]*)', atomname)
        name     = list[1]
        sitetype = list[2]
        charge   = list[3]
        if charge is not None and charge != '':
            c = tkre.Match(r'(\d+)([\+\-])?', charge)
            if c[1] is None:
                pass
            elif c[2] == '-':
                charge = -float(c[1])
            else:
                charge = float(c[1])
        if sitetype is None:
            nametype = name
        else:
            nametype = name + '[' + sitetype + ']'
        return name, nametype, sitetype, charge
    """
        
    def RoundSymmetricPosition(self, x, tol = 0.0001):
#        if x ==  0.3333 or x ==  0.333:
#            return 1.0 / 3.0
        v = 1.0 / 3.0
        if abs(x - v) < tol:
            return v
        v = -1.0 / 3.0
        if abs(x - v) < tol:
            return v
        v = 2.0 / 3.0
        if abs(x - v) < tol:
            return v
        v = -2.0 / 3.0
        if abs(x - v) < tol:
            return v
        v = 1.0 / 6.0
        if abs(x - v) < tol:
            return v
        v = -1.0 / 6.0
        if abs(x - v) < tol:
            return v
        v = 5.0 / 3.0
        if abs(x - v) < tol:
            return v
        v = -5.0 / 3.0
        if abs(x - v) < tol:
            return v
        return x

    def SymmetrizeParameter(self, ls, pos, tol):
        if ls is None or ls == '':
            ls = self.LatticeAxis() 
        if ls == 'cubic' or ls == 'rhombohedral' or ls == 'trigonal':
            pos[0] = pos[1] = pos[2] = (pos[0] + pos[1] + pos[2]) / 3.0
        elif ls == 'tetragonal' or ls == 'hexagonal':
            pos[0] = pos[1] = (pos[0] + pos[1]) / 2.0
        pos[0] = RoundParameter(pos[0], tol);
        pos[1] = RoundParameter(pos[1], tol);
        pos[2] = RoundParameter(pos[2], tol);
        return pos

    def Symmetrize(self, tollatt = 1.0-5, tolangle = 1.0-5, tolpos = 1.0-5):
        latt = self.LatticeParameters()
        if abs(latt[3] - 90.0) < tolangle:
            latt[3] = 90
        if abs(latt[4] - 90.0) < tolangle:
            latt[4] = 90
        if abs(latt[5] - 90.0) < tolangle:
            latt[5] = 90
        self.LatticeParameters(latt)
    
#        $this->SetLatticeSystem('', 1, $tollatt, $tolangle);
#        $SPG = $this->GetCSpaceGroup();
        ls  = SPG.LatticeSystem()

        if ls == 'cubic':
            ar = (latt[0] + latt[1] + latt[2]) / 3.0
            self.SetLatticeParameters([ar, ar, ar, 90, 90, 90])
        elif ls == 'rhombohedral' or ls == 'trigonal':
            ar = (latt[0] + latt[1] + latt[2]) / 3.0
            al = (latt[3] + latt[1] + latt[5]) / 3.0;
            self.SetLatticeParameters([ar, ar, ar, al, al, al])
        elif ls == 'tetragonal':
            ar = (latt[0] + latt[1]) / 2.0
            self.SetLatticeParameters([ar, ar, latt[2], 90, 90, 90])
        elif ls == 'hexagonal':
            ar = (latt[0] + latt[1]) / 2.0
            self.SetLatticeParameters([ar, ar, latt[2], 90, 90, 120])
        elif ls == 'orthorhombic':
            self.SetLatticeParameters([latt[0], latt[1], latt[2], 90, 90, 90])

    def calculate_lattice_parameters_from_vector(self, aij = None):
        if aij is None: 
            aij = self.__aij
            update_self = False
        else:
            update_self = True

        latt = np.empty([6])
        for i in range(3):
            latt[i] = sqrt(np.inner(aij[i], aij[i]))
        latt[3] = np.inner(aij[1], aij[2]) / latt[1] / latt[2]
        latt[4] = np.inner(aij[2], aij[0]) / latt[2] / latt[0]
        latt[5] = np.inner(aij[0], aij[1]) / latt[0] / latt[1]
        for i in range(3, 6):
            latt[i] = self.RoundLatticeParameters(degacos(latt[i]), 1.0e-6)

        if update_self:
            self.__Latt = latt
            V = self.CalculateVolume();

        return latt

    def CalculateLatticeParametersFromVector(self, aij = None):
        return self.calculate_lattice_parameters_from_vector(aij = aij)

    def Metrics(self):
        return self.__gij, self.__Rgij

    def CalcLatticeVectorsFromLatticeParameters(self, latt = None):
        if latt is None:
           latt = self.LatticeParameters()

        ls = self.LatticeSystem()

# Lattice vectors
        aij = np.empty([3, 3], dtype = float)
        if ls == 'trigonal' and abs(latt[4] - 60.0) < 1.0e-3:
            self.ConventionalSPGName = 'F'
            ac = latt[0] * sqrt(2.0)
            self.ConventionalLatticeParameters = [ac, ac, ac, 90.0, 90.0, 90.0]

        if self.ConventionalSPGName is not None \
            and tkre.Match('\s*[FI]', self.ConventionalSPGName, 'i'):
            pass
#            pT = self.GetMatrixConventionalToPrimitiveCell(1)
#            latt = self.ConventionalLatticeParameters
#            ax = a * pT->[0][0];
#            ay = b * pT->[0][1];
#            az = c * pT->[0][2];
#            bx = a * pT->[1][0];
#            by = b * pT->[1][1];
#            bz = c * pT->[1][2];
#            cx = a * pT->[2][0];
#            cy = b * pT->[2][1];
#            cz = c * pT->[2][2];
        elif self.ConventionalSPGName is not None \
            and tkre.Match('\s*[ABC]', self.ConventionalSPGName, 'i'):
            pass
#        pT = $this->GetMatrixConventionalToPrimitiveCell(1);
#        pl = $this->{ConventionalLatticeParameters};
#        my ($a, $b, $c, $alpha, $beta, $gamma) = @$pl;
#        ax = $a * $pT->[0][0];
#        ay = $b * $pT->[0][1];
#        az = $c * $pT->[0][2];
#        bx = $a * $pT->[1][0];
#        by = $b * $pT->[1][1];
#        bz = $c * $pT->[1][2];
#        cx = $a * $pT->[2][0];
#        cy = $b * $pT->[2][1];
#        cz = $c * $pT->[2][2];
        elif tkre.Search('hex', ls, 'i'):
            pass
#        ax = $a / 2.0;
#        ay = -sqrt(3.0) / 2.0 * $a;
#        az = 0.0;
#        bx = $ax;
#        by = -$ay;
#        bz = 0.0;
#        cx = 0.0;
#        cy = 0.0;
#        cz = $c;
        elif tkre.Match('(trig|rhomb)', ls, 'i'):
            pass
#            cosa = cos($torad * $alpha);
#        axy = $a * sqrt( (1.0 - $cosa) / 6.0);
#        ax = sqrt(3.0) * $axy;
#        ay = $axy;
#        az = $axy * sqrt(6.0 / (1.0 - $cosa) - 4.0);
#        bx = -$ax;
#        by = $ay;
#        bz = $az;
#        cx = 0.0;
#        cy = -2.0 * $axy;
#        cz = $az;
        else:
#            cosa = degcos(latt[3])
            cosb = degcos(latt[4])
            cosg = degcos(latt[5])
#            sina = degsin(latt[3])
#            sinb = degsin(latt[4])
            sing = degsin(latt[5])

            aij[0][0] = latt[0]
            aij[0][1] = 0.0;
            aij[0][2] = 0.0;
            aij[1][0] = latt[1] * cosg
            aij[1][1] = latt[1] * sing
            aij[1][2] = 0.0;
            aij[2][0] = latt[2] * cosb
            aij[2][1] = (latt[2] * cosb - aij[2][0] * cosg) / sing
            aij[2][2] = sqrt(latt[2] * latt[2] - aij[2][0] * aij[2][0] - aij[2][1] * aij[2][1])

        return aij

    def CalcMetrics(self, update_aij = True):
        SPG     = self.SpaceGroup()
#        SPGName = SPG.SPGName()
        ls      = self.LatticeSystem()
#        pT      = self.GetMatrixConventionalToPrimitiveCell(0)

        latt = self.LatticeParameters()
        if latt[0] == 0.0:
            print("Error in {}.{}: Invalid Lattice Parameters."
                .format(self.ClassPath(), self.__class__.__name__))
            print("Lattice: {} {} {}   {} {} {}"
                .format(latt[0], latt[1], latt[2], latt[3], latt[4], latt[5]))

        sina = degsin(latt[3])
        sinb = degsin(latt[4])
        sing = degsin(latt[5])
        cosa = degcos(latt[3])
        cosb = degcos(latt[4])
        cosg = degcos(latt[5])
        if abs(cosa) < 1.0e-100:
            cosa = 0.0
        if abs(cosb) < 1.0e-100:
            cosb = 0.0
        if abs(cosg) < 1.0e-100:
            cosg = 0.0
        if abs(sina - 1.0) < 1.0e-100:
            sina = 1.0
        if abs(sinb - 1.0) < 1.0e-100:
            sinb = 1.0
        if abs(sing - 1.0) < 1.0e-100:
            sing = 1.0

        gij = np.empty([3, 3], dtype = float)
        for i in range(3):
            gij[i][i] = latt[i] * latt[i]
        gij[1][0] = gij[0][1] = latt[0] * latt[1] * cosg
        gij[2][0] = gij[0][2] = latt[0] * latt[2] * cosb
        gij[1][2] = gij[2][1] = latt[1] * latt[2] * cosa
        if abs(gij[0][1]) < 1.0e-7:
            gij[1][0] = gij[0][1] = 0.0
        if abs(gij[0][2]) < 1.0e-7:
            gij[2][0] = gij[0][2] = 0.0
        if abs(gij[1][2]) < 1.0e-7:
            gij[1][2] = gij[2][1] = 0.0
        self.__gij = gij

        V = self.CalculateVolume();

        if update_aij == True:
            aij = self.CalcLatticeVectorsFromLatticeParameters(latt)
            self.__aij = aij

        self.__Rgij = np.empty([3, 3], dtype = float)
        self.__Rgij[0][0] = latt[1] * latt[1] * latt[2] * latt[2] * sina * sina / V / V
        self.__Rgij[0][1] = self.__Rgij[1][0] \
            = latt[0] * latt[1] * latt[2] * latt[2] * (cosa * cosb - cosg) / V / V
        if abs(self.__Rgij[0][1]) < 1.0e-7:
            self.__Rgij[0][1] = self.__Rgij[1][0] = 0.0
        self.__Rgij[0][2] = self.__Rgij[2][0] \
            = latt[0] * latt[1] * latt[1] * latt[2] * (cosg * cosa - cosb) / V / V
        if abs(self.__Rgij[0][2]) < 1.0e-7:
            self.__Rgij[0][2] = self.__Rgij[2][0] = 0.0
        self.__Rgij[1][1] = latt[0] * latt[0] * latt[2] * latt[2] * sinb * sinb / V / V
        self.__Rgij[1][2] = self.__Rgij[2][1] \
            = latt[0] * latt[0] * latt[1] * latt[2] * (cosb * cosg - cosa) / V / V
        if abs(self.__Rgij[1][2]) < 1.0e-7:
            self.__Rgij[1][2] = self.__Rgij[2][1] = 0.0
        self.__Rgij[2][2] = latt[0] * latt[0] * latt[1] * latt[1] * sing * sing / V / V

        Ra = np.empty(6, dtype = float)
        for i in range(3):
            Ra[i] = sqrt(self.__Rgij[i][i])
        Ra[3] = degacos(self.__Rgij[1][2] / Ra[1] / Ra[2])
        Ra[4] = degacos(self.__Rgij[2][0] / Ra[2] / Ra[0])
        Ra[5] = degacos(self.__Rgij[0][1] / Ra[0] / Ra[1])
        Ra[3] = self.RoundLatticeParameters(Ra[3], 1.0e-8)
        Ra[4] = self.RoundLatticeParameters(Ra[4], 1.0e-8)
        Ra[5] = self.RoundLatticeParameters(Ra[5], 1.0e-8)
        self.__RLatt = Ra

        RV = self.CalculateReciprocalVolume();

        kV = 1.0 / V
        Ra = kV * np.cross(aij[1], aij[2])
        Rb = kV * np.cross(aij[2], aij[0])
        Rc = kV * np.cross(aij[0], aij[1])
        self.__Raij = np.array([Ra, Rb, Rc])

    def CalMetricsFromLatticeVector(self):
        self.__gij = np.empty([3, 3], dtype = float)
        for i in range(3):
            for j in range(i, 3):
                self.__gij[i][j] = np.inner(self.__aij[i], self.__aij[j])
                self.__gij[j][i] = self.__gij[i][j]

        return self.__gij

    def SetLatticeParameters(self, latt, GuessLatticeSystem = 0, update_aij = True):
#        print("latt=", latt)
        self.__Latt = latt
        SPG = self.SpaceGroup()
        SPG.SetLatticeParameters(latt, GuessLatticeSystem)
        SPG.SetLatticeAxis()
        self.CalcMetrics(update_aij)

    def lattice_parameters(self, UseAtomicUnit = 0):
        latt = list(self.__Latt)
        if UseAtomicUnit:
            return 1.0 / (a0 * 1.0e10) * latt
        return latt

    def LatticeParameters(self, UseAtomicUnit = 0):
        return self.lattice_parameters(UseAtomicUnit = UseAtomicUnit)

    def CalLatticeParametersFromMetrics(self, update_aij = False):
#        print("gij=", self.__gij)
        a     = RoundParameter(sqrt(self.__gij[0][0]), 5.0e-6)
        b     = RoundParameter(sqrt(self.__gij[1][1]), 5.0e-6)
        c     = RoundParameter(sqrt(self.__gij[2][2]), 5.0e-6)
        alpha = self.RoundLatticeParameters(degacos(self.__gij[1][2] / b / c), 1.0e-6)
        beta  = self.RoundLatticeParameters(degacos(self.__gij[2][1] / c / a), 1.0e-6)
        gamma = self.RoundLatticeParameters(degacos(self.__gij[0][1] / a / b), 1.0e-6)

        self.SetLatticeParameters([a, b, c, alpha, beta, gamma], update_aij)

    def set_lattice_vectors(self, aij):
        self.__aij = np.array(aij)
        self.CalMetricsFromLatticeVector();
        self.CalLatticeParametersFromMetrics(update_aij = False);
# CalLatticeParametersFromMetrics re-calculates aij through SetLatticeParameters()
# Recover original aij
        self.__aij = np.array(aij)

    def SetLatticeVectors(self, aij):
        return self.set_lattice_vectors(aij)

    def lattice_vectors(self, UseAtomicUnit = 0):
        if UseAtomicUnit:
            return 1.0 / (a0 * 1.0e10) * self.__aij
        else:
            return self.__aij

    def LatticeVectors(self, UseAtomicUnit = 0):
        return self.lattice_vectors(UseAtomicUnit)

    def count_by_type(self, atype, mode = 'short', target = 'expanded'):
        atype = atype.upper()

        if target == 'expanded':
            AtomSites = self.ExpandedAtomSiteList()
        else:
            AtomSites = self.AtomSiteList()
        nsites = len(AtomSites)
        n = 0
        if mode == 'short':
            for i in range(nsites):
                t = AtomSites[i]
                name = t.AtomNameOnly().upper()
                if name == atype:
                    n += 1
        else:
            for i in range(nsites):
                t = AtomSites[i]
                name = t.AtomName().upper()
                if name == atype:
                    n += 1

        return n

    def normalize_hkl(self, h, k, l):
#        ls = self.lattice_system()
        ls = self.lattice_axis()

        if ls == 'cubic':
            h, k, l = sorted([abs(h), abs(k), abs(l)])
        elif ls == 'hexagonal' or ls == 'trigonal':
            l = abs(l)
            h, k, i = sorted([h, k, -h-k], key = lambda x: abs(x))
            if h == 0 and k < 0:
                h, k, i = [-h, -k, -i]
            elif h < 0:
                h, k, i = [-h, -k, -i]
        elif ls == 'tetragonal':
            h, k = sorted([abs(h), abs(k)])
            l = abs(l)
        elif ls == 'orthorhombic':
            h, k, l = [abs(h), abs(k), abs(l)]
        elif ls == 'rhombohedral':
            h, k, l = sorted([h, k, l])
        elif ls == 'monoclinic':
            lattice_parameters = self.LatticeParameters()
            if abs(lattice_parameters[4] - 90.0) > 1.0e-5:  # alpha
                h = abs(h)
            elif abs(lattice_parameters[5] - 90.0) > 1.0e-5:  # beta
                k = abs(k)
            elif abs(lattice_parameters[6] - 90.0) > 1.0e-5:  # gamma
                l = abs(l)

        return h, k, l

    def multiplicity_cubic(self, h, k,l):
        h = abs(h)
        k = abs(k)
        l = abs(l)
        if h != 0:
            m *= 2
        if k != 0:
            m *= 2
        if l != 0:
            m *= 2
        if h != k and k != l and l != h:
            m *= 6
        elif h == k and k != l:
            m *= 3
        elif k == l and l != h:
            m *= 3
        elif l == h and h != k:
            m *= 3
        return m

    def multiplicity_tetragonal(self, h, k, l):
        h = abs(h)
        k = abs(k)
        if h != 0:
            m *= 2
        if k != 0:
            m *= 2
        if l != 0:
            m *= 2
        if h != k:
            m *= 2
        return m

    def multiplicity_orthorhombic(self, h, k, l):
        m = 1
        if h != 0:
            m *= 2
        if k != 0:
            m *= 2
        if l != 0:
            m *= 2
        return m

    def multiplicity_rhombohedral(self, h, k, l):
        if k == 0:
            h, k, l = k, h, l
        if l == 0:
            h, k, l = l, h, k

        if h == 0:
            if k == 0 and l != 0:
                return 2
            if k != 0 and l == 0:
                return 2
            if k != 0 and l != 0:
                if k != l:
                    return 4
                if k == l:
                    return 2

        if h != k and k != l and h != l:
            return 12

        if h == k and h != l:
            return 6
        if h == l and k != l:
            return 6
        if k == l and h != l:
            return 6

    def multiplicity_trigonal(self, h, k, l):
        return self.multiplicity_hexagonal(h, k, l)

    def multiplicity_hexagonal(self, h, k, l):
        i = -h - k
        if i == 0:
            k = i
        
        if h != 0 and k != 0 and l != 0 and h != k:
            return 24

        if h != 0 and h == k and l != 0:
            return 12
        if h == 0 and k != 0 and l != 0:
            return 12
        if h != 0 and k == 0 and l != 0:
            return 12 
        if h != 0 and k != 0 and h != k and l == 0:
            return 12 

        if h != 0 and h == k and l == 0:
            return 6
        if h == 0 and k != 0 and l == 0:
            return 6
        if h != 0 and k == 0 and l == 0:
            return 6

        if h == 0 and k == 0 and l != 0:
            return 2

        return None

    def multiplicity_hkl(self, h, k, l):
        ls = self.lattice_system()

        if ls == 'cubic':
            return self.multiplicity_cubic(h, k, l)
        elif ls == 'tetragonal':
            return self.multiplicity_tetragonal(h, k, l)
        elif ls == 'trigonal':
            return self.multiplicity_trigonal(h, k, l)
        elif ls == 'rhombohedral':
            return self.multiplicity_rhombohedral(h, k, l)
        elif ls == 'hexagonal':
            return self.multiplicity_hexagonal(h, k, l)
        elif ls == 'orthorhombic':
            return self.multiplicity_orthorhombic(h, k, l)
        elif ls == 'triclinic':
            return 2;
        else:
            print("")
            print(f"Error in tkcrystalobeject.multiplicity_hkl(): Lattice system [{ls}] is not supported.")
        return None
