import time
import numpy as np
from numpy import sin, cos
from xml.etree import ElementTree as ET
import matplotlib
import matplotlib.pyplot as plt

from pymatgen.io.cif import CifParser, CifWriter
from pymatgen.core.structure import Structure
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.core import Element
from pymatgen.core import Composition

from pymatgen.io.vasp.outputs import Vasprun
from pymatgen.io.vasp.inputs import Poscar

from tabulate import tabulate


from tklib.tkobject import tkObject
from tklib.tkutils import get_ext, pint, pfloat
from tklib.tksci.tksci import pi
from tklib.tkwrapper.tktqdm import tqdm


# URL: https://github.com/materialsproject/pymatgen/blob/v2023.5.10/pymatgen/analysis/diffraction/xrd.py
# 'CuKa', 'CuKa2', 'CuKa1', 'CuKb1', 'MoKa', 'MoKa2', 'MoKa1', 'MoKb1',
# 'CrKa', 'CrKa2', 'CrKa1', 'CrKb1', 'FeKa', 'FeKa2', 'FeKa1', 'FeKb1', 'CoKa', 'CoKa2', 'CoKa1', 'CoKb1',
# 'AgKa', 'AgKa2', 'AgKa1', 'AgKb1'
"""
WAVELENGTHS = {
    "CuKa": 1.54184,
    "CuKa2": 1.54439,
    "CuKa1": 1.54056,
    "CuKb1": 1.39222,
    "MoKa": 0.71073,
    "MoKa2": 0.71359,
    "MoKa1": 0.70930,
    "MoKb1": 0.63229,
    "CrKa": 2.29100,
    "CrKa2": 2.29361,
    "CrKa1": 2.28970,
    "CrKb1": 2.08487,
    "FeKa": 1.93735,
    "FeKa2": 1.93998,
    "FeKa1": 1.93604,
    "FeKb1": 1.75661,
    "CoKa": 1.79026,
    "CoKa2": 1.79285,
    "CoKa1": 1.78896,
    "CoKb1": 1.63079,
    "AgKa": 0.560885,
    "AgKa2": 0.563813,
    "AgKa1": 0.559421,
    "AgKb1": 0.497082,
}
"""

def to_str(v):
        eps = 1.0e-6
        if v == 0.0:
            return None
        if v == 1.0:
            return '+'
        if v == -1.0:
            return '-'
        if abs(v - 0.5) < eps:
            return '1/2'
        if abs(v + 0.5) < eps:
            return '-1/2'
        if abs(v - 0.25) < eps:
            return '1/4'
        if abs(v + 0.25) < eps:
            return '-1/4'
        if abs(v - 1.0/3.0) < eps:
            return '1/3'
        if abs(v + 1.0/3.0) < eps:
            return '-1/3'
        if abs(v - 2.0/3.0) < eps:
            return '2/3'
        if abs(v + 2.0/3.0) < eps:
            return '-2/3'
        if abs(v - 1.0/6.0) < eps:
            return '1/6'
        if abs(v + 1.0/6.0) < eps:
            return '-1/6'
        if abs(v - 5.0/6.0) < eps:
            return '5/6'
        if abs(v + 5.0/6.0) < eps:
            return '-5/6'
        if v > 0.0:
            return re.sub(r'^\+', '', v)

        return v
            
def vector_to_str(v):
        sx = to_str(v[0])
        sy = to_str(v[1])
        sz = to_str(v[2])
        s = ''
        if sx:
            s += sx + 'x'
        if sy:
            s += sy + 'y'
        if sz:
            s += sz + 'z'

        return re.sub(r'^\+', '', s)

def matrix_to_str(m):
        x = vector_to_str(m[0])
        y = vector_to_str(m[1])
        z = vector_to_str(m[2])

        return x, y, z

def is_valid_vasprun(path):
    try:
        tree = ET.parse(path)
        return True
    except ET.ParseError:
        return False

class tkPymatgenVASP(tkObject):
    def __init__(self, **args):
        self.poscar = None
        self.vasprun = None
        self.structure = None

    def __del__(self):
        pass

    def read_vasprun(self, infile = 'vasprun.xml', validate = False):
        ext = get_ext(infile).lower()
        if ext == '.xml':
            if validate and not is_valid_vasprun(infile):
                return None
            self.vasprun = Vasprun(infile)
            return self.vasprun
        return None

    def read_structure(self, infile, vasprun = None):
        if vasprun is None:
            if infile == 'vasprun.xml':
                self.vasprun = Vasprun(infile)

        if self.vasprun:
            self.structure = vasprun.final_structure
            return self.vasprun, self.structure
        else:
            return False, False

        self.poscar = Poscar.from_file(infile, check_for_POTCAR = True)
        if not self.poscar:
            return False, False
        self.structure = self.poscar.structure
        return self.poscar, self.structure


#class tkSpaceGroupObject(tkCrystalObject):
class tkPymatgen(tkObject):
    def __init__(self, **args):
        self.structure = None
        self.space_group_analyzer  = None
        self.symmetrized_structure = None
        self.iSPG = None
        self.SPG_name = None

    def __del__(self):
        pass

#https://github.com/materialsproject/pymatgen/blob/v2023.10.4/pymatgen/symmetry/structure.py#L18-L151
    def __str__(self, structure = None) -> str:
        if structure is None:
            structure = self.structure

        if structure is None:
            return self.ClassPath()

        outs = [
            "SymmetrizedStructure",
            f"Full Formula ({structure.composition.formula})",
            f"Reduced Formula: {structure.composition.reduced_formula}",
        ]
        if getattr(structure, 'spacegroup', None):
            outs.append(f"Spacegroup: {structure.spacegroup.int_symbol} ({structure.spacegroup.int_number})")

        def to_str(x):
            return f"{x:>10.6f}"

        outs.append(f"abc   : {' '.join(to_str(val) for val in structure.lattice.abc)}")
        outs.append(f"angles: {' '.join(to_str(val) for val in structure.lattice.angles)}")
        if structure._charge:
            outs.append(f"Overall Charge: {structure._charge:+}")

        if getattr(structure, 'equivalent_sites', None):
            outs.append(f"Sites ({len(structure.equivalent_sites)})")
            data = []
            props = structure.site_properties
            keys = sorted(props)
            for idx, sites in enumerate(structure.equivalent_sites):
                site = sites[0]
                row = [str(idx), site.species_string]
                row.extend([to_str(j) for j in site.frac_coords])
                row.append(structure.wyckoff_symbols[idx])
                for k in keys:
                    row.append(props[k][idx])
                data.append(row)

        outs.append(tabulate(data, headers=["#", "SP", "a", "b", "c", "Wyckoff", *keys]))

        return "\n".join(outs)


    def print_variables(self, obj):
        variables = self.get_member_variables(obj)
        print("variables:")
        for m in variables:
            print(f"  variable: {m}")

    def print_methods(self, obj):
        members = self.get_member_functions(obj)[0]
        print("methods:")
        for m in members:
            print(f"  method: {m}")

    def read(self, path, primitive = False):
        self.path = path
        self.structure = Structure.from_file(path, primitive = primitive)
        return self.structure
    
    def get_structure(self, structure = None):
        if structure is None:
            structure = self.get("symmetrized_structure", None)
            if structure is None:
                structure = self.get("structure", None)
        return structure

    def to(self, path, structure = None, symprec = 1.0e-3, print_level = 0):
        structure = self.get_structure(structure)
        if structure is None:
            print("\nError in tkpymatgen.to(): Can not get structure.")
            return None

        if print_level:
            print("")
            print(f"Save the symmetrized structure to [{path}]")

# This does not save the symmetrized structure. So implement with tkCrytal
#    self.struct.to(filename = convCIFfile, fmt = "cif")
        writer = CifWriter(self.symmetrized_structure, symprec = symprec)
        writer.write_file(path)
        return True


    def get_structure_inf(self):
        s = self.structure
        lattice_inf = s.lattice
        a, b, c, alpha, beta, gamma = lattice_inf.parameters
        volume = lattice_inf.volume
        aij    = lattice_inf.matrix

        inf = {
            "dict": s.as_dict,
            "lattice_parameters": [*s.lattice.abc, *s.lattice.angles],
#            "lattice_vectors": s.lattice,
            "lattice_vectors": s.lattice.matrix,
            "gij"         : s.distance_matrix,
            "total_charge": s.charge,
            "composition" : s.composition,
            "formula"     : s.formula,
            "volume"      : s.volume,
            "density"     : s.density,
            "properties"  : s.properties,
            "cart_coord"  : s.cart_coords,
            "frac_coord"  : s.frac_coords,
#print("atomic num:", s.atomic_numbers)
            }
        return inf
#print("site_properties:", site_properties.sites)
#print("species:", symmetrized_structure.species)
#print("elements:", symmetrized_structure.elements)
#print("types_of_species:", symmetrized_structure.types_of_species)
#print("types_of_specie:", symmetrized_structure.types_of_specie)

    def structure2inf(self, structure = None):
        structure = self.get_structure(structure)
        inf = {
            "latt": [latt.a, latt.b, latt.c, latt.alpha, latt.beta, latt.gamma],
            }
        inf["sites"] = []
        for site in structure:
            element = site.specie
            xyz  = site.frac_coords
            xyzc = site.coords
            inf["sites"].append({"atom_name": element, "xyz": xyz, "xyzc": xyzc})

    def structure2inf_flat(self, structure = None, symprec = 1.0e-4, coordinate = "fc", site_name = "as"):
        structure = self.get_structure(structure)

        latt = structure.lattice
        iSPG, SPG_name, symmetrized_structure = self.symmetrize(structure, symprec)

        formula, reduced_formula = self.chemical_formula(structure)
        natoms               = len(structure)
        composition          = Composition(reduced_formula)
        formula_weight       = pfloat(composition.weight)
        num_atoms_in_formula = composition.num_atoms
        Z                    = int(natoms / num_atoms_in_formula + 1.0e-5)

        inf = {
            "spg_name"       : SPG_name,
            "ispg"           : iSPG,
            "a"              : latt.a,
            "b"              : latt.b,
            "c"              : latt.c,
            "alpha"          : latt.alpha,
            "beta"           : latt.beta,
            "gamma"          : latt.gamma,
            "formula"        : formula,
            "formula_weight" : formula_weight,
            "formula_unit"   : reduced_formula,
            "natoms"         : natoms,
            "natoms_in_formula": num_atoms_in_formula,
            "Z"                : Z,
            }

        atomic_sites = structure.sites
        iatom_dict = {}
        for idx, site in enumerate(atomic_sites):
            name = site.specie.symbol
            xyz  = site.frac_coords
            xyzc = site.coords
        
            if name in iatom_dict.keys():
                iatom_dict[name] += 1
            else:
                iatom_dict[name] = 1

            iatom = iatom_dict[name]
            label = f"{name}_{iatom}"
            inf[f"atom_name{idx+1}"] = name
            if 'f' in coordinate and 's' in site_name:
                inf[f"xf_site_{idx+1}"] = xyz[0]
                inf[f"yf_site_{idx+1}"] = xyz[1]
                inf[f"zf_site_{idx+1}"] = xyz[2]
            if 'c' in coordinate and 's' in site_name:
                inf[f"xc_site_{idx+1}"] = xyzc[0]
                inf[f"yc_site_{idx+1}"] = xyzc[1]
                inf[f"zc_site_{idx+1}"] = xyzc[2]
            if 'f' in coordinate and 'a' in site_name:
                inf[f"xf_{label}"] = xyz[0]
                inf[f"yf_{label}"] = xyz[1]
                inf[f"zf_{label}"] = xyz[2]
            if 'c' in coordinate and 'a' in site_name:
                inf[f"xc_{label}"] = xyzc[0]
                inf[f"yc_{label}"] = xyzc[1]
                inf[f"zc_{label}"] = xyzc[2]

        return inf

    def chemical_formula(self, structure = None):
        structure = self.get_structure(structure)
        if structure is None:
            return None, None

        formula = structure.formula
        reduced_formula = structure.composition.reduced_formula
        return formula, reduced_formula

    def get_site_inf(self, site):
        inf = {
            "composition": site.species_string,
            "species": site.species,
#            "specie": site.specie,
            "label": site.label,
            "x": site.a,
            "y": site.b,
            "z": site.c,
            }
        return inf

    def get_sites(self, structure):
        sites = structure.sites
        return sites

    def get_asymmetric_sites(self, structure):
        sites = self.get_sites(structure)
        asym_sites = []
        for s in sites:
            ls = structure.find_equivalent_sites(s)
            if s is ls[0]:
                asym_sites.append(ls[0])
        return asym_sites
    
    def get_site_occ_inf(self, site):
        inf = {
            "site_name": site.__str__().replace(' ', ''),
            "alphabetical_formula": site.alphabetical_formula,
            "formula": site.formula,
            "reduced_formula": site.reduced_formula,
            "weight": site.weight,
#            "charge": site.charge,
            "element_composition": site.element_composition,
            "elements": site.elements,
            "fractional_composition": site.fractional_composition,
            }
        return inf

    def get_species_and_occu(self, structure):
        species_occ = structure.species_and_occu
        return species_occ
        
    def get_composition_inf(self, composition):
        d = composition.to_data_dict
        comp = d['unit_cell_composition']
        inf = {
            "dict": d,
            "site_name": composition.__str__().replace(' ', ''),
            "composition": comp,
            "unit_cell_composition": comp,
            "natoms": composition.num_atoms,
            "nspecies": len(comp),
            "weight": composition.weight,
            }
        return inf
        '''
    comp = Composition(formula)
    print(f"{comp=}")
    print(f"{comp.as_dict()=}")
    print(f"{comp.get_el_amt_dict()=}")
    print(comp.get_integer_formula_and_factor())
    print(comp.reduced_formula)
    print(comp.reduced_composition)
    print(comp.weight)
    print(comp.num_atoms)
    print(comp.elements)
    print(comp.formula)
    print(comp.alphabetical_formula)
    print(comp.num_atoms)
    print(comp.chemical_system)
    print(comp.fractional_composition)

        '''
    
    def get_ionic_radius(self, atom_name, charge = None, ncoordination = None):
        element = Element(atom_name)
        if charge is None: charge = element.common_oxidation_states[0]

        ionic_radii = element.ionic_radii
        if ncoordination is None:
            r = ionic_radii[charge]
        else:
            r = ionic_radii.get((charge, ncoordination), None)
        if r is None and charge in ionic_radii:
            r = ionic_radii[charge]

        return r

    def get_atom_inf(self, atom_name, charge = None, ncoordination = None):
        element = Element(atom_name)
        if charge is None: charge = element.common_oxidation_states[0]
        if charge is None: charge = 0

        r = self.get_ionic_radius(atom_name, charge, ncoordination)
        electronic_structure = element.electronic_structure
        inf = {
            "name": atom_name,
            "M": element.atomic_mass,
            "Z": element.Z,
            "charge": charge,
            "r": r,
            "electronic_structure": electronic_structure,
            }

        return inf

    def elements(self, structure = None):
        formula, reduced_formula = self.chemical_composition(structure)
        if formula is None:
            return None

        composition = Composition(formula)
        c_dict = composition.as_dict()

        e_list = []
        for i, atom_name in enumerate(c_dict.keys()):
            element = Element(atom_name)
            inf = element.data
#            print(inf)
            M = inf['Atomic mass']
            Z = inf['Atomic no']
            e_list.append({"atom_name": atom_name, "M": M, "Z": Z, "n": c_dict[atom_name]})

        return e_list

    def get_atom_color(self, atom_name, Z):
#        colors = {}
        colors = {"O": "red", "S": "green", "Se": "darkyellow", "Bi": "blue"}
        if atom_name in colors: return colors[atom_name]

#        np.random.seed(Z)
#        r = np.random.rand()
#        g = np.random.rand()
#        b = np.random.rand()
        r = (Z % 9) / 8
        g = (Z % 12) / 11
        b = (Z % 15) / 14
        total = r + g + b
        if total > 0.8:
            r *= 0.8
            b *= 0.8
            g *= 0.8
        
        return (r, g, b)        

    def draw_box(self, axis, structure, origin, ax, ay, az, color = 'black', linewidth = 1.0):
        lattice = structure.lattice
        a000 = origin
        a100 = [origin[i] + ax[i]                 for i in range(3)]
        a101 = [origin[i] + ax[i] + az[i]         for i in range(3)]
        a110 = [origin[i] + ax[i] + ay[i]         for i in range(3)]
        a111 = [origin[i] + ax[i] + ay[i] + az[i] for i in range(3)]
        a010 = [origin[i] + ay[i]                 for i in range(3)]
        a011 = [origin[i] + ay[i] + az[i]         for i in range(3)]
        a001 = [origin[i] + az[i]                 for i in range(3)]
        ls = [a000, a100, a110, a010, a000, a001, a101, a111, a011, a001]
#        ls = lattice.get_cartesian_coords([[0, 0, 0], [1, 0, 0], [1, 1, 0], [0, 1, 0], [0, 0, 0],
#                                           [0, 0, 1], [1, 0, 1], [1, 1, 1], [0, 1, 1], [0, 0, 1]])
        
        for i in range(4):
            axis.plot([ls[i][0],   ls[i+1][0]], [ls[i][1],   ls[i+1][1]], [ls[i][2],   ls[i+1][2]], color = 'blue', linewidth = 0.5)
            axis.plot([ls[i+5][0], ls[i+6][0]], [ls[i+5][1], ls[i+6][1]], [ls[i+5][2], ls[i+6][2]], color = 'blue', linewidth = 0.5)
            axis.plot([ls[i][0],   ls[i+5][0]], [ls[i][1],   ls[i+5][1]], [ls[i][2],   ls[i+5][2]], color = 'blue', linewidth = 0.5)

    def calculate_all_distances(self, structure, max_distance = 2.75):
        natoms = len(structure)
        sites = structure.sites
#        print("natoms=", natoms)
        distances = []
        for i in range(natoms):
            for j in range(i + 1, natoms):
#                distance = structure.get_distance(i, j)
                pos_i = sites[i].coords
                pos_j = sites[j].coords
                distance = np.linalg.norm(np.array(pos_i) - np.array(pos_j))
                if distance <= max_distance:
                    distances.append([i, j, distance])

        return distances
    
    def draw_atoms2(self, structure, draw_range, kr = 400.0):
        xrange, yrange, zrange = draw_range
        structure_draw = Structure(lattice, [], [])
        label_dict = {}
        for site in sites:
            x, y, z = site.coords
            if x < xrange[0] or xrange[1] < x: continue
            if y < yrange[0] or yrange[1] < y: continue
            if z < zrange[0] or zrange[1] < z: continue

            atom_name = site.specie.symbol
            atom_inf = self.get_atom_inf(atom_name)
#            print("atom=", atom_name, atom_inf)
            r = kr * atom_inf["r"]
            color = self.get_atom_color(atom_name, atom_inf["Z"])
            if atom_name not in label_dict.keys():
                label_dict[atom_name] = 1
                ax.scatter(x, y, z, label = atom_name, s = r, color = color)
#                ax.text(x, y, z, atom_name, fontsize=12, ha='center')
            else:
                ax.scatter(x, y, z, s = r, color = color)

            structure_draw.append(atom_name, site.coords, coords_are_cartesian = True)

        return structure_draw, label_dict

    def draw_atoms(self, ax, structure, draw_range, kr = 400.0, atom_alpha = 0.5,
                        vector_r = 3.0, vector_length = 3.0, arrow_length_ratio = 0.3):
        lattice = structure.lattice
        structure_draw = Structure(lattice, [], [])

        eps = 0.1
        xrange, yrange, zrange = draw_range
        sites = structure.sites
        draw_inf = {"atoms": {}}
        arrows_data_list = []
        i = 0
        for isite, site in enumerate(sites):
            x, y, z = site.coords
            if x < xrange[0] - eps or xrange[1] + eps < x: continue
            if y < yrange[0] - eps or yrange[1] + eps < y: continue
            if z < zrange[0] - eps or zrange[1] + eps < z: continue

            atom_name = site.specie.symbol
            atom_inf = self.get_atom_inf(atom_name)
            r = kr * atom_inf["r"]
            color = self.get_atom_color(atom_name, atom_inf["Z"])

            structure_draw.append(atom_name, site.coords, coords_are_cartesian = True)
            structure_draw.sites[i].properties = site.properties.copy()
            i += 1

            v = site.properties.get("velocity", None)
            if v:
                vx, vy, vz = v
            else:
                vx, vy, vz = None, None, None

            if atom_name not in draw_inf["atoms"].keys():
                draw_inf["atoms"][atom_name] = \
                        {"isite": [], "r": r, "color": color, "x": [], "y": [], "z": [], 
                         "vx": [], "vy": [], "vz": []}
            draw_inf["atoms"][atom_name]["isite"].append(isite)
            draw_inf["atoms"][atom_name]["x"].append(x)
            draw_inf["atoms"][atom_name]["y"].append(y)
            draw_inf["atoms"][atom_name]["z"].append(z)
            draw_inf["atoms"][atom_name]["vx"].append(vx)
            draw_inf["atoms"][atom_name]["vy"].append(vy)
            draw_inf["atoms"][atom_name]["vz"].append(vz)

        l0_max = 0.0
        for atom_name, d in draw_inf["atoms"].items():
            vx, vy, vz = d["vx"], d["vy"], d["vz"]
            for i in range(len(vx)):
                l0 = np.linalg.norm([vx[i], vy[i], vz[i]])
                if l0_max < l0: l0_max = l0
        
        kv = vector_length / l0_max
#        print("vector_length=", vector_length, l0_max, kv)

        plot_data_list = []
        for atom_name, d in draw_inf["atoms"].items():
            x, y, z    = d["x"], d["y"], d["z"]
            vx, vy, vz = d["vx"], d["vy"], d["vz"]
            color = d["color"]

            scatter = ax.scatter(x, y, z, label = atom_name, marker = 'o', s = d["r"], 
                            alpha = atom_alpha, facecolor = color, edgecolor = "w") #color = color)
            plot_data_list.append(scatter)

            for i in range(len(x)):
                l0 = np.linalg.norm([vx[i], vy[i], vz[i]])
                length = l0 * kv

#                print(f" {atom_name} at ({x[i]:.4f}, {y[i]:.4f}, {z[i]:.4f})  dx=({vx[i]:10.4g}, {vy[i]:10.4g}, {vz[i]:10.4g}) |dx|={l0:10.4g} l={length:.4f}")
#                print("xyz=", i, x[i], y[i], z[i], vx[i], vy[i], vz[i], length)
                quiver = ax.quiver(x, y, z, vx[i], vy[i], vz[i], length = length, 
                            linewidth = vector_r, arrow_length_ratio = arrow_length_ratio, 
                            color = color, normalize = True)
                arrows_data_list.append(quiver)

        draw_inf["plot_data_list"] = plot_data_list
        draw_inf["arrows_data_list"] = arrows_data_list
        draw_inf["l0_max"] = l0_max
        draw_inf["kv"] = kv

        return structure_draw, draw_inf

    def configure_axis_structure(self, ax, xrange, yrange, zrange, fontsize = 12, legend_fontsize = 12):
        ax.tick_params(axis = 'both', which = 'major', labelsize = fontsize)
        ax.set_xlabel(f'$x$', fontsize = fontsize, labelpad = -5)
        ax.set_ylabel(f'$y$', fontsize = fontsize, labelpad = -5)
        ax.set_zlabel(f'$z$', fontsize = fontsize, labelpad = 0)
#        ax.set_xlabel(f'$a=${lattice.a:.4f} $\AA$', fontsize = fontsize, labelpad = -5)
#        ax.set_ylabel(f'$b=${lattice.b:.4f} $\AA$', fontsize = fontsize, labelpad = -5)
#        ax.set_zlabel(f'$c=${lattice.c:.4f} $\AA$', fontsize = fontsize, labelpad = 0)

        ax.grid(False)
        ax.xaxis.pane.fill = False
        ax.yaxis.pane.fill = False
        ax.zaxis.pane.fill = False
        ax.xaxis.pane.set_edgecolor('w')
        ax.yaxis.pane.set_edgecolor('w')
        ax.zaxis.pane.set_edgecolor('w')
        ax.xaxis.pane.set_alpha(0)
        ax.yaxis.pane.set_alpha(0)
        ax.zaxis.pane.set_alpha(0)
#        ax.w_xaxis.line.set_color('none')
#        ax.w_yaxis.line.set_color('none')
#        ax.w_zaxis.line.set_color('none')
        ax.set_xticks([])
        ax.set_yticks([])
        ax.set_zticks([])
        ax.set_xticklabels([]) 
        ax.set_yticklabels([]) 
        ax.set_zticklabels([]) 

        ax.set_xlim(xrange)
        ax.set_ylim(yrange)
        ax.set_zlim(zrange)
        ax.set_box_aspect([xrange[1] - xrange[0], yrange[1] - yrange[0], zrange[1] - zrange[0]]) 

    def get_xyz_range(self, structure, draw_range):
        if draw_range[0]:
            xrange = [draw_range[0][0] * structure.lattice.a, draw_range[0][1] * structure.lattice.a]
        else:
            xrange = [0, structure.lattice.a]
        if draw_range[1]:
            yrange = [draw_range[1][0] * structure.lattice.b, draw_range[1][1] * structure.lattice.b]
        else:
            yrange = [0, structure.lattice.b]
        if draw_range[2]:
            zrange = [draw_range[2][0] * structure.lattice.c, draw_range[2][1] * structure.lattice.c]
        else:
            zrange = [0, structure.lattice.c]

        return xrange, yrange, zrange

    def draw_bonds(self, ax, structure, distances, bond_r, color = 'gray'):
        sites = structure.sites
        bonds_plot_list = []
        for dis in distances:
            i, j, d = dis
            site_i = sites[i]
            site_j = sites[j]
            pos_i = site_i.coords
            pos_j = site_j.coords
#            print("i,j=", i, j, pos_i, pos_j)
            data, = ax.plot([pos_i[0], pos_j[0]], [pos_i[1], pos_j[1]], [pos_i[2], pos_j[2]], 
                        linestyle = '-', color = color, linewidth = bond_r)
            bonds_plot_list.append(data)
        
        return bonds_plot_list

    def draw_structure(self, tkplt, ax, structure, draw_range = [[0.0, 1.0], [0.0, 1.0], [0.0, 1.0]], 
                    max_distance = 2.75, kr = 200.0, atom_alpha = 0.5, bond_r = 5.0,
                    vector_r = 3.0, vector_length = 3.0, arrow_length_ratio = 0.3,
                    displacement = 1.0, sleep = 0.01, time_step = 0.1, nstep = 100,
                    animation = False, stop_callback = None,
                    fontsize = 16, legend_fontsize = 16):
        if type(tkplt).__module__ == 'matplotlib.pyplot':
            tkplt = None
        else:
            plt = tkplt.plt
    
        lattice = structure.lattice
        aij = structure.lattice.matrix

        xrange, yrange, zrange = self.get_xyz_range(structure, draw_range)
        self.configure_axis_structure(ax, xrange, yrange, zrange)

        ax.plot([], [], label = f'$a=${lattice.a:.4f} $\AA$')
        ax.plot([], [], label = f'$b=${lattice.b:.4f} $\AA$')
        ax.plot([], [], label = f'$c=${lattice.c:.4f} $\AA$')

        xr0 = draw_range[0][0]
        yr0 = draw_range[1][0]
        zr0 = draw_range[2][0]
        self.draw_box(ax, structure, [xr0, yr0, zr0], 
                        [aij[0][0], aij[0][1], aij[0][2]], [aij[1][0], aij[1][1], aij[1][2]], [aij[2][0], aij[2][1], aij[2][2]],
                        color = 'blue', linewidth = 0.5)

        print()
        print("Draw structure 0")
        structure_draw, draw_inf = self.draw_atoms(ax, structure, draw_range = [xrange, yrange, zrange], 
                    kr = kr, atom_alpha = atom_alpha,
                    vector_r = vector_r, vector_length = vector_length, arrow_length_ratio = arrow_length_ratio,
                    )

        distances = self.calculate_all_distances(structure_draw, max_distance = max_distance)
        bonds_plot_list = self.draw_bonds(ax, structure_draw, distances, bond_r, color = 'gray')

        legend = ax.legend(fontsize = legend_fontsize)
#        legend.set_draggable(True)

        if tkplt:
            tkplt.draw()
        else:
            plt.pause(1.0e-5)

# animation
        if not animation: return

        v = structure.sites[0].properties.get("velocity", None)
        if v is None: return

        nsites = len(structure.sites)
        kv     = draw_inf["kv"]
        l0_max = draw_inf["l0_max"]

        plot_data_list   = draw_inf["plot_data_list"]
        arrows_data_list = draw_inf["arrows_data_list"]

        print("Animation...")
        for istep in tqdm(range(1, nstep)):
#            print(f"Draw structure {istep}")
            for quiver in arrows_data_list: quiver.remove()

            phi = 2.0 * pi * time_step * istep
            cosphi = cos(phi)
            sinphi = sin(phi)

            arrows_data_list = []
            positions = [None] * nsites
            for iatom, atom_name in enumerate(draw_inf["atoms"].keys()):
                d = draw_inf["atoms"][atom_name]
                isite_list = d["isite"]
                color      = d["color"]
                x_list  = d["x"]
                y_list  = d["y"]
                z_list  = d["z"]
                vx_list = d["vx"]
                vy_list = d["vy"]
                vz_list = d["vz"]

                x_new = []
                y_new = []
                z_new = []
                for i in range(len(x_list)):
                    isite = isite_list[i]
                    x  = x_list[i]
                    y  = y_list[i]
                    z  = z_list[i]
                    vx = vx_list[i]
                    vy = vy_list[i]
                    vz = vz_list[i]

                    x += kv * vx * displacement * sinphi
                    y += kv * vy * displacement * sinphi
                    z += kv * vz * displacement * sinphi
                    x_new.append(x)
                    y_new.append(y)
                    z_new.append(z)
#                    print("isite=", isite, nsites)
                    positions[isite] = [x, y, z]
                    

                    l0 = np.linalg.norm([vx, vy, vz])
                    length = l0 * kv * cosphi
                    quiver = ax.quiver(x, y, z, vx, vy, vz, length = length, 
                            linewidth = vector_r, arrow_length_ratio = arrow_length_ratio, 
                            color = color, normalize = True)
                    arrows_data_list.append(quiver)

                scatter = plot_data_list[iatom]
                scatter._offsets3d = (x_new, y_new, z_new)

            for ibond, dis in enumerate(distances):
                bonds_plot_list[ibond].remove()

                i, j, d = dis
                pos_i = positions[i]
                pos_j = positions[j]
                data, = ax.plot([pos_i[0], pos_j[0]], [pos_i[1], pos_j[1]], [pos_i[2], pos_j[2]], 
                            linestyle = '-', color = 'gray', linewidth = bond_r)

                bonds_plot_list[ibond] = data

            if tkplt:
                tkplt.draw()
            else:
                plt.pause(1.0e-5)

            if stop_callback and stop_callback():
                print()
                print("Animation temrinated by user")
                break

            time.sleep(sleep)    

    def symmetrize(self, structure, prec = 1.0e-3):
        self.space_group_analyzer  = SpacegroupAnalyzer(structure, symprec = prec)
        self.iSPG     = self.space_group_analyzer.get_space_group_number()
        self.SPG_name = self.space_group_analyzer.get_space_group_symbol()
        self.symmetrized_structure = self.space_group_analyzer.get_symmetrized_structure()
        return self.iSPG, self.SPG_name, self.symmetrized_structure

    def get_symmetrized_structure(self, symprec = 1.0e-3):
        self.symmetry_analyzer = SpacegroupAnalyzer(self.structure, symprec = symprec)
        self.symmetrized_structure = self.symmetry_analyzer.get_symmetrized_structure()
        return self.symmetrized_structure

    def get_spacegroup(self):
#        iSPG = space_group_analyzer.get_space_group_number()
#        SPG_name = space_group_analyzer.get_space_group_symbol()
        spg_name, spg_num = self.structure.get_space_group_info()
        return spg_name, spg_num

    def get_symmetry_inf(self, symprec = 1.0e-3):
        self.symmetry_analyzer = SpacegroupAnalyzer(self.structure, symprec = symprec)
        sa = self.symmetry_analyzer
        inf = {
            "dataset": sa.get_symmetry_dataset(),
            "spg_name": sa.get_space_group_symbol(),
            "spg_num": sa.get_space_group_number(),
            "is_laue": sa.is_laue(),
            "hall": sa.get_hall(),
            "sym_op"         : sa.get_symmetry_operations(cartesian = False),
            "pointgroup_name": sa.get_point_group_symbol(),
            "popintgroup_op" : sa.get_point_group_operations(),
            }
        return inf
