import os
import re
import time
from datetime import datetime
import numpy as np
import pandas as pd
import openpyxl


from tklib.tkfile import tkFile
from tklib.tkutils import pint, pfloat
from tklib.tktemplate import convert_file


def conv_input(template_path, output_path, replace_dict, print_level = 0):
    convert_file(template_path, output_path, replace_dict, print_level = print_level)

def convert_by_scale(v, scale, forward = True, eps = None):
    if scale == 'log':
        if forward:
            if eps is not None and v <= 0.0:
                return log(eps)
            return log(v)
        else:
            return exp(v)

    return v

def save_data(path, labels, data_list):
    if os.path.exists(path):
        try:
            os.remove(path)
        except:
            print(f"Warning: Could not delete [{path}]. Saving data was cancelled.")
            return False

    df = pd.DataFrame(np.array(data_list).T, columns = labels)
    df.to_excel(path, index = False, header = True) 

    if os.path.isfile(path):
        return True

    return False

def add_history_csv(historyfile, varnames, xk, fmin, save_time = False):
    if os.path.exists(historyfile):
        if not os.access(historyfile, os.W_OK):
            return False

    if not os.path.isfile(historyfile):
        print(f"Create [{historyfile}]")
        with open(historyfile, 'w') as out:
            if save_time:
                out.write(f"'-iter,'-time,min:fmin")
            else:
                out.write(f"-iter,min:fmin")
            for i in range(len(xk)):
                out.write(f",{varnames[i]}")
            out.write("\n")

    if os.path.isfile(historyfile):
        print(f"Add fmin and xk to [{historyfile}]")
        iter = -1
        with open(historyfile, 'r') as fp:
            for l in fp:
                aa = l.split(',')
                idx = pint(aa[0], 0)
                if idx > iter:
                    iter = idx
#                iter += 1
        iter += 1
        try:
            out = open(historyfile, 'a')
        except Exception as e:
            print(f"Warning: Could not write to [{historyfile}] due to error [{e}]. Ignore")
            return iter

        if save_time:
            stime = time.time()
            strtime = datetime.fromtimestamp(stime).strftime('%H:%M:%S %Y/%m/%d')
            out.write(f"{iter},{strtime},{fmin}")
        else:
            out.write(f"{iter},{fmin}")
        for x in xk:
            out.write(f",{x}")
        out.write("\n")
        out.close()

    return iter

def add_history_xlsx(historyfile, varnames, xk, fmin, wb = None, ws = None, save_time = False):
    if historyfile is not None:
        if not os.path.isfile(historyfile):
            wb = openpyxl.Workbook()
        else:
            wb = openpyxl.load_workbook(historyfile)

        ws = wb.active

    max_row = ws.max_row
    max_col = ws.max_column
#    print("max_row=", max_row)

    if max_row <= 1 and ws.cell(row = 1, column = 1).value is None:
        ws.cell(row = 1, column = 1, value = '-iter')
        if save_time:
#            print("'-time")
            ws.cell(row = 1, column = 2, value = '-time')
            ioffset = 3
        else:
            ioffset = 2
        ws.cell(row = 1, column = ioffset, value = 'min:fmin')
        ioffset += 1
        for i in range(len(xk)):
            ws.cell(row = 1, column = ioffset + i, value=varnames[i])
        iter = 1
    else:
        iter = ws.cell(row = max_row, column = 1).value
        if type(iter) is str:
            iter = 0
        else:
            iter += 1
    
    ws.cell(row = max_row + 1, column = 1, value = iter)
    if save_time:
        stime = time.time()
        strtime = datetime.fromtimestamp(stime).strftime('%H:%M:%S %Y/%m/%d')
        ws.cell(row = max_row + 1, column = 2, value = strtime)
        ioffset = 3
    else:
        ioffset = 2
    ws.cell(row = max_row + 1, column = ioffset, value = fmin)
    ioffset += 1
    for i in range(len(xk)):
        ws.cell(row = max_row + 1, column = ioffset + i, value = xk[i]) 

    if historyfile is not None:
        try:
            wb.save(historyfile)
        except Exception as e:
            print(f"Warning: Could not write to [{historyfile}] due to error [{e}]. Ignore")

    return iter

def add_history(historyfile, varnames, xk, fmin, wb = None, ws = None, save_time = False):
    if historyfile is None:
        return add_history_xlsx(historyfile, varnames, xk, fmin, wb = wb, ws = ws, save_time = save_time)
    else:
        ext = os.path.splitext(historyfile)[1]
        if ext.lower() == ".csv":
            return add_history_csv(historyfile, varnames, xk, fmin, save_time = save_time)
        elif ext.lower() == ".xlsx":
            return add_history_xlsx(historyfile, varnames, xk, fmin, wb = wb, ws = ws, save_time = save_time)
        else:
            print(f"Warning: Unsupported extension [{ext}] for history file")
            return
