import os
import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt
from scipy.optimize import minimize
import openpyxl
import pandas as pd
from matplotlib import pyplot as plt
import matplotlib.widgets as wg


from tklib.tkutils import terminate, pint, pfloat, getarg, getintarg, getfloatarg, pconv_by_type, str2val
#analyze_varstr, save_csv
from tklib.tkutils import is_exist, is_file, is_dir, val2str
from tklib.tkinifile import tkIniFile
from tklib.tkparams import tkParams
from tklib.tkvariousdata import tkVariousData
from tklib.tkgraphic.tkplotevent import tkPlotEvent

from tklib.tkvariousdata import tkVariousData
from tklib.tkfilter import tkFilter
from tklib.tkexcel import tkExcel
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me, Ea_Arrhenius, log10
from tklib.tksci.tkoptimize import mlsq_general
from tklib.tksci.tkFit_object import tkFit_object
from tklib.tksci.tkFit_lib import save_data, add_history


'''
Fitting and interactive matplotlb helper library
'''


def convert_scale(x, y_scale = 'log', eps = None):
    if y_scale == '': return x

    if x == 0.0:
        if eps is not None: return log(eps) / 2.0
        return -100.0

    if x < 0.0:
        if eps is not None: return log(eps)
        return -log(-x)

    return log(x)

def P(yerr2_sum, yerr_var, P0):
    return P0 * exp(-yerr2_sum / yerr_var / 2.0)

def inf_format_func(plotevent, sel, iline, idata, label, inf, x = 0, y = 0, opt = None, format = format):
    if inf is None:
        return

    fit = inf["fit"]
    index = inf["index"][idata]
    x     = inf["x"][idata]
    y     = inf["y"][idata]
    print(f"{label}: line#{iline} data#{idata}: index={index}")
    for label, xlist in zip(fit.x_labels, fit.x_list):
        print(f"  {label:10}: {xlist[idata]:g}")
    for label, ylist in zip(fit.y_labels, fit.y_list):
        print(f"  {label:10}: {ylist[idata]:g}")
    print(f"  {'x':10}: {x:g}")
    print(f"  {'y':10}: {y:g}")


class tkFit_mxy(tkFit_object):
    def __init__(self, parameter_file = None, app = None, 
                    method = 'nelder-mead', jac = '3-points', tol = 1.0e-5, nmaxiter = 100, nmaxcall = 1000,
                    fplot = 1, fhistory = 0, ffitfiles = 0, print_interval = 1, plot_interval = 1, **args):
        super().__init__(**args)
        
#        super(tkParams, self).__init__(parameter_file, app, args)
#        self.update(**args)

        self.app = app

# tkPlotEvent is initialized in tkFit_object.initial_plot()
#        self.plot_event = tkPlotEvent(plt)
        self.x_labels  = None
        self.x_list    = None
        self.y_labels  = None
        self.y_list    = None
        self.y_scale   = None
        self.w_list    = None
        self.func      = None
        self.iter      = 0
        self.f         = None

        self.ix_plot   = ''

        self.varname  = []
        self.unit     = []
        self.pk       = []
        self.pk_scale = []
        self.dx       = []
        self.optid    = []
        self.kmin     = []
        self.kmax     = []
        self.kpenalty = []
        
        self.__method       = method
        self.jac            = jac
        self.tol            = tol
        self.nmaxiter       = nmaxiter
        self.nmaxcall       = nmaxcall
        self.icall          = 0

        self.fmin           = 1.0e300
        self.fmax_record    = 1.0e300

        self.print_interval = print_interval
        self.plot_interval  = plot_interval
        self.fplot          = fplot
        self.fhistory       = fhistory
        self.ffitfiles      = ffitfiles
        self.stop_flag      = False

        selfhistoryfile = None
        self.fp_history = None
        self.wb_history = None
        self.ws_history = None

    @property
    def method(self):
        return self.__method
        
    @method.setter
    def method(self, method):
        if method == 'simplex':
            self.__method = 'nelder-mead'
        else:
            self.__method = method


    def open_history(self, historyfile = None):
        if historyfile is None: historyfile = self.historyfile
        else: self.historyfile = historyfile

        if not os.path.isfile(historyfile):
            self.wb = openpyxl.Workbook()
        else:
            self.wb = openpyxl.load_workbook(historyfile)

        self.ws = self.wb.active

        return self.wb, self.ws

    def close_history(self, historyfile = None, wb = None):
        if historyfile is None: historyfile = self.historyfile
        if wb is None:
            wb = self.wb

        wb.save(historyfile)

        self.fp_history = None
        self.wb_history = None
        self.ws_history = None

    def add_history(self, historyfile, varnames, xk, fmin, wb = None, ws = None, save_time = False):
        if wb is None:
            if self.wb:
                wb = self.wb
            else:
                if historyfile is None:
                    historyfile = self.historyfile

                if not os.path.isfile(historyfile):
                    wb = openpyxl.Workbook()
                else:
                    wb = openpyxl.load_workbook(historyfile)

        if wb is None: 
            print("Error in tkFit_mxy_flex.add_history(): None wb")
            exit()

        if ws is None:
            ws = wb.active

        return add_history(historyfile, varnames, xk, fmin, wb = wb, ws = ws, save_time = save_time)

    def read_data_from_filter(self, infile, module_dir, module_name, imodule = 0, iregion = 0,
                x_indexes = [], y_indexes = [], app = None, cfg = None, is_print = False):
#        print("")
#        print(f"Load module by {module_dir}/{module_name}.py")
        filter = tkFilter(app = app, cfg = cfg, plugin_dir = module_dir, module_file = f"{module_name}.py")
        self.module_names, self.modules = filter.load(target = "read_data", is_print = True)
        self.inf_input = filter.read_data(imodule, infile, app = app, cfg = cfg, is_print = is_print)

        self.x_labels = []
        self.x_list   = []
        for idx in x_indexes:
            label, data = filter.find_data_array(idx)
            self.x_labels.append(label)
            self.x_list.append(data)

        self.y_labels = []
        self.y_list  = []
        for idx in y_indexes:
            label, data = filter.find_data_array(idx)
            self.y_labels.append(label)
            self.y_list.append(data)


    def read_data(self, infile, x_labels = [0], y_label = 1, xmins = [None], xmaxs = [None], usage = None,
                module_dir = None, module_name = None, imodule = 0, iregion = 0,
                x_indexes = [], y_indexes = [], app = None, cfg = None, is_print = False):
        if module_dir is not None and module_name is not None:
            return self.read_data_from_filter(infile = infile, 
                    module_dir = module_dir, module_name = module_name, imodule = imodule, iregion = iregion,
                    x_indexes = x_indexes, y_indexes = y_indexes, app = app, cfg = cfg, is_print = is_print)

        print("")
        print(f"Read [{infile}]")
        datafile = tkVariousData(infile)
        labels, datalist = datafile.Read_minimum_matrix(close_fp = True, force_numeric = False, usage = usage)
        
        _x_labels = []
        xins      = []
        for l in x_labels:
            x_label, xin = datafile.FindDataArray(l, flag = 'i')
#            print("x l=", l, x_label)
            _x_labels.append(x_label)
            xins.append(xin)
#        xlabel, xin = datafile.FindDataArray(xlabel, flag = 'i')
        ylabel, yin = datafile.FindDataArray(ylabel, flag = 'i')

        self.datafile  = datafile
        self.labels    = labels
        self.datalist  = datalist
        self.x_labels  = _x_labels
        self.y_label   = y_label

        self.nx        = len(x_labels)
        self.ndata_all = len(xin)
#        print("ndata_all=", self.ndata_all)
        self.x_list = []
        for il in range(self.nx):
            self.x_list.append([])
        self.y = []
        self.included_index = []
        for i in range(self.ndata_all):
            is_included = True
            for il in range(len(_x_labels)):
                if xmins[il] is not None and xmins[il] > xins[il][i]:
                    is_included = False
                    continue
                if xmaxs[il] is not None and xmaxs[il] < xins[il][i]:
                    is_included = False
                    continue

                if is_included:
                    self.x_list[il].append(xins[il][i])

            if is_included:
                self.y.append(yin[i])
                self.included_index.append(i)

        self.ndata = len(self.x_list[0])
        self.index = range(self.ndata)

    def cal_ylist(self, pk, x_list = None, run = True, print_level = 1):
        if x_list is None:
            x_list = self.x_list

        ny = len(self.y_list)

#        print("205 func: len(pk)=", len(pk))
        nvars = len(x_list)
        ndata = len(x_list[0])
        y_list = []
        for i in range(ny):
            y_list.append([])

        for i in range(ndata):
# i番目のxの値セット xs を取り出す
            xs = [x_list[j][i] for j in range(nvars)]
# xs に対するyのリストを取り出す
            ret = self.func(xs, pk)
            for i, v in enumerate(ret):
                y_list[i].append(v)

        return y_list

    def cal_difference(self, ymeas, ysim, y_scale = '', eps = None):
        if y_scale == 'log':
            if eps is not None:
                if ymeas <= 0.0:
                    ymeas = eps
                if ysim <= 0.0:
                    ysim = eps

            return log(ymeas) - log(ysim)

        return ymeas - ysim

    def cal_penalty(self, xk_all):
        eps = self.app.cfg.get('eps', 1.0e-10)
        y_scale = self.app.cfg.get('y_scale', '')

        kp   = self.kpenalty
        if kp is None: return 0.0

        kmin = self.kmin
        kmax = self.kmax
        nkp = len(kp)
        p_tot = 0.0
        for i in range(nkp):
            if xk_all[i] < kmin[i]  : origin_value = kmin[i]
            elif xk_all[i] > kmax[i]: origin_value = kmax[i]
            else: continue

            d = convert_scale(xk_all[i], y_scale, eps) - convert_scale(origin_value, y_scale, eps)
            p = kp[i] * d * d

            if xk_all[i] < kmin[i]  :
                print(f"**Warning: [{fit.varname[i]}={xk_all[i]}] is smaller than [{kmin[i]}]. Add penalty [{p:10.3g}] to fmin")
#                print("  p=", y_scale, eps, "  conv=", convert_scale(xk_all[i], y_scale, eps), convert_scale(origin_value, y_scale, eps), "  d=", d, kp[i], p)
            elif xk_all[i] > kmax[i]:
                print(f"**Warning: [{fit.varname[i]}={xk_all[i]}] is larger than [{kmax[i]}]. Add penalty [{p:10.3g}] to fmin")
#                print("  p=", y_scale, eps, "  conv=", convert_scale(xk_all[i], y_scale, eps), convert_scale(origin_value, y_scale, eps), "  d=", d, kp[i], p)

            p_tot += p
            xk_all[i] = origin_value

        return p_tot

    def cal_fmin(self, xk_all, x_list, y_list, w_list, fit = None, run = True, print_level = 1):
        eps = self.app.cfg.get('eps', 1.0e-10)
        y_scale = self.app.cfg.get('y_scale', '')

        if fit is None: fit = self

#        yc_list = self.ycal_list(x_list, xk_all, run = run, print_level = print_level)
        yc_list = self.cal_ylist(xk_all, x_list, run = run, print_level = print_level)
        if yc_list is None: return None

        nx = len(x_list[0])
        nc = len(yc_list[0])
        if nc < nx: 
            print(f"\nWarning in cal_fmin(): # of data in yc_list[0] is smaller than that of x_list[0] (nx={nx})\n")
            return None

        fmin = 0.0
        yerr_list = []
        for i in range(nx):
            ymeas = fit.y_list[0][i]
            ysim  = yc_list[0][i]
            d = convert_scale(ymeas, y_scale, eps) - convert_scale(ysim, y_scale, eps)
    
            yerr_list.append(d)
            if w_list:
                fmin += w_list[0][i] * d * d
            else:
                fmin += d * d

        fit.yerr2_sum = fmin
        fmin = sqrt(fmin / nx)

        fit.yc_list = yc_list
        fit.yerr_list = yerr_list

        return fmin

    def cal_total_fmin(self, xk_all, x_list, y_list, w_list, run):
        p_tot = self.cal_penalty(xk_all)
        fmin  = self.cal_fmin(xk_all, x_list, y_list, w_list, run)
        if fmin is None: return None

        fmin  = fmin + p_tot

        return fmin

    def save_history(self, xk_all, fmin, save_func = None):
        if save_func is None: save_func = save_data

        if self.cfg.fhistory:
            print(f"Add fmin and parameters to [{self.historyfile}]")
            iter = add_history(self.historyfile, self.varname, xkr0, fmin)
            if iter is not None:
                print(f"DB #{iter}: fmin={fmin:12.4g} ", end = '')
                for v in xk:
                    print(f" {v:12.6g}", end = '')
                print()
            else:
                iter = None
                print(f"  Warning: Failed to add to {self.historyfile}]")
                print(f"           Check write permission / disk space etc")

        if self.cfg.ffitfiles and iter is not None:
            save_path = f'fit{iter:04}.xlsx'
            print(f"Save last input and calculation data  to [{save_path}]")
            if hasattr(self, 'yini_list'):
                ret = save_func(save_path, ['x', 'input', 'initial', 'final'], [self.x_list[0], self.y_list[0], self.yini_list[0], yc_list[0]])
            else:
                ret = save_func(save_path, ['x', 'input', 'final'], [self.x_list[0], self.y_list[0], yc_list[0]])
            if not ret:
                print(f"  Warning: Failed to save to [{save_path}]")
                print(f"           Check write permission / disk space etc")

    def minimize_func(self, xk, x_list = None, y_list = None, w_list = None, run = True, print_level = 1):
        if not self.hasattr('icall'): self.icall = 0
        if not self.hasattr('nmaxcall'): self.nmaxcall = -1
        if not self.hasattr('fmax_record'): self.fmax_record = 1.0e300

        if self.icall > self.nmaxcall:
            if print_level:
                print(f"\nMessage: # of minimize_func calls [{self.icall}] exceesds nmaxcall [{self.nmaxcall}].")
                print(f"           Will stop optimization\n")
            self.stop_flag = True

        if self.stop_flag: return 1.0e300

        if x_list is None: x_list = self.x_list
        if y_list is None: y_list = self.y_list
        if w_list is None: w_list = self.w_list

        if print_level: print()
        nx = len(x_list[0])
        xk_all = self.recover_parameters(xk, set_member = False)
        fmin = self.cal_total_fmin(xk_all, x_list, y_list, w_list, run)

        if fmin is None: return 1.0e10

        if print_level: print(f"\n  fmin={fmin}")

        if fmin < fit.fmax_record:
            save_history(xk_all, fmin, fit)
        else:
            print(f"Warning: fmin={fmin:g} is larger than fmax_record={fit.fmax_record:g}. Skip recording")

        self.icall += 1
        self.fmin = fmin

        return fmin

    def _minimize_func(self, pk, x_list = None, y_list = None, w_list = None):
        if x_list is None:
            x_list = self.x_list
        if y_list is None:
            y_list = self.y_list
        if w_list is None:
            w_list = self.w_list

        ny = len(y_list)
        if w_list is None:
            w_list = []
            for i in range(ny):
                w_list.append(1.0)
        if self.y_scale is None:
            self.y_scale = []
            for i in range(ny):
                self.y_scale.append('')
        if self.pk_scale is None:
            self.pk_scale = []
            for i in range(ny):
                self.pk_scale.append('')

        def convert_val(v_list, conv_list, forward = True):
            ret = []
            for v, c in zip(v_list, conv_list):
                if c == 'log':
                    if forward:
                        if v > 0.0:
                            v = log(v)
                        elif v < 0.0:
                            v = -log(v)
                        else:
                            v = -100.0
                    else:
                        v = exp(v)
                ret.append(v)
            return ret

        recoverpk = self.recover_parameters(pk, set_member = False)

        f = 0.0
        kp   = self.kpenalty
        p_tot = 0.0
        if kp is not None:
            kmin = self.kmin
            kmax = self.kmax
            nkp = len(kp)
            for i in range(nkp):
                if recoverpk[i] < kmin[i]:
                    d = recoverpk[i] - kmin[i]
                    p = kp[i] * d * d
                    p_tot += p
                    recoverpk[i] = kmin[i]
                    print(f"**Warning: [{self.varname[i]}] is smaller than [{kmin[i]}]. Add penalty [{p:10.3g}] to f")
                elif recoverpk[i] > kmax[i]:
                    d = recoverpk[i] - kmax[i]
                    p = kp[i] * d * d
                    p_tot += p
                    recoverpk[i] = kmax[i]
                    print(f"**Warning: [{self.varname[i]}] is larger than [{kmax[i]}]. Add penalty [{p:10.3g}] to f")

        nx    = len(x_list)
        ndata = len(x_list[0])
        ny    = len(y_list)
        for i in range(ndata):
            xs = [x_list[j][i] for j in range(nx)]
            ycal_list = self.func(xs, recoverpk)
            yobs_list = [y_list[j][i] for j in range(ny)]
            yobs_list = convert_val(yobs_list, self.y_scale)
            ycal_list = convert_val(ycal_list, self.y_scale)
            for j in range(ny):
                d = ycal_list[j] - yobs_list[j]
                f += w_list[j] * d * d

        f = sqrt(f / ndata) + p_tot

        return f

    def minimize(self, method = None, jac = None, tol = None, nmaxiter = None, initial_simplex = None):
        if method is not None:
            self.method = method
        if jac is None:
            jac = self.get('jac', '3-point')
        if tol is None:
            tol = self.tol
        if nmaxiter is None:
            nmaxiter = self.nmaxiter

        print("")
        print("Optimizing parameters:")
        optpk = self.extract_parameters()
        print("   method  =", self.method)
        print("   nmaxiter=", nmaxiter)
        print("   tol     =", tol)
        print("   optpk=", optpk)
        print("")
        print("Start minimization:")
        if self.method == 'nelder-mead':
            if initial_simplex and len(initial_simplex) > 0:
                options = {'maxiter': nmaxiter, "disp": True, "initial_simplex": initial_simplex}
            else:
                options = {'maxiter': nmaxiter, "disp": True}
            res = minimize(self.minimize_func, optpk, 
                    method = self.method, 
                    callback = self.callback,
                    tol = tol, options = options)
        else:
            res = minimize(self.minimize_func, optpk, 
                    method = self.method, jac = jac, 
                    callback = self.callback,
                    tol = tol, options = {'maxiter': nmaxiter, "disp": True})

        self.finalpk = self.recover_parameters(res.x)
        self.iter = res.nit
        if self.get('fun', None) is not None:
            self.ffin = res.fun
        else:
            self.ffin = self.minimize_func(self.finalpk)
            
        return self.finalpk, res.fun, res.success, res

    def callback(self, pk, run = True):
        cfg = self.cfg
        exist_stopfile = os.path.isfile(cfg.stopfile)
        fplot  = getattr(self, 'fplot', True) # and self.fplot
        xk_all = self.recover_parameters(pk, set_member = False)
        n = len(xk_all)

        if self.stop_flag:
            return False

        print("exist_stopfile:", cfg.stopfile, exist_stopfile)
        if exist_stopfile:
            print()
            print(f"Message: Found stop file [{cfg.stopfile}] for early stop.")
            print()
            self.stop_flag = True
            return False

        if fplot:
            w = self.plt.get_current_fig_manager().window
            if w != self.window:
                print()
                print(f"Message: Graph window is closed for early stop.")
                print()
                self.stop_flag = True
                return False

        if self.iter % self.print_interval == 0:
            print(f"iter: {self.iter}")
            for i in range(n):
                print(f"  {self.varname[i]:10}: {xk_all[i]:10.4g} {self.unit[i]}")
            f = self.minimize_func(pk)
            print(f"    f={f:12.6g}")

            if self.get('fmin_list', None) is not None:
                self.iter_list.append(self.iter + 1)
                self.fmin_list.append(f)

        if fplot and self.iter % self.plot_interval == 0:
            ycal_list = self.cal_ylist(xk_all, run = run)

            x_list   = self.__dict__.get('x_list', None)
            x_labels = self.__dict__.get('x_labels', None)
            nx = len(x_list)
            if nx == 1 or type(self.ix_plot) is int:
                if type(self.ix_plot) is int and self.ix_plot < nx:
                    ix_plot = self.ix_plot
                else:
                    ix_plot = 0
                for i in range(len(self.y_list)):
                    axis = self.data_axes[i]
                    self.fit_data_list[i][0].set_data(x_list[ix_plot], ycal_list[i])
            else:
                for i in range(len(self.y_list)):
                    axis = self.data_axes[i]
                    self.fit_data_list[i][0].set_data(range(len(ycal_list[i])), ycal_list[i])

            if self.get('fmin_list', None) is not None:
                self.error_data[0].set_data(self.iter_list, self.fmin_list)
                self.error_axis.set_xlim([min(self.iter_list), max(self.iter_list)])
                self.error_axis.set_ylim([min(self.fmin_list), max(self.fmin_list)])

#            self.plt.tight_layout()
            self.plt.subplots_adjust(top = self.plot_region[0], bottom = self.plot_region[1])
            self.plt.pause(0.0001)

        self.iter += 1
    
        self.retrieve_parameter_list(xk_all, self.varname, target = cfg)
        self.save_parameters(cfg.parameterfile, cfg)

        return True
    
    def _callback(self, pk):
        if self.stop_flag:
            return False
    
        fplot = hasattr(self, 'fplot') and self.fplot
        
        w = self.plt.get_current_fig_manager().window
        if fplot and hasattr(self, 'window') and w != self.window:
            self.stop_flag = True
            return False

        recoverpk = self.recover_parameters(pk, set_member = False)

        if self.iter % self.print_interval == 0:
            print(f"iter: {self.iter}")
            n = len(recoverpk)
            for i in range(n):
                print(f"  {self.varname[i]:10}: {recoverpk[i]:10.4g} {self.unit[i]}")
            f = self.minimize_func(pk)
            print(f"    f={f:12.6g}")

            if self.get('fmin_list', None) is not None:
                self.iter_list.append(self.iter + 1)
                self.fmin_list.append(f)

        if fplot and self.iter % self.plot_interval == 0:
            ycal_list = self.cal_ylist(recoverpk)

            x_list   = self.__dict__.get('x_list', None)
            x_labels = self.__dict__.get('x_labels', None)
            nx = len(x_list)
            if nx == 1 or type(self.ix_plot) is int:
                if type(self.ix_plot) is int and self.ix_plot < nx:
                    ix_plot = self.ix_plot
                else:
                    ix_plot = 0
                for i in range(len(self.y_list)):
                    axis = self.data_axes[i]
                    self.fit_data_list[i][0].set_data(x_list[ix_plot], ycal_list[i])
            else:
                for i in range(len(self.y_list)):
                    axis = self.data_axes[i]
                    self.fit_data_list[i][0].set_data(range(len(ycal_list[i])), ycal_list[i])
#                    self.fit_data_list[i][0].set_data(self.index, ycal_list[i])

            if self.get('fmin_list', None) is not None:
                self.error_data[0].set_data(self.iter_list, self.fmin_list)
                self.error_axis.set_xlim([min(self.iter_list), max(self.iter_list)])
                self.error_axis.set_ylim([min(self.fmin_list), max(self.fmin_list)])

#            plt.tight_layout()
            self.plt.subplots_adjust(top = self.plot_region[0], bottom = self.plot_region[1])
            plt.pause(0.0001)

        self.iter += 1

        return True

    def button_click(self, e):
        self.stop_flag = True

    def initial_plot(self, data_axes, error_axis = None, 
                        label_input = None,
                        yini_list = None, label_ini = None,
                        label_fit = None,
                        ix_plot = '', x_scale = '',
                        label_error = None, fmin = None,
                        fontsize = 16, legend_fontsize = None,
                        plt = None, fig = None, use_pause = 0.00001, 
                        savefig_path = None,
                        button_region = [0.15, 0.95, 0.10, 0.03], plot_region = [0.92, 0.15]):
#                        button_region = [0.15, 0.95, 0.10, 0.03], plot_region = [0.90, 0.10]):
        has_fplot = hasattr(self, 'fplot')
        if has_fplot and not self.fplot:
            return
        
        if label_input is None:
            label_input = 'input'
        if label_ini is None:
            label_ini = 'initial'
        if label_fit is None:
            label_fit = 'fit'
        if label_error is None:
            label_error = 'error'
        if legend_fontsize is None:
            legend_fontsize = fontsize

        self.error_axis = error_axis

        self.plot_event = tkPlotEvent(plt)
        self.plot_event.prepare_annotation()
        self.plt = plt
        self.fig = fig
        self.ix_plot = ix_plot
        self.x_scale = x_scale
        self.use_pause     = use_pause
        self.button_region = button_region
        self.plot_region   = plot_region

        self.yini_list = yini_list
        self.error_axis = error_axis
        self.data_axes = data_axes
        for axis in data_axes:
            axis.tick_params(labelsize = fontsize)

        nx    = len(self.x_list)
        index = range(len(self.x_list[0]))
        if nx == 1 or type(ix_plot) is int:
            if type(ix_plot) is int and ix_plot < nx:
                pass
            else:
                ix_plot = 0
            self.fit_data_list = []
            for i in range(len(self.y_list)):
                axis = data_axes[i]
#                self.input_data   = axis.plot(self.x_list[ix_plot], self.y_list[i], label = self.y_labels[i], linestyle = '',  marker = 'o', markersize = 5.0)
#                self.initial_data = axis.plot(self.x_list[ix_plot], yini_list[i],   label = label_ini,       linestyle = '-', linewidth = 0.5, color = 'red')
                line_input,   = axis.plot(self.x_list[ix_plot], self.y_list[i], label = self.y_labels[i], linestyle = '',  marker = 'o', markersize = 5.0)
                line_initial, = axis.plot(self.x_list[ix_plot], yini_list[i],   label = label_ini,       linestyle = '-', linewidth = 0.5, color = 'red')

                format = "{label}: line#{iline} data#{idata}: x={x:g} y={y:g}"
                self.plot_event.annotation.add_line(f"input y#{i}", axis, axis, self.x_list[ix_plot], self.y_list[i], line_input, 
                        inf_list = {"x": self.x_list[ix_plot], "y": self.y_list[i]},
                        annotation_format = format, inf_format = format)
                self.plot_event.annotation.add_line(f"initial y#{i}", axis, axis, self.x_list[ix_plot], yini_list[i], line_initial, 
                        inf_list = {"x": self.x_list[ix_plot], "y": self.y_list[i]},
                        annotation_format = format, inf_format = format)

                fit_data = axis.plot([], [], label = label_fit, linestyle = '-', linewidth = 0.5, color = 'blue')
                self.fit_data_list.append(fit_data)

            axis.set_xlabel(self.x_labels[ix_plot], fontsize = fontsize)
            axis.set_ylabel(self.y_labels[0], fontsize = fontsize)
            if self.x_scale == 'log':
                axis.set_xscale('log')
            if self.y_scale == 'log':
                axis.set_yscale('log')

            axis.legend(fontsize = legend_fontsize)

#            self.plot_event_input_data   = self.plot_event.add_data({"label": label_input, "plot_type": "2D", "axis": axis, 
#                                        "data": self.input_data})
#            self.plot_event_initial_data = self.plot_event.add_data({"label": label_ini,   "plot_type": "2D", "axis": axis, 
#                                        "data": self.initial_data})
#            self.plot_event_fit_data     = self.plot_event.add_data({"label": label_fit,   "plot_type": "2D", "axis": axis, 
#                                        "data": fit_data})
        else:
            self.fit_data_list = []
            for i in range(len(self.y_list)):
                axis = data_axes[i]
#                self.input_data   = axis.plot(index, self.y_list[i], label = self.y_labels[i], linestyle = '',  marker = 'o', markersize = 5.0)
#                self.initial_data = axis.plot(index, yini_list[i],   label = label_ini,   linestyle = '-', linewidth = 0.5, color = 'red')
                line_input,   = axis.plot(index, self.y_list[i], label = self.y_labels[i], linestyle = '',  marker = 'o', markersize = 5.0)
                line_initial, = axis.plot(index, yini_list[i],   label = label_ini,   linestyle = '-', linewidth = 0.5, color = 'red')

                format = "{label}: line#{iline} data#{idata}: index={index} x={x:g} y={y:g}"
                self.plot_event.annotation.add_line(f"input", axis, axis, index, self.y_list[i], line_input, 
                        inf_list = {"index": index, "x": self.x_list[i], "y": self.y_list[i], "fit": self},
                        annotation_format = format, inf_format = inf_format_func)
                self.plot_event.annotation.add_line(f"initial", axis, axis, index, yini_list[i], line_initial, 
                        inf_list = {"index": index, "x": self.x_list[i], "y": yini_list[i], "fit": self},
                        annotation_format = format,
                        inf_format = inf_format_func)

                fit_data     = axis.plot([], [], label = label_fit, linestyle = '-', linewidth = 0.5, color = 'blue')
                self.fit_data_list.append(fit_data)
       
            axis.set_xlabel('index', fontsize = fontsize)
            axis.set_ylabel(self.y_labels[0], fontsize = fontsize)
            if self.y_scale == 'log':
                axis.set_yscale('log')
            axis.legend(fontsize = legend_fontsize)
#            self.plot_event_input_data   = self.plot_event.add_data({"label": label_input, "plot_type": "2D", "axis": axis, 
#                                        "data": self.input_data,   "x_list": self.x_list, "x_labels": self.x_labels})
#            self.plot_event_initial_data = self.plot_event.add_data({"label": label_ini,   "plot_type": "2D", "axis": axis, 
#                                        "data": self.initial_data, "x_list": self.x_list, "x_labels": self.x_labels})
#            self.plot_event_fit_data     = self.plot_event.add_data({"label": label_fit,   "plot_type": "2D", "axis": axis, 
#                                        "data": fit_data,     "x_list": self.x_list, "x_labels": self.x_labels})

        self.view_xlim_data_list = []
        self.view_ylim_data_list = []
        for i in range(len(self.y_list)):
            axis = data_axes[i]
            self.view_xlim_data_list.append(axis.get_xlim())
            self.view_ylim_data_list.append(axis.get_ylim())

#        if self.fig is None:
#            print("")
#            print("Warning: figure object is not given.")
#            print("    Mouse click event is not bound")
#            print("")
#        else:
#            self.plot_event.register_event(self.fig, event = "button_press_event", 
#                        callback = lambda event: self.plot_event.onclick(event))

        if error_axis:
            self.error_axis = error_axis
            error_axis.tick_params(labelsize = fontsize)
            if fmin is None:
                fmin = self.minimize_func(self.pk)
            self.iter_list = [0]
            self.fmin_list = [fmin]

            self.error_data = error_axis.plot(self.iter_list, self.fmin_list, label = label_error, 
                        linestyle = '-', linewidth = 0.5, color = 'black',
                        marker = 'o', markersize = 5.0, markerfacecolor = 'black', markeredgecolor = 'black')
            error_axis.set_xlabel('iteration', fontsize = fontsize)
            error_axis.set_ylabel('error', fontsize = fontsize)
            error_axis.set_yscale('log')
            error_axis.legend(fontsize = legend_fontsize)

#            self.plot_event_error_data = self.plot_event.add_data(
#                        {"label": label_error, "plot_type": "2D", "axis": self.error_axis, "data": self.error_data})

#        self.plt.tight_layout()
        self.plt.subplots_adjust(top = plot_region[0], bottom = plot_region[1])

        self.button_region = button_region
        self.ax_button = plt.axes(button_region)
        self.stop_button = wg.Button(self.ax_button, 'stop', color = '#f8e58c', hovercolor = '#38b48b')
        self.stop_button.on_clicked(self.button_click)

        if savefig_path is not None and savefig_path != '':
            plt.savefig(savefig_path)

        if use_pause is not None:
            self.plt.pause(use_pause)
        else:
            self.plt.show()
        self.window   = plt.get_current_fig_manager().window

    def finalize_plot(self, yfin_list, iter = None, fmin = None, 
                savefig_path = 'final.png', use_pause = 0.00001, button_region = None, plot_region = None):
        fplot = hasattr(self, 'fplot') and self.fplot
        if not fplot:
            return

        if button_region is None:
            button_region = self.button_region
        if plot_region is None:
            plot_region = self.plot_region

        x_list   = self.__dict__.get('x_list', None)
        x_labels = self.__dict__.get('x_labels', None)
        nx = len(x_list)
        if nx == 1 or type(self.ix_plot) is int:
            if type(self.ix_plot) is int and self.ix_plot < nx:
                ix_plot = self.ix_plot
            else:
                ix_plot = 0
            for i in range(len(self.y_list)):
                self.fit_data_list[i][0].set_data(x_list[ix_plot], yfin_list[i])

                axis = self.data_axes[ix_plot]
                format = "{label}: line#{iline} data#{idata}: x={x:g} y={y:g}"
                self.plot_event.annotation.add_line(f"final", axis, axis, x_list[ix_plot], yfin_list[i], self.fit_data_list[i][0], 
                        inf_list = {"x": x_list[ix_plot], "y": yfin_list[i], "fit": self},
                        annotation_format = format, inf_format = inf_format_func)

                self.data_axes[ix_plot].set_xlim([min(self.x_list[0]), max(self.x_list[0])])
                self.data_axes[ix_plot].set_ylim([min([self.view_ylim_data_list[i][0], min(yfin_list[i])]), 
                                 max([self.view_ylim_data_list[i][1], max(yfin_list[i])])])
        else:
            self.index = list(range(len(yfin_list[0])))
            for i in range(len(self.y_list)):
                self.fit_data_list[0][0].set_data(self.index, yfin_list[i])

                axis = self.data_axes[0]
                format = "{label}: line#{iline} data#{idata}: index={index} x={x:g} y={y:g}"
                self.plot_event.annotation.add_line(f"final", axis, axis, x_list[i], yfin_list[i], self.fit_data_list[i][0], 
                        inf_list = {"index": self.index, "x": self.x_list[i], "y": self.y_list[i], "fit": self},
                        annotation_format = format, inf_format = inf_format_func)

                self.data_axes[0].set_xlim([min(self.index), max(self.index)])
                self.data_axes[0].set_ylim([min([self.view_ylim_data_list[i][0], min(yfin_list[i])]), 
                                 max([self.view_ylim_data_list[i][1], max(yfin_list[i])])])

        if self.error_axis:
            format = "error: iter={iteration}  fmin={fmin:g}"
            self.plot_event.annotation.add_line(f"error", self.error_axis, self.error_axis, self.iter_list, self.fmin_list, self.error_data[0], 
                        inf_list = {"iteration": self.iter_list, "fmin": self.fmin_list},
                        annotation_format = format, inf_format = format)

        if self.y_scale == 'log':
            self.data_axes[0].set_yscale('log')

        if iter is None:
            iter = self.iter
        if fmin is None:
            fmin = self.ffin
            self.iter_list.append(iter)
            self.fmin_list.append(ffin)

        if self.get('error_data', None) is not None:
            self.error_data[0].set_data(self.iter_list, self.fmin_list)
            self.error_axis.set_xlim([min(self.iter_list), max(self.iter_list)])
            self.error_axis.set_ylim([min(self.fmin_list), max(self.fmin_list)])

        self.stop_button.label.set_text('finished')
#        self.ax_button = plt.axes(self.button_region)
#        self.stop_button = wg.Button(self.ax_button, 'finished', color = '#f8e58c', hovercolor = '#38b48b')
#        self.stop_button.on_clicked(self.button_click)

        self.plot_event.register_annotation_event(self.fig, activate = True, print_level = 0)

#        plt.tight_layout()
        self.plt.subplots_adjust(top = plot_region[0], bottom = plot_region[1])

        if savefig_path is not None and savefig_path != '':
            plt.savefig(savefig_path)

        if use_pause is not None:
            self.plt.pause(use_pause)
        else:
            self.plt.show()

    def S2_with_x_ivar(self, x, ivar, variance, run = True, print_level = 0):
        pk = self.pk.copy()
        if ivar is not None:
            pk[ivar] = x
        fmin = self.cal_fmin(pk, self.x_list, self.y_list, self.w_list, fit = self, run = run, print_level = print_level)
        
        return self.yerr2_sum / variance

    def search_xlim_p(self, variance, ivar, xstep, xlim_max, S2_0, S2_target, nmax_iter, S2_eps):
        kstep = 2.0

        x0 = self.pk[ivar]
        x_min  = x0
        S2_min = S2_0
        while True:
            x_new = x0 + xstep
            if xlim_max <= x_new:
                x_new = xlim_max
            S2_new = self.S2_with_x_ivar(x_new, ivar, variance)
            if S2_target < S2_new or xlim_max == x_new:
                x_max = x_new
                S2_max = S2_new
                break

            x_min = x_new
            S2_min = S2_new
            xstep *= kstep
    
        if abs(S2_max - S2_target) < S2_eps:
            return x_max, S2_max
        else:
            for iter in range(10):
                x_half = (x_min + x_max) / 2.0
                S2_half = self.S2_with_x_ivar(x_half, ivar, variance)
                if abs(S2_half - S2_target) < S2_eps:
                    break

                if (S2_min - S2_target) * (S2_half - S2_target) <= 0:
                    x_max = x_half
                    S2_max = S2_half
                else:
                    x_min = x_half
                    S2_min = S2_half

        return x_half, S2_half

    def search_xlim_m(self, variance, ivar, xstep, xlim_min, S2_0, S2_target, nmax_iter, S2_eps):
        kstep = 2.0

        x0 = self.pk[ivar]
        x_max  = x0
        S2_max = S2_0
        while True:
            x_new = x0 - xstep
            if x_new <= xlim_min:
                x_new = xlim_min
            S2_new = self.S2_with_x_ivar(x_new, ivar, variance)
            if S2_target < S2_new or x_new == xlim_min:
                x_min = x_new
                S2_min = S2_new
                break

            x_max = x_new
            S2_max = S2_new
            xstep *= kstep

        if abs(S2_max - S2_target) < S2_eps:
            return x_min, S2_min
        else:
            for iter in range(10):
                x_half = (x_min + x_max) / 2.0
                S2_half = self.S2_with_x_ivar(x_half, ivar, variance)
                if abs(S2_half - S2_target) < S2_eps:
                    break
            
                if (S2_min - S2_target) * (S2_half - S2_target) <= 0:
                    x_max = x_half
                    S2_max = S2_half
                else:
                    x_min = x_half
                    S2_min = S2_half

        return x_half, S2_half

    def build_likelihood_func(self, xk, ivar, 
                x_list = None, y_list = None, w_list = None,
                xlim_min = None, xlim_max = None, kvariance = 6.0, ncal = 101, 
                nmax_iter = 10, S2_eps = 0.1, scale = None, run = True, print_level = 0):
        if x_list is None:
            x_list = self.x_list
        if y_list is None:
            y_list = self.y_list
        if w_list is None:
            w_list = self.w_list
        if xlim_min is None:
            xlim_min = self.kmin[ivar]
        if xlim_max is None:
            xlim_max = self.kmax[ivar]
        if scale is None:
            scale = self.pk_scale[ivar]

        fmin = self.cal_fmin(xk, x_list, y_list, w_list, fit = self, run = run, print_level = print_level)
        yerr_list = self.yerr_list

        nx = len(yerr_list)
        yerr2_sum0 = sum([d * d for d in yerr_list])
        yerr_mean = np.mean(yerr_list)
        variance  = np.var(yerr_list)

        if print_level:
            print(f"Error statistics for {self.varname[ivar]}:")
            print(f"  # of data   : {nx}")
            print(f"  mean        : {yerr_mean:12.6g}")
            print(f"  variance    : {variance:12.6g}")
            print(f"  S2=sum(error^2): {yerr2_sum0:12.6g}")
#    print(f"  fmin=sqrt({yerr2_sum0:12.6g}/{nx}={fmin:12.6g}")

        x0 = xk[ivar]
        S2_0 = self.S2_with_x_ivar(x0, ivar, variance)
        p0 = exp(-S2_0 / 2.0)
        S2_target = S2_0 + kvariance
#       p0 = exp(-yerr2_sum0 / variance / 2.0)
#        print(f"  p0={p0:12.6g}")
#       print(f"  x0={x0:g}")
        if print_level:
            print(f"  xrange: {xlim_min:g} - {xlim_max:g}")
            print(f"  S2={S2_0:12.6g}")
            print(f"  p0={p0:12.6g}")
            print(f"  S2_target={S2_target:g}")

        x_list = [x0]
        p_list = [1.0]

        xstep = (xlim_max - x0) / 10.0
        xmax, S2_max = self.search_xlim_p(variance, ivar, xstep, xlim_max, S2_0, S2_target, nmax_iter, S2_eps)
        if print_level:
            print(f"    calculate in xrange: {x0:g} - {xmax:g} at S2={S2_max:g}")

        if scale == 'log':
            xstep = (log(xmax) - log(x0)) / (ncal - 1)
        else:
            xstep = (xmax - x0) / (ncal - 1)
        for i in range(1, ncal):
            if scale == 'log':
                x = x0 * exp(i * xstep)
            else:
                x = x0 + i * xstep
            _S2 = self.S2_with_x_ivar(x, ivar, variance)
            x_list.append(x)
            p_list.append(exp(-_S2 / 2.0) / p0)

        xstep = (x0 - xlim_min) / 10.0
        xmin, S2_min = self.search_xlim_m(variance, ivar, xstep, xlim_min, S2_0, S2_target, nmax_iter, S2_eps)
        if print_level:
            print(f"    calculate in xrange: {xmin:g} - {x0:g} at S2={S2_min:g}")
        if scale == 'log':
            xstep = (log(x0) - log(xmin)) / (ncal - 1)
        else:
            xstep = (x0 - xmin) / (ncal - 1)
        for i in range(1, ncal):
            if scale == 'log':
                x = x0 * exp(-i * xstep)
            else:
                x = x0 - i * xstep
            _S2 = self.S2_with_x_ivar(x, ivar, variance)
            x_list.append(x)
            p_list.append(exp(-_S2 / 2.0) / p0)

#zipを使って、x_listとp_listをx_listの値準にソートする
        x_list, p_list = zip(*sorted(zip(x_list, p_list)))
#       print("x  y")
#       for x, p in zip(x_list, p_list):
#           print(f"{x:g}\t{p:g}")

        return x_list, p_list

    def get_xrange_list(self, itarget, nx, varname = None, xmin = None, xmax = None, scale = None, print_level = 1):
        if varname is None:
            varname = self.varname[itarget]
        if xmin is None:
            xmin = self.kmin[itarget]
        if xmax is None:
            xmax = self.kmax[itarget]
        if scale is None:
            scale = self.pk_scale[itarget]

        if scale == 'log':
            if print_level:
                print(f"   range {xmin} - {xmax} for {nx} points in {scale} scale")
            logx0 = log(xmin)
            logx1 = log(xmax)
            logxstep = (logx1 - logx0) / (nx - 1)
            xrange_list = [exp(logx0 + i * logxstep) for i in range(nx)]
        else:
            if print_level:
                print(f"   range {xmin} - {xmax} for {nx} points in linear scale")
            xstep = (xmax - xmin) / (nx - 1)
            xrange_list = [xmin + i * xstep for i in range(nx)]

        if xmax < xrange_list[nx - 1]:
            xrange_list[nx - 1] = xmax
            if print_level:
                print(f"    xrange_list for {varname}:", [f"{x:10.4g}" for x in xrange_list])

        return xrange_list
