from math import exp, sqrt, log, gamma
from scipy import integrate         # 数値積分関数 integrateを読み込む
import openpyxl


from tklib.tkfile import tkFile
from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me
from tklib.tktransport.tkDOS_FEA import integrate_Simpson, integrate_Simpson_list, tkDOS
from tklib.tktransport.tkTransport import integrate_Simpson_list, fe, fh, meff2NC_FEA, meff2DC0_FEA, BMShift_FEA


# sigma (S/m), S (V/K)
# muw (m^2/(Vs)), nw (m^-3)
def weighted_mobility(sigma, S, T, unit = None):
    if unit is None:
        print("\nWarning in tktransport.tkTansport.weighted_mobility(): unit is not given. Use unit='m'")

    if unit == 'cm':
        sigma /= 0.01
        S *= 1.0e-6

    if S < 0.0:
        sign = -1
        carriertype = 'e'
    else:
        sign = 1
        carriertype = 'h'

    absS  = abs(S)
    kBe   = kB / e
    SkBe = absS / kBe

    m2kBT = 2.0 * me * kB * T
    A1b = 8.0 * pi * e * pow(m2kBT, 1.5)
    A1 = 3.0 * h**3 * sigma / A1b
    A2 = 3.0 / (pi * pi) / kBe * absS
    SkBem2 = SkBe - 2.0
    expp1m2 = exp(SkBem2)

    expm5m1 = exp(-5.0 * (SkBe - 1.0))
    expp5m1 = exp( 5.0 * (SkBe - 1.0))
    fdeg  = 1.0 / (1.0 + expp5m1)
    fndeg = 1.0 / (1.0 + expm5m1)
    
    muw = A1 * (expp1m2 * fndeg + A2 * fdeg)
    
    nw  = sigma / e / muw

    if unit == 'cm':
        return muw * 1.0e4, nw * 1.0e-6, sign, carriertype  # cm2/Vs, cm^-3

    return muw, nw, sign, carriertype    # m2/Vs, m^-3

# sigma (S/m), S (V/K)
# muw (m^2/(Vs)), nw (m^-3)
def weighted_mobility_new(sigma, S, T, meff = 1.0, r = 0.0, unit = None):
    if unit is None:
        print("\nWarning in tktransport.tkTansport.weighted_mobility(): unit is not given. Use unit='m'")

    if S < 0.0:
        sign = -1
        carriertype = 'e'
    else:
        sign = 1
        carriertype = 'h'

    Ske = abs(S) / (kB / e)
    expS = exp(5.0 * (abs(Ske) - 2.0))
    fdeg  = 1.0 / (1.0 + expS)
    fndeg = 1.0 / (1.0 + 1.0 / expS)

    mkT2 = 2.0 * kB * me * meff * T
    kSnyder = 3.0 * h * h * h / 8.0 / pi / pow(mkT2, 1.5)
    kNndeg = h * h * h / 2.0 / pow(pi * mkT2, 1.5)
    kNdeg  = 3.0 * h * h * h / 8.0 / pi / pi / pi / pi

#    Ninvndeg = kSnyder * exp(Ske - r - 2.0)
#    Ninvdeg  = kSnyder * 3.0 / pi / pi * Ske
    Ninvndeg = kNndeg * exp(Ske - r - 2.0)
    Ninvdeg = kNdeg * pow(Ske * 3.0 / mkT2 / (r + 1.0), 1.5)

    muw = sigma / e * (Ninvndeg * fndeg + Ninvdeg * fdeg)
    nw  = sigma / e / muw
#    print("f=", fndeg, sigma / e * Ninvndeg, fdeg, sigma / e * Ninvdeg)

    if unit == 'cm':
        return muw * 1.0e4, nw * 1.0e-6, sign, carriertype  # cm2/Vs, cm^-3

    return muw, nw, sign, carriertype

def weighted_mobility_exact(sigma, S, T, meff = 1.0, r = 0.0, polarity = 'h', 
                EFmin = -5.0, EFmax = 5.0, eps = 1.0e-10, nmaxiter = 1000, print_level = 1, unit = None):
    if unit is None:
        print("\nWarning in tktransport.tkTansport.weighted_mobility(): unit is not given. Use unit='m'")

    if S < 0.0:
        sign = -1
        carriertype = 'e'
    else:
        sign = 1
        carriertype = 'h'

    dos = tkDOS()
    dos.EV = 0.0
    dos.EC = 3.0
    dos.mheff = meff
    dos.meeff = meff
    dos.NV = meff2NC_FEA(dos.mheff, T)
    dos.NC = meff2NC_FEA(dos.meeff, T)
    dos.DV0 = meff2DC0_FEA(dos.mheff, T)
    dos.DC0 = meff2DC0_FEA(dos.meeff, T)
    dos.EA = 0.05
    dos.NA = 0.0
    dos.ED = 2.95
    dos.ND = 0.0
    dos.EF0 = 0.1

    EF = dos.EF_from_S(T, abs(S), r, polarity = polarity, EFmin = EFmin, EFmax = EFmax, eps = eps, nmaxiter = nmaxiter, print_level = print_level)
    N = dos.Nh(T, EF) * 1.0e6
    muw = sigma / N / e
    
    if unit == 'cm':
        return muw * 1.0e4, N * 1.0e-6, sign, carriertype  # cm2/Vs, cm^-3

    return muw, N, sign, carriertype, EF
