import sys
import numpy as np
from numpy import log, log10, sqrt
import pandas as pd
import openpyxl
from sklearn.decomposition import PCA
import matplotlib.pyplot as plt
from sklearn.cluster import KMeans
from sklearn.linear_model import LinearRegression
from sklearn.neural_network import MLPRegressor
from sklearn.model_selection import train_test_split


infile = "property_db.xlsx"
outfile = "property_clustering.xlsx"

target_key = "kzz"  #target key for objective function of clustering
prop_min = 300.0    #minimum value of target_key
n_clusters = 5      #number of clusters
n_components = 2    #number of principal components
ksize = 1.0         #marker size for scatter plot


nargs = len(sys.argv)
if nargs > 1:
    target_key = sys.argv[1]
if nargs > 2:
    prop_min = float(sys.argv[2])
if nargs > 3:
    n_clusters = int(sys.argv[3])
if nargs > 4:
    n_components = int(sys.argv[4])
if nargs > 5:
    ksize = float(sys.argv[5])

pca_descriptors = ["kx","ky","kz","f(THz)","kzz"]
pcr_descriptors = ["kx","ky","kz","f(THz)"]
boxplot_descriptors = ["kx","ky","kz","f(THz)", "v(m/s)", "tau(ps)", "MFP(nm)", "w*target", target_key]

print()
print("##############################################")
print("#  Clustering for ALMODE result")
print("##############################################")
print(f"target      : {target_key}")
print(f"n_clusters  : {n_clusters}")
print(f"n_components: {n_components}")
print(f"prop_min    : {prop_min}")
print(f"ksize       : {ksize}")

#pandsを使ってinfileを読み込む
df = pd.read_excel(infile)
#dfから無効データを削除する
df = df.dropna()
#dfのtarget_key列の値がprop_minより小さい行を削除する
df = df[df[target_key] > prop_min]   
kx = df["kx"]
ky = df["ky"]
kz = df["kz"]
kxc = df["kx,c"]
kyc = df["ky,c"]
kzc = df["kz,c"]
_is = df["is"]
f = df["f(THz)"]
target = df[target_key]
df["w*target"] = [w * t for w, t in zip(df["w"], target)]

#kxc,kyc,kzcの絶対値のリストを作り、dfに追加
df["|kx,c|"] = [abs(ikx) for ikx in kxc]
df["|ky,c|"] = [abs(iky) for iky in kyc]
df["|kz,c|"] = [abs(ikz) for ikz in kzc]

print(df)
#dfから"kx","ky","kz","is","kzz"の列を取り出し、PCA解析を行う
X = df[pca_descriptors]
#standardize X
X = (X - X.mean()) / X.std()

pca = PCA(n_components = n_components)
pca.fit(X)
X_pca = pca.transform(X)

print()
print("x_pca:")
print(X_pca)

pcr_objective = "kzz"
X_pcr = df[pcr_descriptors]
logy_list = [log(y + 1.0e-6) for y in df[pcr_objective]]
Y_pcr = pd.DataFrame(np.array([logy_list]).T, columns = [pcr_objective])
#X_pcr を記述子として、Y_pcrを目的変数としてPCR解析を行う
pcr_model = LinearRegression()
#X_pcrを標準化する
X_pcr = (X_pcr - X_pcr.mean()) / X_pcr.std()
pcr_model.fit(X_pcr, Y_pcr)
Y_pcr_predicted = pcr_model.predict(X_pcr)

X_train, X_test, Y_train, Y_test = train_test_split(X_pcr, Y_pcr, test_size=0.2, random_state=42)
#X_pcr を記述子として、Y_pcrを目的変数としてMLP回帰を行う
mlp_model = MLPRegressor(hidden_layer_sizes=(10, 5))
mlp_model.fit(X_train, Y_train)
Y_train_predicted_mlp = mlp_model.predict(X_train)
Y_test_predicted_mlp = mlp_model.predict(X_test)

#Xを使ってclusteringを行う
kmeans = KMeans(n_clusters=n_clusters)
kmeans.fit(X)
y_kmeans = kmeans.predict(X)

#クラスタリング結果をdfに追加する
df["cluster"] = y_kmeans

#DataFrame()型の変数cluster_dfを作る
cluster_list = []
for icluster in range(n_clusters):
    data = df[df["cluster"]==icluster]
    print("data=", data)
    cluster_list.append(data)

#cluster毎にkx,ky,kz,_is,kzzの絶対値の平均、標準偏差を出力
print()
print("y_kmeans:")
for icluster in range(n_clusters):
    print(f"cluster#{icluster}:")
    data = df[df["cluster"]==icluster]
    for icolumn in ["|kx,c|","|ky,c|","|kz,c|","is","kzz"]:
        print(f"  {icolumn}: mean={data[icolumn].mean():.3f}, std={data[icolumn].std():.3f}")

df = df.sort_values(target_key, ascending = False)
print()
print(f"Save to {outfile}")
#df.to_excel()でopenpyxlsxを指定する
df.to_excel(outfile, engine="openpyxl", index=False)

print()
print("plot:")

#kx, ky, kzに関する3D散布図を作成する。markerの大きさはtargetの値に比例する
fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
for icluster in range(n_clusters):
    data = df[df["cluster"]==icluster]
    ax.scatter(data["|kx,c|"], data["|ky,c|"], data["|kz,c|"], s=data["w*target"]*ksize, label = f"cluster#{icluster}")
ax.set_xlabel("kx")
ax.set_ylabel("ky")
ax.set_zlabel("kz")
#ax.set_xlim([0.0, 0.5])
#ax.set_ylim([0.0, 0.5])
#ax.set_zlim([0.0, 0.5])
ax.set_aspect("auto")
ax.legend()
plt.pause(1.0e-5)

fig = plt.figure()
ax = fig.add_subplot(111, projection='3d')
for icluster in range(n_clusters):
    data = df[df["cluster"]==icluster]
    logtau = [log10(val) for val in data["tau(ps)"]]
    ax.scatter(data["f(THz)"], logtau, data["v(m/s)"], s=data["w*target"]*ksize, label = f"cluster#{icluster}")
ax.set_xlabel("f(THz)")
ax.set_ylabel("log$_{10}$(tau/ps)")
ax.set_zlabel("v(m/s)")
ax.set_aspect("auto")
ax.legend()
plt.pause(1.0e-5)

for icluster in range(n_clusters):
    fig, ax = plt.subplots(1,1)
    cdata = cluster_list[icluster][boxplot_descriptors]
    cdata = (cdata - cdata.mean()) / cdata.std()
    cdata.boxplot()
    ax.set_title(f"Boxplot for Cluster#{icluster}")
    ax.set_xlabel("Distribution")
    ax.set_ylabel("Values")
    plt.pause(1.0e-5)

fig, ax = plt.subplots(1,1)
ax.scatter(X_pca[:,0],X_pca[:,1])
ax.set_xlabel("First principal component")
ax.set_ylabel("Second principal component")
plt.pause(1.0e-5)

#x_pcaについての負荷量グラフを作成する
nx = int(sqrt(n_components) + 0.9999)
ny = int(n_components / nx + 0.9999)
if nx * ny < n_components:
    ny += 1
fig, axes = plt.subplots(nx, ny)
axes = axes.flatten()
#x軸の目盛り文字はpca_descriptorsリスト変数の内容にする
plt.title(f"PC Loading")
for i in range(n_components):
    ax = axes[i]
    ax.set_title(f"PC#{i+1}")
 
    ax.bar(range(len(pca.components_[i])), pca.components_[i])
    ax.set_xticks(range(len(pca_descriptors)))
    ax.set_xticklabels(pca_descriptors)
    ax.set_xlabel("Feature")
    ax.set_ylabel("Loadings")
plt.tight_layout()
plt.pause(1.0e-5)

'''
#Y_pcrとY_pcr_predictedの関係をプロットする
fig, ax = plt.subplots(1,1)
plt.title("PCR")
ax.scatter(Y_pcr, Y_pcr_predicted)
ax.set_xlabel("Observed")
ax.set_ylabel("Predicted")
plt.pause(1.0e-5)
'''

fig, ax = plt.subplots(1,1)
plt.title("MLP")
ax.scatter(Y_train, Y_train_predicted_mlp, label = "train")
ax.scatter(Y_test, Y_test_predicted_mlp, label = 'test')
ax.set_xlabel("Observed")
ax.set_ylabel("Predicted")
ax.legend()
plt.pause(1.0e-5)

#クラスタリング結果をプロットする
fig, ax = plt.subplots(1,1)
#clusterの値毎にプロットし、legendを表示する。kzzの大きさによってmarkerのサイズを変える
for icluster in range(n_clusters):
    data = df[df["cluster"]==icluster]
    sizes = data["kzz"] * 1.0e-3
#    sizes = [log(val + 1.0e-6) for val in data["kzz"]]
    ax.scatter(X_pca[y_kmeans==icluster,0], X_pca[y_kmeans==icluster,1], label=f"cluster#{icluster}", s=sizes) 
ax.set_xlabel("First principal component")
ax.set_ylabel("Second principal component")
ax.set_xscale("log")
ax.set_yscale("log")
ax.legend()
plt.pause(1.0e-5)

input("\n>> Press Enter to termninate>>")
