import os
import sys
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkvariousdata import tkVariousData
from tklib.tkcrystal.tkalamode import read_input, read_taudata, read_kpoints, read_properties, cal_contrib, convert_k_list


"""
Convert ALMODE anphon files to Excel
"""


plugin_ver = "[[x, y]]:0.2"   # data_list_type = [[x, y]]


usage_str = '''
f"  (i) usage: python {sys.argv[0]} infile logfile taufile outfile"
f"         ex: python {sys.argv[0]} {cparams.infile} {cparams.anphonlog_file} {cparams.tau_file} {cparams.outfile}"
'''[1:-1]



#================================
# global parameters
#================================
default_ext = '.in'
input_type  = "ALMODE anphon input .in"
output_type = "Excel .xlsx"

figsize = (8, 4)
fontsize = 12
legend_fontsize = 12


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0

    cparams.infile = 'rta.in'
    cparams.anphonlog_file = 'k15.log'
    cparams.tau_file = 'tau_300K-k15.dat'
    cparams.outfile  = None


def update_vars(app = None, cparams = None):
    if getarg(2, None) is None:
        app.terminate(usage = usage)
    
    cparams.infile          = getarg(1, cparams.infile)
    cparams.anphonlog_file  = getarg(2, cparams.anphonlog_file)
    cparams.tau_file        = getarg(3, cparams.tau_file)
    cparams.outfile         = getarg(4, cparams.outfile)

    if cparams.outfile is None:
        cparams.outfile = replace_path(cparams.infile, os.path.join("{dirname}", "{filebody}-property.xlsx"))


def check_file_type(infile, inf = None, app = None, cparams = None):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != default_ext:
        return f"Error: file extension [{ext}] in [{infile}] is invalid"

    fp = tkFile(infile, 'r', encoding = 'shift_jis')
    found_general = False
    found_kpoint = False
    found_analysis = False
    found_mode = False
    for line in fp.fp:
        line = line.lower()
#        print("line: ", line)
        if '&general' in line:
            found_general = True
        if '&kpoint' in line:
            found_kpoint = True
        if '&analysis' in line:
            found_analysis = True

        line = line.strip()
        if '=' in line:
            _aa = line.split('=', 1)
            if _aa[0].strip() == 'mode':
                found_mode = True

#        print("f:", found_general, found_kpoint, found_analysis, found_mode)
        if found_general and found_kpoint and found_analysis and found_mode:
            fp.close()
            return True

    fp.close()

    return f"Error: File type of [{infile}] is inconsistent with [{input_type}]"

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, is_print = False):
    if is_print:
        print_level = 1
    else:
        print_level = 0
        
    if not os.path.isfile(infile):
        if is_print:
            print(f"Error in almode_property2xlsx.read_data(): Can not read [{infile}]")
#        app.terminate(f"Error: Can not read [{infile}]")
        return False

    ftype =  check_file_type(infile)
    if ftype is False:
        print(f"Error in almode_property2xlsx.read_data(): "
            + f"File type is inconsistent with {input_type}")
        return False
    elif type(ftype) is str and 'Error' in ftype:
        print(ftype)
        return False
        
    if is_print:
        print("")
        print(f"Read [{infile}]")
    inf_input = read_input(cparams.infile)
    if inf_input is None:
        print(f"Warning: Can not read [{cparams.infile}]. Skip")

    if is_print:
        print(f"Read [{cparams.anphonlog_file}]")
        print(f"Read [{cparams.tau_file}]")
    sn_list, ik_list, is_list, kx_list, ky_list, kz_list, w_list, fk_list, fTHz_list, _tau_list, v_list, MFP_list, \
                 kxx_list, kxy_list, kxz_list, kyx_list, kyy_list, kyz_list, kzx_list, kzy_list, kzz_list, inf_log \
                 = read_properties(cparams.anphonlog_file, cparams.tau_file, print_level = print_level)
    if sn_list is None:
        print(f"Warning: Can not read [{cparams.anphonlog_file}] or [{cparams.tau_file}]. Skip")

    praij = inf_log['praij']
    kxc_list, kyc_list, kzc_list = convert_k_list(praij, kx_list, ky_list, kz_list)

    labels    = ['sn', 'ik', 'is', 'kx', 'ky', 'kz', 'kx,c', 'ky,c', 'kz,c', 
                 'w', 'f(cm-1)', 'f(THz)', 'tau(ps)', 'v(m/s)', 'MFP(nm)', 
                 'kxx(W/mK)', 'kxy', 'kxz', 'kyx', 'kyy', 'kyz', 'kzx', 'kzy', 'kzz']
    data_list = [sn_list, ik_list, is_list, kx_list, ky_list, kz_list, kxc_list, kyc_list, kzc_list, 
                 w_list, fk_list, fTHz_list, _tau_list, v_list, MFP_list,
                 kxx_list, kxy_list, kxz_list, kyx_list, kyy_list, kyz_list, kzx_list, kzy_list, kzz_list]

    inf = {}
    dirname, basename, filebody, ext = split_file_path(infile)
    inf["data_list_type"] = "[[x, y]]"
    inf["meta"]        = []
    inf["filename"]    = infile
    inf["filenames"]   = [infile, cparams.anphonlog_file, cparams.tau_file]
    inf["sample_name"] = inf_input.get("KD", "no name")
    inf["labels"]      = [labels]
    inf["data_list"]   = data_list
    inf["inf_input"]   = inf_input

    return inf

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-property.xlsx"))

def print_data(inf, app = None, cparams = None):
    print("")
    print("filenames:", inf["filenames"])
    print("sample   :", inf["sample_name"])

    labels = inf["labels"][0]
    ncol   = len(labels)
    data_list = inf["data_list"]
    ndata = len(data_list[0])

    label_line = ''
    for l in inf["labels"][0]:
        label_line += f"{l:10} "
    label_line.rstrip()
    print(label_line)

    nprint = 50
    nskip = int(ndata / nprint + 0.9999)
    for i in range(0, ndata, nskip):
        line = ''
        for icol in range(ncol):
            val = data_list[icol][i]
            try:
                val = float(val)
                s = f"{val:10.4g}"
            except:
                try:
                    val = int(val)
                    s = f"{val:10d}"
                except:
                    s = val

            line += s + ' '
        line.rstrip()
        print(line)

def save_data(outfiles, inf, app = None, cparams = None, is_print = False):
    labels    = inf["labels"][0]
    data_list = inf["data_list"]
    if type(outfiles) is str:
        outfile = outfiles
    else:
        outfile = outfiles[0]
    
    print("")
    print(f"Save to [{outfile}]")
    print("labels:", labels)
    tkVariousData().to_excel(outfile, labels, data_list)

    return True


def plot_data(inf, app = None, cparams = None):
    print()
    print("No plot function in this plugin")

def convert(inf, app = None, cparams = None):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Convert ALMODE anphon files to MS Excel property file")
    print( "==========================================================================")
    print(f"Input file             : {cparams.infile}")
    print(f"anphon console log file: {cparams.anphonlog_file}")
    print(f"property file          : {cparams.tau_file}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    if cparams.outfile is None:
        cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)

    inf = read_data(cparams.infile, cparams = cparams, is_print = False)
    print_data(inf)
    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, cparams = cparams)
#    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
