import os
import sys
import json
from monty.json import MontyEncoder

from mp_api.client import MPRester
from pymatgen.core.structure import Structure
from pymatgen.core import Composition
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.io.cif import CifWriter, CifParser


formula = None
output_format = "cif"
prec = 1.0e-3


def usage():
    print()
    print(f"Usage: python {sys.argv[0]} formula [output_format] [prec]\n")

def update_vars():
    global formula, output_formuat, prec

    argv = sys.argv
    narg = len(argv)
    if narg <= 1:
        print("\nError: Chemical formula must be given as the first arg\n")
        input("Press ENTER to terminate>>\n")
        exit()

    formula = argv[1]
    if narg >= 3: output_format = argv[2]
    if narg >= 4: prec = float(argv[3])


def get_MPReser(API_KEY = None):
    if not API_KEY: API_KEY = os.getenv('MP_APIKEY')
    if API_KEY is None:
        print("\nError: Can not get MP API Key from the environment var MP_APIKEY\n")
        return None

    mpr = MPRester(API_KEY)
    if mpr is None:
        print(f"\nError: Can not get MPRester using the given API_KEY [{API_KEY}]\n")
        return None

    return mpr

def get_material_data(formula, mpr):
    if ',' in formula:
        elements_in = formula.replace(' ', '').split(',')
        search_results = mpr.materials.summary.search(elements = elements_in, fields = [])
    elif '-' in formula:
        search_results = mpr.materials.summary.search(chemsys = formula, fields = [])
    else:
        search_results = mpr.materials.summary.search(formula = formula, fields = [])

    if not search_results:
        print(f"No data found for {formula}")
        return None, None

    mid_list = []
    material_list = []
    for res in search_results:
        material_list.append(res)
        mid_list.append(res.material_id)

    return mid_list, material_list

def get_structure(formula):
    mpr = get_MPReser()
    if not mpr: exit()

    available_fields = mpr.materials.summary.available_fields
    print()
    print("available_fields:")
    for f in sorted(available_fields):
        print("  ", f)

    if ',' in formula:
        elements_in = formula.replace(' ', '').split(',')
    else:
        elements_in = None

    mid_list, material_list = get_material_data(formula, mpr)
    if mid_list is None: return None

    print()
    print("Material IDs:", mid_list)

    structures = []
    for material_id, material in zip(mid_list, material_list):
        material_data = mpr.materials.summary.search(material_ids = [material_id], fields = available_fields)
        if not material_data:
            print(f"No structure data found for material_id {material_id}")
            continue

        structure_dict = material_data[0].structure.as_dict()
        structure = Structure.from_dict(structure_dict)

        cformula     = structure.reduced_formula    #formula.replace(" ", "")
        composition  = Composition(cformula)
        elements_res = [element.symbol for element in composition.elements]

        if elements_in:
            not_included = []
            for e in elements_res:
                if e not in elements_in: not_included.append(e)

            if len(not_included) >= 1:
                print(f"  ** One or more elements are not included for [{cformula}] in the given condition [{formula}]:", not_included)
                continue
            else:
                print(f"  ** Found [{cformula}] [{material_id}]")

        structure.cmaterial   = material
        structure.cformula    = cformula
        structure.celements   = elements_res
        structure.material_id = material_id

        structures.append(structure)
        print(f"Found material_id for {formula}: {material_id}: {cformula}")

    return structures


def save_structures(structures, output_format, formula):
    mpr = get_MPReser()
    keys = mpr.materials.summary.available_fields

    print()
    for i, structure in enumerate(structures):
        cformula = structure.cformula
        if output_format == "poscar":
            if i == 0:
                output_path = f"POSCAR"
            else:
                output_path = f"{cformula}_{structure.material_id}.POSCAR"
        else:
            output_path = f"{cformula}_{structure.material_id}.{output_format}"

#        symmetry_analyzer = SpacegroupAnalyzer(structure, symprec = prec) #, angle_tolerance = 5.0)
#        symmetrized_structure = symmetry_analyzer.get_symmetrized_structure()
#        print()
#        print("symmetrized_structure:")
#        print(symmetrized_structure)
        
        print(f"Saving {cformula} structure to [{output_path}]...")
#        structure.to(filename = output_path, fmt = output_format)
        writer = CifWriter(structure, symprec = prec) #, angle_tolerance = 5.0)
#        writer = CifWriter(symmetrized_structure, symprec = prec) #, angle_tolerance = 5.0)
        writer.write_file(output_path)

        outfile = f'meta_{cformula}_{structure.material_id}.txt'
        print(f"  meta data to [{outfile}]...")
        material = structure.cmaterial
        fp = open(outfile, 'w')
        for key in keys:
            val = getattr(material, key, None)
            if val is not None:
                fp.write(f"{key}: {val}\n")
#                print(f"  {key}: ", val)
        fp.close()

        outfile = f'meta_{cformula}_{structure.material_id}.json'
        print(f"  meta data to [{outfile}]...")
        with open(outfile, 'w') as f:
            json.dump(material.dict(), f, cls = MontyEncoder, indent = 2)


def get_strcutures():
    structures = get_structure(formula)
    if structures:
        save_structures(structures, output_format, formula)


def main():
    update_vars()

    get_strcutures()

    usage()
    print()
    input("Press ENTER to terminate>>\n")
    exit()


if __name__ == "__main__":
    main()
