import os
import sys
import shutil
import glob
import csv
import re
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from scipy.interpolate import interp1d
from pprint import pprint
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkvariousdata import tkVariousData
from tklib.tkutils import colors, graph_style_idx, sort_lists
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, pconv, is_list, split_file_path, replace_path, del_quote
from tklib.tkgraphic.tkplotevent import tkPlotEvent


"""
Read MDR HAXPES .csv file
"""

usage_str = '''
"  (i) usage: python {} mode input_file xmin xmax".format(sys.argv[0])
"         ex: python {} {} {} {}".format(sys.argv[0], cparams.infile, cparams.xmin, cparams.xmax)
'''[1:-1]

plugin_ver = "data:0.2"   # data_list_type = [[x, y]]


#================================
# global parameters
#================================
default_ext = '.csv'
extensions  = ['.csv']
input_type  = "MDR HAXPES .csv"
output_type = "MSExcel .xlsx"

figsize = (8, 4)
fontsize = 12
legend_fontsize = 12


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.infile = f'*{default_ext}'

    cparams.xmin  = '*'
    cparams.xmax  = '*'


def update_vars(app = None, cparams = None):
    if getarg(1, None) is None:
        app.terminate("\nNo argument given", usage = usage)
    
    cparams.infile = getarg(1, cparams.infile)
    cparams.xmin   = getfloatarg(2, cparams.xmin)
    cparams.xmax   = getfloatarg(3, cparams.xmax)


def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def check_file_type(infile, inf = None, app = None, cparams = None):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != default_ext:
        return None

    fp = tkFile(infile, mode = 'r', encoding = None)
    line = fp.readline()
    fp.close()
    
    if '_EB_eV' in line:
        return {"file_type": input_type}

    return None

def read_data(infile, app = None, cparams = None, is_print = False):
    if is_print:
        print("")
        print(f"Read [{infile}]")

    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not read [{infile}]")

    fp = tkFile(cparams.infile, mode = 'r', encoding = None)
    line = fp.readline()
    labels = line.split(',')
    labels = [l.strip() for l in labels]

    data_list0 = []
    while 1:
        line = fp.readline()
        if not line:
            break

        a = line.split(',')
        if len(a) < 2:
            continue

        a = [pfloat(s, defval = None) for s in a]
        data_list0.append(a)

    fp.close()

    E_indexes = []
    for i, s in enumerate(labels):
        if '_EB_eV' in s:
            E_indexes.append(i)
    
    spc_list = []
    labels_list = []
    for i in range(len(E_indexes)):
        spc_list.append([])

    ndata = len(data_list0)
    ncol  = len(labels)
    nE = len(E_indexes)
    data_list = []
    for ispc in range(nE):
        idxs = E_indexes[ispc]
        if ispc == nE - 1:
            idxe = ncol - 1
        else:
            idxe = E_indexes[idxs + 1] - 1

        if labels[idxe] == '':
            idxe -= 1

        labels_list.append(labels[idxs:idxe+1])

        nc = idxe - idxs + 1
        dlist = []
        for ic in range(nc):
            dlist.append([])

        for i in range(ndata):
            data = data_list0[i]
            if data[idxs] is None or data[idxs] == '':
                continue

            for idx in range(idxs, idxe + 1):
                dlist[idx - idxs].append(data[idx])

#        dlist = sort_lists(dlist)
        data_list.append(dlist)

    if is_print:
        print("labels_list=", labels_list)
        print("labels=", labels)
        print("len=", ncol, len(data_list), len(data_list[0]))
        print("E indexes: ", E_indexes)

    inf = {}
    dirname, basename, filebody, ext = split_file_path(cparams.infile)
    inf["data_list_type"] = "[[x, y]]"
    inf["filename"]    = infile
    inf["sample_name"] = filebody
    inf["yscale"]      = 'linear'
    inf["meta"]        = []
    inf["data_list"]   = data_list
    inf["nregion"]     = len(data_list)
    inf["labels"]      = labels_list

    return inf

def save_data(outfiles, inf, app = None, cparams = None, is_print = False):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    if not is_list(labels[0]):
        data_list = [data_list]
        labels    = [labels]

    outfile = outfiles[0]

    print("")
    print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, labels, data_list)

    return True

def print_data(inf, app = None, cparams = None):
    labels = inf["labels"]
    print("labels=", labels)
    if not is_list(labels[0]):
        labels = [labels]
    label_str = ''
    for l in labels:
        print("l=", l)
        label_str += "\t".join(l)

    print("")
    print("filename:", inf["filename"])
    print("sample  :", inf["sample_name"])
    print("yscale  :", inf["yscale"])
    print("labels  :", label_str)
    print("nregion :", inf["nregion"])

    meta      = inf["meta"]
    data_list = inf["data_list"]
    nregion   = inf["nregion"]

    for k in range(nregion):
        data = data_list[k]
        ndata = len(data[0])

        print("")
        print(f"Region #{k+1}")
        print(f"  ndata: {ndata}")
        print(label_str)
        for i in range(0, ndata, ndata // 50):
            print(f"  {data[0][i]:12.4g}", end = '')
            for j in range(1, len(data)):
                print(f"\t{data[j][i]:12.4g}", end = '')
            print("")

def plot_data(inf, app = None, cparams = None):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    if not is_list(labels[0]):
        data_list = [data_list]
        labels    = [labels]

    print("")
    print("filename:", inf["filename"])
    print("sample  :", inf["sample_name"])
    print("yscale  :", inf["yscale"])
    print("labels  :", inf["labels"])
    print("nregion :", inf["nregion"])
    
    meta      = inf["meta"]
    labels    = inf["labels"]
    data_list = inf["data_list"]
    nregion = inf["nregion"]

    print("")
    print("Plot")

    fig, axes = plt.subplots(1, nregion, figsize = figsize, sharey = True, gridspec_kw = {'wspace': 0})
    plot_event = tkPlotEvent(plt)
    if nregion == 1:
        axes = [axes]

    plt.title(f"{cparams.infile}")
    for i in range(nregion):
        ax = axes[i]
        ax.tick_params(labelsize = fontsize)

        data = data_list[i]
        _labels = labels[i]
        x = data[0]
        ndata = len(x)
        nspectrum = len(data)
        for j in range(1, nspectrum):
            color, style = graph_style_idx(j - 1)
            ax.plot(x, data[j],picker = True, label = _labels[j], color = color, linestyle = style, linewidth = 1.0)

        ax.legend(fontsize = legend_fontsize)
        if i == 0:
            ax.set_ylabel(labels[0][1], fontsize = fontsize)
        ax.invert_xaxis()

    icenter = nregion // 2
    axes[icenter].set_xlabel(labels[0][0], fontsize = fontsize)

    plt.tight_layout()

#    plot_event.register_click(fig) # callback = lambda event: plot_event.onclick(event))
    plot_event.register_pick(fig) # callback = lambda event: plot_event.onclick(event))
    plt.pause(0.1)
    
    print("")
    input("Press ENTER to terminate>>\n")

def convert(inf, app = None, cparams = None):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Convert CIF file to powder XRD pattern")
    print( "==========================================================================")
    print(f"Input file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams)
    print_data(inf)
    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, cparams = cparams)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
