import os
import sys
import shutil
import glob
import pprint
import csv
import re
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from scipy.interpolate import interp1d
from pprint import pprint
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkexcel_db import tkExcelDB
from tklib.tkvariousdata import tkVariousData
from tklib.tkutils import colors, graph_style_idx, sort_lists
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, pconv, is_list, split_file_path, replace_path, del_quote
from tklib.tkgraphic.tkplotevent import tkPlotEvent


"""
Read Photoionization cross-section for HAXPES 6 keV
"""

usage_str = '''
"  (i) usage: python {} mode input_file".format(sys.argv[0])
'''[1:-1]

plugin_ver = "data:0.2"   # data_list_type = [[x, y]]


#================================
# global parameters
#================================
default_ext = '.xlsx'
extensions  = ['.xlsx']
input_type  = "Photo ionization cross-section HAXPES .xlsx"
output_type = "MS-Excel .xlsx"

figsize = (8, 4)
fontsize = 12
legend_fontsize = 12


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.infile = f'*{default_ext}'


def update_vars(app = None, cparams = None):
    if getarg(1, None) is None:
        app.terminate("\nNo argument given", usage = usage)
    
    cparams.infile = getarg(1, cparams.infile)


def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 0):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != default_ext:
        return None

    xls = tkExcelDB(infile, mode = 'r', encoding = None)
    xls.close()
    
    v = xls.get(1, 2)
    if 'ver' in v:
        return {"file_type": input_type}

    return None

def read_data(infile, app = None, cparams = None, is_print = False, print_level = 0):
    if is_print:
        print("")
        print(f"Read [{infile}]")

    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not read [{infile}]")

    xls = tkExcelDB(infile, mode = 'r', encoding = None)
    xls.close()
    
    nmaxcol = xls.max_column()
    nmaxrow = xls.max_row()

    labels = []
    for icol in range(1, nmaxcol + 1):
        v = xls.get(2, icol)
        if v is None or v == '':
            break
        
        labels.append(v)

    ncol = len(labels)
    data_list = []
    for i in range(ncol):
        data_list.append([])

    an = None
    for irow in range(3, nmaxrow + 1):
        v = xls.get(irow, 2)
        if v is None or v == '':
            continue

        v = xls.get(irow, 1)
        if v is None or v == '':
            v = an

        an = v

        if v is None or v == '':
            continue
            

        data_list[0].append(an)
        for icol in range(2, ncol + 1):
            v = xls.get(irow, icol)
            data_list[icol - 1].append(pfloat(v, v))

    inf = {}
    dirname, basename, filebody, ext = split_file_path(cparams.infile)
    inf["data_list_type"] = "[[data]]"
    inf["filename"]    = infile
    inf["sample_name"] = filebody
    inf["yscale"]      = 'linear'
    inf["meta"]        = []
    inf["data_list"]   = [data_list]
    inf["nregion"]     = 1
    inf["labels"]      = [labels]

    return inf

def save_data(outfiles, inf, app = None, cparams = None, is_print = False):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    if not is_list(labels[0]):
        data_list = [data_list]
        labels    = [labels]

    outfile = outfiles[0]

    print("")
    print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, labels, data_list)

    return True

def print_data(inf, app = None, cparams = None):
    print("")
    print("filename:", inf["filename"])

    labels    = inf["labels"][0]
    data_list = inf["data_list"][0]
    ndata     = len(data_list[0])
    ncol      = len(data_list)

    print("")
    for l in labels:
        print(f"{l[:10]:10} ", end = '')
    print()
    for i in range(ndata):
        for j in range(ncol):
            v = data_list[j][i]
            if v is None:
                v = 'None'
                print(f"{v:10} ", end = '')
            elif type(v) is float:
                print(f"{v:10.3g} ", end = '')
            else:
                print(f"{v:10} ", end = '')
        print()

def plot_data(inf, app = None, cparams = None):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    if not is_list(labels[0]):
        data_list = [data_list]
        labels    = [labels]

    print("")
    print("filename:", inf["filename"])
    
    labels    = inf["labels"][0]
    data_list = inf["data_list"][0]
    
    idx_an = None
    idx_xs = None
    for i, l in enumerate(labels):
        l = l.lower()
        if 'at' in l and 'n' in l:
            idx_an = i
        if 'cross' in l and 'section' in l:
            idx_xs = i

    print("")
    print("Plot")

    fig, ax = plt.subplots(1, 1, figsize = figsize)
    ax.tick_params(labelsize = fontsize)
    plot_event = tkPlotEvent(plt)

    plt.title(f"Photo ionization cross section")
    ax.plot(data_list[idx_an], data_list[idx_xs], linestyle = '', marker = 'o')
#    ax.legend(fontsize = legend_fontsize)
    ax.set_xlabel('Atomic number', fontsize = fontsize)
    ax.set_ylabel('Cross section', fontsize = fontsize)

    plt.tight_layout()

#    plot_event.register_click(fig) # callback = lambda event: plot_event.onclick(event))
#    plot_event.register_pick(fig) # callback = lambda event: plot_event.onclick(event))
    plt.pause(0.1)
    
    print("")
    input("Press ENTER to terminate>>\n")

def convert(inf, app = None, cparams = None):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Read photo ionization cross sections")
    print( "==========================================================================")
    print(f"Input file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams)
    print_data(inf)
    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, cparams = cparams)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
