import os
import sys
import numpy as np
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkvariousdata import tkVariousData
from tklib.tkgraphic.tkplotevent import tkPlotEvent
from tklib.tksci.tkconvolution import convolution, convolve_func
from tklib.tkcrystal.tkvasp import tkVASP
from tklib.tkutils import colors, graph_style_idx
import tklib.tkgui.Ryu_colormap as ryu_cm

import PIX2xlsx as pix


"""
Convert VASP output files to PES
"""

usage_str = '''
"  (i) usage: python {} output_dir".format(sys.argv[0])
"         ex: python {} {}".format(sys.argv[0], cparams.infile)
'''[1:-1]


plugin_ver = "dos:0.2"  # data_list_type = [[x, y]]


#================================
# global parameters
#================================
default_ext = ""
extensions = [""]
input_type  = "VASP output"
output_type = "PES"

figsize = (8, 8)
fontsize = 12
legend_fontsize = 8

linestyles = ['solid', 'dashed', 'dashdot', 'dotted']
markers = ['o', 's', '+', 'x', 'D', 'v', '^', '<', '>', '8', 'h', 'H']
#colors  = ['black', 'red', 'blue', 'darkgreen', 'darkorange', 'hotpink', 'lightgreen', 'cyan', 'yellow', 'magenta', 'chocolate', 
#           'navy', 'slategray', 'olive' ]
colors = ryu_cm.colors
colors = ["black", "red", "blue", colors[0], colors[5], colors[8], colors[2], colors[9]]


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0

    cparams.infile = ''

    cparams.pix_path = ''

# Energy to search edge energies
    cparams.EF0  = 0.1
# Critera DOS value to search band edges
    cparams.DOSth = 1.0e-5

# Max number of electrons occupies one state
    cparams.Nemax = 1.0

# width of Gaussian functio for convolution
    cparams.width = 0.0

# Plot configuration
    cparams.band_marker_size       = 4
    cparams.band_marker_edge_width = 0.5

    cparams.xmin = -10.0  # eV
    cparams.xmax =  10.0  # eV

# Occupancy threshold to separate HOMO and LUMO
    cparams.occ_th = 0.5

    cparams.save_figure = 1
    cparams.plot_figure = 1
    cparams.plot_figure_Ne = 1


def update_vars(app = None, cparams = None):
    if getarg(1, None) is None:
        app.terminate(usage = usage)
    
    cparams.infile     = getarg(1, cparams.infile)
    cparams.pix_path   = getarg(2, cparams.pix_path)
    cparams.xmin   = getfloatarg(3, cparams.xmin)
    cparams.xmax   = getfloatarg(4, cparams.xmax)
    cparams.width  = getfloatarg(5, cparams.width)

def get_base_path(infile):
    vasp = tkVASP()
    if os.path.isfile(infile):
        base_path = vasp.getdir(infile)
    else:
        base_path = infile

    return vasp, base_path
    
def check_exist(path, print_level = 0):
    f = os.path.isfile(path)
    if not f:
        if print_level:
            print(f"Error in check_file_type(): [{path}] does not exist")
        return None

    return True

def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 0):
    if not os.path.isfile(infile) and not os.path.isdir(infile):
        return f"Error: file [{infile}] does not exist"

    vasp, base_path = get_base_path(infile)

    if not check_exist(vasp.get_INCAR(base_path), print_level):
        return None
    if not check_exist(vasp.get_POSCAR(base_path), print_level):
        return None
    if not check_exist(vasp.get_POTCAR(base_path), print_level):
        return None
    if not check_exist(vasp.get_OUTCAR(base_path), print_level):
        return None
    if not check_exist(vasp.get_DOSCAR(base_path), print_level):
        return None
    if not check_exist(vasp.get_VASPPath(base_path, 'EIGENVAL'), print_level):
        return None

    return {"file_type": "VASP DOS"}

def get_output_path(infile, inf = None, app = None, cparams = None):
    vasp, base_path = get_base_path(infile)
    return os.path.join(base_path, "vasp2pes-pes.xlsx")
#    return replace_path(infile, os.path.join("{dirname}", "{filebody}-dos.xlsx"))

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, print_level = 0):
    vasp, base_path = get_base_path(infile)
    INCAR_path   = vasp.get_INCAR(base_path)
    POSCAR_path  = vasp.get_POSCAR(base_path)
    OUTCAR_path  = vasp.get_OUTCAR(base_path)
    DOSCAR_path  = vasp.get_VASPPath(base_path, 'DOSCAR')
    EIGENVAL_path  = vasp.get_VASPPath(base_path, 'EIGENVAL')

    if print_level:
        print("")
        print(f"infile: {infile}")
        print("base_path : ", base_path)
        print("  INCAR   : ", INCAR_path)
        print("  POSCAR  : ", POSCAR_path)
        print("  OUTCAR  : ", OUTCAR_path)
        print("  DOSCAR  : ", DOSCAR_path)
        print("  EIGENVAL: ", EIGENVAL_path)

    if not os.path.isfile(INCAR_path):
        app.terminate(f"Error: Can not read [{INCAR_path}]")

    if print_level:
        print("")
        print("*** Read crystal structure from [{}]".format(POSCAR_path))
    
    cry1 = vasp.read_poscar(POSCAR_path)
    if cry1 is None:
        app.terminate(f"Error: Can not read [{POSCAR_path}]", usage = usage)

    a1, b1, c1, alpha1, beta1, gamm1 = cry1.LatticeParameters()
    Vcell1 = cry1.Volume()
    if cry1 is None:
        cry1.PrintInf("cell")

    incarinf = vasp.read_incar_inf(INCAR_path)
    outcarinf = vasp.read_outcar_inf(OUTCAR_path)
    EF    = outcarinf["EF"]
    ISPIN = outcarinf["ISPIN"]
    if cry1 is None:
        print("")
        print("Information in OUTCAR:")
        print("  ISPIN: ", ISPIN)
        print("  EF = {} eV corrected to 0".format(EF))

    doscarinf = vasp.read_doscar(DOSCAR_path, unit = '/cm3', print_level = print_level)
    nE     = doscarinf["nE"]
    E      = doscarinf["E"]
    if nE == 0:
        Edmin = 0.0
        Edmax = 0.0
    else:
        Edmin  = min(E)
        Edmax  = max(E)

    if nE > 0 and print_level:
        print("")
        print("DOS E range: {:10.6g} - {:10.6g} eV, {} points".format(Edmin, Edmax, nE))

    eigenvalinf = vasp.read_eigenval(EIGENVAL_path, EF = EF)
    nk      = eigenvalinf["nk"]
    nLevels = eigenvalinf["nLevels"]
    bandedgeinf  = vasp.find_band_edges_from_eigenval(EF0 = 0.0, eigenvalinf = eigenvalinf, ISPIN = ISPIN, occ_th = cparams.occ_th)
#    bandedgeinf  = vasp.find_band_edges_from_eigenval(EF0 = EF + cparams.EF0, eigenvalinf = eigenvalinf, ISPIN = ISPIN, occ_th = cparams.occ_th)
#    bandedgeinf2 = vasp.gbandedges(fn_eig = EIGENVAL_path, ef = EF)

    if print_level >= 2:
        print("k points in EIGENVAL:")
        print("nk=", nk)
        print("nLevels=", nLevels)
        for i in range(nk):
            el = eigenvalinf['EList'][i]
            kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns = el
            print("  ({:8.4f} {:8.4f} {:8.4f}) w={:8.4g}  {:8.4f} {:12.4f}".format(kx, ky, kz, wk, dk, ktot))

    EV   = bandedgeinf["EV"]
    EC   = bandedgeinf["EC"]
    Eg   = bandedgeinf["Eg"]
#    EV1    = bandedgeinf2["EVBM"] - EF
#    EC1    = bandedgeinf2["ECBM"] - EF
#    Eg1    = bandedgeinf2["Eg"]
    if print_level:
        print(f"Band edge from EIGENVAL:")
        print(f"EF0={cparams.EF0:10.6f} eV")
        print(f"  band_edges: EV={EV:10.6f}  EC={EC:10.6f}  Eg={Eg:10.6f} eV")
#        print(f"  find_band_edges: EV={EV:10.6f}  EC={EC:10.6f}  Eg={Eg:10.6f} eV")
#        print(f"  gbandedges()   : EV={EV1:10.6f}  EC={EC1:10.6f}  Eg={Eg1:10.6f} eV")

    sample_name = doscarinf["SampleName"]
    data_list = doscarinf["data_list"]
    ndatalist = len(data_list)
    labels = doscarinf["labels"]
    Ne     = doscarinf["Ne"]
    Neup   = doscarinf["Neup"]
    Nedn   = doscarinf["Nedn"]

    if print_level >= 2:
        print("")
        print("Convolution:")
    for i in range(1, ndatalist):
        if print_level >= 2:
            print(f"  {i}-th data / {ndatalist-1}")
        data_list[i] = convolution(E, data_list[i], cparams.width, func_type = 'gauss')

    dirname, basename, filebody, ext = split_file_path(infile)
    inf = {}
    inf["data_list_type"] = "[[x, y]]"
    inf["filename"]    = infile
    inf["sample_name"] = sample_name
    inf["yscale"]      = 'linear'
    inf["meta"]        = [] #[inf_list]
    inf["data_list"]   = [data_list]
    inf["nregion"]     = 1
    inf["nspectrum"]   = len(data_list) - 1
    inf["ndata"]       = len(data_list[0])
    inf["xmin"]        = data_list[0][0]
    inf["xmax"]        = data_list[0][inf["ndata"] - 1]
    inf["xstep"]       = data_list[0][1] - data_list[0][0]
    inf["labels"]      = [labels]

    inf["Ne"]   = Ne
    inf["Neup"] = Neup
    inf["Nedn"] = Nedn

    inf["npdos"]          = doscarinf["npdos"]
    inf["atom_sites"]     = doscarinf["atom_sites"]
    inf["PDOS_atom_list"] = doscarinf["PDOS_atom_list"]
    inf["atom_types"]     = doscarinf["atom_types"]
    inf["lmlabels"]       = doscarinf["lmlabels"]
    inf["PDOS_list"]      = doscarinf["PDOS_list"]
    
    inf["IsSpinPolarized"] = doscarinf["IsSpinPolarized"]
    inf["ISPIN"]           = doscarinf["ISPIN"]
    inf["SpinOrbit"]       = doscarinf["SpinOrbit"]
    inf["IsNonCollinear"]  = doscarinf["IsNonCollinear"]
    inf["ks"]              = doscarinf["ks"]
    inf["EV"] = EV
    inf["EC"] = EC
    inf["Eg"] = Eg

    inf["cry"] = doscarinf["cry"]
    inf["eigenvalinf"] = eigenvalinf
    inf["bandedgeinf"] = bandedgeinf
    inf["doscarinf"] = doscarinf

    return inf

def print_data(inf, app = None, cparams = None, print_level = 1):
    doscarinf = inf["doscarinf"]
    print("")
    print("filename :", inf["filename"])
    print("IsSpinPolarized   :", doscarinf["IsSpinPolarized"])
    print("SpinOrbit   :", doscarinf["SpinOrbit"])
    print("IsSpinPolarized   :", doscarinf["IsSpinPolarized"])
    print("IsNonCollinear   :", doscarinf["IsNonCollinear"])
    print("ks:", doscarinf["ks"])

    print("")
    print("sample   :", inf["sample_name"])
    print("yscale   :", inf["yscale"])
    print("labels   :", inf["labels"])
    print("nspectrum:", inf["nspectrum"])
    print("ndata    :", inf["ndata"])
    print(f"E range : {inf['xmin']} - {inf['xmax']} degrees, {{inf['xstep']}} step")
    
    nspectrum = inf["nspectrum"]
    data_list = inf["data_list"][0]
    print("\t".join(inf["labels"][0]))
    for i in range(0, inf["ndata"], 50):
        v = [f"{data_list[ispectrum][i]:10.3g}" for ispectrum in range(nspectrum + 1)]
        print("\t".join(v))

def plot_data(inf, app = None, cparams = None, print_level = 1):
    vasp = tkVASP()
    
    data_list = inf["data_list"][0]
    nspectrum = len(data_list)
    npdos     = inf["npdos"]
    labels    = inf["labels"][0][:nspectrum]
    xE      = data_list[0]
    y_list  = data_list[1:]
    ndata   = len(xE)
    xlabel  = labels[0]
    ylabels = labels[1:]

    atom_types  = inf["atom_types"]
    natom_types = len(atom_types)
    pdos_list  = inf["PDOS_list"]
    lmlabels   = inf["lmlabels"]

    atom_sites = inf["atom_sites"]
    pdos_atom_list = inf["PDOS_atom_list"]

    PES       = inf["PES_spectrum"]
    PES_labels = [f'{atom} (PES)' for atom in PES.keys()]
    PES_data   = [PES[atom] for atom in PES.keys()]

    if cparams.xmin == '*':
        xmin = inf["xmin"]
    else:
        xmin = max([inf["xmin"], cparams.xmin])
    if cparams.xmax == '*':
        xmax = inf["xmax"]
    else:
        xmax = min([inf["xmax"], cparams.xmax])
    
    x0idx = 0
    for i in range(ndata):
        if xE[i] < cparams.xmin:
            x0idx = i
        else:
            break

    x1idx = ndata
    for i in range(ndata):
        if xE[i] > cparams.xmax:
            x1idx = i
        else:
            break

    print("")
    print("plot")
    print("ndata    :", ndata)
    print("ylabels  :", ylabels)
    print("E index range:", x0idx, x1idx)
    if npdos > 0:
        print("plot lm-decomposed DOS")
        print("  atom_types=", atom_types)
        print("  lmlabels=", lmlabels)

    fig, axes = plt.subplots(natom_types + 2, 1, figsize = figsize, sharex = True, gridspec_kw = {'hspace': 0})
    if npdos == 0:
        ax = axes
    else:
        ax = axes[0]

    plot_event = tkPlotEvent(plt)

    ax.set_title(f"{cparams.infile}")
    ax.tick_params(labelsize = fontsize)

    ymin_plot =  1.0e100
    ymax_plot = -1.0e100
    nlinestyles = len(linestyles)
    ncolors = len(colors)
    for i in range(len(ylabels)):
        y_plot = y_list[i][x0idx:x1idx - 1]
        ymin_plot = min([ymin_plot, min(y_plot)])
        ymax_plot = max([ymax_plot, max(y_plot)])
        color, ls = graph_style_idx(i)
        ax.plot(xE, y_list[i], label = ylabels[i], picker = True, 
                    linestyle = ls, color = color, linewidth = 1.0)

    print("E range:", xmin, xmax)
    print("y range:", ymin_plot, ymax_plot)
    ax.set_xlim(xmin, xmax)
    ax.set_ylim(ymin_plot, ymax_plot)

    ax.set_xlabel(r'Energy (eV)', fontsize = fontsize)
    ax.set_ylabel('DOS (states/cm$^3$)', fontsize = fontsize)
    ax.legend(fontsize = legend_fontsize)

    for i in range(natom_types):
        ax2 = axes[i + 1]
        ax2.tick_params(labelsize = fontsize)
        atom_name = atom_types[i]
        lmlabels = [f'{atom_name} tot', 's', 'p', 'd']
        ymin_plot =  1.0e100
        ymax_plot = -1.0e100
        for ilm in range(len(lmlabels)):
            lm = lmlabels[ilm]
            lmdos = vasp.cal_lmdecomposed_dos(inf, atom_name, lm, is_print = True)
            if lmdos is None:
                continue
    
            lmdos = convolution(xE, lmdos, cparams.width, func_type = 'gauss')
            y_plot = lmdos[x0idx:x1idx - 1]
            ymin_plot = min([ymin_plot, min(y_plot)])
            ymax_plot = max([ymax_plot, max(y_plot)])
            if 'tot' in lm:
                label = lm
            else:
                label = f"{atom_name} {lm}"
            color, ls = graph_style_idx(ilm)
            ax2.plot(xE, lmdos, label = label, picker = True, 
                linestyle = ls, color = color, linewidth = 1.0)
    
        ax2.set_ylabel(f"{atom_name} PDOS", fontsize = fontsize)
        ax2.set_ylim(ymin_plot, ymax_plot)
        ax2.legend(fontsize = legend_fontsize)

    ax = axes[-1]
    PES       = inf["PES_spectrum"]
    PES_labels = [f'{atom} (PES)' for atom in PES.keys()]
    PES_data   = [PES[atom] for atom in PES.keys()]
    ymin_plot =  1.0e100
    ymax_plot = -1.0e100
    for i, label in enumerate(PES_labels):
        color, ls = graph_style_idx(i)
        y_plot = PES_data[i][x0idx:x1idx - 1]
        ax.plot(xE, PES_data[i], label = label, picker = True, 
                linestyle = ls, color = color, linewidth = 1.0)
        ymin_plot = min([ymin_plot, min(y_plot)])
        ymax_plot = max([ymax_plot, max(y_plot)])
    ax.set_ylabel(f"PES", fontsize = fontsize)
    ax.set_ylim(ymin_plot, ymax_plot)
    ax.legend(fontsize = legend_fontsize)

    def on_legend_label_click(event):
        if hasattr(event.artist, 'get_text'):
            clicked_label = event.artist.get_text()  # クリックされた凡例ラベルのテキストを取得
        elif hasattr(event.artist, 'get_label'):
            clicked_label = event.artist.get_label()  # クリックされた凡例ラベルのテキストを取得
        else:
            return

        print(f"データ [{clicked_label}] がクリックされました")

#    plot_event.register_click(fig) # callback = lambda event: plot_event.onclick(event))
    plot_event.register_pick(fig) # callback = lambda event: plot_event.onclick(event))

    plt.tight_layout()
    plt.pause(0.1)
    
    input("Press ENTER to terminate>>\n")
    

def save_data(outfiles, inf, app = None, cparams = None, print_level = 0):
    data_list = inf["data_list"][0]
    nspectrum = len(data_list)
    labels    = inf["labels"][0][:nspectrum]
    PES       = inf["PES_spectrum"]
    PES_labels = [f'{atom} (PES)' for atom in PES.keys()]
    PES_data   = [PES[atom] for atom in PES.keys()]
    
#    print("nspectrum=", nspectrum, labels)
    outfile   = outfiles[0]

    if print_level:
        print("")
        print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, [labels, PES_labels], [data_list, PES_data])

    return True

def convert(inf, app = None, cparams = None, print_level = 0):
    vasp = tkVASP()
    
    print("")
    print("Convert PDOS to PES")

    data_list = inf["data_list"][0]
    nspectrum = len(data_list)
    npdos     = inf["npdos"]
    labels    = inf["labels"][0][:nspectrum]

    atom_types  = inf["atom_types"]
    natom_types = len(atom_types)
    pdos_list  = inf["PDOS_list"]
    lmlabels   = inf["lmlabels"]

    atom_sites = inf["atom_sites"]
    pdos_atom_list = inf["PDOS_atom_list"]

    xE      = data_list[0]
    y_list  = data_list[1:]
    ndata   = len(xE)
    xlabel  = labels[0]
    ylabels = labels[1:]

    print("")
    print(f"Read photo ionization cross section from [{cparams.pix_path}]")
    file_type = pix.check_file_type(cparams.pix_path, inf = None, app = app, cparams = cparams, print_level = 1)
    if file_type is None:
        app.terminate(f"Error: [{cparams.pix_path}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf_pix = pix.read_data(cparams.pix_path, cparams = cparams, print_level = 2)
    labels_pix = inf_pix["labels"][0]
    idx_an  = None
    idx_orb = None
    idx_xs  = None
    for i, l in enumerate(labels_pix):
        l = l.lower()
        if 'at' in l and 'n' in l:
            idx_an = i
        if 'orb' in l:
            idx_orb = i
        if 'cross' in l and 'section' in l:
            idx_xs = i

    print("labels=", labels)
    print("labels=", inf["labels"][0])
    print("lmlabels=",lmlabels)
    print("atom_types=", atom_types)
    print("pix labels=", inf_pix["labels"][0])
    print(f"pix idx of atomic number: {idx_an}")
    print(f"pix idx of orbital      : {idx_orb}")
    print(f"pix idx of cross section: {idx_xs}")

    an       = inf_pix["data_list"][0][idx_an]
    orbitals = inf_pix["data_list"][0][idx_orb]
    xs       = inf_pix["data_list"][0][idx_xs]
    print("pix orbitals=", orbitals)

    pix_dict = {}
    for at in atom_types:
        for i, o in enumerate(orbitals):
            aa = o.split()
            if aa[0] == at:
                if pix_dict.get(at, None) is None:
                    pix_dict[at] = {aa[1]: xs[i]}
                else:
                    pix_dict[at][aa[1]] = xs[i]

    print("")
    print("Photo ionization cross section found:")
    for atom in pix_dict.keys():
        print(f"  {atom}:")
        for o in pix_dict[atom].keys():
            print(f"    {o}: {pix_dict[atom][o]}")

    print("")
    print("Calculate PES")
    PES = {'total': np.zeros(ndata, dtype = float)}
    for atom in atom_types:
        PES[atom] = np.zeros(ndata, dtype = float)
    for i in range(natom_types):
        atom = atom_types[i]
        lmlabels = ['s', 'p', 'd', 'f']
        for ilm in range(len(lmlabels)):
            lm = lmlabels[ilm]
            if lm[0] == 's' or lm[0] == 'p' or lm[0] == 'd' or lm[0] == 'f':
                pass
            else:
                print(f"  oribital {lm} is skipped")
                continue

            print(f"PDOS for {atom} {lm}:")
            xs = None
            if pix_dict.get(atom, None) is None:
                print(f"  Error: Can not find atom [{atom}] in pix_dict. Terminate")
                exit()

            for o in pix_dict[atom].keys():
                if lm[0] in o:
                    xs = pix_dict[atom][o]

            if xs is None:
                print(f"  Warning: Can not find orbital [{atom} {lm}] in pix_dict. Skipped.")
                continue
            
            print(f"Cross section for {atom} {lm} is taken from pix_dict {o}: {xs}")

            lmdos = vasp.cal_lmdecomposed_dos(inf, atom, lm, is_print = True)
            if lmdos is None:
                continue
    
            lmdos = convolution(xE, lmdos, cparams.width, func_type = 'gauss')
            
            for i, d in enumerate(lmdos):
                I = xs * d
                PES['total'][i] += I
                PES[atom][i] += I

#    for atom in atom_types:
#        print(f"{atom}:", max(PES[atom]), PES[atom])

    inf["PES_spectrum"] = PES

    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    vasp, base_path = get_base_path(cparams.infile)

    print("")
    print( "==========================================================================")
    print(" Convert VASP PDOS output to PES spectrum")
    print( "==========================================================================")
    print(f"Input file: {cparams.infile}")
    print(f"base_path : {base_path}")

    logfile = os.path.join(base_path, "vasp2dos-out.txt")
#    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile, print_level = 1)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams, print_level = 2)
    print_data(inf)
    inf = convert(inf, cparams = cparams, print_level = 1)
    save_data([cparams.outfile], inf, cparams = cparams, print_level = 1)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
