import os
import sys
import numpy as np


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkcrystal.tkvasp import tkVASP
from tklib.tkcrystal.tkcif import tkCIF, tkCIFData


"""
Convert VASP POSCAR to CIF
"""


usage_str = '''
"  (i) usage: python {} POSCAR outfile".format(sys.argv[0])
"         ex: python {} {} {}".format(sys.argv[0], cparams.infile, cparams.outfile)
'''[1:-1]


plugin_ver = "structure:0.2"


#================================
# global parameters
#================================
default_ext = ""
input_type  = "VASP POSCAR"
output_type = "CIF"


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0

    cparams.infile  = 'POSCAR'
    cparams.outfile = 'POSCAR.cif'


def update_vars(app = None, cparams = None):
#    if getarg(1, None) is None:
#        app.terminate(usage = usage)
    
    cparams.infile  = getarg(1, cparams.infile)
    cparams.outfile = getarg(2, cparams.outfile)

def get_base_path(infile):
    vasp = tkVASP()
    if os.path.isfile(infile):
        base_path = vasp.getdir(infile)
    else:
        base_path = infile

    return vasp, base_path
    
def check_exist(path, print_level = 0):
    f = os.path.isfile(path)
    if not f:
        if print_level:
            print(f"Error in check_file_type(): [{path}] does not exist")
        return None

    return True

def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 0):
    if not os.path.isfile(infile) and not os.path.isdir(infile):
        return None  #f"Error: file [{infile}] does not exist"

    if infile != 'POSCAR' and infile != 'CONTCAR':
        return None

    return {"file_type": "VASP POSCAR"}

def get_output_path(infile, inf = None, app = None, cparams = None):
    return cparams.outfile

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, print_level = 0):
    if print_level >= 1:
        print("")
        print("POSCAR path: ", infile)

    vasp = tkVASP()
    cry = vasp.read_poscar(infile, is_print = 1)
#    a, b, c, alpha, beta, gamm = cry.LatticeParameters()
    if print_level >= 1:
        cry.PrintInf("cell")

    inf = {}
    inf["data_list_type"] = "[tkCIF]"
    inf["filename"]    = infile
    inf["sample_name"] = cry.ChemicalComposition()
    inf["meta"]        = [] #[inf_list]
    inf["cry"]         = [cry]

    return inf

def print_data(inf, app = None, cparams = None, print_level = 1):
    cry = inf["cry"][0]
    if print_level:
        cry.PrintInf()

def plot_data(inf, app = None, cparams = None, print_level = 1):
    print()
    print("Not plot data in this plugin")

def save_data(outfiles, inf, app = None, cparams = None, print_level = 0):
    cry = inf["cry"][0]

    if type(outfiles) is str:
        cifpath = outfiles
    else:
        cifpath = outfiles[0]

    print("")
    print("Write to [{}]".format(cifpath))
    cif = tkCIFData()
    cif.CreateCIFFileFromCCrystal(cry, cifpath)

    return True

def convert(inf, app = None, cparams = None, print_level = 0):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

#    vasp, base_path = get_base_path(cparams.infile)
    base_path = os.path.dirname(cparams.infile)

    print("")
    print( "==========================================================================")
    print(" Convert VASP POSCAR to CIF")
    print( "==========================================================================")
    print(f"Input POSCAR file: {cparams.infile}")
#    print(f"base_path : {base_path}")

    logfile = os.path.join(base_path, "poscar2cif-out.txt")
#    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile, cparams = cparams)
    print(f"Output CIF file : {cparams.outfile}")

    file_type = check_file_type(cparams.infile, print_level = 1)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams, print_level = 2)
    print_data(inf)
#    inf = convert(inf, cparams = cparams, print_level = 1)
    save_data([cparams.outfile], inf, cparams = cparams, print_level = 1)
#    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
