import os
import sys
import numpy as np
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import colors
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkvariousdata import tkVariousData
from tklib.tkgraphic.tkplotevent import tkPlotEvent
from tklib.tksci.tkconvolution import convolution, convolve_func
from tklib.tkcrystal.tkvasp import tkVASP
from tklib.tkcrystal.tkbandstructure import plot_band


"""
Convert VASP output files to Band structure
"""

usage_str = '''
"  (i) usage: python {} output_dir".format(sys.argv[0])
"         ex: python {} {}".format(sys.argv[0], cparams.infile)
'''[1:-1]


plugin_ver = "band:0.2"  # data_list_type = [[x, y]]


#================================
# global parameters
#================================
default_ext = ""
input_type  = "VASP output"
output_type = "Band"

figsize = (6, 8)
#figsize = (6, 4)

#colors = ['#000000', '#ff0000', '#00aa00', '#0000ff', '#aaaa00', '#ff00ff', '#00ffff', '#aa0000', '#00aa00', '#0000aa']
#colors = ['k', 'r', 'g', 'b', 'y', 'm', 'c']
fontsize        = 16
labelfontsize   = 12
legend_fontsize = 8


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0

    cparams.infile = ''

# Energy to search edge energies
    cparams.EF0  = 0.1
# Max number of electrons occupies one state
    cparams.Nemax = 1.0
#    cparams.mode = 'bandocc'
    cparams.occ_th = 0.5

# mode: 'band', 'bandline', 'bandocc'
    cparams.validmodes = ['band', 'bandline', 'bandocc']
    cparams.mode = 'band'

# Plot configuration
    cparams.band_marker_size       = 4
    cparams.band_marker_edge_width = 0.5

    cparams.xmin = -10.0  # eV
    cparams.xmax =  10.0  # eV

# Occupancy threshold to separate HOMO and LUMO
    cparams.occ_th = 0.5

    cparams.save_figure = 1
    cparams.plot_figure = 1


def update_vars(app = None, cparams = None):
    if getarg(1, None) is None:
        app.terminate(usage = usage)
    
    cparams.infile = getarg(1, cparams.infile)
    cparams.xmin   = getfloatarg(2, cparams.xmin)
    cparams.xmax   = getfloatarg(3, cparams.xmax)

def get_base_path(infile):
    vasp = tkVASP()
    if os.path.isfile(infile):
        base_path = vasp.getdir(infile)
    else:
        base_path = infile

    return vasp, base_path
    
def check_exist(path, print_level = 0):
    f = os.path.isfile(path)
    if not f:
        if print_level:
            print(f"Error in check_file_type(): [{path}] does not exist")
        return None

    return True

def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 0):
    if not os.path.isfile(infile) and not os.path.isdir(infile):
        return f"Error: file [{infile}] does not exist"

    vasp, base_path = get_base_path(infile)

    if not check_exist(vasp.get_INCAR(base_path), print_level):
        return None
    if not check_exist(vasp.get_POSCAR(base_path), print_level):
        return None
    if not check_exist(vasp.get_POTCAR(base_path), print_level):
        return None
    if not check_exist(vasp.get_OUTCAR(base_path), print_level):
        return None
    if not check_exist(vasp.get_VASPPath(base_path, 'EIGENVAL'), print_level):
        return None

    return {"file_type": "VASP Band"}

def get_output_path(infile, inf = None, app = None, cparams = None):
    vasp, base_path = get_base_path(infile)
    return os.path.join(base_path, "vasp2dos-band.xlsx")
#    return replace_path(infile, os.path.join("{dirname}", "{filebody}-band.xlsx"))

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, print_level = 0):
    vasp = tkVASP()
    
    vasp, base_path = get_base_path(infile)
    INCAR_path    = vasp.get_INCAR(base_path)
    POSCAR_path   = vasp.get_POSCAR(base_path)
    KPOINTS_path  = vasp.get_VASPPath(base_path, 'KPOINTS')
    CONTCAR_path  = vasp.get_CONTCAR(base_path)
    OUTCAR_path   = vasp.get_OUTCAR(base_path)
    EIGENVAL_path = vasp.get_VASPPath(base_path, 'EIGENVAL')
    DOSCAR_path   = vasp.get_VASPPath(base_path, 'DOSCAR')
    if print_level:
        print("")
        print("CAR dir(ideal) : ", base_path)
        print("  INCAR   : ", INCAR_path)
        print("  POSCAR  : ", POSCAR_path)
        print("  KPOINTS : ", KPOINTS_path)
        print("  CONTCAR : ", CONTCAR_path)
        print("  OUTCAR  : ", OUTCAR_path)
        print("  EIGENVAL: ", EIGENVAL_path)
        print("  DOSCAR  : ", DOSCAR_path)
        print("")
        print(f"Occupancy threshold to seprate HOMO and LUMO: {cparams.occ_th}")
        print(f"save_figure: {cparams.save_figure}")
        print(f"plot_figure: {cparams.plot_figure}")

        print("")
        print("EF0: {} eV".format(cparams.EF0))

    if print_level:
        print("")
        print("*** Read crystal structure from [{}]".format(POSCAR_path))
    cry1 = vasp.read_poscar(POSCAR_path)
    if cry1 is None:
        terminate("Error: Can not read [{}]".format(POSCAR_path), usage = usage)

    a1, b1, c1, alpha1, beta1, gamm1 = cry1.LatticeParameters()
    cry1.PrintInf("cell")

    incarinf = vasp.read_incar_inf(INCAR_path)
    outcarinf = vasp.read_outcar_inf(OUTCAR_path)
    EF    = outcarinf["EF"]
    ISPIN = outcarinf["ISPIN"]
    IsHF  = outcarinf["IsHF"]
    if print_level:
        print("")
        print("Information in OUTCAR:")
        print("  ISPIN  : ", ISPIN)
        print("  IsHF: ", IsHF)
        print("  EF = {} eV corrected to 0".format(EF))

    kpoints_inf = vasp.read_kpoints(KPOINTS_path)
#    print("kpoints: ", kpoints_inf)
#    print("kpoints: ", kpoints_inf['kpoints'])
    nk = kpoints_inf['nk']
    kpoints = kpoints_inf['kpoints']
    kptdic  = kpoints_inf['kpointsdic']

    if print_level:
        print("")
        print("Special k points in KPOINTS:")
        if IsHF == 0:
            kp = kptdic[0]
            kx0, ky0, kz0, kname0 = kp["kx0"], kp["ky0"], kp["kz0"], kp["kname0_conv"]
            kx1, ky1, kz1, kname1 = kp["kx1"], kp["ky1"], kp["kz1"], kp["kname1_conv"]
            w, dk, ktot = kp["kw"], 0.0, 0.0
            if kname0 == '':
                kname0 = 'no_name'
            print("  {:10} ({:8.4f} {:8.4f} {:8.4f}) w={:.3f}  {:8.4f} {:12.4f}".format(kname0, kx0,  ky0, kz0, w, dk, ktot))
        for i in range(len(kptdic)):
            kp = kptdic[i]
            kx0, ky0, kz0, kname0 = kp["kx0"], kp["ky0"], kp["kz0"], kp["kname0_conv"]
            kx1, ky1, kz1, kname1 = kp["kx1"], kp["ky1"], kp["kz1"], kp["kname1_conv"]
            w, dk, ktot = kp["kw"], kp["dk"], kp["ktot"]
            if kname1 == '':
                kname1 = 'no_name'
            print("  {:10} ({:8.4f} {:8.4f} {:8.4f}) w={:.3f}  {:8.4f} {:12.4f}".format(kname1, kx1,  ky1, kz1, w, dk, ktot))

    if print_level:
        print("")
        print("k points in KPOINTS:")

    if IsHF == 0:
        kp = kpoints[0]
        kx0, ky0, kz0, kname0 = kp["kx0"], kp["ky0"], kp["kz0"], kp["kname0_conv"]
        kx1, ky1, kz1, kname1 = kp["kx1"], kp["ky1"], kp["kz1"], kp["kname1_conv"]
        w, dk, ktot = kp["kw"], 0.0, 0.0
        if print_level:
            print("  {:10} ({:8.4f} {:8.4f} {:8.4f}) w={:.3f}  {:8.4f} {:12.4f}".format(kname0, kx0, ky0, kz0, w, dk, ktot))

    if print_level:
        for i in range(nk):
            kp = kpoints[i]
            kx0, ky0, kz0, kname0 = kp["kx0"], kp["ky0"], kp["kz0"], kp["kname0_conv"]
            kx1, ky1, kz1, kname1 = kp["kx1"], kp["ky1"], kp["kz1"], kp["kname1_conv"]
            w, dk, ktot = kp["kw"], kp["dk"], kp["ktot"]
            print("  {:10} ({:8.4f} {:8.4f} {:8.4f}) w={:.3f}  {:8.4f} {:12.4f}".format(kname1, kx1,  ky1, kz1, w, dk, ktot))

    eigenvalinf = vasp.read_eigenval(EIGENVAL_path, KPOINTS_inf = kpoints_inf, ISPIN = ISPIN, IsHF = IsHF, EF = EF)
    if print_level:
        print("eigenvalinf=", eigenvalinf)
    nk      = eigenvalinf["nk"]
    nLevels = eigenvalinf["nLevels"]
    EList = eigenvalinf['EList']
    nk_band = len(EList)
    if print_level:
        print("")
        print("k points in EIGENVAL:")
        print("nk(band)=", nk_band)
        print("nLevels =", nLevels)
    if print_level >= 2:
        for i in range(nk_band):
            el = EList[i]
            kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns = el
            print(f"  ({kx:8.4f} {ky:8.4f} {kz:8.4f}) w={wk:8.4g}  dk={dk:8.4f} ktot={ktot:8.4f}")

    bandedgeinf = vasp.find_band_edges_from_eigenval(
            EF0 = cparams.EF0, eigenvalinf = eigenvalinf, ISPIN = ISPIN, occ_th = cparams.occ_th)
#    bandedgeinf2 = vasp.gbandedges(CAR_path, OUTCAR_path, EIGENVAL_path, EF)
#    if print_level:
#        print("bandedgeinf=", bandedgeinf)
    EV   = bandedgeinf["EV"]
    EC   = bandedgeinf["EC"]
    Eg   = bandedgeinf["Eg"]
    EHOMO = bandedgeinf["EHOMO"]
    ELUMO = bandedgeinf["ELUMO"]
#    EV    = bandedgeinf2["EVBM"] - EF
#    EC    = bandedgeinf2["ECBM"] - EF
#    Eg    = bandedgeinf2["Eg"]
    print("")
    print(f"Band edge from EIGENVAL:")
    print(f"EF from OUTCAR: {EF:10.6f} eV")
    print(f"EF0={cparams.EF0:10.6f} eV")
    print(f"  band_edges: EV={EV:10.6f}  EC={EC:10.6f}  Eg={Eg:10.6f} eV")
    print(f"             HOMO:{EHOMO:10.6f} eV")
    print(f"             LUMO:{ELUMO:10.6f} eV")
#    print(f"  gbandedges()   : EV={EV:10.6f}  EC={EC:10.6f}  Eg={Eg:10.6f} eV")
#    print(f"                         EHOMO={EHOMO+EF:10.6f}  ELUMO={ELUMO+EF:10.6f}")

    xk   = []
    kvec = []
    Eup  = []
    Nup  = []
    Edn  = []
    Ndn  = []
    for i in range(len(EList)):
        el = EList[i]
        kx0, ky0, kz0, wk, dk, ktot, Eups, occups, Edns, occdns = [*el]
        xk.append(ktot)
        kvec.append([kx0, ky0, kz0, dk, ktot])
        Eup.append(Eups)
        Nup.append(occups)
        if ISPIN == 2:
            Edn.append(Edns)
            Ndn.append(occdns)

    sample_name = incarinf["SYSTEM"]
    dirname, basename, filebody, ext = split_file_path(infile)
    inf = {}
    inf["data_list_type"] = "[[x, y]]"
    inf["filename"]    = infile
    inf["sample_name"] = sample_name
    inf["yscale"]      = 'linear'
    inf["meta"]        = [] #[inf_list]
    """
    inf["data_list"]   = [data_list]
    inf["nregion"]     = 1
    inf["nspectrum"]   = len(data_list) - 1
    inf["ndata"]       = len(data_list[0])
    inf["xmin"]        = data_list[0][0]
    inf["xmax"]        = data_list[0][inf["ndata"] - 1]
    inf["xstep"]       = data_list[0][1] - data_list[0][0]
    inf["labels"]      = []
    """
    
    inf["nk"]      = nk
    inf["nLevels"] = nLevels
    inf["nk_band"] = nk_band
    inf["kpoints"] = kpoints
    inf["kptdic"]  = kptdic

    inf["EList"]   = EList
    inf["kvec"] = kvec
    inf["ktot"] = xk
    inf["Eup"] = Eup
    inf["Edn"] = Edn
    inf["Nup"] = Nup
    inf["Ndn"] = Ndn

    inf["EF"]    = EF
    inf["ISPIN"] = ISPIN
    inf["IsHF"]  = IsHF

    inf["incarinf"]    = incarinf
    inf["outcarinf"]   = outcarinf
    inf["kpoints_inf"] = kpoints_inf
    inf["eigenvalinf"] = eigenvalinf
    inf["cry"]         = cry1

    inf["EV"] = EV
    inf["EC"] = EC
    inf["Eg"] = Eg
    inf["ELUMO"] = ELUMO
    inf["EHOMO"] = EHOMO

    return inf

def print_data(inf, app = None, cparams = None, print_level = 1):
    nk      = inf["nk"]
    nLevels = inf["nLevels"]
    EList   = inf['EList']
    nk_band = len(EList)
    EC    = inf['EC']
    EV    = inf['EV']
    Eg    = inf['Eg']
    ELUMO = inf["ELUMO"]
    EHOMO = inf["EHOMO"]

    print("")
    print("filename :", inf["filename"])
    print("ISPIN   :", inf["ISPIN"])
    print("IsHF   :", inf["IsHF"])

    print("")
    print("k points in EIGENVAL:")
    print("nk(band)=", nk_band)
    print("nLevels =", nLevels)
    print(f"band_edges: EV={EV:10.6f}  EC={EC:10.6f}  Eg={Eg:10.6f} eV")
    if print_level:
        for i in range(nk_band):
            el = EList[i]
            kx, ky, kz, wk, dk, ktot, Eups, occups, Edns, occdns = el
            print(f"  ({kx:8.4f} {ky:8.4f} {kz:8.4f}) w={wk:8.4g}  dk={dk:8.4f} ktot={ktot:8.4f}")

def plot_data(inf, app = None, cparams = None, print_level = 1):
    vasp = tkVASP()

    ISPIN = inf["ISPIN"]
    EC    = inf['EC']
    EV    = inf['EV']
    Eg    = inf['Eg']
    ELUMO = inf["ELUMO"]
    EHOMO = inf["EHOMO"]
    Emin = cparams.xmin
    Emax = cparams.xmax

    kptdic = inf["kptdic"]
    EList = inf["EList"]
    xk    = inf["ktot"]
    kvec  = inf["kvec"]
    yEup  = inf["Eup"]
    yNup  = inf["Nup"]
    yEdn  = inf["Edn"]
    yNdn  = inf["Ndn"]

    print("")
    print("plot band structure")

    fig = plt.figure(figsize = figsize)
    axband = fig.add_subplot(1, 1, 1)

# バンド構造をプロット
# xk: プロットするk点の蓄積距離のリスト
# yE: E(xk[i])。入れ子になったリストで構わない
# ktotallist: k点境界における、最初のk点からの距離の和のリスト
# ktotal_namelist: k点境界における、k点の名称
#def plot_band(axis, xk, yE, Erange, ktotallist, ktotal_namelist):
    if len(kptdic) == 0:
        ktotallist      = []
        ktotal_namelist = []
    else:
        kp = kptdic[0]
        kname0 = kp["kname0"]
        ktotallist      = [0.0]
        ktotal_namelist = [vasp.convert_kname(kname0)]
        for i in range(len(kptdic)):
            kp = kptdic[i]
            kx0, ky0, kz0, kname0 = kp["kx0"], kp["ky0"], kp["kz0"], kp["kname0"]
            kx1, ky1, kz1, kname1 = kp["kx1"], kp["ky1"], kp["kz1"], kp["kname1"]
            ktot = kp["ktot"]
            ktotallist.append(ktot)
            ktotal_namelist.append(vasp.convert_kname(kname1))

    if cparams.mode == 'bandocc':
        plot_band(plt, axband, ISPIN, xk, yEup, yEdn, [Emin, Emax], occups = yNup, occdns = yNdn, 
                ktotallist = ktotallist, ktotal_namelist = ktotal_namelist, 
                EV = EV, EC = EC, EF = cparams.EF0, EHOMO = EHOMO, ELUMO = ELUMO,
                EFlabel = '$E_{F0}$',
                marker = 'o', linestyle = '',
                markersize = cparams.band_marker_size, markeredgewidth = cparams.band_marker_edge_width,
                legendloc = 'lower right')
    elif cparams.mode == 'bandline':
        plot_band(plt, axband, ISPIN, xk, yEup, yEdn, [Emin, Emax], 
                ktotallist = ktotallist, ktotal_namelist = ktotal_namelist, 
                EV = EV, EC = EC, EF = cparams.EF0, EHOMO = EHOMO, ELUMO = ELUMO,
                EFlabel = '$E_{F0}$',
                marker = None, linestyle = '-',
                legendloc = 'lower right')
    elif cparams.mode == 'band':
        plot_band(plt, axband, ISPIN, xk, yEup, yEdn, [Emin, Emax],
                ktotallist = ktotallist, ktotal_namelist = ktotal_namelist, 
                EV = EV, EC = EC, EF = cparams.EF0, EHOMO = EHOMO, ELUMO = ELUMO,
                EFlabel = '$E_{F0}$',
                marker = 'o', linestyle = '',
                markersize = cparams.band_marker_size, markeredgewidth = cparams.band_marker_edge_width,
                legendloc = 'lower right')
    
# Rearange the graph axes so that they are not overlapped
    plt.tight_layout()

    if cparams.save_figure:
        outfig_path = 'band.png'
        if print_level:
            print("")
            print(f"Save band structure to figure file [{outfig_path}]")
        plt.savefig(outfig_path, dpi = 300, transparent = True)

    cparams.use_pause = True
    if cparams.plot_figure:
        if cparams.use_pause:
            plt.pause(0.1)
            print("")
            input("Press ENTER to exit>>")
        else:
            plt.show()
            print("")
            print("Close graph window to terminate")

def save_data(outfiles, inf, app = None, cparams = None, print_level = 0):
    outfile = outfiles[0]

    ISPIN = inf["ISPIN"]

    xk   = inf["ktot"]
    kvec = inf["kvec"]
    Eup  = inf["Eup"]
    Edn  = inf["Edn"]
#    Nup = inf["Nup"]
#    Ndn = inf["Ndn"]
    nk = len(kvec)
    nL = len(Eup[0])

    labels = ["kx", "ky", "kz", "dk", "ktot"]
    for iL in range(nL):
        labels.extend([f"Eup({iL})"])
    if ISPIN == 2:
        for iL in range(nL):
            labels.extend([f"Edn({iL})"])

    kx   = [kvec[ik][0] for ik in range(nk)]
    ky   = [kvec[ik][1] for ik in range(nk)]
    kz   = [kvec[ik][2] for ik in range(nk)]
    dk   = [kvec[ik][3] for ik in range(nk)]
    ktot = [kvec[ik][4] for ik in range(nk)]
    data_list = [kx, ky, kz, dk, ktot]
    for iL in range(nL):
        yup = [Eup[ik][iL] for ik in range(nk)]
        data_list.extend([yup])
        if ISPIN == 2:
            ydn = [Edn[ik][iL] for ik in range(nk)]
            data_list.extend([ydn])

    print("")
    print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, labels, data_list)

    return True

def convert(inf, app = None, cparams = None, print_level = 0):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    vasp, base_path = get_base_path(cparams.infile)

    print("")
    print( "==========================================================================")
    print(" Convert CIF file to powder XRD pattern")
    print( "==========================================================================")
    print(f"Input file: {cparams.infile}")
    print(f"base_path : {base_path}")

    logfile = os.path.join(base_path, "vasp2dos-out.txt")
#    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile, print_level = 1)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams, print_level = 2)
    print_data(inf)
    inf = convert(inf, cparams = cparams, print_level = 1)
    save_data([cparams.outfile], inf, cparams = cparams, print_level = 1)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
