import os
import sys
import numpy as np
import matplotlib.pyplot as plt

#from pymatgen.core import Composition


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import get_ext, getarg, getintarg, getfloatarg, pint, pfloat, terminate
from tklib.tkutils import split_file_path, replace_path
from tklib.tkcrystal.tkpymatgen import tkPymatgenVASP, tkPymatgen

"""
Convert VASP POSCAR/vasprun.xml to CIF
"""


usage_str = '''
"  (i) usage: python {} POSCAR/vasprun.xml outfile".format(sys.argv[0])
"         ex: python {} {} {}".format(sys.argv[0], cparams.infile, cparams.outfile)
'''[1:-1]


plugin_ver = "structure:0.2"


#================================
# global parameters
#================================
default_ext = ""
input_type  = "VASP POSCAR/vasprun.xml"
output_type = "CIF"


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0

    cparams.infile  = 'POSCAR'
    cparams.outfile = 'POSCAR.cif'

    cparams.symprec = 1.0e-5

def update_vars(app = None, cparams = None):
#    if getarg(1, None) is None:
#        app.terminate(usage = usage)
    
    cparams.infile         = getarg(1, cparams.infile)
    cparams.outfile        = getarg(2, cparams.outfile)
    cparams.symprec        = getfloatarg(3, cparams.symprec)
    cparams.add_relaxation = getintarg(4, 0) #cparams.add_relaxation)
    cparams.min_dE         = getfloatarg(5, 1.0e-4) #cparams.add_relaxation)
    cparams.coordinate     = getfloatarg(6, "fc")   # f for fractional coordinate, c for cartesian
    cparams.site_name      = getfloatarg(7, "as")   # a for atom name + index, s for site + index
    cparams.validate_vasprun = getfloatarg(8, False)   # a for atom name + index, s for site + index

def get_base_path(infile):
    vasp = tkVASP()
    if os.path.isfile(infile):
        base_path = vasp.getdir(infile)
    else:
        base_path = infile

    return vasp, base_path

def check_exist(path, print_level = 0):
    f = os.path.isfile(path)
    if not f:
        if print_level:
            print(f"Error in check_file_type(): [{path}] does not exist")
        return None

    return True

def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 0):
    if not os.path.isfile(infile) and not os.path.isdir(infile):
        return None  #f"Error: file [{infile}] does not exist"

    ext = get_ext(infile).lower()
    if ext == '.xml':
        return {"file_type": "VASP vasprun.xml"}

    if infile != 'POSCAR' and infile != 'CONTCAR':
        return None

    return {"file_type": "VASP POSCAR"}

def get_output_path(infile, inf = None, app = None, cparams = None):
    cparams.outfile = app.replace_path(infile, template = ["{dirname}", "{filebody}.cif"])
    return cparams.outfile

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, print_level = 0):
    infile = os.path.abspath(infile)
    dirname, basename, filebody, ext = split_file_path(infile, check_dir = True)

    add_relaxation = cparams.get("add_relaxation", 0)
    min_dE = cparams.get("min_dE", 1.0e-4)

    if print_level >= 1:
        print("")
        print("VASP structure file path: ", infile)

    vasp = tkPymatgenVASP()
    pm   = tkPymatgen()

    vasprun = vasp.read_vasprun(infile, validate = cparams.get("validate_vasprun", False))
    if not vasprun:
        print()
        app.terminate("Error in vasprun2inputs.read_data(): Could not read [{infile}]", pause = True)
        
    inf, structure = vasp.read_structure(infile, vasprun)
    
    symprec = cparams.get("symprec", 1.0e-4)
    iSPG, SPG_name, symmetrized_structure = pm.symmetrize(structure, symprec)
    if print_level:
        print(f"add_relaxation   : {add_relaxation}")
        print(f"min_dE           : {min_dE} eV")
        print("Space group number: ", iSPG)
        print("Space group name  : ", SPG_name)
        print("symmetrized_structure:")
        print(symmetrized_structure)

    incar_params = vasprun.incar
    system = incar_params.get("SYSTEM", "blank SYSTEM")

    final_structure = vasprun.final_structure
    formula, reduced_formula = pm.chemical_formula(final_structure)
    natoms = len(final_structure)
    '''
    latt = structure.lattice
    composition     = Composition(reduced_formula)
    formula_weight  = composition.weight
    num_atoms_in_formula = composition.num_atoms
    '''

    inf = {}
    inf["data_list_type"] = "[tkCIF]"
    inf["filename"]    = infile
    inf["sample_name"] = reduced_formula
    inf["meta"]        = [] #[inf_list]
    inf["pymatgen"]    = pm
    inf["vasp"]        = vasp
    inf["vasprun"]     = vasprun
    inf["input_inf"]   = inf
    inf["initial_structure"] = vasprun.initial_structure
    inf["final_structure"]   = final_structure
    inf["structure_original"] = [structure]
    inf["structure"]          = [symmetrized_structure]

    keyval = []
    if add_relaxation:
        nsteps = len(vasprun.ionic_steps)
        prev_E = None
        for i, step in enumerate(vasprun.ionic_steps):
            Efree    = pfloat(step['e_fr_energy'])
            Etot_woS = pfloat(step['e_wo_entrp'])
            Etot     = pfloat(step['e_0_energy'])
            _inf = pm.structure2inf_flat(step['structure'], symprec, 
                        coordinate = cparams.coordinate, site_name = cprams.site_name)

            _inf.update({
                "file_path"      : infile,
                "dir_name"       : dirname,
                "file_name"      : basename,
                "system"         : system,
                "converged"      : vasprun.converged,
                "converged_electronic": vasprun.converged_electronic,
                "converged_ionic": vasprun.converged_ionic,
                "i_ion_step"     : i+1,
                "Etot"           : Etot,
                "Efree"          : Efree,
                "Etot_woS"       : Etot_woS,
                "Etot_atom"      : Etot / natoms,
                "Efree_atom"     : Efree / natoms,
                "Etot_woS_atom"  : Etot_woS / natoms,
                })

            if prev_E is None or i == nsteps - 1:
                keyval.append(_inf)
            elif abs(Etot - prev_E) > min_dE:
                keyval.append(_inf)
            prev_E = Etot
    else:
        Etot = pfloat(vasprun.final_energy)
        _inf = pm.structure2inf_flat(final_structure, symprec,
                        coordinate = cparams.get("coordinate", "fc"), site_name = cparams.get("site_name", "as"))
        _inf.update({
            "file_path"      : infile,
            "dir_name"       : dirname,
            "file_name"      : basename,
            "system"         : system,
            "converged"      : vasprun.converged,
            "converged_electronic": vasprun.converged_electronic,
            "converged_ionic": vasprun.converged_ionic,
            "i_ion_step"     : 0,
            "Etot"           : Etot,
            "Etot_atom"      : Etot / natoms,
            "EF"             : vasprun.efermi,
            })
        keyval.append(_inf)

    inf["filenames"]   = [infile]
    inf["keyval_list"] = keyval

    return inf

def print_data(inf, app = None, cparams = None, print_level = 1):
    vasprun   = inf["vasprun"]
    keyval    = inf["keyval_list"]
    structure = inf["structure"]
    print()
    print("Final structure:")
#    print(structure[-1])

    latt = structure[-1].lattice
    print(f"Lattice constants: {latt.a:12.6g} {latt.b:12.6g} {latt.c:12.6g}  {latt.alpha:12.6g} {latt.beta:12.6g} {latt.gamma:12.6g}")
    print("Atomic sites:")
    for site in structure[-1]:
        element = site.specie
        xyz  = site.frac_coords
        xyzc = site.coords
        print(f"  {element:2} {xyz}, {xyzc} A")

    print()
    print("Etot(final)=", keyval[-1]["Etot"])
    print("efermi=", keyval[-1].get("EF", None))
#    print("dielectric_data=", vasprun.dielectric_data)
#    print("actual_kpoints=", vasprun.actual_kpoints)
#    print("actual_kpoints_weights=", vasprun.actual_kpoints_weights)

    if print_level >= 2:
        print()
        print("initial_structure:")
        print(vasprun.initial_structure)
        if hasattr(vasprun, "final_structure"):
            print()
            print("final_structure:")
            print(vasprun.final_structure)

    print()
    print("Ionic states:")
#    print("vasprun.ionic_steps=", vasprun.ionic_steps)
    for i, step in enumerate(vasprun.ionic_steps):
        e_fr_energy = step['e_fr_energy']
        e_wo_entrp  = step['e_wo_entrp']
        e_0_energy  = step['e_0_energy']
        structure   = step['structure']
        print(f"#{i:3}: Efree={e_fr_energy:20.16g} Efree/woS={e_wo_entrp:20.16g} E0={e_0_energy:20.16g}")
#        print(structure)

#    print("incar=", vasprun.incar)
#    print("kpoints=", vasprun.kpoints)
#    print("vasprun=", vasprun.__dict__)
#    print("vasprun=", vasprun.__dict__.keys())

def save_data(outfiles, inf, app = None, cparams = None, print_level = 1):
    if type(outfiles) == str:
        outfile = outfiles
    else:
        outfile = outfiles[0]

    structure = inf["structure"]
    pm = inf["pymatgen"]
    pm.to(outfile, structure, symprec = cparams.symprec, print_level = print_level)

    return True

def plot_data(inf, app = None, cparams = None, print_level = 0):
    pm   = tkPymatgen()

    final_structure = inf.get("final_structure", None)
    if final_structure is None:
        print()
        print("Error inplot_data(): final_structure is not given")
        return

    print()
    print("Draw final structre:")
    print(final_structure)

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111, projection='3d')
    pm.draw_structure(ax, final_structure)
#    ax.legend()

    plt.pause(1.0e-4)
    input("\nPress ENTER to terminate>>\n")

def convert(inf, app = None, cparams = None, print_level = 0):
    return inf


def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

#    vasp, base_path = get_base_path(cparams.infile)
    base_path = os.path.dirname(cparams.infile)

    print("")
    print( "==========================================================================")
    print(" Convert VASP POSCAR/vasprun.xml to CIF")
    print( "==========================================================================")
    print(f"Input VASP structure file: {cparams.infile}")
#    print(f"base_path : {base_path}")

#    logfile = os.path.join(base_path, "poscar2cif-out.txt")
    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile, app = app, cparams = cparams)
    print(f"Output CIF file : {cparams.outfile}")

    file_type = check_file_type(cparams.infile, print_level = 1)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams, print_level = 2)
    print_data(inf)
#    inf = convert(inf, cparams = cparams, print_level = 1)
    save_data([cparams.outfile], inf, cparams = cparams, print_level = 1)
    plot_data(inf, cparams = cparams, print_level = 0)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
