import os
import sys
import re
import numpy as np
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import colors
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkutils import find_range_indexes
from tklib.tkvariousdata import tkVariousData
from tklib.tkgraphic.tkplotevent import tkPlotEvent
from tklib.tksci.tkconvolution import convolution, convolve_func
from tklib.tkcrystal.tkvasp import tkVASP


"""
Convert VASPKIT DOS files to DOS
"""


plugin_ver = "dos:0.2"  # data_list_type = [[x, y]]


#================================
# global parameters
#================================
default_ext = ".dat"
extensions  = ['.dat']
input_type  = "VASPKIT output"
output_type = "DOS"

figsize = (8, 8)
fontsize = 12
legend_fontsize = 8

markers = ['o', 's', '+', 'x', 'D', 'v', '^', '<', '>', '8', 'h', 'H']
colors  = ['black', 'red', 'blue', 'darkgreen', 'darkorange', 'hotpink', 'cyan', 'magenta', 'chocolate', 
           'navy', 'slategray', 'olive', 'lightgreen', 'yellow' ]


#=============================
# Treat argments
#=============================
def initialize(app = None, cparams = None):
    cparams.debug = 0

    cparams.infiles = []

# width of Gaussian functio for convolution
    cparams.width = 0.0
    cparams.nxmin_plot = 500

# Plot configuration
    cparams.band_marker_size       = 4
    cparams.band_marker_edge_width = 0.5

    cparams.xmin = -10.0  # eV
    cparams.xmax =  10.0  # eV

# Occupancy threshold to separate HOMO and LUMO
    cparams.occ_th = 0.5

    cparams.save_figure = 1
    cparams.plot_figure = 1
    cparams.plot_figure_Ne = 1


def update_vars(app = None, cparams = None):
    if getarg(1, None) is None:
        app.terminate(usage = app.usage)

    app.add_argument(opt = "--xmin", type = "float", var_name = 'xmin',  opt_str = "--xmin=val", desc = 'x minimum to read',
                     defval = None, optional = True);
    app.add_argument(opt = "--xmax", type = "float", var_name = 'xmax',  opt_str = "--xmax=val", desc = 'x maxmum to read',
                     defval = None, optional = True);
    app.add_argument(opt = "--width", type = "float", var_name = 'width',  opt_str = "--width=val(>0.0)", desc = 'width for convolution',
                     defval = None, optional = True);
    app.add_argument(opt = "--nxmin_plot", type = "int", var_name = 'nxmin_plot',  opt_str = "--nxmin_plot=int(>1)", desc = 'minimum numberof x to plot',
                     defval = None, optional = True);
    app.add_argument(opt = None, type = "list", var_name = 'infiles', opt_str = "infiles",  desc = 'Input files', 
                     defval = None, optional = False);

    args_opt, args_idx, args_vars = app.read_args(vars = cparams.__dict__, check_allowed_args = True)
    if args_opt is None:
        error_no = args_idx
        error_message = args_vars
        app.terminate("\n\n"
                   +  "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
                   + f"!  {error_message}\n"
                   +  "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!",
                   usage = app.usage)


def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 0):
    if not os.path.isfile(infile) and not os.path.isdir(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != default_ext:
        return None

    fp = tkFile(infile, 'r', enconding = None)
    line = fp.readline()
    fp.close()
    
    if not '#Energy' in line:
        return None

    return {"file_type": "VASPKIT DOS"}

def get_output_path(infile, inf = None, app = None, cparams = None):
    dirname, basename, filebody, ext = split_file_path(infile)
    return os.path.join(dirname, f"{basename}-out.xlsx")

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infiles, app = None, cparams = None, print_level = 0):
    print("")
    print(f"Read ", infiles)
    data_list = []
    labels = []
    for i, f in enumerate(infiles):
        dirname, basename, filebody, ext = split_file_path(f)
        m = re.search(r'_(\w+)', filebody)
        if m:
            atomname = m[1]
        else:
            atomname = f'atom{i+1}'

        if print_level:
            print(f"  {f}")
            print(f"    atom: {atomname}")
        
        fp = tkFile(f, 'r')
        if fp.fp is None:
            continue

        line = fp.readline()
        aa = line.split()
        _labels = [f"{atomname} {s}" for s in aa]
        _labels[0] = 'Energy'

        ny = len(_labels)
        data = []
        for i in range(ny):
            data.append([])
        while 1:
            line = fp.readline()
            if line is None or line == '':
                break
            
            aa = line.split()
            if len(aa) <= 2:
                continue

            for j, s in enumerate(aa):
                data[j].append(pfloat(s))

        fp.close()

        data_list.append(data)
        labels.append(_labels)

    dirname, basename, filebody, ext = split_file_path(infiles[0])
    sample_name = basename
    inf = {}
    inf["data_list_type"] = "[[x, y]]"
    inf["filename"]    = infiles[0]
    inf["filenames"]   = infiles
    inf["sample_name"] = sample_name
    inf["yscale"]      = 'linear'
    inf["meta"]        = [] #[inf_list]
    inf["data_list"]   = data_list
    inf["nregion"]     = len(data_list)
    inf["nspectrum"]   = len(data_list[0]) - 1
    inf["labels"]      = labels

    return inf

def print_data(inf, app = None, cparams = None, print_level = 1):
    print("")
    print("filenames:", inf["filenames"])

    print("")
    print("sample   :", inf["sample_name"])
    print("yscale   :", inf["yscale"])
    print("labels   :", inf["labels"])
    
    nregion   = inf["nregion"]
    nspectrum = inf["nspectrum"]
    data_list = inf["data_list"]
    print("")
    for ir in range(nregion):
        data  = data_list[ir]
        ndata = len(data[0])
        nskip = max([int(ndata / 50), 1])

        print("")
        for l in inf['labels'][ir]:
            print(f"{l:8} ", end = '')
        print("")
        for i in range(0, ndata, nskip):
            for ic in range(len(data)):
                v = data[ic][i]
                print(f"{v:8.3g} ", end = '')
            print("")

def plot_data(inf, app = None, cparams = None, print_level = 1):
    data_list = inf["data_list"]
    nspectrum = inf["nspectrum"]
    nregion   = inf["nregion"]
    labels    = inf["labels"]

    if cparams.xmin == '*':
        xmin = min(data_list[0][0])
    else:
        xmin = max([min(data_list[0][0]), cparams.xmin])
    if cparams.xmax == '*':
        xmax = max(data_list[0][0])
    else:
        xmax = min([max(data_list[0][0]), cparams.xmax])

    print("")
    print("plot")
    fig, axes = plt.subplots(nregion, 1, figsize = figsize, sharex = True, gridspec_kw = {'hspace': 0})
    plot_event = tkPlotEvent(plt)
    axes[0].set_title(f"{cparams.infiles[0]}")
    ncolors = len(colors)

    for i, data in enumerate(data_list):
        ny      = len(data)
        xE      = data[0]
        ndata   = len(xE)
        xlabel  = labels[i][0]
        ylabels = labels[i][1:]
        dx = xE[1] - xE[0]
        nskip = max([1, int(cparams.width / dx / 5.0 + 1.0e-5)])
        if cparams.nxmin_plot is not None and int(ndata / nskip) < cparams.nxmin_plot:
            nskip = int(ndata / cparams.nxmin_plot)
        print(f"ndata    : {ndata}")
        print(f"ylabels  : {ylabels}")
        print(f"dx       : {dx}")
        print(f"width for convolution: {cparams.width}")
        print(f"nxmin_plot: {cparams.nxmin_plot}")
        print(f"nskip for convolution: {nskip}")

        ax = axes[i]
        ax.tick_params(labelsize = fontsize)

        ymin =  1.0e100
        ymax = -1.0e100
        for j in range(1, len(ylabels)):
            x_plot, y_plot = convolution(xE, data[j], cparams.width, func_type = 'gauss', nskip = nskip, nxmin_plot = cparams.nxmin_plot)
            ax.plot(x_plot, y_plot, label = ylabels[j], picker = True, color = colors[j % ncolors], linewidth = 0.5)
            x0idx, x1idx = find_range_indexes(x_plot, cparams.xmin, cparams.xmax)
            ymin = min([ymin, min(y_plot[x0idx:x1idx-1])])
            ymax = max([ymax, max(y_plot[x0idx:x1idx-1])])

        ymin = min([0.0, ymin])

        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)
        ax.set_xlabel(r'Energy (eV)', fontsize = fontsize)
        ax.set_ylabel('DOS (states/cm$^3$)', fontsize = fontsize)
        ax.legend(fontsize = legend_fontsize)

    def on_legend_label_click(event):
        if hasattr(event.artist, 'get_text'):
            clicked_label = event.artist.get_text()  # クリックされた凡例ラベルのテキストを取得
        elif hasattr(event.artist, 'get_label'):
            clicked_label = event.artist.get_label()  # クリックされた凡例ラベルのテキストを取得
        else:
            return

        print(f"データ [{clicked_label}] がクリックされました")

    fig.canvas.mpl_connect('pick_event', on_legend_label_click)

    plt.tight_layout()
    plt.pause(0.1)
    
    input("Press ENTER to terminate>>")

def save_data(outfiles, inf, app = None, cparams = None, print_level = 0):
    return True
    outfile   = outfiles[0]
    data_list = inf["data_list"]
    nspectrum = len(data_list[0])

    _labels = []
    for i in range(len(inf["labels"])):
        _labels.extend(inf["labels"][i])

    _data_list = []
    for i in range(len(inf["data_list"])):
        _data_list.extend(inf["data_list"][i])

    if print_level:
        print("")
        print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, _labels, _data_list)

    return True

def convert(inf, app = None, cparams = None, print_level = 0):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str = None, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Convert CIF file to powder XRD pattern")
    print( "==========================================================================")
    print(f"Input files: {cparams.infiles}")

    logfile = app.replace_path(cparams.infiles[0], template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infiles[0])

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infiles[0], print_level = 1)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infiles[0]}] invalid file type", usage = app.usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = app.usage)

    inf = read_data(cparams.infiles, cparams = cparams, print_level = 2)
    print_data(inf)
    inf = convert(inf, cparams = cparams, print_level = 1)
    save_data([cparams.outfile], inf, cparams = cparams, print_level = 1)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = app.usage)


if __name__ == "__main__":
    main()
