import os
import sys
import numpy as np
import matplotlib.pyplot as plt

from pymatgen.core import Composition
from pymatgen.io.vasp.inputs import Incar, Poscar, Kpoints, Potcar


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import get_ext, getarg, getintarg, getfloatarg, pint, pfloat, terminate
from tklib.tkutils import split_file_path, replace_path
from tklib.tkcrystal.tkpymatgen import tkPymatgenVASP, tkPymatgen


"""
Convert vasprun.xml to VASP input files (INCAR and POSCAR)
"""


usage_str = '''
"  (i) usage: python {} POSCAR/vasprun.xml outdir".format(sys.argv[0])
"         ex: python {} {} {}".format(sys.argv[0], cparams.infile, cparams.outfile)
'''[1:-1]


plugin_ver = "structure:0.2"


#================================
# global parameters
#================================
default_ext = ".xml"
input_type  = "vasprun.xml"
output_type = "VASP input files"


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0

    cparams.infile  = 'POSCAR'
    cparams.outfile = ''

    cparams.symprec = 1.0e-5

def update_vars(app = None, cparams = None):
#    if getarg(1, None) is None:
#        app.terminate(usage = usage)
    
    cparams.infile  = getarg(1, cparams.infile)
    cparams.outfile = getarg(2, cparams.outfile)
    cparams.symprec = getfloatarg(3, cparams.symprec)
    cparams.validate_vasprun = getfloatarg(4, False)

def get_base_path(infile):
    vasp = tkVASP()
    if os.path.isfile(infile):
        base_path = vasp.getdir(infile)
    else:
        base_path = infile

    return vasp, base_path

def check_exist(path, print_level = 0):
    f = os.path.isfile(path)
    if not f:
        if print_level:
            print(f"Error in check_file_type(): [{path}] does not exist")
        return None

    return True

def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 0):
    if not os.path.isfile(infile) and not os.path.isdir(infile):
        return None  #f"Error: file [{infile}] does not exist"

    ext = get_ext(infile).lower()
    if ext == '.xml':
        return {"file_type": "VASP vasprun.xml"}

    return None

def get_output_path(infile, inf = None, app = None, cparams = None):
    INCAR   = app.replace_path(infile, template = ["{dirname}", "{filebody}.INCAR"])
    POSCAR  = app.replace_path(infile, template = ["{dirname}", "{filebody}.POSCAR"])
    KPOINTS = app.replace_path(infile, template = ["{dirname}", "{filebody}.KPOINTS"])
    POTCAR  = app.replace_path(infile, template = ["{dirname}", "{filebody}.POTCAR"])
    return [INCAR, POSCAR, KPOINTS, POTCAR]

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, print_level = 0):
    infile = os.path.abspath(infile)
    dirname, basename, filebody, ext = split_file_path(infile, check_dir = True)

    if print_level >= 1:
        print("")
        print("vasprun.xml path: ", infile)

    vasp = tkPymatgenVASP()
    pm   = tkPymatgen()
    vasprun = vasp.read_vasprun(infile, validate = cparams.get("validate_vasprun", False))
    if not vasprun:
        print()
        app.terminate("Error in vasprun2inputs.read_data(): Could not read [{infile}]", pause = True)
        
    inf, structure = vasp.read_structure(infile, vasprun)
    
    iSPG, SPG_name, symmetrized_structure = pm.symmetrize(structure, cparams.symprec)
    if print_level:
        print("Space group number: ", iSPG)
        print("Space group name: ", SPG_name)
        print("symmetrized_structure:")
        print(symmetrized_structure)

    incar_params = vasprun.incar
    system = incar_params.get("SYSTEM", "blank SYSTEM")

    final_structure = vasprun.final_structure
    latt = structure.lattice
    formula, reduced_formula = pm.chemical_formula(final_structure)

    natoms = len(final_structure)
    composition     = Composition(reduced_formula)
    formula_weight  = composition.weight
    num_atoms_in_formula = composition.num_atoms

    inf = {}
    inf["data_list_type"] = "[tkCIF]"
    inf["filename"]    = infile
    inf["sample_name"] = reduced_formula
    inf["meta"]        = [] #[inf_list]
    inf["pymatgen"]    = pm
    inf["vasp"]        = vasp
    inf["vasprun"]     = vasprun
    inf["input_inf"]   = inf
    inf["initial_structure"] = vasprun.initial_structure
    inf["final_structure"]   = final_structure
    inf["structure_original"] = [structure]
    inf["structure"]          = [symmetrized_structure]

    return inf

def print_data(inf, app = None, cparams = None, print_level = 1):
    vasprun = inf["vasprun"]
    incar_params = vasprun.incar
    incar = Incar.from_dict(incar_params)
    structure = vasprun.final_structure
    poscar = Poscar(structure)
    kpoints = vasprun.kpoints
    potcar_symbols = vasprun.potcar_symbols
    print()
    print("INCAR:")
    print(incar)
    print()
    print("POSCAR:")
    print(poscar)
    print()
    print("KPOINTS:")
    print(kpoints)
    print()
    print("POTCAR:")
    print(potcar_symbols)

def save_data(outfiles, inf, app = None, cparams = None, print_level = 0):
    if type(outfiles) == str:
        INCAR = outfiles
        POSCAR = None
        KPOINTS = None
        POTCAR = None
    else:
        INCAR, POSCAR, KPOINTS, POTCAR = outfiles

    vasprun = inf["vasprun"]
    incar_params = vasprun.incar
    incar = Incar.from_dict(incar_params)
    structure = vasprun.final_structure
    poscar = Poscar(structure)
    kpoints = vasprun.kpoints
    potcar_symbols = vasprun.potcar_symbols

    print()
    if INCAR:
        print(f"Save INCAR to [{INCAR}]:")
        incar.write_file(INCAR)
    if POSCAR:
        print(f"Save POSCAR to [{POSCAR}]:")
        poscar.write_file(POSCAR)
    if KPOINTS:
        print(f"Save KPOINTS to [{KPOINTS}]:")
        kpoints.write_file(KPOINTS)
    if POTCAR:
         print(f"POTCAR is not generated automatically")
         print(potcar_symbols)

    return True

def plot_data(inf, app = None, cparams = None, print_level = 1):
    pm   = tkPymatgen()

    final_structure = inf.get("final_structure", None)
    if final_structure is None:
        print()
        print("Error inplot_data(): final_structure is not given")
        return

    print()
    print("Draw final structre:")
#    print(final_structure)

    fig = plt.figure(figsize=(8, 8))
    ax = fig.add_subplot(111, projection='3d')
    pm.draw_structure(ax, final_structure)
#    ax.legend()

    plt.pause(1.0e-4)
    input("\nPress ENTER to terminate>>\n")

def convert(inf, app = None, cparams = None, print_level = 0):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

#    vasp, base_path = get_base_path(cparams.infile)
    base_path = os.path.dirname(cparams.infile)

    print("")
    print( "==========================================================================")
    print(" Convert VASP POSCAR/vasprun.xml to CIF")
    print( "==========================================================================")
    print(f"Input VASP structure file: {cparams.infile}")
#    print(f"base_path : {base_path}")

#    logfile = os.path.join(base_path, "poscar2cif-out.txt")
    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfiles = get_output_path(cparams.infile, app = app, cparams = cparams)
    print(f"Output files: {cparams.outfiles}")

    file_type = check_file_type(cparams.infile, print_level = 1)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams, print_level = 2)
    print_data(inf)
#    inf = convert(inf, cparams = cparams, print_level = 1)
    save_data(cparams.outfiles, inf, cparams = cparams, print_level = 1)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
