import os
import sys
import chardet
import json


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkcrystal.tkcrystal import tkCrystal
from tklib.tkcrystal.tkcif import tkCIF, tkCIFData


"""
Convert JSON crystal structure file to CIF
"""


usage_str = '''
"  (i) usage: python {} infile(.json) outfile".format(sys.argv[0])
"         ex: python {} {} {}".format(sys.argv[0], cparams.infile, cparams.outfile)
'''[1:-1]


plugin_ver = "structure:0.2"


#================================
# global parameters
#================================
default_ext = ".json"
input_type  = "JSON crystal structure"
output_type = "CIF"


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0

    cparams.infile  = 'GaN.json'
    cparams.outfile = ''


def update_vars(app = None, cparams = None):
#    if getarg(1, None) is None:
#        app.terminate(usage = usage)
    
    cparams.infile  = getarg(1, cparams.infile)
    cparams.outfile = getarg(2, cparams.outfile)

def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 0):
    if not os.path.isfile(infile) and not os.path.isdir(infile):
        return None  #f"Error: file [{infile}] does not exist"

    filebody, ext = os.path.splitext(infile)
    if ext.lower() != default_ext:
        return None

    return {"file_type": input_type}

def get_output_path(infile, inf = None, app = None, cparams = None):
    filebody, ext = os.path.splitext(infile)
    return filebody + ".cif" 

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, print_level = 0):
    if print_level >= 1:
        print("")
        print("Input JSON path: ", infile)

    cry = tkCrystal()
#    cry.SetSampleName(sample_name)
#    cry.SetCrystalName(sample_name)

    try:
        with open(infile, 'rb') as f:
            result = chardet.detect(f.read())
            encoding = result['encoding']
        print("encoding=", encoding)
    except Exception as e:
        print(f"Error in read_data() to check charcode: {e}")
        return False

    try:
        with open(infile, "r", encoding=encoding) as f:
            json_data = json.load(f)
    except Exception as e:
        print(f"Error in read_data() to read [{infile}]: {e}")
        return False

#    print(json.dumps(json_data, indent=4))

    lattice = json_data["crystal_structure"]["lattice"]
    a = pfloat(lattice["a"].split()[0])
    b = pfloat(lattice["b"].split()[0])
    c = pfloat(lattice["c"].split()[0])
    alpha = pfloat(lattice["α"].split()[0])
    beta = pfloat(lattice["β"].split()[0])
    gamma = pfloat(lattice["ɣ"].split()[0])
#    if print_level:
#        print("latt=", a, b, c, alpha, beta, gamma)

    cry.SetLatticeParameters([a, b, c, alpha, beta, gamma])

    symmetry = json_data["crystal_structure"]["symmetry"]
    SPGName = symmetry["symbol"]
    iSPG    = pint(symmetry["international_number"])
#    if print_level:
#        print(f"Space group: {SPGName} ({iSPG})")   
    cry.SetSpaceGroup(name = SPGName, number = iSPG, set = 1)

    sites = json_data["crystal_structure"]["atomic_positions"]
    for s in sites:
        atom_name = s["element"]
        x = eval(s["x"])
        y = eval(s["y"])
        z = eval(s["z"])
        if print_level:
            print(f"{atom_name} {x} {y} {z}")

        cry.AddAtomSite(name = atom_name, pos = [x, y, z])

    cry.ExpandCoordinates()
    if print_level:
        cry.PrintInf()

    inf = {}
    inf["data_list_type"] = "[tkCIF]"
    inf["filename"]    = infile
    inf["sample_name"] = cry.ChemicalComposition()
    inf["meta"]        = [] #[inf_list]
    inf["cry"]         = [cry]
    inf["json"]        = [json_data]

    return inf

def print_data(inf, app = None, cparams = None, print_level = 1):
    cry = inf["cry"][0]
    if print_level:
        cry.PrintInf()

def plot_data(inf, app = None, cparams = None, print_level = 1):
    print()
    print("Not plot data in this plugin")

def save_data(outfiles, inf, app = None, cparams = None, print_level = 0):
    cry = inf["cry"][0]

    if type(outfiles) is str:
        cifpath = outfiles
    else:
        cifpath = outfiles[0]

    print("")
    print("Write to [{}]".format(cifpath))
    cif = tkCIFData()
    cif.CreateCIFFileFromCCrystal(cry, cifpath)

    return True

def convert(inf, app = None, cparams = None, print_level = 0):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

#    vasp, base_path = get_base_path(cparams.infile)
    base_path = os.path.dirname(cparams.infile)

    print("")
    print( "==========================================================================")
    print(" Convert VASP POSCAR to CIF")
    print( "==========================================================================")
    print(f"Input POSCAR file: {cparams.infile}")
#    print(f"base_path : {base_path}")

    logfile = os.path.join(base_path, "poscar2cif-out.txt")
#    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile, cparams = cparams)
    print(f"Output CIF file : {cparams.outfile}")

    file_type = check_file_type(cparams.infile, print_level = 1)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams, print_level = 2)
    print_data(inf)
#    inf = convert(inf, cparams = cparams, print_level = 1)
    save_data([cparams.outfile], inf, cparams = cparams, print_level = 1)
#    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
