import os
import sys
import shutil
import glob
import csv
import re
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from scipy.interpolate import interp1d
from pprint import pprint
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import colors
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkvariousdata import tkVariousData
from tklib.tkexcel_db import tkExcelDB
from tklib.tkgraphic.tkplotevent import tkPlotEvent


"""
Convert XLSX file to XRD
"""


plugin_ver = "data:0.2"   # data_list_type = [[x, y]]


usage_str = '''
"  (i) usage: python {} mode Bi2OS2.xlsx xmin xmax".format(sys.argv[0])
"         ex: python {} {} {} {}".format(sys.argv[0], cparams.infile, cparams.xmin, cparams.xmax)
'''[1:-1]



#================================
# global parameters
#================================
default_ext = ".xlsx"
extensions  = [".xlsx", ".csv"]
input_type  = "S-sigma-T .xlsx"
output_type = "MS-Excel .xlsx"

figsize = (8, 4)
fontsize = 12
legend_fontsize = 12


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0
    cparams.findvalidstructure = True

    cparams.infile = None

    cparams.xmin  = 20.0
    cparams.xmax  = 120.0


def update_vars(app = None, cparams = None):
    if getarg(2, None) is None:
        app.terminate(usage = usage)
    
    cparams.infile = getarg(1, cparams.infile)
    cparams.xmin   = getfloatarg(2, cparams.xmin)
    cparams.xmax   = getfloatarg(3, cparams.xmax)


def check_file_type(infile, inf = None, app = None, cparams = None):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    is_passed = False
    for _ext in extensions:
        if ext.lower() == _ext.lower():
            is_passed = True
            break

    if not is_passed:
        return None

    xls = tkExcelDB(infile, table_name = None, OpenFile = True, CloseFile = True, 
                description = 'MS-Excel for XRD', IsPrint = True)
    xls.close()
    T_idx     = xls.get_icolumn_from_label_regex(r'^T\s*\(K', column_org = 1, row_org = 1, sheet = None, flags = 0)
    S_idx     = xls.get_icolumn_from_label_regex(r'^S\s*\(.*V/K', column_org = 1, row_org = 1, sheet = None, flags = 0)
    sigma_idx = xls.get_icolumn_from_label_regex(r'^sigma\s*\(S/cm\)', column_org = 1, row_org = 1, sheet = None, flags = 0)
#    print(f"indexes in [{infile}]")
#    print(f"  T    : {T_idx}")
#    print(f"  S    : {S_idx}")
#    print(f"  sigma: {sigma_idx}")

    if T_idx is None or S_idx is None or sigma_idx is None:
        return None
    
    return {"file_type": "MSExcel .xlsx"}

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, is_print = False):
    if is_print:
        print("")
        print(f"Read [{infile}]")

    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not read [{infile}]")

    xls = tkExcelDB(infile, table_name = None, OpenFile = True, CloseFile = True, 
                description = 'MS-Excel for sigma - S - T data', IsPrint = True)
    xls.close()

    max_column = xls.max_column()
    max_row    = xls.max_row()
    labels_all, data_all = xls.read_data(idata = None, flags = '', convert_type = float, def_val = "input", is_print = True)

    T_idx     = xls.get_icolumn_from_label_regex(r'^T\s*\(K', column_org = 1, row_org = 1, sheet = None, flags = 0)
    S_idx     = xls.get_icolumn_from_label_regex(r'^S\s*\(.*V/K', column_org = 1, row_org = 1, sheet = None, flags = 0)
    sigma_idx = xls.get_icolumn_from_label_regex(r'^sigma\s*\(S/cm\)', column_org = 1, row_org = 1, sheet = None, flags = 0)
    T_data     = xls.get_col_data(r'^T\s*\(K',          ws = None)
    S_data     = xls.get_col_data(r'^S\s*\(.*V/K',      ws = None)
    sigma_data = xls.get_col_data(r'^sigma\s*\(S/cm\)', ws = None)
    data_list = [T_data, sigma_data, S_data]
    _labels   = xls.get_labels()
    labels    = [_labels[T_idx-1], _labels[sigma_idx-1], _labels[S_idx-1]]
    if is_print:
        print(f"indexes in [{infile}]")
        print(f"  T     : {T_idx}")
        print(f"  S     : {S_idx}")
        print(f"  sigma : {sigma_idx}")
        print(f"  labels:", labels)
        print(f"  labels(all):", labels_all)


    dirname, basename, filebody, ext = split_file_path(infile)
    inf = {}
    inf["data_list_type"] = "[[x, y]]"
    inf["meta"]        = []
    inf["filename"]    = infile
    inf["sample_name"] = basename
    inf["labels"]      = [labels]
    inf["yscale"]      = 'linear'
    inf["data_list"]   = [data_list]
    inf["nregion"]     = 1
    inf["nspectrum"]   = len(data_list) - 1
    inf["ndata"]       = len(data_list[0])
    inf["xmin"]        = data_list[0][0]
    inf["xmax"]        = data_list[0][inf["ndata"] - 1]

    inf["labels_all"]    = labels_all
    inf["data_list_all"] = data_all

    return inf

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def print_data(inf, app = None, cparams = None):
    labels = inf["labels"][0]
    data_list = inf["data_list"][0]
    print("")
    print("filename :", inf["filename"])
    print("sample   :", inf["sample_name"])
    print("yscale   :", inf["yscale"])
    print("labels   :", labels)
    print("ndata    :", inf["ndata"])
    
    print("")
    for l in labels:
        print(f"{l:15}\t", end = '')
    print()
    for i in range(inf["ndata"]):
        for icol in range(len(data_list)):
            v = data_list[icol][i]
            if v is None:
                print(f"{'None':15}\t", end = '')
            elif type(v) is str:
                print(f"{v:15}\t", end = '')
            else:
                print(f"{v:8.4g}\t", end = '')
        print()

def plot_data(inf, app = None, cparams = None, is_print = False):
    labels  = inf["labels"][0]
    ylabels = labels[1:]
    data_list = inf["data_list"][0]
    
    if is_print:
        print("")
        print("plot")
        print("filename :", inf["filename"])
        print("sample   :", inf["sample_name"])
        print("yscale   :", inf["yscale"])
        print("labels   :", labels)
        print("ndata    :", inf["ndata"])

    fig, axes = plt.subplots(1, 2, figsize = figsize)
    ax = axes[0]
    dirname, basename, filebody, ext = split_file_path(cparams.infile)
    ax.set_title(f"{basename}")

    ncolors = len(colors)
    for i in range(len(ylabels)):
        ax = axes[i]
        ax.tick_params(labelsize = fontsize)
        ax.plot(data_list[0], data_list[i], label = ylabels[i], color = colors[i % ncolors], linewidth = 0.5,
                    marker = 'o')

        if pfloat(cparams.xmin, None) is None:
            xmin = inf["xmin"]
        else:
            xmin = max([inf["xmin"], cparams.xmin])
        if pfloat(cparams.xmax, None) is None:
            xmax = inf["xmax"]
        else:
            xmax = min([inf["xmax"], cparams.xmax])
#        ax.set_xlim(xmin, xmax)
    
        ax.set_xlabel(labels[0],   fontsize = fontsize)
        ax.set_ylabel(labels[i+1], fontsize = fontsize)
        ax.legend(fontsize = legend_fontsize)

    plt.tight_layout()
    plt.pause(0.1)
    
    input("Press ENTER to terminate>>")
    

def save_data(outfiles, inf, app = None, cparams = None, is_print = False):
    labels_all    = inf["labels_all"]
    data_list_all = inf["data_list_all"]
    outfile   = outfiles[0]

    print("")
    print(f"Save to [{outfile}]")
    print(f"  labels:", labels_all)
    tkVariousData().to_excel(outfile, labels_all, data_list_all)

    return True

def convert(inf, app = None, cparams = None, is_print = True):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Convert CIF file to powder XRD pattern")
    print( "==========================================================================")
    print(f"Input file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, cparams = cparams, is_print = True)
    print_data(inf)
    inf = convert(inf, cparams = cparams, is_print = True)
    save_data([cparams.outfile], inf, cparams = cparams, is_print = True)
    plot_data(inf, cparams = cparams, is_print = True)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
