import os
import sys
import glob
import re
import csv
import datetime
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me, Ea_Arrhenius
from tklib.tkutils import colors
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkutils import del_quote, split_two, split_quoted_args
from tklib.tkfile import tkFile
from tklib.tkvariousdata import tkVariousData
from tklib.tkgraphic.tkplotevent import tkPlotEvent



"""
Convert Resitest8300 .csv T-dependent Hall data to MS-Excel
"""

usage_str = '''
"  (i) usage: python {} infile(.hdt) Tmin Tmax".format(sys.argv[0])
"         ex: python {} {} {} {}".format(sys.argv[0], cparams.infile, cparams.xmin, cparams.xmax)
'''[1:-1]


plugin_ver = "data:0.2"  # data_list_type = [[x, y]]


#================================
# global parameters
#================================
default_ext = ".csv"
input_type  = "Resitest8300 .csv"
output_type = "Hall T"

figsize         = (8, 8)
fontsize        = 12
legend_fontsize = 12


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0
    cparams.findvalidstructure = True

    cparams.infile = f'*{default_ext}'

    cparams.xmin  = '*'
    cparams.xmax  = '*'


def update_vars(app = None, cparams = None):
    if getarg(1, None) is None:
        app.terminate(usage = usage)
    
    cparams.infile      = getarg(1, cparams.infile)
    cparams.xmin       = getfloatarg(2, cparams.xmin)
    cparams.xmax       = getfloatarg(3, cparams.xmax)


def normalize_str(str):
    str = str.replace('～', '-')
    str = str.replace('年', '/')
    str = str.replace('月', '/')
    str = str.replace('日', '')
    str = str.replace('-', '/')

    return str

def get_sec_passed(daystr, timestr):
#    date_obj = datetime.datetime.strptime(f"{daystr} {timestr}", '%Y年%m月%d日 %H:%M:%S')
    date_obj = datetime.datetime.strptime(f"{daystr} {timestr}", '%Y/%m/%d %H:%M:%S')
    delta = date_obj - datetime.datetime(2022, 1, 1, 0, 0, 0)
    dsec = delta.days * 24 * 60 * 60 + delta.seconds
    return dsec

def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 1):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != default_ext:
        return None

    fp = tkFile(infile, mode = 'r', encoding = None)
    is_ok = 0
    for i in range(10):
        line = fp.readline()
        if not line:
            break

        key, val = split_two(line, sep = ',', quotation = 'remove')
#        print("a=", key, val)
        if key == '試料名':
            is_ok += 1
        elif key == '測定者':
            is_ok += 1
        elif key == 'コメント':
            is_ok += 1
            
    fp.close()
    
    if is_ok <= 2:
        return None
        
    return {"file_type": input_type}

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, print_level = 0):
    if print_level:
        print("")
        print(f"Read from [{infile}]")

    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not read [{infile}]")

    fp0 = tkFile(infile, mode = 'r', encoding = None)
    fp = csv.reader(fp0.fp, delimiter = ',')

    target = None
    meta_top = {}
    while 1:
        line = next(fp)
        if not line:
            break

        if line[0] == '':
            continue
        
        key = line[0]
        if len(line) <= 1:
            val = None
        else:
            val = " ".join(line[1:])

#        print(f"key=[{key}][{val}]")
        if val is None:
            key = del_quote(key)
            if "比抵抗測定結果>>>" in key:
                target = 'rho'
                break

        if val is not None:
            meta_top[key] = val

    target = None
    time_rho  = []
    T_rho     = []
    T_se_rho  = []
    F         = []
    rho       = []
    rho_sheet = []
    sigma     = []

    time_Hall = []
    T_Hall    = []
    T_se_Hall = []
    VHall    = []
    phase    = []
    VHall    = []
    noise    = []
    polarity = []
    RH       = []
    N        = []
    N_sheet  = []
    mu       = []
    meta_rho  = []
    meta_Hall = []

    def add_Hall_data():
         print("Hall meta=", meta)
         time_Hall.append([meta["測定開始時刻"], meta["測定終了時刻"]])
         Ts = pfloat(meta["測定開始温度"])
         Te = pfloat(meta["測定終了温度"])
         T_Hall.append((Ts + Te) * 0.5)
         T_se_Hall.append([Ts, Te])
         noise.append(pfloat(meta["ﾉｲｽﾞ"]))
         VHall.append(pfloat(meta["ﾎｰﾙ起電圧"]))
         RH.append(pfloat(meta["ﾎｰﾙ係数"]))
         phase.append(pfloat(meta["位相"]))
         polarity.append(pint(meta["ｷｬﾘｱﾀｲﾌﾟ"]))
         N.append(pfloat(meta["ｷｬﾘｱ濃度"]))
         N_sheet.append(pfloat(meta["ｼｰﾄｷｬﾘｱ濃度"]))
         mu.append(pfloat(meta["移動度"]))
         meta_Hall.append(meta)
    
    def add_rho_data():
         print("rho meta=", meta)
         Ts = pfloat(meta["測定開始温度"])
         Te = pfloat(meta["測定終了温度"])
         T_rho.append((Ts + Te) * 0.5)
         T_se_rho.append([Ts, Te])
         F.append(pfloat(meta["Ｆ値"]))
         res = pfloat(meta["比抵抗値"])
         rho.append(res)
         rho_sheet.append(pfloat(meta["シート抵抗値"]))
         sigma.append(1.0 / res)
         meta_rho.append(meta)
    
    meta = {}
    no_record = False
    while 1:
        try:
            line = next(fp)
        except:
            print("last:", target)
            if target == 'Hall':
                add_Hall_data()
            elif target == 'rho':
                add_rho_data()
            break

        if len(line) == 0:
            continue

#        print("line=", line)
        if "比抵抗測定結果>>>" in line[0]:
            target = 'rho'
            no_record = False
            print("rho")
            if meta:
                print("meta=", meta)
                add_Hall_data()
            meta = {}
            continue
        elif "ホール測定結果>>>" in line[0]:
            target = 'Hall'
            no_record = False
            print("Hall")
            if meta:
                print("meta=", meta)
                add_rho_data()
            meta = {}
            continue
#        elif "測定詳細結果>>>" in line[0]:
#            no_record = True
#            print("details")
#            pass

        if no_record:
            continue

#        key, val = split_two(line, sep = ',', quotation = 'remove', defval = None)
        key = line[0]
        if len(line) > 1:
            val = line[1]
        else:
            val is None
        
#        print(f"key0=[{key}][{val}]")
        if val is None:
            continue
        
        if key != '' and pfloat(key, defval = None) is None:
#            print(f"key1=[{key}][{val}]")
            meta[key] = val

    fp0.close()

    data_list = [T_rho, T_Hall, F, noise, phase, polarity, VHall, RH, rho, sigma, mu, N, N_sheet]
    labels    = ["T_rho", "T_Hall", "F", "noise", "phase", "polarity", "VHall (V)", "RH (cm^3)", 
                 "rho (ohm.cm)", "sigma (S/cm)", "mu (cm2/Vs)", "N (cm-3)", "N_sheet (cm-2)"]

    sample_name = meta_top["試料名"]
    meta = {
        "sample_name": sample_name,
        "thickness": meta_top["試料厚み"],
        "person": meta_top["測定者"],
        "measurement_date": normalize_str(meta_top["測定年月日"]),
        "comment": meta_top["コメント"],
        }
                
    dirname, basename, filebody, ext = split_file_path(cparams.infile)
    inf = {}
    inf["data_list_type"] = "[[x, y]]"
    inf["filename"]    = infile
    inf["sample_name"] = sample_name
    inf["yscale"]      = 'log'
    inf["meta"]        = [meta]
    inf["data_list"]   = [data_list]
    inf["nregion"]     = 1
    inf["nx"]          = 2
    inf["nspectrum"]   = len(data_list) - 1
    inf["ndata"]       = len(data_list[0])
    inf["xmin"]        = min([min(T_rho), min(T_Hall)])
    inf["xmax"]        = max([max(T_rho), max(T_Hall)])
    inf["labels"]      = [labels]

    return inf

def print_data(inf, app = None, cparams = None, print_level = 2):
    data_list = inf["data_list"][0]
    labels    = inf["labels"][0]
    nspectrum = inf["nspectrum"]

    if print_level:
        print("")
        print("filename :", inf["filename"])
        print("sample   :", inf["sample_name"])
        print("yscale   :", inf["yscale"])
        print("labels   :", inf["labels"])
        print("nspectrum:", inf["nspectrum"])
        print("ndata    :", inf["ndata"])
        print(f"T range : {inf['xmin']:8.3g} - {inf['xmax']:8.3g} degrees")

    if print_level:
        for i in range(len(labels)):
            print(f"{labels[i]:10}  ", end = '')
        print("")
    if print_level >= 2:
        for i in range(inf["ndata"]):
            for k in range(len(labels)):
                print(f"{data_list[k][i]:10.3g}  ", end = '')
            print("")

def plot_data(inf, app = None, cparams = None):
    labels    = inf["labels"][0]
    data_list = inf["data_list"][0]
    T       = data_list[0]
    Tinv    = [1000.0 / T for T in T]
    xlabel  = labels[0]
    sigma   = data_list[9]
    mu      = data_list[10]
    N       = data_list[11]

    fig, axes = plt.subplots(3, 2, figsize = figsize)#, sharex = True, gridspec_kw = {'hspace': 0}) #, 'vspace': 0})

    axes[0][0].set_title(f"{cparams.infile}")
    for i in range(3):
        for j in range(2):
            axes[i][j].tick_params(labelsize = fontsize)

    ncolors = len(colors)
    axes[0][0].plot(T,    sigma, label = labels[9], color = 'black', linewidth = 0.5, marker = 'o')
    axes[0][1].plot(Tinv, sigma, label = labels[9], color = 'black', linewidth = 0.5, marker = 'o')
    axes[0][0].set_ylabel('$\sigma$ (S/cm)', fontsize = fontsize)
    axes[0][0].set_yscale('log')
    axes[0][1].set_yscale('log')
    axes[1][0].plot(T,    mu,    label = labels[10], color = 'black', linewidth = 0.5, marker = 'o')
    axes[1][1].plot(Tinv, mu,    label = labels[10], color = 'black', linewidth = 0.5, marker = 'o')
    axes[1][0].set_ylabel('$\mu$ (cm$^2$/Vs)', fontsize = fontsize)
    axes[2][0].plot(T,    N,     label = labels[11], color = 'black', linewidth = 0.5, marker = 'o')
    axes[2][1].plot(Tinv, N,     label = labels[11], color = 'black', linewidth = 0.5, marker = 'o')
    axes[2][0].set_yscale('log')
    axes[2][1].set_yscale('log')
    axes[2][0].set_ylabel('$N$ (cm$^{-3}$)', fontsize = fontsize)
    axes[2][0].set_xlabel(xlabel, fontsize = fontsize)
    axes[2][1].set_xlabel("1000/T (K$^{-1}$)", fontsize = fontsize)
#    axes[2].legend(fontsize = legend_fontsize)

    if cparams.xmin == '*':
        Tmin = min(T)
    else:
        Tmin = max([min(T), cparams.xmin])
    if cparams.xmax == '*':
        Tmax = max(T)
    else:
        Tmax = min([max(T), cparams.xmax])
    axes[0][0].set_xlim(Tmin, Tmax)
    axes[1][0].set_xlim(Tmin, Tmax)
    axes[2][0].set_xlim(Tmin, Tmax)
    axes[0][1].set_xlim(1000.0 / Tmax, 1000.0 / Tmin)
    axes[1][1].set_xlim(1000.0 / Tmax, 1000.0 / Tmin)
    axes[2][1].set_xlim(1000.0 / Tmax, 1000.0 / Tmin)

    plt.tight_layout()
    plt.pause(0.1)
    
    input("Press ENTER to terminate>>")
    

def save_data(outfiles, inf, app = None, cparams = None, print_level = 0):
    data_list = inf["data_list"][0]
    labels    = inf["labels"][0]
    outfile   = outfiles[0]

    print("")
    print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, labels, data_list)

    return True

def convert(inf, app = None, cparams = None, print_level = 0):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print("###################################################################")
    print("# Read T-dependent Hall data measured by ResiTest 8300 from .csv file")
    print("###################################################################")
    print(f"Input file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

#    cparams.outfile = get_output_path(cparams.infile)
    cparams.outfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.xlsx"])

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, app = app, cparams = cparams, print_level = 1)
    print_data(inf, print_level = 2)
    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, app = app, cparams = cparams)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
