import os
import sys
import glob
import re
import datetime
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tksci.tksci import pi, h, hbar,c, e, kB, me, Ea_Arrhenius
from tklib.tkutils import colors
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkexcel import tkExcel
from tklib.tkexcel_db import tkExcelDB
from tklib.tkvariousdata import tkVariousData
from tklib.tkgraphic.tkplotevent import tkPlotEvent
import tklib.tkre as tkre



"""
Convert XLSX file to XRD
"""

usage_str = '''
"  (i) usage: python {} mode infile(.xlsx) Tmin Tmax".format(sys.argv[0])
"         ex: python {} {} {} {}".format(sys.argv[0], cparams.infile, cparams.xmin, cparams.xmax)
'''[1:-1]


plugin_ver = "data:0.2"  # data_list_type = [[x, y]]


#================================
# global parameters
#================================
default_ext = ".xlsx"
input_type  = "Resitest8400 .xlsx"
output_type = "Hall T"

figsize         = (8, 8)
fontsize        = 12
legend_fontsize = 12


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0
    cparams.findvalidstructure = True

    cparams.infile = f'*{default_ext}'

    cparams.xmin  = '*'
    cparams.xmax  = '*'


def update_vars(app = None, cparams = None):
    if getarg(1, None) is None:
        app.terminate(usage = usage)
    
    cparams.infile      = getarg(1, cparams.infile)
    cparams.xmin       = getfloatarg(2, cparams.xmin)
    cparams.xmax       = getfloatarg(3, cparams.xmax)


def read_value_range2(irow, key, inf, sheet):
    inf[f"{key}_1"] = pfloat(sheet[irow][2].value)
    val = sheet[irow][3].value.split()
    inf[f"{key}_1_range"] = [pfloat(val[0]), pfloat(val[2])]

    inf[f"{key}_2"] = pfloat(sheet[irow][4].value)
    val = sheet[irow][5].value.split()
    inf[f"{key}_2_range"] = [pfloat(val[0]), pfloat(val[2])]

def normalize_str(str):
    str = str.replace('～', '-')
    str = str.replace('年', '/')
    str = str.replace('月', '/')
    str = str.replace('日', '')

    return str
    
def get_sec_passed(daystr, timestr):
#    date_obj = datetime.datetime.strptime(f"{daystr} {timestr}", '%Y年%m月%d日 %H:%M:%S')
    date_obj = datetime.datetime.strptime(f"{daystr} {timestr}", '%Y/%m/%d %H:%M:%S')
    delta = date_obj - datetime.datetime(2022, 1, 1, 0, 0, 0)
    dsec = delta.days * 24 * 60 * 60 + delta.seconds
    return dsec

def read_resitest_excel(f, print_level = 1):
    print(f"Read [{f}]")

    inf = {}

    fp = tkExcel(f)
    if fp is None:
        print("")
        print(f"Error in read_resitest_excel(): Can not read [{f}]")
        return None

    inf['file_path'] = f

    s_config = fp.Read_sheet(sheet_name = 0)
    try:
        inf["measurement_type"] = s_config[9][1].value
        inf["comment"]          = s_config[16][1].value
        inf["source_path"]      = s_config[2][0].value
        inf["thickness"]        = pfloat(s_config[10][1].value)
    except:
        fp.Close()
        return None

    val = s_config[10][0].value
    if 'nm' in val:
        inf["thickness"] *= 1.0e-7    # nm => cm
        inf["thickness_unit"] = 'nm'
    elif 'mm' in val:
        inf["thickness"] *= 0.01      # mm => cm
        inf["thickness_unit"] = 'mm'
    elif 'um' in val:
        inf["thickness"] *= 1.0e-4    # um => cm
        inf["thickness_unit"] = 'um'
    else:
        print("")
        print("Error in read_resitest_excel(): Invalid unit [{}]".format(inf["thickness_unit"]))
        print("")
        exit()

    s_summary = fp.Read_sheet(sheet_name = 2)

    inf["carrier_type"] = s_summary[23][2].value
    val = s_summary[23][3].value.split('%')
    inf["carrier_type_accuracy"] = pfloat(val[0]) * 0.01  # % => float

    def read_value_range(irow, key, inf, sheet):
        inf[key] = pfloat(sheet[irow][2].value)
        val = sheet[irow][3].value.split()
        inf[f"{key}_range"] = [pfloat(val[0]), pfloat(val[2])]
    
    read_value_range(22, 'mu_Hall',  inf, s_summary)
    read_value_range(24, 'N',        inf, s_summary)
    read_value_range(25, 'Nsheet',   inf, s_summary)
    read_value_range(26, 'RH',       inf, s_summary)
    read_value_range(27, 'RHsheet',  inf, s_summary)
    read_value_range(28, 'rho',      inf, s_summary)
    read_value_range(29, 'rhosheet', inf, s_summary)
    read_value_range(29, 'VH',       inf, s_summary)
    inf['VH_phase'] = pfloat(s_summary[31][2].value)

    s_rho = fp.Read_sheet(sheet_name = 3)
    inf['day_start']  = normalize_str(s_rho[14][2].value)
    inf['day_end']    = normalize_str(s_rho[14][3].value)
    inf['time_start'] = s_rho[15][2].value
    inf['time_end']   = s_rho[15][3].value
    inf['T_start']    = pfloat(s_rho[16][2].value)
    inf['T_end']      = pfloat(s_rho[16][3].value)
    inf['T_average']  = pfloat(s_rho[16][4].value)

    inf["dsec_start"] = get_sec_passed(inf['day_start'], inf['time_start'])
    inf["dsec_end"]   = get_sec_passed(inf['day_end'],   inf['time_end'])

    read_value_range2(24, 'rho_final',      inf, s_rho)
    read_value_range2(25, 'rhosheet_final', inf, s_rho)
    inf["F_1"] = pfloat(s_rho[26][2].value)
    inf["F_2"] = pfloat(s_rho[26][4].value)

    inf["rho2134"] = pfloat(s_rho[29][2].value)
    inf["rho3241"] = pfloat(s_rho[29][3].value)
    inf["rho4312"] = pfloat(s_rho[29][4].value)
    inf["rho1423"] = pfloat(s_rho[29][5].value)
    inf["rho1423"] = pfloat(s_rho[29][5].value)

    def read_vanderPauw(icolumn, key, inf, sheet):
        i0 = 38
        ic = icolumn
        inf[f"{key}_+_averageV"] = pfloat(sheet[i0  ][ic].value)
        inf[f"{key}_+_errorV"]   = pfloat(sheet[i0+1][ic].value)
        inf[f"{key}_+_averageI"] = pfloat(sheet[i0+2][ic].value)
        inf[f"{key}_+_errorI"]   = pfloat(sheet[i0+3][ic].value)
        inf[f"{key}_+_currentV"] = pfloat(sheet[i0+4][ic].value)

        i0 = 44
        inf[f"{key}_-_averageV"] = pfloat(sheet[i0  ][ic].value)
        inf[f"{key}_-_errorV"]   = pfloat(sheet[i0+1][ic].value)
        inf[f"{key}_-_averageI"] = pfloat(sheet[i0+2][ic].value)
        inf[f"{key}_-_errorI"]   = pfloat(sheet[i0+3][ic].value)
        inf[f"{key}_-_currentV"] = pfloat(sheet[i0+4][ic].value)

    read_vanderPauw(2, 'R2134', inf, s_rho)
    read_vanderPauw(3, 'R3241', inf, s_rho)
    read_vanderPauw(4, 'R4312', inf, s_rho)
    read_vanderPauw(5, 'R1423', inf, s_rho)

    s_Hall = fp.Read_sheet(sheet_name = 4)
    inf['day_start_Hall']  = normalize_str(s_Hall[19][2].value)
    inf['day_end_Hall']    = normalize_str(s_Hall[19][3].value)
    inf['time_start_Hall'] = s_Hall[20][2].value
    inf['time_end_Hall']   = s_Hall[20][3].value
    inf['T_start_Hall']    = pfloat(s_Hall[21][2].value)
    inf['T_end_Hall']      = pfloat(s_Hall[21][3].value)
    inf['T_average_Hall']  = pfloat(s_Hall[21][4].value)
    inf['B_start_Hall']    = pfloat(s_Hall[22][2].value)
    inf['B_end_Hall']      = pfloat(s_Hall[22][3].value)

    inf["dsec_start_Hall"] = get_sec_passed(inf['day_start_Hall'], inf['time_start_Hall'])
    inf["dsec_end_Hall"]   = get_sec_passed(inf['day_end_Hall'],   inf['time_end_Hall'])

    inf["C_averageVH"] = pfloat(s_Hall[29][2].value)
    val = s_Hall[29][3].value.split()
    inf["C_VH_range"]  = [pfloat(val[0]), pfloat(val[2])]
    inf["D_averageVH"] = pfloat(s_Hall[29][5].value)
    val = s_Hall[29][6].value.split()
    inf["D_VH_range"]  = [pfloat(val[0]), pfloat(val[2])]

    
    """
    R2134 = rho_vanderPaw(inf['thickness'], 
                    inf['R2134_+_averageV'] - inf['R2134_-_averageV'],
                    inf['R2134_+_averageI'] - inf['R2134_-_averageI'])
    print("R2134=", R2134)
    """

    fp.Close()
    
    return inf


def check_file_type(infile, inf = None, app = None, cparams = None, print_level = 1):
    if not os.path.isfile(infile):
        return f"Error: file [{infile}] does not exist"

    dirname, basename, filebody, ext = split_file_path(infile)
    if ext.lower() != default_ext:
        return None

    if re.match(r'^~\$', filebody): # or '-out' in filebody:
        return None

    fp = tkExcelDB(infile, table_name = None, OpenFile = True, CloseFile = True, 
                description = 'MS-Excel for XRD', IsPrint = True)
    v = fp.ws.cell(1, 1).value
    if v != 'プログラム設定':
        return None

    return {"file_type": input_type}

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_data(infile, app = None, cparams = None, print_level = 0):
    cparams.input_dir = app.replace_path(infile, template = "{dirname}")

    if print_level:
        print("")
        print(f"Read from [{cparams.input_dir}]")

    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not read [{infile}]")

    if not os.path.isdir(cparams.input_dir):
        app.terminate(f"Error: Can not access to dir [{cparams.input_dir}]")

    fmask = os.path.join(cparams.input_dir, "*.xlsx")
    files = glob.glob(fmask)
    read_files = []
    if print_level:
        print(f"Input files in {cparams.input_dir}:")
    for f in files:
        ret = check_file_type(f, print_level = 0)
        if ret is None:
            continue

        dirname, basename, filebody, ext = split_file_path(f)
        read_files.append(f)

#    if print_level:
#        for f in read_files:
#            print(f"  {basename}")

    list = []
    for f in read_files:
        dirname, basename, filebody, ext = split_file_path(infile)
#        print(f"** Read [{basename}]")
        inf = read_resitest_excel(f, print_level = 1)
        if inf is None:
            if print_level:
                print(f"Warning from read_data(): File [{f}] is not a ResiTest 8400 file.")
            continue

        list.append(inf)

#    print("list=", list)
    list_sorted = sorted(list, key = lambda x: x['dsec_start'])
 
    t0 = list_sorted[0]['dsec_start']
    xt       = []
    yT       = []
    yF       = []
    yRR      = []

    xtav     = []
    xTav     = []
    yrhoav   = []
    yrho_err = []
    ysigmaav = []
    ymuav    = []
    ymu_err  = []
    yNav     = []
    yN_err   = []
    yRHav    = []
    yR2134   = []
    yV_Ip2134 = []
    yV_Ip3241 = []
    yV_Ip4312 = []
    yV_Ip1423 = []
    yV_Im2134 = []
    yV_Im3241 = []
    yV_Im4312 = []
    yV_Im1423 = []
    yV_IpCur2134 = []
    yV_IpCur3241 = []
    yV_IpCur4312 = []
    yV_IpCur1423 = []
    yV_ImCur2134 = []
    yV_ImCur3241 = []
    yV_ImCur4312 = []
    yV_ImCur1423 = []
    
    xtHall  = []
    yVH     = []
    yVH_err = []
    for inf in list_sorted:
        f = inf['file_path']
        if print_level:
            print(f"{f}: {inf['day_start']} {inf['time_start']}")

        xt.append((inf['dsec_start'] - t0) / 60.0 / 60.0)
        yT.append(inf['T_start'])
        yF.append(inf['F_1'])
        yRR.append(inf['rho2134'] / inf['rho3241'])
        yRR.append(inf['rho4312'] / inf['rho1423'])

        xt.append((inf['dsec_end'] - t0) / 60.0 / 60.0)
        yT.append(inf['T_end'])
        yF.append(inf['F_2'])

        xtav.append(((inf['dsec_start'] + inf['dsec_end']) / 2.0 - t0) / 60.0 / 60.0)
        xTav.append((inf['T_start'] + inf['T_end'] + inf['T_start_Hall'] + inf['T_end_Hall']) / 4.0)

        yrhoav.append(inf['rho'])
        ysigmaav.append(1.0 / inf['rho'])
        maxerr = max([abs(inf['rho'] - inf['rho_range'][0]), abs(inf['rho'] - inf['rho_range'][1])])
        yrho_err.append(maxerr)

        ymuav.append(inf['mu_Hall'])
        ymu_err.append(max([abs(inf['mu_Hall'] - inf['mu_Hall_range'][0]), abs(inf['mu_Hall'] - inf['mu_Hall_range'][1])]))

        yNav.append(inf['N'])
        if inf['carrier_type'] == 'N':
            yRHav.append(-1.0 / e / inf['N'])
        else:
            yRHav.append(1.0 / e / inf['N'])
        yN_err.append(max([abs(inf['N'] - inf['N_range'][0]), abs(inf['N'] - inf['N_range'][1])]))

        yV_Ip2134.append(inf['R2134_+_averageV'] / inf['R2134_+_averageI'])
        yV_Ip3241.append(inf['R3241_+_averageV'] / inf['R3241_+_averageI'])
        yV_Ip4312.append(inf['R4312_+_averageV'] / inf['R4312_+_averageI'])
        yV_Ip1423.append(inf['R1423_+_averageV'] / inf['R1423_+_averageI'])
        yV_Im2134.append(inf['R2134_-_averageV'] / inf['R2134_-_averageI'])
        yV_Im3241.append(inf['R3241_-_averageV'] / inf['R3241_-_averageI'])
        yV_Im4312.append(inf['R4312_-_averageV'] / inf['R4312_-_averageI'])
        yV_Im1423.append(inf['R1423_-_averageV'] / inf['R1423_-_averageI'])

        yV_IpCur2134.append(inf['R2134_+_currentV'] / inf['R2134_+_averageI'])
        yV_IpCur3241.append(inf['R3241_+_currentV'] / inf['R3241_+_averageI'])
        yV_IpCur4312.append(inf['R4312_+_currentV'] / inf['R4312_+_averageI'])
        yV_IpCur1423.append(inf['R1423_+_currentV'] / inf['R1423_+_averageI'])
        yV_ImCur2134.append(inf['R2134_-_currentV'] / inf['R2134_-_averageI'])
        yV_ImCur3241.append(inf['R3241_-_currentV'] / inf['R3241_-_averageI'])
        yV_ImCur4312.append(inf['R4312_-_currentV'] / inf['R4312_-_averageI'])
        yV_ImCur1423.append(inf['R1423_-_currentV'] / inf['R1423_-_averageI'])

        xtHall.append((inf['dsec_start_Hall'] - t0) / 60.0 / 60.0)
        yVH.append(inf['C_averageVH'])
        yVH_err.append(max([abs(inf['C_averageVH'] - inf['C_VH_range'][0]), abs(inf['C_averageVH'] - inf['C_VH_range'][1])]))

        xtHall.append((inf['dsec_end_Hall'] - t0) / 60.0 / 60.0)
        yVH.append(inf['D_averageVH'])
        yVH_err.append(max([abs(inf['D_averageVH'] - inf['D_VH_range'][0]), abs(inf['D_averageVH'] - inf['D_VH_range'][1])]))

    labels    = ["T (K)", "RH (cm^3)", "sigma (S/cm)", "mu (cm^2/Vs)", "N (cm^-3)"]
    data_list = [xTav, yRHav, ysigmaav, ymuav, yNav]

    dirname, basename, filebody, ext = split_file_path(cparams.infile)
    inf = {}
    inf["data_list_type"] = "[[x, y]]"
    inf["filename"]    = infile
    inf["sample_name"] = basename
    inf["yscale"]      = 'linear'
    inf["meta"]        = [] #[inf_list]
    inf["data_list"]   = [data_list]
    inf["nregion"]     = 1
    inf["nspectrum"]   = len(data_list) - 1
    inf["ndata"]       = len(data_list[0])
    inf["xmin"]        = data_list[0][0]
    inf["xmax"]        = data_list[0][inf["ndata"] - 1]
    inf["labels"]      = [labels]

    return inf

def print_data(inf, app = None, cparams = None, print_level = 2):
    data_list = inf["data_list"][0]
    labels    = inf["labels"][0]
    nspectrum = inf["nspectrum"]

    if print_level:
        print("")
        print("filename :", inf["filename"])
        print("sample   :", inf["sample_name"])
        print("yscale   :", inf["yscale"])
        print("labels   :", inf["labels"])
        print("nspectrum:", inf["nspectrum"])
        print("ndata    :", inf["ndata"])
        print(f"T range : {inf['xmin']:8.3g} - {inf['xmax']:8.3g} degrees")

    if print_level:
        for i in range(len(labels)):
            print(f"{labels[i]:10}\t", end = '')
        print("")
    if print_level >= 2:
        for i in range(inf["ndata"]):
            for k in range(len(labels)):
                print(f"{data_list[k][i]:10.4g}\t", end = '')
            print("")

def plot_data(inf, app = None, cparams = None):
    labels    = inf["labels"][0]
    data_list = inf["data_list"][0]
    T       = data_list[0]
    Tinv    = [1000.0 / T for T in T]
    xlabel  = labels[0]
    sigma   = data_list[2]
    mu      = data_list[3]
    N       = data_list[4]

#    fig = plt.figure(figsize = figsize)
    fig, axes = plt.subplots(3, 2, figsize = figsize)#, sharex = True, gridspec_kw = {'hspace': 0}) #, 'vspace': 0})

    axes[0][0].set_title(f"{cparams.infile}")
    for i in range(3):
        for j in range(2):
            axes[i][j].tick_params(labelsize = fontsize)

    ncolors = len(colors)
    axes[0][0].plot(T,    sigma, label = labels[2], color = 'black', linewidth = 0.5, marker = 'o')
    axes[0][1].plot(Tinv, sigma, label = labels[2], color = 'black', linewidth = 0.5, marker = 'o')
    axes[0][0].set_ylabel('$\sigma$ (S/cm)', fontsize = fontsize)
    axes[0][0].set_yscale('log')
    axes[0][1].set_yscale('log')
    axes[1][0].plot(T,    mu,    label = labels[3], color = 'black', linewidth = 0.5, marker = 'o')
    axes[1][1].plot(Tinv, mu,    label = labels[3], color = 'black', linewidth = 0.5, marker = 'o')
    axes[1][0].set_ylabel('$\mu$ (cm$^2$/Vs)', fontsize = fontsize)
    axes[2][0].plot(T,    N,     label = labels[4], color = 'black', linewidth = 0.5, marker = 'o')
    axes[2][1].plot(Tinv, N,     label = labels[4], color = 'black', linewidth = 0.5, marker = 'o')
    axes[2][0].set_yscale('log')
    axes[2][1].set_yscale('log')
    axes[2][0].set_ylabel('$N$ (cm$^{-3}$)', fontsize = fontsize)
    axes[2][0].set_xlabel(xlabel, fontsize = fontsize)
    axes[2][1].set_xlabel("1000/T (K$^{-1}$)", fontsize = fontsize)
#    axes[2].legend(fontsize = legend_fontsize)

    if cparams.xmin == '*':
        Tmin = min(T)
    else:
        Tmin = max([min(T), cparams.xmin])
    if cparams.xmax == '*':
        Tmax = max(T)
    else:
        Tmax = min([max(T), cparams.xmax])
    axes[0][0].set_xlim(Tmin, Tmax)
    axes[1][0].set_xlim(Tmin, Tmax)
    axes[2][0].set_xlim(Tmin, Tmax)
    axes[0][1].set_xlim(1000.0 / Tmax, 1000.0 / Tmin)
    axes[1][1].set_xlim(1000.0 / Tmax, 1000.0 / Tmin)
    axes[2][1].set_xlim(1000.0 / Tmax, 1000.0 / Tmin)

    plt.tight_layout()
    plt.pause(0.1)
    
    input("Press ENTER to terminate>>")
    

def save_data(outfiles, inf, app = None, cparams = None, print_level = 0):
    data_list = inf["data_list"][0]
    labels    = inf["labels"][0]
    outfile   = outfiles[0]

    print("")
    print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, labels, data_list)

    return True

def convert(inf, app = None, cparams = None, print_level = 0):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print("###################################################################")
    print("# Read T-dependent Hall data measured by ResiTest 8400 from directory")
    print("###################################################################")
    print(f"Input file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "resitest8400_T2xlsx-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

#    cparams.outfile = get_output_path(cparams.infile)
    cparams.outfile = app.replace_path(cparams.infile, template = ["{dirname}", "resitest8400_T2xlsx-summary.xlsx"])

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, app = app, cparams = cparams, print_level = 1)
    print_data(inf, print_level = 2)
    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, app = app, cparams = cparams)
    plot_data(inf, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
