import os
import platform
import re
import numpy as np
from numpy import log, abs, sqrt
from scipy import interpolate
import pandas as pd
import openpyxl
import time
from datetime import datetime


from tklib.tkfile import tkFile
from tklib.tkutils import pint, pfloat
from tklib.tkparams import tkParams
from tklib.tksci.tkFit_lib import conv_input, save_data, add_history


config = tkParams()

def initialize_minimize_func():
    os = platform.system()
    
    if os == 'Linux':
        config.atlas_cmd = '/home/share/sedatools/bin/deckbuild'
        config.atlas_options = '-run -ascii'
    elif os == 'Windows':
        config.atlas_cmd = 'c:/sedatools/exe/deckbuild'
        config.atlas_options = '-run'
    else:
        config.atlas_cmd = '/home/share/sedatools/bin/deckbuild'
        config.atlas_options = '-run -ascii'
#    config.atlas_options = ''
#    config.atlas_options = '-V 5.30.0.R -run -ascii'

    
def get_cmd(xk_all, fit):
    infile  = fit.atlasfile  # './303K_Vd01.in'
    outfile = fit.outfile #'./303K_Vd01.out'
    return f'{config.atlas_cmd} {config.atlas_options} "{infile}" -outfile "{outfile}"'

#    cmd = 'python'
#    script = 'cal_mu.py'
#    mode = 'fit'
#    infile = fit.infile
#    transferfile = fit.transferfile
#    fplot = 0
#    return f"{cmd} {script} {mode} {infile} {transferfile} {xk_all[0]} {xk_all[1]} {xk_all[2]} {xk_all[3]}"

def read_ATLAS_DOS_file(infile, app = None, cparams = None, print_level = 0):
    fp = open(infile, 'r')
    if fp is None:
        terminate(f"Error in read_data(): Can not read [{f}]", usage = usage)
    
    E  = []
    DOS_exp   = []
    DOS_gauss = []
    DOS_total = []
    while 1:
        line = fp.readline()
        if not line:
            break

        if len(line) > 1 and line[0] == 'd':
            values = line.split()
            E.append(pfloat(values[1]))
            DOS_exp.append(pfloat(values[2]))
            DOS_gauss.append(pfloat(values[3]))
            DOS_total.append(pfloat(values[4]))

    fp.close()
    
    return {"E": E, "DOS_exp": DOS_exp, "DOS_gauss": DOS_gauss, "DOS_total": DOS_total}

def read_ATLAS_log_file(infile, app = None, cparams = None, print_level = 0):
    fp = open(infile, 'r')
    if fp is None:
        return None
    
    VG = []
    VD = []
    ID = []
    while 1:
        line = fp.readline()
        if not line:
            break

        if len(line) > 1 and line[0] == 'd':
            values = line.split()
            VG.append(pfloat(values[1]))
            VD.append(pfloat(values[8]))
            ID.append(pfloat(values[9]))

    fp.close()
    
    return {"VG": VG, "VD": VD, "ID": ID}

def read_data(infile, filetype = ''):
    print(f"Read data from [{infile}]")

    if not os.path.isfile(infile):
        return None, None, None

    if filetype == 'calc':
        ret = read_ATLAS_log_file(infile)
        return ['Vg (V)', 'Vd (V)', 'Id(cal) (A)'], [ret["VG"], ret["VD"]], [ret["ID"]]
    else:
        df = pd.read_excel(infile) #, header = True)
        labels = df.columns.tolist()
        data_list = df.values.T

        return ['Vg (V)', 'Vd (V)', 'Id(obs) (A)'], [list(data_list[0]), list(data_list[6])], [list(data_list[1])]

def update_files(xk_all, fit):
    replace_dict = {
        "Cfile"  : fit.Cfile,
        "defectaccfile": fit.defectaccfile,
        "defectdonfile": fit.defectdonfile,
        "log_file": fit.atlaslogfile,
        "cal_file": fit.calfile,
        "Temperature": fit.T,
        "Eg": fit.cparams.Eg,
        }
    for i, varname in enumerate(fit.varname):
        replace_dict[varname] = xk_all[i]

    if os.path.exists(fit.atlasfile):
        print(f"Delete [{fit.atlasfile}]")
        os.remove(fit.atlasfile)
    if os.path.exists(fit.Cfile):
        print(f"Delete [{fit.Cfile}]")
        os.remove(fit.Cfile)
    source_atlasfile = os.path.join(fit.templatepath, fit.atlasfile)
    source_Cfile     = os.path.join(fit.templatepath, fit.Cfile)
    print(f"Make [{fit.atlasfile}] from template [{source_atlasfile}]")
    conv_input(source_atlasfile, fit.atlasfile, replace_dict, print_level = fit.print_level)
    print(f"Make [{fit.Cfile}] from template [{source_Cfile}]")
    conv_input(source_Cfile, fit.Cfile, replace_dict, print_level = fit.print_level)

# ATLASの場合、VG,VDはATLAS .inファイルで定義されているので、xd_listは使わない
def ycal_list(xd_list, xk_all, fit = None, run = True):
    xk = xk_all.copy()
    if fit and fit.kpenalty:
        kp = fit.kpenalty
        kmin = fit.kmin
        kmax = fit.kmax
        nkp = len(kp)
        for i in range(nkp):
            if xk[i] < kmin[i]:
                print(f"**Warning: [{fit.varname[i]}={xk[i]}] is smaller than [{kmin[i]}]. Replace with [{kmin[i]}].")
                xk[i] = kmin[i]
            elif xk[i] > kmax[i]:
                print(f"**Warning: [{fit.varname[i]}={xk[i]}] is larger than [{kmax[i]}]. Replace with [{kmax[i]}].")
                xk[i] = kmax[i]

    if run:
        update_files(xk, fit)

    cmd = get_cmd(xk, fit)
    stime = time.time()
    strtime = datetime.fromtimestamp(stime).strftime('%Y/%m/%d %H:%M:%S')
    if run:
        print(f"Run [{cmd}] at {strtime}")
    else:
        print(f"Runned [{cmd}] at {strtime}")
    print(f"  vars: ", end = '')
    for i, x in enumerate(xk):
        print(f" {x:10.4g}", end = '')
    print()
    if run:
        ret = os.system(cmd)
        etime = time.time()
        if ret != 0:
            print()
            print(f"Error: Could not execute [{cmd}]")
            print()
            exit()
        
        strtime = datetime.fromtimestamp(etime).strftime('%Y/%m/%d %H:%M:%S')
        print(f"  Ended at {strtime}. Collapsed time: {etime - stime} s")

#    labels, x_list, y_list = read_data(fit.transferfile, 'calc')
    labels, x_list, y_list = read_data(fit.atlaslogfile, 'calc')
    if fit:
        fit.xc_list = x_list
        fit.yc_list = y_list
    
    return [y_list[0]]

def extract_data(xs, ys, xcs, ycs):
    nobs = len(xs) * len(xs[0])
    ncal = len(xcs) * len(xcs[0])
    Vgmin_obs = min(xs[0])
    Vgmax_obs = max(xs[0])
    Vdmin_obs = min(xs[1])
    Vdmax_obs = max(xs[1])
    Vgmin_cal = min(xcs[0])
    Vgmax_cal = max(xcs[0])
    Vdmin_cal = min(xcs[1])
    Vdmax_cal = max(xcs[1])
#    print("n=", nobs, ncal)
#    print(f"VG range: obs {Vgmin_obs} - {Vgmax_obs}   cal {Vgmin_cal} - {Vgmax_cal}")
#    print(f"VD range: obs {Vdmin_obs} - {Vdmax_obs}   cal {Vdmin_cal} - {Vdmax_cal}")

# VD, VG を与えて、測定値から内挿してIdを取り出す
    def extract(Vd_v, Vg, Vd, Id):
        eps = 1.0e-4
        Vg_list = []
        Id_list = []
        for Vg, Vd, Id in zip(Vg, Vd, Id):
            if abs(Vd - Vd_v) < eps:
                Vg_list.append(Vg)
                Id_list.append(Id)
#        print("267 Vg_list=", Vg_list)
#        print("268 Id_list=", Id_list)

        return Vg_list, Id_list

    Vgi_list = []
    Vdi_list = []
    Idi_obs_list = []
    Idi_cal_list = []
    for VDcal in sorted(set(xcs[1])):
        Vgs, Ids = extract(VDcal, xs[0], xs[1], ys[0])
        fobs = interpolate.interp1d(Vgs, Ids)
        for VGcal, Idcal in zip(xcs[0], ycs[0]):
            if VGcal < Vgmin_obs or Vgmax_obs < VGcal:
                continue
            
# fobsの戻り値がarray()になってしまうので、0.0を足してfloatにする
            Idi_obs = fobs(VGcal) + 0.0
            Vgi_list.append(VGcal)
            Vdi_list.append(VDcal)
            Idi_obs_list.append(Idi_obs)
            Idi_cal_list.append(Idcal)

    return Vgi_list, Vdi_list, Idi_obs_list, Idi_cal_list

def save_parameters(fit, cparams):
    print(f"Save parameters to [{cparams.parameterfile}]")
#    cparams.save_parameters(cparams.parameterfile, section = "Preferences")
    flag_keys       = ["fplot", "fhistory", "ffitfiles"]
    files_keys      = ["templatepath", "IVfile", "atlasfile", "atlaslogfile", "Cfile", "defectaccfile", "defectdonfile",
                       "file1", "file2", "file3", "file4", "file5"]
    condiction_keys = ["mode", "method", "jac", "nmaxiter", "nmaxcall", "tol", "y_convert", "Eg", "T"]
    scan_keys       = ["target_var", "x0", "x1", "nx"]
    exclude_keys    = condiction_keys + flag_keys + files_keys + scan_keys + fit.varname \
                    + ["infile", "calfile", "logfile", "outfile", "parameterfile", "stopfile", "historyfile", "fitfile"]

    cparams.save_parameters(cparams.parameterfile, section = "Preferences", exclude_keys = exclude_keys)
    cparams.save_parameters(cparams.parameterfile, section = "Flags",      keys = flag_keys)
    cparams.save_parameters(cparams.parameterfile, section = "Files",      keys = files_keys)
    cparams.save_parameters(cparams.parameterfile, section = "Scan",       keys = scan_keys)
    cparams.save_parameters(cparams.parameterfile, section = "Condition",  keys = condiction_keys)
    fit.save_parameters    (cparams.parameterfile, section = "Parameters", keys = fit.varname)

def callback(pk, fit, run = True):
    cparams = fit.cparams

    if fit.stop_flag:
        return False

    exist_stopfile = os.path.isfile(cparams.stopfile)
#    print("exist_stopfile:", cparams.stopfile, exist_stopfile)
    if exist_stopfile:
        print()
        print(f"Message: Found stop file [{cparams.stopfile}] for early stop.")
        print()
        fit.stop_flag = True
        return False

    fplot = getattr(fit, 'fplot', True) # and fit.fplot
    if fplot:
        w = fit.plt.get_current_fig_manager().window
        if w != fit.window:
#        if w != getattr(fit, 'window', None): # and w != fit.window:
            print()
            print(f"Message: Graph window is closed for early stop.")
            print()
            fit.stop_flag = True
            return False

    recoverpk = fit.recover_parameters(pk, set_member = False)

    if fit.iter % fit.print_interval == 0:
        print(f"iter: {fit.iter}")
        n = len(recoverpk)
        for i in range(n):
            print(f"  {fit.varname[i]:10}: {recoverpk[i]:10.4g} {fit.unit[i]}")
        f = fit.minimize_func(pk)
        print(f"    f={f:12.6g}")

        if fit.get('fmin_list', None) is not None:
            fit.iter_list.append(fit.iter + 1)
            fit.fmin_list.append(f)

    if fplot and fit.iter % fit.plot_interval == 0:
        ycal_list = fit.cal_ylist(recoverpk, run = run)

        x_list  = fit.__dict__.get('x_list', None)
        xlabels = fit.__dict__.get('xlabels', None)
        nx = len(x_list)
        if nx == 1 or type(fit.ix_plot) is int:
            if type(fit.ix_plot) is int and fit.ix_plot < nx:
                ix_plot = fit.ix_plot
            else:
                ix_plot = 0
            for i in range(len(fit.y_list)):
                axis = fit.data_axes[i]
                fit.fit_data_list[i][0].set_data(x_list[ix_plot], ycal_list[i])
        else:
            for i in range(len(fit.y_list)):
                axis = fit.data_axes[i]
                fit.fit_data_list[i][0].set_data(range(len(ycal_list[i])), ycal_list[i])

        if fit.get('fmin_list', None) is not None:
            fit.error_data[0].set_data(fit.iter_list, fit.fmin_list)
            fit.error_axis.set_xlim([min(fit.iter_list), max(fit.iter_list)])
            fit.error_axis.set_ylim([min(fit.fmin_list), max(fit.fmin_list)])

        fit.plt.tight_layout()
        fit.plt.subplots_adjust(top = fit.plot_region[0], bottom = fit.plot_region[1])
        fit.plt.pause(0.0001)

    fit.iter += 1

    fit.retrieve_parameter_list(recoverpk, fit.varname, target = cparams)
    save_parameters(fit, cparams)

    return True

def cal_penalty(xk_all, fit):
    kp   = fit.kpenalty
    if kp is None:
        return 0.0

    kmin = fit.kmin
    kmax = fit.kmax
    nkp = len(kp)
    p_tot = 0.0
    for i in range(nkp):
        if xk_all[i] < kmin[i]:
            d = xk_all[i] - kmin[i]
            p = kp[i] * d * d
            p_tot += p
            print(f"**Warning: [{fit.varname[i]}={xk_all[i]}] is smaller than [{kmin[i]}]. Add penalty [{p:10.3g}] to fmin")
            xk_all[i] = kmin[i]
        elif xk_all[i] > kmax[i]:
            d = xk_all[i] - kmax[i]
            p = kp[i] * d * d
            p_tot += p
            print(f"**Warning: [{fit.varname[i]}={xk_all[i]}] is larger than [{kmax[i]}]. Add penalty [{p:10.3g}] to fmin")
            xk_all[i] = kmax[i]

    return p_tot

def cal_fmin(xk_all, x_list, y_list, w_list, fit, run):

    yc_list = ycal_list(x_list, xk_all, fit, run = run)
    Vgi_list, Vdi_list, Idi_obs_list, Idi_cal_list = extract_data(fit.x_list, fit.y_list, fit.xc_list, fit.yc_list)

    nx = len(x_list[0])
    fmin = 0.0
    eps = 1.0e-20
    for i in range(nx):
        if fit.y_convert == 'log':
            if Idi_obs_list[i] <= 0.0:
                Iobs = eps
            else:
                Iobs = Idi_obs_list[i]
            if Idi_cal_list[i] <= 0.0:
                Ical = eps
            else:
                Ical = Idi_cal_list[i]
            d  = log(Iobs) - log(Ical)
        else:
            d  = Idi_obs_list[i] - Idi_cal_list[i]
        fmin += d * d

    fmin = sqrt(fmin / nx)

    fit.yc_list = yc_list
    fit.Vgi_list = Vgi_list
    fit.Vdi_list = Vdi_list
    fit.Idi_obs_list = Idi_obs_list
    fit.Idi_cal_list = Idi_cal_list

    return fmin

def cal_total_fmin(xk_all, x_list, y_list, w_list, fit, run):
    p_tot = cal_penalty(xk_all, fit)
    fmin = cal_fmin(xk_all, x_list, y_list, w_list, fit, run)
    fmin = fmin + p_tot

    return fmin

def minimize_func(xk_extracted, x_list = None, y_list = None, w_list = None, fit = None, run = True):
    if fit.icall > fit.nmaxcall:
        print(f"\nMessage: # of minimize_func calls [{fit.icall}] exceesds nmaxcall [{fit.nmaxcall}].")
        print(f"           Will stop optimization\n")
        fit.stop_flag = True

    if fit.stop_flag:
        return 1.0e300

    if x_list is None:
        x_list = fit.x_list
    if y_list is None:
        y_list = fit.y_list

    print()
    xk_all = fit.recover_parameters(xk_extracted, set_member = False)
    fmin = cal_total_fmin(xk_all, x_list, y_list, w_list, fit, run)

    if fit.cparams.fhistory:
        print(f"Add fmin and parameters to [{fit.historyfile}]")
        iter = add_history(fit.historyfile, fit.varname, xk_all, fmin)
        if iter is not None:
            print(f"#{iter}: fmin={fmin:12.4g} ", end = '')
            for v in xk_all:
                print(f" {v:12.6g}", end = '')
            print()
        else:
            iter = None
            print(f"  Warning: Failed to add to {fit.historyfile}]")
            print(f"           Check write permission / disk space etc")

    if fit.cparams.ffitfiles and iter is not None:
        save_path = f'fit{iter:04}.xlsx'
        print(f"Save last input and calculation data  to [{save_path}]")
        if hasattr(fit, 'yini_list'):
            ret = save_data(save_path, ['VG', 'VD', 'Id(in)', 'Id(ini)', 'Id(fin)'], [fit.x_list[0], fit.x_list[1], fit.y_list[0], fit.yini_list[0], fit.yc_list[0]])
        else:
            ret = save_data(save_path, ['VG', 'VD', 'Id(in)', 'Id(fin)'], [fit.x_list[0], fit.x_list[1], fit.y_list[0], fit.yc_list[0]])
        if not ret:
            print(f"  Warning: Failed to save to [{save_path}]")
            print(f"           Check write permission / disk space etc")

    fit.icall += 1
    fit.fmin = fmin

    return fmin
