import os
import platform
import sys
import csv
import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt
import re
from matplotlib import pyplot as plt
import pandas as pd

try:
    from ctypes import CDLL, c_int, c_double, POINTER, byref
except:
    print()
    print("#======================================================")
    print("# Warning: Failed to import CDLL libry.")
    print("# mode=plot DLL will not work")
    print("#======================================================")


from tklib.tksci.tksci import log10, e, kB
from tklib.tkutils import getarg, getintarg, getfloatarg, pconv_by_type, print_line, print_data, joinf
from tklib.tkapplication import tkApplication
from tklib.tkparams import tkParams
from tklib.tkvariousdata import tkVariousData
#from tklib.tkfilter import tkFilter
from tklib.tkgraphic.tkplotevent import tkPlotEvent
from tklib.tksci.tkFit_mxy import tkFit_mxy

from tklib.tksci.tkFit_lib import conv_input, save_data, add_history
from minimize_func_ATLAS import read_ATLAS_log_file, read_ATLAS_DOS_file, read_data, ycal_list, save_parameters
from minimize_func_ATLAS import initialize_minimize_func, ycal_list, minimize_func, callback, extract_data, update_files


"""
Fitting by SIMPLEX and Gaussian Process regressions
"""


ProgramName = "ATLAS_Fit_flex"


def initialize():
    app          = tkApplication()
    argv, narg   = app.get_argv()
    app.cparams  = tkParams()
    cparams      = app.cparams
    cparams.debug       = 0
    cparams.print_level = 0

    initialize_minimize_func()

    app.add_argument(opt = None, type = "str", var_name = 'file1',  opt_str = "file1 (for mode=plot: [DOS|IV|history|fitDLL]", desc = 'additional file #1', defval = '')
    app.add_argument(opt = None, type = "str", var_name = 'file2',  opt_str = "file2", desc = 'additional file #2', defval = '')
    app.add_argument(opt = None, type = "str", var_name = 'file3',  opt_str = "file3", desc = 'additional file #3', defval = '')
    app.add_argument(opt = None, type = "str", var_name = 'file4',  opt_str = "file4", desc = 'additional file #4', defval = '')
    app.add_argument(opt = None, type = "str", var_name = 'file5',  opt_str = "file5", desc = 'additional file #5', defval = '')

    app.add_argument(opt = "--help", type = "str", defval = 0, optional = True)
    app.add_argument(opt = "--mode", type = "str", var_name = 'mode',  opt_str = "--mode=[init|update|sim|scan|fit|plot]", desc = 'task mode',
                     defval = 'fit', optional = True)

    app.add_argument(opt = '--infile', type = "str", var_name = 'infile', opt_str = "--infile=path",  desc = 'reference (observed IV) Excel file', 
                     defval = 'Vd_0.1V_L8_W20.xlsx', optional = True)
    app.add_argument(opt = '--IVfile', type = "str", var_name = 'IVfile', opt_str = "--IVfile=path",  desc = 'observed IV output Excel file', 
                     defval = 'IVobs.xlsx', optional = True)
    app.add_argument(opt = "--fitfile", type = "str", var_name = 'fitfile', opt_str = "--fitfile=path",  desc = 'Fiting result file', 
                     defval = 'IVfit.xlsx', optional = True)

#    app.add_argument(opt = "--transferfile", type = "str", var_name = 'transferfile', opt_str = "--transferfile=path",  desc = 'Data file for transfer', 
#                     defval = 'IVfit.xlsx', optional = True)

    app.add_argument(opt = "--atlasfile", type = "str", var_name = 'atlasfile', opt_str = "--atlasfile=path",  desc = 'ATLAS input file', 
                     defval = '303K_Vd01.in', optional = True)
    app.add_argument(opt = '--templatepath', type = "str", var_name = 'templatepath', opt_str = "--templatepath=path",  desc = 'template file path (*.in)', 
                     defval = './template', optional = True)
    app.add_argument(opt = "--outfile", type = "str", var_name = 'outfile', opt_str = "--outfile=path",  desc = 'output Excel file', 
                     defval = '303K_Vd01.out', optional = True)
    app.add_argument(opt = "--Cfile", type = "str", var_name = 'Cfile', opt_str = "--Cfile=path",  desc = 'ATLAS C-Interpreter file', 
                     defval = 'mobility.c', optional = True)
#                     defval = 'cal_mu_model240312_IGOX.c', optional = True)
    app.add_argument(opt = "--defectaccfile", type = "str", var_name = 'defectaccfile', opt_str = "--defectaccfile=path",  desc = 'ATLAS acceptor-type DOS file', 
                     defval = 'defacc.out', optional = True)
    app.add_argument(opt = "--defectdonfile", type = "str", var_name = 'defectdonfile', opt_str = "--defectdonfile=path",  desc = 'ATLAS donor-type DOS file', 
                     defval = 'defdon.out', optional = True)
    app.add_argument(opt = "--atlaslogfile", type = "str", var_name = 'atlaslogfile', opt_str = "--atlaslogfile=path",  desc = 'ATLAS output (.log) file', 
                     defval = 'idvg_vd001.log', optional = True)
    app.add_argument(opt = "--calfile", type = "str", var_name = 'calfile', opt_str = "--calfile=path",  desc = 'ATLAS output cal (.dat) file', 
                     defval = "idvg_calc.dat", optional = True)

    app.add_argument(opt = "--target_var", type = "str",   defval = 'conc')
    app.add_argument(opt = "--x0",         type = "float", defval = 0.0)
    app.add_argument(opt = "--x1",         type = "float", defval = 1.0)
    app.add_argument(opt = "--nx",         type = "int",   defval = 5)

    app.add_argument(opt = "--Eg", type = "float", desc = 'Band gap', defval = 2.8)
    app.add_argument(opt = "--T", type = "float", desc = 'Temperature', defval = 303.0)

    app.add_argument(opt = "--ix_plot", type = "int|str", defval = 0, desc = 'index of x_data to plot as x-axis', optional = True)
    app.add_argument(opt = "--x_scale", type = "str", defval = "", desc = 'x scale to plot [blank|log]', optional = True)
    app.add_argument(opt = "--fhistory", type = "int", var_name = 'fhistory',  opt_str = "--fhisotry=[0|1]", desc = 'flag to save hisotry file',
                     defval = 1, optional = True)
    app.add_argument(opt = "--ffitfiles", type = "int", var_name = 'ffitfiles',  opt_str = "--ffitfiles=[0|1]", desc = 'flag to save fit files',
                     defval = 1, optional = True)
    app.add_argument(opt = "--fplot", type = "int", var_name = 'fplot',  opt_str = "--fplot=[0|1]", desc = 'flag to plot graph',
                     defval = 1, optional = True)

    app.varname    = ["conc",     "qf1",    "nta1", "wta1",    "nga1", "wga1", "ega1",    "nga2", "wga2", "ega2",    "nga3", "wga3", "ega3",    "nga4", "wga4", "ega4"]
    app.unit       = [    "",        "", "cm-3/eV",   "eV", "cm-3/eV",   "eV",   "eV", "cm-3/eV",   "eV",   "eV", "cm-3/eV",   "eV",   "eV", "cm-3/eV",   "eV",   "eV"]
    app.pk_convert = [ "log",        "",     "log",     "",     "log",     "",     "",     "log",     "",     "",     "log",     "",     "",     "log",     "",     ""]
    app.optid      = [     1,         1,         0,      0,         0,      1,      1,         1,      1,      1,         1,      1,      1,         1,      1,      1]
    app.x0         = [ 1.0e17, -5.08362,    0.0e20,  0.020,    3.0e16,   0.00,    0.0,    5.8e15,   0.41,   0.20,    5.8e15,   0.21,   0.30,    5.8e15,   0.21,   0.50]
    app.dx         = [ 5.0e17,      5.0,    2.0e20, 1.0e-5,    1.0e19,   0.10,   0.00,    2.0e17,   0.10,   0.05,    2.0e16,   0.20,   0.30,    2.0e17,   0.10,   0.05]
    app.kmin       = [ 1.0e13,  -1.0e23,    0.0e10, 1.0e-6,    0.0e10,  0.001,  -0.20,    0.0e10,   0.01,   0.01,    0.0e10,   0.01,   0.20,    0.0e10,   0.01,   0.40]
    app.kmax       = [ 1.0e20,   1.0e23,    1.0e23,    2.0,    1.0e23,    2.0,    2.0,    1.0e23,    2.0,    0.3,    1.0e23,    2.0,    2.0,    1.0e23,    2.0,    2.0]
    app.kpenalty   = [    1.0,      1.0,       1.0,    1.0,       1.0,    1.0,    1.0,       1.0,    1.0,    1.0,       1.0,    1.0,    1.0,       1.0,    1.0,    1.0]

    for i, varname in enumerate(app.varname):
        app.add_argument(opt = f"--{varname}", type = "float", defval = app.x0[i])
    app.add_argument(opt = f"--y_convert", type = "str", defval = "log")
    app.add_argument(opt = f"--f_max_plot", type = "float", defval = 10.0)

    app.add_argument(opt = "--method", type = "str", var_name = 'method',  opt_str = "--method=[nelder-mead]", desc = 'optimization algorism',
                     defval = 'nelder-mead', optional = True)
    app.add_argument(opt = "--jac", type = "str", var_name = 'jac',  opt_str = "--jac=[3-points|2-points|func]", desc = 'First differential',
                     defval = '3-points', optional = True)
    app.add_argument(opt = "--nmaxiter", type = "int", var_name = 'nmaxiter',  opt_str = "--nmaxiter=int(>1)", desc = 'maximum interation number for optimization',
                     defval = 1000, optional = True)
    app.add_argument(opt = "--nmaxcall", type = "int", var_name = 'nmaxcall',  opt_str = "--nmaxcall=int(>1)", desc = 'maximum number for minimize_func calls',
                     defval = 10000, optional = True)
    app.add_argument(opt = "--tol", type = "float", var_name = 'tol',  opt_str = "--tol=val", desc = 'eps for optimization',
                     defval = 1.0e-5, optional = True)

    app.add_argument(opt = "--print_interval", type = "int", var_name = 'print_interval',  opt_str = "--print_interval=val", desc = 'print interval',
                     defval = 5, optional = True)
    app.add_argument(opt = "--plot_interval", type = "int", var_name = 'plot_interval',  opt_str = "--plot_interval=val", desc = 'plot interval',
                     defval = 5, optional = True)

#=============================
# Graph configuration
#=============================
    cparams.figsize = [8, 6]
    app.add_argument(opt = "--fontsize",            type = "int", defval = 16, optional = True)
    app.add_argument(opt = "--legend_fontsize",     type = "int", defval = 12, optional = True)
    app.add_argument(opt = "--graphupdateinterval", type = "int", defval = 10, optional = True)

    arg0 = app.check_arg(0, None, vartype = 'str')
    arg_help = app.check_arg('--help', None, vartype = 'int')
    if arg0 == '?' or arg0 == 'help' or arg_help:
        app.terminate(usage = 'build-in', pause = True)
    
    return app, cparams


#=============================
# Treat argments
#=============================
def update_vars(app, cparams, apply_default = True):
    args_opt, args_idx, args_vars = app.read_args(vars = cparams, check_allowed_args = True, apply_default = apply_default)
    if args_opt is None:
        error_no = args_idx
        error_message = args_vars
        app.terminate("\n\n"
                   +  "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!\n"
                   + f"!  {error_message}\n"
                   +  "!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!",
                   usage = app.usage)

    return args_opt, args_idx, args_vars

#==========================================
# functions
#==========================================
def init_fit(app, cparams):
#    cparams.print_parameters(heading = "\ncparams:")
    print("mode                 : {}".format(cparams.mode))
    print("fplot                : {}".format(cparams.fplot))
    print("fhistory             : {}".format(cparams.fhistory))
    print("ffitfiles            : {}".format(cparams.ffitfiles))
    print("stop path            : {}".format(cparams.stopfile))
    print("template path        : {}".format(cparams.templatepath))
    print("Observed IV infile   : {}".format(cparams.infile))
    print("Observed IV outfile  : {}".format(cparams.IVfile))
    print("ATLAS input file     : {}".format(cparams.atlasfile))
    print("ATLAS log file       : {}".format(cparams.atlaslogfile))
    print("ATLAS cal file       : {}".format(cparams.calfile))
    print("outfile              : {}".format(cparams.outfile))
    print("Fitting file         : {}".format(cparams.fitfile))
#    print("data file (transfer) : {}".format(cparams.transferfile))
    print("parameter file       : {}".format(cparams.parameterfile))
#    print("parameter backup file: {}".format(cparams.parameterbkfile))
    print("history file         : {}".format(cparams.historyfile))
    print()
    print(f"Fitting configuration")
    print(f"  method   : {cparams.method}")
    print(f"  tol      : {cparams.tol}")
    print(f"  nmaxiter : {cparams.nmaxiter}")
    print(f"  nmaxcall : {cparams.nmaxcall}")
    print(f"  y_convert: {cparams.y_convert}")

    fit = tkFit_mxy(method = cparams.method, jac = cparams.jac, tol = cparams.tol, nmaxiter = cparams.nmaxiter, nmaxcall = cparams.nmaxcall,
                fplot = cparams.fplot, print_interval = cparams.print_interval, plot_interval = cparams.plot_interval)
    fit.configure(print_level = 0, cparams = cparams)
    fit.copy_attributes(cparams, ["T", "y_convert", "infile", "outfile", "historyfile", "fitfile", 
                                  "templatepath", "IVfile", "atlasfile", "Cfile", "defectaccfile", "defectdonfile", "atlaslogfile", "calfile"])

# fitのフィッティング変数・条件の定義
    fit.copy_attributes(app, ["varname", "unit", "pk_convert", "optid", "dx", "kmin", "kmax", "kpenalty"])
    x0 = fit.build_parameter_list(fit.varname, source_obj = cparams)
# 起動時引数のパラメターを設定
    fit.pk         = x0.copy()
    nvars = len(fit.pk)
# iniファイルのパラメターを設定
    fit.read_parameters(cparams.parameterfile, section = "Parameters", keys = fit.varname)
#    fit.print_variables()

    fit.initial_simplex = None
    if fit.dx:
        fit.initial_simplex = fit.build_initial_simplex()
        for i, splx in enumerate(fit.initial_simplex):
            print(f"  {i}: ", end = '')
            for v in splx:
                print(f" {v:12.4g}", end = '')
            print()
    
# fitから呼び出される関数の定義
# xsはxの値のセット
# x_list[i][j] はi番目のx変数のj番目の値
    if cparams.jac == 'func':
        fit.jac = lambda xk: diff1(xk, fit)
    else:
        fit.jac = cparams.jac
    fit.configure(cal_ylist = lambda pk, run = True: ycal_list(None, pk, fit, run = run),
                  minimize_func = lambda pk: minimize_func(pk, fit.x_list, fit.y_list, fit.w_list, fit),
                  callback = lambda pk: callback(pk, fit)
                 )
#    fit.func = lambda xs, pk: ycal(xs, pk)
#    fit.cal_ylist = lambda pk: ycal_list(fit.x_list, pk, fit)
#    fit.cal_ylist = lambda pk: ycal_list(xd_list, pk, fit)
#    fit.minimize_func = lambda pk: minimize_func(pk, fit.x_list, fit.y_list, fit.w_list, fit)
#    fit.callback = lambda pk: callback(pk, fit)
#    fit.jac = '3-points'
#    fit.jac = '2-points'
#    fit.jac = lambda xk: diff1(xk, fit)

# fitのグラフ関連変数の定義
    fit.configure(fplot = fit.fplot, plt = plt, iter = 0, xiter = [], yfmin = [], ycal_list = None, yfin_list = None)

    print()
    fit.print_variables(heading = "Fitting parameters:")

    return fit

def init(app, cparams):
    print("")
    print("Make ini file")
    fit = init_fit(app, cparams)

#    cparams.parameterfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}.in"])
    save_parameters(fit, cparams)

    app.terminate(pause = True)

def scan(app, cparams):
    print("")
    print("Scan parameter for TFT simulations")
    fit = init_fit(app, cparams)

    itarget = None
    for i, varname in enumerate(fit.varname):
        if varname == cparams.target_var:
            itarget = i
            break
    else:
        print(f"Error: target var [{cparams.target_var}] is not in the varname list ", fit.varname)
        app.terminate(pause = True)

    print()
    print(f"Read input data from [{cparams.infile}]")
    fit.labels, xd_list, yd_list = read_data(cparams.infile)
    if xd_list is None:
        print()
        print(f"Error in scan(): Can not read [{cparams.infile}]")
        print()
        app.terminate(pause = True)
    print(f"Save input data repeat to [{fit.IVfile}]")
    save_data(fit.IVfile, fit.labels, xd_list + yd_list)

    print()
    print("Configure fit (tkFit_mxy object):")
    fit.configure(xlabels = [fit.labels[0]], ylabels = [fit.labels[1]], w_list = None)

# 測定データとシミュレーションデータでデータ数が違うので、
# シミュレーションデータの x_list(VG,VD)に合わせて測定データを抜き出し、.x_list, .y_listに入れる
    print()
    print(f"Read simulation data from [{fit.atlaslogfile}]")
    yini_list = fit.cal_ylist(fit.pk, run = True)
    labels, xini_list, yini_list = read_data(fit.atlaslogfile, 'calc')
    print(f"Extract measurement data for VG and VD included in the simulation data")
    Vgi_list, Vdi_list, Idi_obs_list, Idi_ini_list = extract_data(xd_list, yd_list, xini_list, yini_list)
    fit.x_list    = [Vgi_list, Vdi_list]
    fit.y_list    = [Idi_obs_list]
    fit.xini_list = [Vgi_list, Vdi_list]
    fit.yini_list = [Idi_ini_list]

    print(f"Save extracted input data to [{fit.fitfile}]")
    save_data(fit.fitfile, fit.labels, [fit.x_list[0], fit.x_list[1], fit.y_list[0]])

    x_list = []
    fmin_list = []
    ysim_all_list = []
    if fit.fplot:
        print()
        print("plot")
        fig, axes = plt.subplots(1, 2, figsize = cparams.figsize)
        ax   = axes[0]
        axIV = axes[1]

        ax.tick_params(labelsize = cparams.fontsize)
        ax.plot(x_list, fmin_list, label = 'fmin', marker = 'o')
        ax.set_xlabel(cparams.target_var, fontsize = cparams.fontsize)
        ax.set_ylabel('fmin', fontsize = cparams.fontsize)
        if fit.pk_convert[itarget] == 'log':
            ax.set_xscale('log')
        ax.set_yscale('log')
        ax.legend(fontsize = cparams.legend_fontsize)

        axIV.tick_params(labelsize = cparams.fontsize)
        axIV.plot(xd_list[0], yd_list[0], label = 'measured', linestyle = '-')
        axIV.plot(Vgi_list, Idi_obs_list, label = 'measured (extracted)', linestyle = '', marker = 'o')
        axIV.set_xlabel(labels[0], fontsize = cparams.fontsize)
        axIV.set_ylabel(labels[2], fontsize = cparams.fontsize)
        axIV.legend(fontsize = cparams.legend_fontsize)

#        plt.tight_layout()
        plt.pause(1.0e-5)

    eps = max([1.0e-5, abs(cparams.x1 - cparams.x0) * 1.0e-5])
    x_initial = fit.pk[itarget]
    print()
    print(f"Scanning parameter {cparams.target_var}")
    if fit.pk_convert[itarget] == 'log':
        print(f"   range {cparams.x0} - {cparams.x1} for {cparams.nx} points in {fit.pk_convert[itarget]} scale")
        logx0 = log(cparams.x0)
        logx1 = log(cparams.x1)
        logxstep = (logx1 - logx0) / (cparams.nx - 1)
        xrange_list = [x_initial] + [exp(logx0 + i * logxstep) for i in range(cparams.nx)]
    else:
        print(f"   range {cparams.x0} - {cparams.x1} for {cparams.nx} points in linear scale")
        xstep = (cparams.x1 - cparams.x0) / (cparams.nx - 1)
        xrange_list = [x_initial] + [cparams.x0 + i * xstep for i in range(cparams.nx)]
    print("    xrange_list:", [f"{x:10.4g}" for x in xrange_list])

    fmin_min = 1.0e300
    ix_fmin = None
    x_min = None
    xk_fmin = None
    pk0 = fit.pk.copy()
    for ix, x in enumerate(xrange_list):
        fit.pk = pk0.copy()
        fit.pk[itarget] = x

        print()
        print(f"next x({cparams.target_var})={x:12.6g}")
        print(f"  ix={ix:03d}: xk=", fit.pk)
        optpk = fit.extract_parameters(fit.pk)
        fmin = minimize_func(optpk, fit = fit, run = True)
        labels    = ['Vg', 'Vd', 'Id(obs)', 'Id(sim)']
        data_list = [fit.Vgi_list, fit.Vdi_list, fit.Idi_obs_list, fit.Idi_cal_list]
        print_data(labels = labels, data_list = data_list,
                label_format = '{:^15}', data_format = '{:>15.4g}', header = "Simulated IV:", nmax = 20, print_level = 0)
        print(f"  fmin={fmin:12.6g}")
        print(f"Save input and simulated data to [{fit.fitfile}]")
        save_data(fit.fitfile, labels, data_list)

        x_list.append(x)
        fmin_list.append(fit.fmin)
        ysim_all_list.append(fit.Idi_cal_list)
        if fmin < fmin_min:
            ix_fmin = ix
            x_min = x
            xk_fmin = fit.pk.copy()
            fmin_min = fmin

        if fit.fplot:
            ax.lines[0].set_xdata(x_list)
            ax.lines[0].set_ydata(fmin_list)
            ax.set_xlim([min(x_list), max(x_list)])
            ax.set_ylim([min(fmin_list), max(fmin_list)])

            axIV.plot(fit.Vgi_list, fit.Idi_cal_list, label = f'cal#{ix} {cparams.target_var}={x:10.4g}', linewidth = 0.5, linestyle = 'dashed', marker = '^')
            axIV.legend(fontsize = cparams.legend_fontsize)

#            plt.tight_layout()
            plt.pause(1.0e-5)

    print()
    print(f"Take minimum fmin={fmin} at iter={ix_fmin} {cparams.target_var}={x_min:10.4g} with ", xk_fmin)
    print(f"Save parameters to [{cparams.parameterfile}]")
    fit.pk = xk_fmin
    fit.print_variables(heading = "Optimum variables")
    fit.save_parameters(cparams.parameterfile, section = "Parameters", keys = fit.varname)

    if fit.fplot:
#        fig, axes = plt.subplots(1, 1, figsize = cparams.figsize)
#        ax = axes
#        ax.tick_params(labelsize = cparams.fontsize)
#        axIV.plot(xd_list[0], yd_list[0], label = 'measured', linestyle = '', marker = 'o', markersize = 1.5)
#        for i, ylist in enumerate(ysim_all_list):
#            axIV.plot(Vgi_list, ylist, label = f'{cparams.target_var}={x_list[i]:12.4g}', marker = '^')
        plt.pause(1.0e-5)

        app.terminate(pause = True)

def plot_DOS(app, cparams):
    output_xlsx = "ATLAS_DOS.xlsx"
    files = []
    if cparams.file1 == '' or cparams.file1.upper() == 'DOS':
        files.append(cparams.defectaccfile)
    else:
        files.append(cparams.file1)
    if cparams.file2 == '':
        files.append(cparams.defectdonfile)
    else:
        files.append(cparams.file2)
    if cparams.file3 != '':
        files.append(cparams.file3)
    if cparams.file4 != '':
        files.append(cparams.file4)
    if cparams.file3 != '':
        files.append(cparams.file5)
            
    print("")
    print("Plot DOS files")
    print(f"Input DOS files:")
    print("Number of files:", len(files))
    for f in files:
        print(f"  {f}")
    print(f"Output file: {output_xlsx}")

    E_list = []
    DOS_list = []
    print()
    print("Read DOS files:")
    for f in files:
        print(f"  Read [{f}]")
        inf = read_ATLAS_DOS_file(f, app, cparams, print_level = 1)
        E_list = inf["E"]
        DOS = inf["DOS_total"]
        DOS_list.append(DOS)
        print("    ndata:", len(DOS))

    print()
    print(f"Save DOS to [{output_xlsx}]")
    save_data(output_xlsx, ["E(eV)"] + files, [E_list] + DOS_list)


#=========================================================
# Plot graph
#=========================================================
    fig, axes = plt.subplots(1, 1, figsize = cparams.figsize)
    ax = axes

    ax.set_title(files[0])
    ax.tick_params(labelsize = cparams.fontsize)
    for i, dos in enumerate(DOS_list):
        ax.plot(E_list, dos, label = files[i])
    ax.set_xlabel(r"$E - E_V$ (eV)", fontsize = cparams.fontsize)
    ax.set_ylabel(r"DOS /cm$3$", fontsize = cparams.fontsize)
    ax.set_yscale('log')
    ax.set_ylim([1.0e10, 1.0e21])
    ax.legend(fontsize = cparams.legend_fontsize)

    plt.tight_layout()

    plt.pause(0.0001)
    app.terminate(pause = True)


def plot_DLL(app, cparams):
    system = platform.system()
    if system == "Windows":
        dll_path = os.path.splitext(cparams.Cfile)[0] + ".dll"
    else:
        dll_path = os.path.splitext(cparams.Cfile)[0] + ".so"
    
    dll_path = os.path.abspath(dll_path)
    print(f"Run {dll_path}")
    print(f"Eg={cparams.Eg} eV")
    print(f"T={cparams.T} K")
    ATLAS = CDLL(dll_path)

# 関数の引数と戻り値の型を設定
    ATLAS.tofimun.argtypes = [
        c_double, c_double, c_double, c_double, c_double, c_double, c_double, c_double, 
        POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double), POINTER(c_double)
        ]
    ATLAS.tofimun.restype = c_int

    ATLAS.tftacc.argtypes = [
        c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)
        ]
    ATLAS.tftdon.restype = c_int
    ATLAS.tftdon.argtypes = [
        c_double, c_double, POINTER(c_double), POINTER(c_double), POINTER(c_double)
        ]
    ATLAS.tftdon.restype = c_int

    Eperp = 0.0
    Na    = 0.0e0
    Nd    = 1.0e17
    Epar  = 0.0
    xcomp = 0.0
    ycomp = 0.0
    mun       = c_double()
    munp      = c_double()
    munm      = c_double()
    dmundep   = c_double()
    dmundepar = c_double()
    dmundl    = c_double()
    dmundn    = c_double()

    h = 1.0e-4
    ncal = 301
    nconc  = 1.0e16
    nconc2 = 1.0e20
    logn_step = (log(nconc2) - log(nconc)) / (ncal - 1)
    k = exp(logn_step)
    T = 300.0
    N_list = []
    mu_list = []
    mud_list = []
    mud2_list = []
    print("")
    print("cal_mobility() - N:")
    for i in range(ncal):
        ATLAS.tofimun(Eperp, Na, Nd, nconc, Epar, T, xcomp, ycomp, 
                            byref(mun), byref(dmundep), byref(dmundepar), byref(dmundl), byref(dmundn))
        ATLAS.tofimun(Eperp, Na, Nd, nconc * (1.0 + h), Epar, T, xcomp, ycomp, 
                            byref(munp), byref(dmundep), byref(dmundepar), byref(dmundl), byref(dmundn))
        ATLAS.tofimun(Eperp, Na, Nd, nconc * (1.0 - h), Epar, T, xcomp, ycomp, 
                            byref(munm), byref(dmundep), byref(dmundepar), byref(dmundl), byref(dmundn))
        mu = mun.value
        mud = dmundn.value
        mud2 = (munp.value - munm.value) / (2.0 * h * nconc)
        print(f"TL={T:8.3g}  Ne={nconc:8.4g}   mue={mu:8.4g}   dmue/dNe={mud:8.4g} ({mud2:8.4g})")
        N_list.append(nconc)
        mu_list.append(mu)
        mud_list.append(mud)
        mud2_list.append(mud2)

        nconc *= k

    E0 = 0.0
    E1 = cparams.Eg
    nE = 101
    Estep = (E1 - E0) / (nE - 1)
    E_list = [E0 + i * Estep for i in range(nE)]
    DOSacc_list = []
    DOSdon_list = []
    dosacc = c_double()
    dosdon = c_double()
    sigmae = c_double()
    sigmah = c_double()
    print("")
    print("ATLAS.tftacc()/.tftdon() - E:")
    for E in E_list:
        ATLAS.tftacc(E, cparams.Eg, byref(dosacc), byref(sigmae), byref(sigmah))
        ATLAS.tftdon(cparams.Eg - E, cparams.Eg, byref(dosdon), byref(sigmae), byref(sigmah))
        DOSacc_list.append(dosacc.value)
        DOSdon_list.append(dosdon.value)

# Plot graphs
    fig, axes = plt.subplots(3, 1, figsize = (8, 6))

    ax = axes[0]
    ax.tick_params(labelsize = cparams.fontsize)
    ax.plot(N_list, mu_list, label = 'cal', linestyle = '-', linewidth = 0.5, color = 'black')
    ax.set_xlabel("$N$ (cm$^{-3}$)", fontsize = cparams.fontsize)
    ax.set_ylabel("$\mu$ (cm$^2$/(Vs))", fontsize = cparams.fontsize)
    ax.set_xscale('log')

    ax = axes[1]
    ax.tick_params(labelsize = cparams.fontsize)
    ax.plot(N_list, mud_list, label = r'exact', linestyle = '-', linewidth = 0.5, color = 'black')
    ax.plot(N_list, mud2_list,  label = r'approx', linestyle = '', marker = 'o', markersize = 1.0, color = 'red')
    ax.set_xlabel(r"$N$ (cm$^{-3}$)", fontsize = cparams.fontsize)
    ax.set_ylabel(r"$d\mu/dN_e$ (cm$^5$/(Vs))", fontsize = cparams.fontsize)
    ax.set_xscale('log')

    ax = axes[2]
    ax.tick_params(labelsize = cparams.fontsize)
    ax.plot(E_list, DOSacc_list, label = r'acceptor type', linestyle = '-', linewidth = 0.5, color = 'black')
    ax.plot(E_list, DOSdon_list, label = r'donor type',    linestyle = '-', linewidth = 0.5, color = 'blue')
    ax.set_xlabel(r"$E$ (eV)", fontsize = cparams.fontsize)
    ax.set_ylabel(r"DOS (cm$^{-3}$)", fontsize = cparams.fontsize)
    ax.set_yscale('log')
    ax.set_ylim([1.0e10, 1.0e21])
    ax.legend(fontsize = cparams.legend_fontsize)

    plt.tight_layout()

    plt.pause(0.0001)
    app.terminate(pause = True)

def plot_history(app, cparams):
    cparams.historyfile   = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-history.xlsx"])
    if cparams.file2 == '':
        infile = cparams.historyfile
    else:
        infile = cparams.file2

    print()
    print("Plot hisotry")
    print(f"f_max_plot: {cparams.f_max_plot}")
    print(f"History file: [{infile}]")

    df = pd.read_excel(infile)#, sheet_name = None)
    df = df[df['min:fmin'] <= cparams.f_max_plot]
    labels = df.columns.tolist()
    iter_list = df.iloc[:, 0].values.tolist()
    fmin_list = df.iloc[:, 1].values.tolist()
    y_labels = labels[2:]
    data_list = df.iloc[:, 2:].to_numpy().T.tolist()

    iter_file_list = []
    fmin_file_list = []
    for idx, iter in enumerate(iter_list):
        fitfile = f'fit{iter:04}.xlsx'
        if not os.path.isfile(fitfile):
            continue

        print(f"fitfile found for #{iter} [{fitfile}]")
        iter_file_list.append(iter)
        fmin_file_list.append(fmin_list[idx])

    ngraph = len(labels) - 1
    nx = int(sqrt(ngraph) + 0.99)
    ny = int(ngraph / nx) + 1

# Plot graphs
    fig, axes = plt.subplots(1, 1, figsize = (8, 6))
    ax = axes
#    fig, axes = plt.subplots(nx, ny, figsize = (8, 6))
#    axes = axes.flatten()

#    ax = axes[0]
    ax.tick_params(labelsize = cparams.fontsize)
    ax.plot(iter_list,      fmin_list,      label = 'fmin',            linestyle = '-', linewidth = 0.5, marker = 'o', markersize = 1.0)
    ax.plot(iter_file_list, fmin_file_list, label = 'fmin(with file)', linestyle = '',  marker = '*', markersize = 5.0)
    ax.set_xlabel("iteration", fontsize = cparams.fontsize)
    ax.set_ylabel("fmin", fontsize = cparams.fontsize)
    ax.set_yscale('log')
    ax.legend(fontsize = cparams.legend_fontsize)

    plt.tight_layout()

    def motion(event):
        x = event.xdata
        y = event.ydata
        ln_v.set_xdata(x)
        ln_h.set_ydata(y)
        plt.draw()

    ln_v = plt.axvline(0, linestyle = 'dashed', linewidth = 0.5, color = 'red')
    ln_h = plt.axhline(0, linestyle = 'dashed', linewidth = 0.5, color = 'red')
    plt.connect('motion_notify_event', motion)
    
    plt.pause(0.0001)
    
    app.terminate(pause = True)

def plot_fit(app, cparams):
    if cparams.file2 == '':
        app.terminate("\nError: Fit file must be given.\n", pause = True)
    else:
        infile = cparams.file2

    print()
    print(f"Read fit data from [{infile}]")
    #文字列infileの終わりにある数値をreを使って、取り出す
    m = re.search(r'\d+$', os.path.splitext(infile)[0])
    if m:
        num = int(m.group())
    else:
        num = None
    print(f"  hisotry #: {num}")

    df = pd.read_excel(infile)#, sheet_name = None)
    labels = df.columns.tolist()
    VG_list = df.iloc[:, 0].values.tolist()
    VD_list = df.iloc[:, 1].values.tolist()
    data_list = df.iloc[:, 2:].to_numpy().T.tolist()
    ndata = len(data_list)
    print("labels=", labels)

    if num is not None:
        cparams.historyfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-history.xlsx"])
        print()
        print(f"Read parameters for #{num} from [{cparams.historyfile}]")
        df_h = pd.read_excel(cparams.historyfile)#, sheet_name = None)
        #df_hの '-iter'列から、値がnumの行を取り出す
        df_num = df_h[df_h['-iter'] == num]
 #       print("df_num=", df_num)
        labels_h = df_h.columns.tolist()
        for i, label in enumerate(labels_h):
            val = df_num.iloc[0, i]
            print(f"  {label:15}: {val}")

# Plot graphs
    fig, axes = plt.subplots(1, 1, figsize = (8, 6))
    ax = axes

    ax.tick_params(labelsize = cparams.fontsize)
    if ndata >= 1:
        ax.plot(VG_list, data_list[0], label = labels[2], linestyle = '', marker = 'o', markersize = 3.0)
    if ndata >= 2:
        ax.plot(VG_list, data_list[1], label = labels[3], linestyle = '-', color = 'blue', linewidth = 0.5)
    if ndata >= 3:
        ax.plot(VG_list, data_list[2], label = labels[4], linestyle = '-', color = 'red', linewidth = 0.5)
    ax.set_xlabel(r"$V_g$ (V)", fontsize = cparams.fontsize)
    ax.set_ylabel(r"$I_d$ (A)", fontsize = cparams.fontsize)
    ax.set_yscale('log')
    ax.legend(fontsize = cparams.legend_fontsize)

    plt.tight_layout()

    plt.pause(0.0001)
    app.terminate(pause = True)

def plot_IV(app, cparams):
    output_xlsx = "ATLAS_IV.xlsx"
    files = []
    if cparams.file1.upper() == 'IV' and cparams.file2 != '':
        pass
    elif cparams.file1.upper() == 'IV':
        files.append(cparams.atlaslogfile)
    else:
        files.append(cparams.file1)
    if cparams.file2 != '':
        files.append(cparams.file2)
    if cparams.file3 != '':
        files.append(cparams.file3)
    if cparams.file4 != '':
        files.append(cparams.file4)
    if cparams.file3 != '':
        files.append(cparams.file5)
            
    print("")
    print("Plot IV files")
    print(f"Input IV files:")
    print("Number of files:", len(files))
    for f in files:
        print(f"  {f}")
    print(f"Output file: {output_xlsx}")

    VG_list = []
    VD_list = []
    ID_list = []
    print()
    print("Read IV files:")
    for f in files:
        print(f"  Read [{f}]")
        inf = read_ATLAS_log_file(f, app, cparams, print_level = 1)
        VG_list = inf["VG"]
        VD_list = inf["VD"]
        ID = inf["ID"]
        ID_list.append(ID)
        print("    ndata:", len(ID))

    print()
    print("Save IV to [{output_xlsx}]")
    save_data(output_xlsx, ["VG (V)", "VD (V)"] + files, [VG_list, VD_list] + ID_list)


#=========================================================
# Plot graph
#=========================================================
    fig, axes = plt.subplots(2, 1, figsize = cparams.figsize)
    ax1 = axes[0]
    ax2 = axes[1]

    ax1.tick_params(labelsize = cparams.fontsize)
    ax1.set_title(files[0])
    for i, iv in enumerate(ID_list):
        ax1.plot(VG_list, iv, label = files[i])
    ax1.set_xlabel(r"$V_g$ (V)", fontsize = cparams.fontsize)
    ax1.set_ylabel(r"$I_d$ (A)", fontsize = cparams.fontsize)
    ax1.set_yscale('log')
#    ax1.set_ylim([1.0e12, 1.0e21])
    ax1.legend(fontsize = cparams.legend_fontsize)

    ax2.tick_params(labelsize = cparams.fontsize)
    ax2.set_title(files[0])
    for i, iv in enumerate(ID_list):
        ax2.plot(VD_list, iv, label = files[i])
    ax2.set_xlabel(r"$V_d$ (V)", fontsize = cparams.fontsize)
    ax2.set_ylabel(r"$I_d$ (A)", fontsize = cparams.fontsize)
#    ax2.set_yscale('log')
#    ax2.set_ylim([1.0e12, 1.0e21])
    ax2.legend(fontsize = cparams.legend_fontsize)\

    plt.tight_layout()

    plt.pause(0.0001)
    app.terminate(pause = True)

def plot(app, cparams):
    arg1 = cparams.file1.lower()

    if arg1 == '' or cparams.file1 == 'dos':
        plot_DOS(app, cparams)
    elif arg1 == 'iv':
        plot_IV(app, cparams)
    elif arg1 == 'history':
        plot_history(app, cparams)
    elif arg1 == 'fit':
        plot_fit(app, cparams)
    elif arg1 == 'dll':
        plot_DLL(app, cparams)
    else:
        plot_DOS(app, cparams)

    
def sim(app, cparams):
    print("")
    print("Perform TFT simulation")
    print(f"Eg={cparams.Eg} eV")
    print(f"T={cparams.T} K")
    fit = init_fit(app, cparams)

    print()
    print(f"Read input data from [{cparams.infile}]")
    fit.labels, xd_list, yd_list = read_data(cparams.infile)
    print(f"Save input data repeat to [{fit.IVfile}]")
    save_data(fit.IVfile, fit.labels, xd_list + yd_list)

    print()
    print("Configure fit (tkFit_mxy object):")
    fit.configure(xlabels = [fit.labels[0]], ylabels = [fit.labels[1]], w_list = None)

# 測定データとシミュレーションデータでデータ数が違うので、
# シミュレーションデータの x_list(VG,VD)に合わせて測定データを抜き出し、.x_list, .y_listに入れる
    print()
    print(f"Read simulation data from [{fit.atlaslogfile}]")
    yini_list = fit.cal_ylist(fit.pk, run = True)
    labels, xini_list, yini_list = read_data(fit.atlaslogfile, 'calc')
    print(f"Extract measurement data for VG and VD included in the simulation data")
    Vgi_list, Vdi_list, Idi_obs_list, Idi_ini_list = extract_data(xd_list, yd_list, xini_list, yini_list)
    fit.x_list    = [Vgi_list, Vdi_list]
    fit.y_list    = [Idi_obs_list]
    fit.xini_list = [Vgi_list, Vdi_list]
    fit.yini_list = [Idi_ini_list]
    
    print(f"Save extracted input data to [{fit.fitfile}]")
    save_data(fit.fitfile, fit.labels, [fit.x_list[0], fit.x_list[1], fit.y_list[0]])

    optpk = fit.extract_parameters()
    minimize_func(optpk, fit = fit, run = False)

    print_data(labels = fit.labels + ['ini'], 
            data_list = fit.x_list + fit.y_list + fit.yini_list,
            label_format = '{:^15}', data_format = '{:>15.4g}', header = "Simulated IV:", nmax = 20, print_level = 0)

    print(f"Save input and initial data to [{fit.fitfile}]")
    save_data(fit.fitfile, fit.labels + ['ini'], fit.x_list + fit.y_list + fit.yini_list)

    if fit.fplot:
        print("plot")
        fig, axes = plt.subplots(1, 1, figsize = cparams.figsize)
        ax = axes
        ax.tick_params(labelsize = cparams.fontsize)
        ax.plot(xd_list[0], yd_list[0], label = 'measured', linestyle = '', marker = 'o', markersize = 1.5)
        ax.plot(Vgi_list, Idi_obs_list, label = 'measured (extracted)', linestyle = '', marker = 'o')
        ax.plot(fit.x_list[0], fit.yini_list[0], label = 'simulated', marker = '^')
        ax.set_xlabel(labels[0], fontsize = cparams.fontsize)
        ax.set_ylabel(labels[2], fontsize = cparams.fontsize)
        ax.legend(fontsize = cparams.legend_fontsize)
        plt.pause(0.01)
        input("\nPress ENTER to terminate>>")
        print()


def fit(app, cparams):
    print("")
    print("Fitting to TFT simulation")
    print(f"Eg={cparams.Eg} eV")
    print(f"T={cparams.T} K")
    fit = init_fit(app, cparams)

    print()
    print(f"Read input data from [{cparams.infile}]")
    fit.labels, xd_list, yd_list = read_data(cparams.infile)
    print(f"Save input data repeat to [{fit.IVfile}]")
    save_data(fit.IVfile, fit.labels, xd_list + yd_list)

    print()
    print("Configure fit (tkFit_mxy object):")
    nx = len(xd_list)
    ny = len(yd_list)
    print("nx:", nx)
    print("ny:", ny)
    fit.xlabels = fit.labels[0:nx]
    fit.ylabels = fit.labels[nx:]
    fit.configure(xlabels = [fit.labels[0]], ylabels = [fit.ylabels[0]], w_list = None)

# 測定データとシミュレーションデータでデータ数が違うので、
# シミュレーションデータの x_list(VG,VD)に合わせて測定データを抜き出し、.x_list, .y_listに入れる
    print()
    print(f"Read simulation data from [{fit.atlaslogfile}]")
    yini_list = fit.cal_ylist(fit.pk, run = True)
    labels, xini_list, yini_list = read_data(fit.atlaslogfile, 'calc')
    print(f"Extract measurement data for VG and VD included in the simulation data")
    Vgi_list, Vdi_list, Idi_obs_list, Idi_ini_list = extract_data(xd_list, yd_list, xini_list, yini_list)
    fit.x_list    = [Vgi_list, Vdi_list]
    fit.y_list    = [Idi_obs_list]
    fit.xini_list = [Vgi_list, Vdi_list]
    fit.yini_list = [Idi_ini_list]
    
    print(f"Save extracted input data to [{fit.fitfile}]")
    save_data(fit.fitfile, fit.labels, [fit.x_list[0], fit.x_list[1], fit.y_list[0]])

    print("")
    print(f"Calculate initial IV curve")
    optpk = fit.extract_parameters()
    fini = fit.minimize_func(optpk)

    print(f"Initial function: fmin={fini:10.4g}")
    print_data(labels = ["x", "y_obs", "y_ini"], 
            data_list = [fit.x_list[0], fit.y_list[0], fit.yini_list[0]], 
            label_format = '{:^15}', data_format = '{:>15.4g}', header = "Initial data:", nmax = 20, print_level = 0)
    print("fini=", fini)
    
    print(f"Save input and initial data to [{fit.fitfile}]")
    save_data(fit.fitfile, fit.labels + ['ini'], fit.x_list + fit.y_list + fit.yini_list)
    
#========================================
# plot
#========================================
    if fit.fplot:
        fig, axes = plt.subplots(len(fit.y_list), 2, figsize = cparams.figsize)
#    axes = [axes]
#    axes = axes.flatten()

        fit.initial_plot(data_axes = [axes[0], axes[1]], error_axis = axes[1], yini_list = fit.yini_list, label_ini = 'initial',
                    ix_plot = cparams.ix_plot, x_scale = cparams.x_scale, fmin = fini, plt = plt, fig = fig,
                    fontsize = cparams.fontsize, legend_fontsize = cparams.legend_fontsize)
        fit.plot_event.remove('error')

#========================================
# Optimize
#========================================
    print("")
    print("Optimize:")
    fit.print_variables(heading = "Variables")
    pfin, ffin, success, res = fit.minimize(cparams.method, jac = fit.jac, initial_simplex = fit.initial_simplex)
    fit.pk = pfin.copy()
    if success:
        print(f"Converged at iteration: {res.nit}")
    else:
        print(f"Function did not converge")

#========================================
# Final result
#========================================
    fit.retrieve_parameter_list(pfin, fit.varname, target = fit.cparams)

    fit.print_variables(heading = "Final parameters:", fmin = ffin)

    print("")
    print(f"Save parameters to [{cparams.parameterfile}]")
    save_parameters(fit, cparams)

    yfin_list = fit.cal_ylist(pfin, run = True)
    print("")
    print_data(labels = [fit.xlabels[0], "y_obs", "y_cal"], 
            data_list = [fit.x_list[0], fit.y_list[0], yfin_list[0]], 
                         label_format = '{:^15}', data_format = '{:>15.4g}', header = "Final data:", nmax = 20, print_level = 0)
#    fit.print_scores(heading = "\nScores between y(input) and y(fit)", y1 = fit.y, y2 = yfin)
    print(f"Save input, initial, and final data to [{cparams.fitfile}]")
    save_data(fit.fitfile, fit.labels + ['ini', 'fin'], fit.x_list +  fit.y_list + fit.yini_list + yfin_list)


#=============================
# グラフの表示
#=============================
    if fit.fplot:
        print("")
        print("Plot optimized")
        fit.finalize_plot(yfin_list, iter = res.nit, fmin = ffin)

        fit.layout()
#       plt.tight_layout()
        plt.pause(0.01)

    print()
    input("Press ENTER to terminate>>")

def update_files0(app, cparams):
    print("")
    print("Update files from templates")
    fit = init_fit(app, cparams)
    print(f"Eg={cparams.Eg} eV")
    print(f"C-Interpreter:", fit.Cfile)
    print(f"ATLAS input file:", fit.atlasfile)
    print(f"defectaccfile: ", fit.defectaccfile)
    print(f"defectdonfile: ", fit.defectdonfile)
    print(f"ATLAS log file:", fit.atlaslogfile)
    print(f"Calculation result file:", fit.calfile)
    fit.print_variables(heading = "Parameters:")

    update_files(fit.pk, fit)


def main():
    print()
    print( "#========================================================")
    print(f"#  {ProgramName}")
    print( "#========================================================")

    app, cparams = initialize()

#--infile引数をチェックし、読み込みファイル名を最初に確定する
#    update_vars(app, cparams,apply_default = True)
    cparams.infile = app.check_arg('--infile', None, vartype = 'int')

    cparams.parameterfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}.in"])
    cparams.historyfile   = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-history.xlsx"])
    cparams.stopfile      = app.replace_path(cparams.infile, template = ["{dirname}", "stop"])
#    cparams.parameterbkfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-back.in"])
    print(f"infile: {cparams.infile}")
    print(f"parameterfile: {cparams.parameterfile}")

    cparams.logfile = app.replace_path(cparams.infile)
    print(f"Open logfile [{cparams.logfile}]")
    app.redirect(targets = ["stdout", cparams.logfile], mode = 'w')
#    app.redirect(targets = ["stdout"], mode = 'a', 
#            redirect_traceback = True, output_traceback = 'stdout', display_type_traceback = 'colored')
#    app.redirect_exception()

#起動時引数で与えられたパラメータを設定
#    update_vars(app, cparams,apply_default = False)
    update_vars(app, cparams,apply_default = True)
#パラメータファイルの読み込み。Parametersセクション以外のみ
    print(f"Read parameters from [{cparams.parameterfile}]")
    for section in ["Preferences", "Flags", "Files", "Scan", "Condition"]:
        cparams.read_parameters(cparams.parameterfile, section = section,
                ignore_keys = ["logfile", "outfile", "parameterfile", "parameterbkfile"])
#起動時引数で与えられたパラメータを再設定
    app.force_given_args(cparams)

    if cparams.mode == 'init':
        init(app, cparams)
    elif cparams.mode == 'update':
        update_files0(app, cparams)
    elif cparams.mode == 'sim':
        sim(app, cparams)
    elif cparams.mode == 'scan':
        scan(app, cparams)
    elif cparams.mode == 'fit':
        fit(app, cparams)
    elif cparams.mode == 'plot':
        plot(app, cparams)
    else:
        app.terminate("Error in main: Invalide mode [{}]".format(cparams.mode), usage = app.usage, pause = True)

#    app.terminate("", usage = app.usage, pause = True)


if __name__ == "__main__":
    main()
