#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <malloc.h>
#include <string.h>

#if !defined MAKE_DLL
#include <template.h>
#endif

#if defined LINUX
#define EXPORT_FUNC __attribute__((visibility("default")))
#else
#define EXPORT_FUNC
#endif

extern double pow(double x, double p);
extern double log(double x);
extern double log10(double x);
extern double exp(double x);


/*
 * General field dependent mobility model for electrons.
 * TOtal FIeld Mobility (parallel and perpendicular field components)
 * Statement: MATERIAL/MOBILITY
 * Parameter: F.TOFIMUN
 * Arguments:
 * Eperp    [in] - perpendicular electric field (V/cm)
 * Na       [in] - acceptor concentration  (/cm^3)
 * Nd       [in] - donor concentration     (/cm^3)
 * nconc    [in] - electron concentration  (/cm^3)
 * Epar     [in] - parallel electric field (V/cm)
 * TL       [in] - lattice temperature     (K)
 * xcomp    [in] - x-species fraction      (0-1)
 * ycomp    [in] - y-species fraction      (0-1)
 * 
 * *mun     [return] - hole mobility (cm^2/Vs)
 * *dmundep [return] - derivative of *mun wrt Eperp
 * *dmundepar [return] - derivative of *mun wrt Epar
 * *dmundl  [return] - derivative of *mun wrt TL
 * *dmundn  [return] - derivative of *mun wrt nconc
*/
EXPORT_FUNC int tofimun(double Eperp, double Na, double Nd, double nconc, double Eparl, 
		    double TL, double xcomp, double ycomp, 
		    double *mun, double *dmundep, double *dmundepar, double *dmundl, double *dmundn)
{
/* physical constant */
static double q  = 1.602e-19 ;       /* q - elementary charge (C) */
static double e  = 1.602176634e-19;  /* C */
static double kB = 1.380649e-23;     /* JK<sup>-1</sup> */
static double kb = 8.617e-5 ;        /* kb = kB / e */


static double xc = 0.3e18;
static double yc = 15.0;

static int norder_a  = 0;
static int norder_b0 = 2;
static int norder_b  = 3;

static double ca[1] = {0.0};
static double cb[2] = {1.2996515567552953, 0.21349493414970788};

    double log_x, log_xc;
    double dlogx;
    
    log_xc = log(xc);
    log_x  = log(nconc);
    dlogx  = log_x - log_xc;

    *dmundep   = 0.0;
    *dmundepar = 0.0;
    *dmundl = 0.0;

    if(dlogx < 0.0) 
    {
        *mun = yc;
        *dmundn = 0.0;
        return 1;
    }
    else
    {
        dlogx = log_x  - log_xc;
        *mun = yc + cb[0] * pow(dlogx, 2.0) + cb[1] * pow(dlogx, 3.0);
        *dmundn = 2.0 * cb[0] * pow(dlogx, 1.0) / nconc + 3.0 * cb[1] *pow(dlogx, 2.0) / nconc;

/*
        *mun = yc;
        *dmundn = 0.0;
        for(i = norder_b0 ; i <= norder_b ; i++)
        {
            c = cb[i - norder_b0];
            *mun += c * pow(dlogx, i);
            *dmundn += i * c * pow(dlogx, i - 1) / nconc;
        }
*/
    }
    
    return 1;
}

/*
 * TFT acceptor trap density as function of energy
 * Statement: DEFECTS
 * Parameter: F.TFTACC
 * Arguments:
 * energy     energy relative to conduction band (eV)
 * eg         band gap (eV)
 * *density   tft acceptor trap density (cm^-3)
 * *sigmae    donor trap electron capture cross-section
 * *sigmah    donor trap hole capture cross-section
 */
int tftacc(double energy, double eg, double *density, double *sigmae, double *sigmah)
{
	double nta1 = {{nta1}};
	double wta1 = {{wta1}};
	double nga1 = {{nga1}};
	double wga1 = {{wga1}};
	double ega1 = {{ega1}};
	double nga2 = {{nga2}};
	double wga2 = {{wga2}};
	double ega2 = {{ega2}};
	double nga3 = {{nga3}};
	double wga3 = {{wga3}};
	double ega3 = {{ega3}};
	double nga4 = {{nga4}};
	double wga4 = {{wga4}};
	double ega4 = {{ega4}};
	double ke1, ke2, ke3, ke4;

	*sigmae = 1.0e-18;
	*sigmah = 1.0e-18;

    ke1 = (ega1 - energy) / wga1;
    ke2 = (ega2 - energy) / wga2;
    ke3 = (ega3 - energy) / wga3;
    ke4 = (ega4 - energy) / wga4;
    *density = nta1 * exp(-energy / wta1)
             + nga1 * exp(-ke1 * ke1)
             + nga2 * exp(-ke2 * ke2)
             + nga3 * exp(-ke3 * ke3)
             + nga4 * exp(-ke4 * ke4);

	return(0);
}

/*
 * TFT donor trap density as function of energy
 * Statement: DEFECTS
 * Parameter: F.TFTDON
 * Arguments:
 * energy     energy relative to valence band (eV)
 * eg         band gap (eV)
 * *density   tft donor trap density (cm^-3)
 * *sigmae    donor trap electron capture cross-section
 * *sigmah    donor trap hole capture cross-section
 */
int tftdon(double energy, double eg, double *density, double *sigmae, double *sigmah)
{
	tftacc(energy, eg, density, sigmae, sigmah);
	return(0);
}

int main(int argc, char **argv)
{
    double Eperp = 0.0;
    double Na    = 0.0e0;
    double Nd    = 1.0e17;
    double nconc = 0.0e0;
    double Epar  = 0.0;
    double TL    = 300.0;
    double xcomp = 0.0;
    double ycomp = 0.0;
    double mun, dmundep, dmundepar, dmundl, dmundn;
    double h, munp, munm, mund;
    double Eg = 3.0;
    double EC = 0.0;
    double EV = 3.0;
    int nE = 51;
    double Estep;
    int i;
    double E;
    double ados, ddos;
    double sigmae, sigmah;


    TL = 300.0;
    h = 1.0e-4;
    printf("\n");
    printf("Return from tofimun():\n");
    for(nconc = 1.0e15 ; nconc <= 1.0e22 ; nconc *= 2.0) 
    {
        tofimun(Eperp, Na, Nd, nconc, Epar, TL, xcomp, ycomp, &mun, &dmundep, &dmundepar, &dmundl, &dmundn);
        tofimun(Eperp, Na, Nd, nconc * (1.0 + h), Epar, TL, xcomp, ycomp, &munp, &dmundep, &dmundepar, &dmundl, &dmundn);
        tofimun(Eperp, Na, Nd, nconc * (1.0 - h), Epar, TL, xcomp, ycomp, &munm, &dmundep, &dmundepar, &dmundl, &dmundn);
        mund = (munp - munm) / (2.0 * h * nconc);
        printf("TL=%8.3f  Ne=%8.4g   mue=%8.4g   dmundn=%8.4g (%8.4g)\n", TL, nconc, mun, dmundn, mund);
    }

	Estep = (EV - EC) / (nE - 1);
    printf("\n");
    printf("Return from tftacc() and tftdon():\n");
    for(i = 0 ; i < nE ; i++) 
    {
        E = EC + i * Estep;
        tftacc(E, Eg, &ados, &sigmae, &sigmah);
        tftdon(E, Eg - E, &ddos, &sigmae, &sigmah);
        printf("E=%8.3f  ados=%8.4g   ddos=%8.4g\n", E, ados, ddos);
	}
}

