import os
import platform
import re
import numpy as np
from numpy import log, abs, sqrt, exp
from scipy import interpolate
import pandas as pd
import openpyxl
import time
from datetime import datetime


from tklib.tkfile import tkFile
from tklib.tkutils import pint, pfloat
from tklib.tkparams import tkParams
from tklib.tksci.tkFit_lib import conv_input, save_data, add_history

import cal_mu_model240312_IGOX as cmu
from cal_mu_model240312_IGOX import cal_mu_list


config = tkParams()

def initialize_minimize_func():
    os = platform.system()
    config.script = 'cal_mu_model240312_IGOX.py'

def get_cmd(xk, fit):
    config.mode = 'fit'
    config.infile = fit.infile
    config.datafile = fit.datafile
    config.fplot = 0
    head = f"python {config.script} {config.mode} {config.infile} {config.datafile}"
    s = ''
    for x in xk:
        s += f" {x}"
    return head + s

def read_data(infile, Tmin, Tmax, Nmin, Nmax, filetype = ''):
    print(f"Read data from [{infile}]")
    print(f"  T range: {Tmin} - {Tmax} K")
    print(f"  N range: {Nmin} - {Nmax} cm-3")

    if not os.path.isfile(infile):
        return None, None, None

    if filetype == 'calc':
        df = pd.read_excel(infile) #, header = True)
        labels = df.columns.tolist()
        data_list = df.values.T
#        print(df)
#        print("labels=", labels, infile)
        if len(labels) >= 4:
            return ['T (K)', 'N (cm-3)', 'mu(obs) (cm2/Vs)', 'mu(sim)'], data_list[0:2], data_list[2:4]

        return ['T (K)', 'N (cm-3)', 'mu(obs) (cm2/Vs)'], data_list[0:2], data_list[2:]
    else:
        data_dict = cmu.read_data(infile, 'IGO+X')

        N_all = []
        T_all = []
        mu_obs_all = []
        for key, d in data_dict.items():
            for i in range(len(d['T'])):
                T = d['T'][i]
                N = d['N'][i]
                mu_obs = d['mu'][i]

                if T < Tmin or Tmax < T:
                    continue
                if N < Nmin or Nmax < N:
                    continue

                T_all.append(T)
                N_all.append(N)
                mu_obs_all.append(mu_obs)
        #zipを使って、T_all,N_all,mu_obs_allを、N_allの昇順に並び替える
        N_all, T_all, mu_obs_all = zip(*sorted(zip(N_all, T_all, mu_obs_all), key=lambda x: x[1]))

#    print("T_all=", T_all)
#    print("N_all=", N_all)
#    print("mu_obs_all=", mu_obs_all)

        return ['T (K)', 'N (cm-3)', 'mu(obs) (cm2/Vs)'], [T_all, N_all], [mu_obs_all]

def ycal_list(x_list, xk, fit = None, run = True):
#    N_II_deg, N_NI, Nii_ndeg, NT = xk
#    mu_cal_all = cal_mu_list(x_list[0], x_list[1], N_II_deg, N_NI, Nii_ndeg, NT)
    mu_cal_all = cal_mu_list(x_list[0], x_list[1], *xk)

    return [mu_cal_all]




    cmd = get_cmd(xk, fit)
    if run:
        print(f"Run [{cmd}]")
        os.system(cmd)
    else:
        print(f"Don't run [{cmd}]")

    labels, x_list, y_list = read_data(fit.datafile, None, fit.cparams.Tmin, fit.cparams.Tmax, fit.cparams.Tmin, fit.cparams.Nmax, 'calc')
    return [y_list[1]]

def save_parameters(fit, cparams):
    print(f"Save parameters to [{cparams.parameterfile}]")
#    cparams.save_parameters(cparams.parameterfile, section = "Preferences")
    flag_keys = ["fplot", "fhistory", "ffitfiles"]
    condiction_keys = ["mode", "method", "jac", "nmaxiter", "nmaxcall", "tol", "y_convert"]
    exclude_keys = flag_keys + condiction_keys + fit.varname \
                    + ["infile", "calfile", "logfile", "outfile", "datafile", "parameterfile", "stopfile", "historyfile", "fitfile"]
    cparams.save_parameters(cparams.parameterfile, section = "Preferences", exclude_keys = exclude_keys)
    cparams.save_parameters(cparams.parameterfile, section = "Flags",      keys = flag_keys)
    cparams.save_parameters(cparams.parameterfile, section = "Condition",  keys = condiction_keys)
    fit.save_parameters    (cparams.parameterfile, section = "Parameters", keys = fit.varname)

def callback(pk, fit, run = True):
    cparams = fit.cparams

    if fit.stop_flag:
        return False

    exist_stopfile = os.path.isfile(cparams.stopfile)
    print("exist_stopfile:", cparams.stopfile, exist_stopfile)
    if exist_stopfile:
        print()
        print(f"Message: Found stop file [{cparams.stopfile}] for early stop.")
        print()
        fit.stop_flag = True
        return False

    fplot = getattr(fit, 'fplot', True) # and fit.fplot
    if fplot:
        w = fit.plt.get_current_fig_manager().window
        if w != fit.window:
            print()
            print(f"Message: Graph window is closed for early stop.")
            print()
            fit.stop_flag = True
            return False

    recoverpk = fit.recover_parameters(pk, set_member = False)

    if fit.iter % fit.print_interval == 0:
        print(f"iter: {fit.iter}")
        n = len(recoverpk)
        for i in range(n):
            print(f"  {fit.varname[i]:10}: {recoverpk[i]:10.4g} {fit.unit[i]}")
        f = fit.minimize_func(pk)
        print(f"    f={f:12.6g}")

        if fit.get('fmin_list', None) is not None:
            fit.iter_list.append(fit.iter + 1)
            fit.fmin_list.append(f)

    if fplot and fit.iter % fit.plot_interval == 0:
        ycal_list = fit.cal_ylist(recoverpk, run = run)

        x_list  = fit.__dict__.get('x_list', None)
        xlabels = fit.__dict__.get('xlabels', None)
        nx = len(x_list)
        if nx == 1 or type(fit.ix_plot) is int:
            if type(fit.ix_plot) is int and fit.ix_plot < nx:
                ix_plot = fit.ix_plot
            else:
                ix_plot = 0
            for i in range(len(fit.y_list)):
                axis = fit.data_axes[i]
                fit.fit_data_list[i][0].set_data(x_list[ix_plot], ycal_list[i])
        else:
            for i in range(len(fit.y_list)):
                axis = fit.data_axes[i]
                fit.fit_data_list[i][0].set_data(range(len(ycal_list[i])), ycal_list[i])

        if fit.get('fmin_list', None) is not None:
            fit.error_data[0].set_data(fit.iter_list, fit.fmin_list)
            fit.error_axis.set_xlim([min(fit.iter_list), max(fit.iter_list)])
            fit.error_axis.set_ylim([min(fit.fmin_list), max(fit.fmin_list)])

        fit.plt.tight_layout()
        fit.plt.subplots_adjust(top = fit.plot_region[0], bottom = fit.plot_region[1])
        fit.plt.pause(0.0001)

    fit.iter += 1

    fit.retrieve_parameter_list(recoverpk, fit.varname, target = cparams)
    save_parameters(fit, cparams)

    return True

def minimize_func(xk, x_list = None, y_list = None, w_list = None, fit = None, run = True):
    if fit.icall > fit.nmaxcall:
        print(f"\nMessage: # of minimize_func calls [{fit.icall}] exceesds nmaxcall [{fit.nmaxcall}].")
        print(f"           Will stop optimization\n")
        fit.stop_flag = True

    if fit.stop_flag:
        return 1.0e300

    if x_list is None:
        x_list = fit.x_list
    if y_list is None:
        y_list = fit.y_list

    print()
    nx = len(x_list[0])
    xkr = fit.recover_parameters(xk, set_member = False)
    xkr0 = xkr.copy()
    kp   = fit.kpenalty
    p_tot = 0.0
    if kp is not None:
        kmin = fit.kmin
        kmax = fit.kmax
        nkp = len(kp)
        for i in range(nkp):
            if xkr[i] < kmin[i]:
                d = xkr[i] - kmin[i]
                p = kp[i] * d * d
                p_tot += p
                print(f"**Warning: [{fit.varname[i]}={xkr[i]}] is smaller than [{kmin[i]}]. Add penalty [{p:10.3g}] to fmin")
                xkr[i] = kmin[i]
            elif xkr[i] > kmax[i]:
                d = xkr[i] - kmax[i]
                p = kp[i] * d * d
                p_tot += p
                print(f"**Warning: [{fit.varname[i]}={xkr[i]}] is larger than [{kmax[i]}]. Add penalty [{p:10.3g}] to fmin")
                xkr[i] = kmax[i]

    yc_list = ycal_list(x_list, xkr, fit, run = run)
    fmin = 0.0
    eps = 1.0e-20
    for i in range(nx):
        ymeas = fit.y_list[0][i]
        ysim  = yc_list[0][i]
        if fit.y_convert == 'log':
            if ymeas <= 0.0:
                ymeas = eps
            if ysim <= 0.0:
                ysim = eps
            d  = log(ymeas) - log(ysim)
        else:
            d  = ymeas - ysim
        fmin += d * d

    fmin = sqrt(fmin / nx) + p_tot

    if fit.cparams.fhistory:
        print(f"Add fmin and parameters to [{fit.historyfile}]")
        iter = add_history(fit.historyfile, fit.varname, xkr0, fmin)
        if iter is not None:
            print(f"#{iter}: fmin={fmin:12.4g} ", end = '')
            for v in xk:
                print(f" {v:12.6g}", end = '')
            print()
        else:
            iter = None
            print(f"  Warning: Failed to add to {fit.historyfile}]")
            print(f"           Check write permission / disk space etc")

    if fit.cparams.ffitfiles and iter is not None:
        save_path = f'fit{iter:04}.xlsx'
        print(f"Save last input and calculation data  to [{save_path}]")
        if hasattr(fit, 'yini_list'):
            ret = save_data(save_path, ['E', 'DOS(in)', 'DOS(ini)', 'DOS(fin)'], [fit.x_list[0], fit.y_list[0], fit.yini_list[0], yc_list[0]])
        else:
            ret = save_data(save_path, ['E', 'DOS(in)', 'DOS(fin)'], [fit.x_list[0], fit.y_list[0], yc_list[0]])
        if not ret:
            print(f"  Warning: Failed to save to [{save_path}]")
            print(f"           Check write permission / disk space etc")

    fit.icall += 1
    fit.fmin = fmin
    fit.yc_list = yc_list

    return fmin
