import os
import sys
import re
import random

try:
    import tklib.tkimport as imp
except Exception as e:
    print()
    print("######################################################################")
    print("###########  ERROR ERROR ERROR ERROR ERROR ERROR #####################")
    print("######################################################################")
    print(f"# Failed to import [tklib.tkimport] module ({e}).")
    print(f"#  Add [tkProg]{os.sep}tklib{os.sep}python to PYTHONPATH variable")
    print(f"#  Current PYTHONPATH:", sys.path)
    print("######################################################################")
    input("Press ENTER to terminate>>")
    exit()

np  = imp.import_lib("numpy", stop_by_error = False)
mpl = imp.import_lib("matplotlib", stop_by_error = False)
pd  = imp.import_lib("pandas", stop_by_error = False)
skl = imp.import_lib("tkinter", stop_by_error = False)
tk  = imp.import_lib("sklearn", stop_by_error = False)
imp.messages(stop_by_error = True)

from numpy import sin, cos, tan, pi, exp, log, sqrt
from matplotlib import pyplot as plt
import matplotlib.widgets as wg
from tkinter import ttk
from tkinter import simpledialog



from tklib.tkutils import print_data, pint, pfloat, index2val, pconv_by_type
from tklib.tkapplication import tkApplication
from tklib.tkparams import tkParams
from tklib.tkvariousdata import tkVariousData
from tklib.tksci.tkFit_lib import save_data
from tklib.tkgraphic.tkplotevent import tkPlotEvent, RangeSelector
from tklib.tkgraphic.tkplot_pyplot import select_plt, tkPlot_pyplot
from tklib.tkgraphic.tkplot_tkinter import tkPlot_tkinter
from tklib.tkgui.tksimple_gui import get_window_from_plt, CustomDialog_with_config
from tklib.tksci.tkmlr import tkMLData, tkMLR

import minimize_func_mu as mf

"""
Fitting by SIMPLEX and Gaussian Process regressions
"""


# plot configurations
widths  = [1.0, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5, 0.5]
markers = ['o', 's', '^', '<', '>', '+', '^', 'x']
msizes  = [5.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0, 3.0]


def init(app, cfg):
    print("")
    print(f"Make ini file [{cfg.parameterfile}]")
    fit = mf.init_fit(app, cfg)
    mf.save_parameters(cfg.parameterfile, fit, cfg)

    app.terminate(pause = True)

def inf(app, cfg):
    print()
    print("Information:")
    fit = mf.init_fit(app, cfg)
    fit.print_variables()
    
def plot(app, cfg):
    global msizes
    
    if cfg.file1 == '':
        if os.path.isfile(cfg.fitfile):
            filetype = 'fitfile'
            cfg.file1 = cfg.fitfile
        elif os.path.isfile(cfg.outfile):
            filetype = 'inputfile'
            cfg.file1 = cfg.outfile
    elif cfg.file1 == 'history':
        filetype = 'history'
        cfg.file1 = cfg.historyfile
    elif cfg.file1 == 'input':
        filetype = 'inputfile'
        cfg.file1 = cfg.outfile
    elif cfg.file1 == 'fit':
        filetype = 'fitfile'
        cfg.file1 = cfg.fitfile

    print("")
    print("Plot")
    print(f"file1   : {cfg.file1}")
    print(f"filetype: {filetype}")

    print()
    fit.x_labels, fit.y_labels, fit.x_list, fit.y_list, fit.w_list = mf.read_data(cfg.infile, 
                    cfg.sample, cfg.Tmin, cfg.Tmax, cfg.Nmin, cfg.Nmax, ix_sort = cfg.ix_sort)
    fit.labels = fit.x_labels + fit.y_labels
    if filetype == 'history':
        fit.labels = fit.labels[0:2]
        fit.y_list = [fit.y_list[0]]

    tkplt, root, tkpyplot = select_plt(use_tkinter = cfg.use_tkplt, plt = plt, parent = None, 
                title = "Plot data")
    fig, axes = tkplt.subplots(1, 1, figsize = cfg.figsize, dpi = 100, tight_layout = True)
#    fig, axes = plt.subplots(1, 1, figsize = cfg.figsize)
    if cfg.use_tkplt:
        tkplt.add_toolbar()
        canvas = tkplt.add_canvas(fig)

    plot_event = tkPlotEvent(plt)
    plot_event.prepare_annotation()

    ax = axes
    ax.tick_params(labelsize = cfg.fontsize)
    if filetype == 'history':
        msizes = [0.3]

    for i in range(len(fit.y_list)):
        width = widths[i % len(widths)]
        marker = markers[i % len(markers)]
        msize  = msizes[i % len(msizes)]
        label = fit.labels[i+1]
        line, = ax.plot(fit.x_list[0], fit.y_list[i], label = label, linewidth = width, marker = marker, markersize = msize)

        format = "{label}\n data#{idata}:\n T={T:8.4f}\n N={N:10.4g}\n mu={mu:10.4g}"
        plot_event.annotation.add_line(label, ax, ax, fit.x_list[0], fit.y_list[i], line, 
                    inf_list = {"T": fit.x_list[0], "N": fit.x_list[1], "mu": fit.y_list[0]},
                    annotation_format = format, inf_format = format)

    plot_event.register_annotation_event(fig, activate = True, print_level = 0)

    if filetype == 'history':
        x_label = 'iteration'
        y_label = 'error'
        ax.set_yscale('log')
    else:
        x_label = fit.x_labels[0]
        y_label = fit.y_labels[0]

    ax.set_xlabel(x_label, fontsize = cfg.fontsize)
    ax.set_ylabel(y_label, fontsize = cfg.fontsize)
    ax.legend(fontsize = cfg.legend_fontsize)

    tkplt.pause(0.01)
    app.terminate(pause = True)

def sim(app, cfg):
    print("")
    print("Calculate mobility")
    fit = mf.init_fit(app, cfg)

    print()
    print(f"Read input data from [{cfg.infile}]")
    fit.x_labels, fit.y_labels, fit.x_list, fit.y_list, fit.w_list = mf.read_data(cfg.infile, 
                    cfg.sample, cfg.Tmin, cfg.Tmax, cfg.Nmin, cfg.Nmax, ix_sort = cfg.ix_sort)
    fit.labels = fit.x_labels + fit.y_labels
    print(f"Save input data repeat to [{fit.outfile}]")
    save_data(fit.outfile, fit.labels, fit.x_list + fit.y_list)

    fit.open_history()
    optpk = fit.extract_parameters()
    mf.minimize_func(optpk, fit = fit, run = False)
    fit.close_history()
    fit.yini_list = fit.yc_list

    print_data(labels = fit.labels + ['ini'], 
            data_list = fit.x_list + fit.y_list + fit.yini_list,
            label_format = '{:^15}', data_format = '{:>15.4g}', header = "Simulated IV:", nmax = 20, print_level = 0)

    print(f"Save input and initial data to [{fit.fitfile}]")
    save_data(fit.fitfile, [fit.x_labels[0], 'input', 'initial'], [fit.x_list[0], fit.y_list[0], fit.yini_list[0]])

    if fit.fplot:
        print("plot")
        tkplt, root, tkpyplot = select_plt(use_tkinter = cfg.use_tkplt, plt = plt, parent = None, 
                title = "Simulation")
        fig, axes = tkplt.subplots(1, 1, figsize = cfg.figsize, dpi = 100, tight_layout = True)
#        fig, axes = plt.subplots(1, 1, figsize = cfg.figsize)
        if cfg.use_tkplt:
            tkplt.add_toolbar()
            canvas = tkplt.add_canvas(fig)

        plot_event = tkPlotEvent(plt)
        plot_event.prepare_annotation()

        ax = axes
        ax.tick_params(labelsize = cfg.fontsize)
        line_obs, = ax.plot(fit.x_list[0], fit.y_list[0],    label = 'input',  linestyle = '', marker = 'o', markersize = 1.5)
        line_sim, = ax.plot(fit.x_list[0], fit.yini_list[0], label = 'simulated', linewidth = 0.5)
        ax.set_xlabel(fit.labels[0], fontsize = cfg.fontsize)
        ax.set_ylabel(fit.labels[1], fontsize = cfg.fontsize)
        ax.legend(fontsize = cfg.legend_fontsize)

        format = "{label}\n data#{idata}:\n T={T:8.4f}\n N={N:10.4g}\n mu={mu:10.4g}"
        plot_event.annotation.add_line("input", ax, ax, fit.x_list[0], fit.y_list[0], line_obs, 
                    inf_list = {"T": fit.x_list[0], "N": fit.x_list[1], "mu": fit.y_list[0]},
                    annotation_format = format, inf_format = format)

        plot_event.annotation.add_line("sim", ax, ax, fit.x_list[0], fit.yini_list[0], line_sim, 
                    inf_list = {"T": fit.x_list[0], "N": fit.x_list[1], "mu": fit.yini_list[0]},
                    annotation_format = format, inf_format = format)

        plot_event.register_annotation_event(fig, activate = True, print_level = 0)

        tkplt.pause(0.01)

    app.terminate(pause = True)

def fit(app, cfg, print_level = True):
    print("")
    print("Fitting to mu-T relation")
    fit = mf.init_fit(app, cfg, print_level = False)
    nvars = len(fit.pk)

    print()
    print(f"Read input data from [{cfg.infile}]")
    fit.x_labels, fit.y_labels, fit.x_list, fit.y_list, fit.w_list \
            = mf.read_data(cfg.infile, cfg.sample, cfg.Tmin, cfg.Tmax, cfg.Nmin, cfg.Nmax, ix_sort = cfg.ix_sort)
    if fit.x_labels is None:
        app.terminate(f"\nError in fit(): Can not read [{cfg.infile}]", pause = True)

    fit.labels = fit.x_labels + fit.y_labels
    print(f"Save input data to [{fit.datafile}]")
    save_data(fit.datafile, fit.labels, fit.x_list + fit.y_list)

    print()
    fit.print_variables(heading = "Fitting parameters:")

    print()
    if type(cfg.ix_plot) is int:
        print(f"ix_plot: {cfg.ix_plot} {fit.x_labels[cfg.ix_plot]}")

    print()
    print(f"Calculate initial relation")
    fit.open_history()
    optpk = fit.extract_parameters()
    fini = fit.minimize_func(optpk)
    fit.yini_list = fit.yc_list

    print(f"Initial function: fmin={fini:10.4g}")
    print_data(labels = [fit.x_labels[0], fit.x_labels[1], "input", "initial"], 
            data_list = [fit.x_list[0], fit.x_list[1], fit.y_list[0], fit.yini_list[0]], 
            label_format = '{:^15}', data_format = '{:>15.4g}', header = "Initial data:", nmax = 20, print_level = 0)
    print("fini=", fini)

    print(f"Save input and initial data to [{fit.datafile}]")
    save_data(fit.datafile, fit.labels + ['ini'], fit.x_list + fit.y_list + fit.yini_list)

#========================================
# plot
#========================================
    if fit.fplot:
        tkplt, root, tkpyplot = select_plt(use_tkinter = cfg.use_tkplt, plt = plt, parent = None, title = "fit")
        fig, axes = tkplt.subplots(len(fit.y_list), 2, figsize = cfg.figsize, dpi = 100, tight_layout = True)
#        fig, axes = plt.subplots(len(fit.y_list), 2, figsize = cfg.figsize)
#    axes = [axes]
#    axes = axes.flatten()
        if cfg.use_tkplt:
            tkplt.create_window(fig)
            tkplt.add_toolbar()
            canvas = tkplt.add_canvas(fig)

        plot_event = tkPlotEvent(plt)
        plot_event.add_stop_button()
        plot_event.prepare_annotation()
        plot_event.prepare_popup_menu(fig, parent = root)

        fit.initial_plot(data_axes = [axes[0], axes[1]], error_axis = axes[1], yini_list = fit.yini_list, label_ini = 'initial',
                    ix_plot = cfg.ix_plot, x_scale = cfg.x_scale, fmin = fini, plt = tkplt, fig = fig,
                    fontsize = cfg.fontsize, legend_fontsize = cfg.legend_fontsize)
        fit.plot_event.remove('error')

        tkplt.tight_layout()
        tkplt.subplots_adjust(top = fit.plot_region[0], bottom = fit.plot_region[1])
        tkplt.pause(0.0001)

#========================================
# Optimize
#========================================
    print("")
    print("Optimize:")
    fit.print_variables(heading = "Variables")
    pfin, ffin, success, res = fit.minimize(cfg.method, jac = fit.jac, initial_simplex = fit.initial_simplex)
    fit.close_history()
    fit.pk = pfin.copy()

    if success:
        print(f"Converged at iteration: {res.nit}")
    else:
        print(f"Function did not converge")

#========================================
# Final result
#========================================
    fit.retrieve_parameter_list(pfin, fit.varname, target = fit.cfg)
    fit.print_variables(heading = "Final parameters:", fmin = ffin)

    print("")
#    print(f"Save parameters to [{cfg.parameterfile}]")
    mf.save_parameters(cfg.parameterfile, fit, cfg)

    yfin_list = fit.cal_ylist(pfin, run = True)
    print("")
    print_data(labels = [fit.x_labels[0], "input", "final"], 
            data_list = [fit.x_list[0], fit.y_list[0], yfin_list[0]], 
                         label_format = '{:^15}', data_format = '{:>15.4g}', header = "Final data:", nmax = 20, print_level = 0)
#    fit.print_scores(heading = "\nScores between y(input) and y(fit)", y1 = fit.y, y2 = yfin)
    print(f"Save input, initial, and final data to [{cfg.datafile}]")
    save_data(fit.datafile, fit.labels + ['ini', 'fin'], fit.x_list +  fit.y_list + fit.yini_list + yfin_list)


#=============================
# グラフの表示
#=============================
    if fit.fplot:
        print("")
        print("Plot optimized")
        fit.finalize_plot(yfin_list, iter = res.nit, fmin = ffin, savefig_path = 'final_fit.png')

        fit.layout()
        tkplt.tight_layout()
        tkplt.subplots_adjust(top = fit.plot_region[0], bottom = fit.plot_region[1])
        tkplt.pause(0.01)

    app.terminate(pause = True, input_text = "\nPress ENTER to terminate>>\n")

def error(app, cfg):
    print()
    print("Estimate errors for fitted parameters:")

    print("Fitting to mu-T relation")
    fit = mf.init_fit(app, cfg)

    print(f"Read input data from [{cfg.infile}]")
    fit.x_labels, fit.y_labels, fit.x_list, fit.y_list, fit.w_list \
            = mf.read_data(cfg.infile, cfg.sample, cfg.Tmin, cfg.Tmax, cfg.Nmin, cfg.Nmax, 
                           ix_sort = cfg.ix_sort)
    fit.labels = fit.x_labels + fit.y_labels
    print(f"Save input data to [{fit.datafile}]")
    save_data(fit.datafile, fit.labels, fit.x_list + fit.y_list)

    print()
    fit.print_variables(heading = "Fitting parameters:")


# For plot
    ngraph = 0
    for ivar in range(len(fit.pk)):
        if fit.optid[ivar] == 1:
            ngraph += 1

    nx_graph = int(sqrt(ngraph) + 0.9999)
    ny_graph = ngraph // nx_graph
    if nx_graph * ny_graph < ngraph:
        ny_graph += 1
    tkplt, root, tkpyplot = select_plt(use_tkinter = cfg.use_tkplt, plt = plt, parent = None, 
                title = "Estimate errors for optimized parameters")
    fig, axes = tkplt.subplots(ny_graph, nx_graph, figsize = cfg.figsize, dpi = 100, tight_layout = True)
#    fig, axes = plt.subplots(ny_graph, nx_graph, figsize = cfg.figsize)
    if cfg.use_tkplt:
        tkplt.add_toolbar()
        canvas = tkplt.add_canvas(fig)

    plot_event = tkPlotEvent(plt)
    plot_event.prepare_annotation()

    if nx_graph == 1:
        axes = [axes]
    elif nx_graph > 1 and ny_graph > 1:
        axes = axes.flatten()

    print()
    print("Error estimation:")
    p_sigma = exp(-0.5)
    ig = 0
    result_list = []
    for ivar in range(len(fit.pk)):
        if fit.optid[ivar] == 0:
            continue

        scale = fit.pk_scale[ivar]
        xlim_min = fit.kmin[ivar]
        xlim_max = fit.kmax[ivar]
        print()
        print(f"{ivar}: {fit.varname[ivar]}={fit.pk[ivar]:12.6g}")
        x_list, p_list = fit.build_likelihood_func(fit.pk, ivar, 
                        xlim_min = xlim_min, xlim_max = xlim_max, 
                        kvariance = cfg.kvariance, ncal = cfg.ncal_error, scale = scale)

        #p_listが最大値を取るインデックスを取得する
        nlist = len(p_list)
        imax  = np.argmax(p_list)
#        print("p_list=", p_list)
        print("  index at maximum p in p_list:", imax)

        x_min_s = None
        if 0 < imax and min(p_list[0:imax]) < p_sigma:
            x, y = zip(*sorted(zip(p_list[0:imax+1], x_list[0:imax+1])))
            x_min_s = np.interp(p_sigma, x, y)
            print(f"  x_min_s at p(sigma): {x_min_s:12.6g}")

        x_max_s = None
        if imax < nlist - 1 and min(p_list[imax:]) < p_sigma:
            x, y = zip(*sorted(zip(p_list[imax:], x_list[imax:])))
            x_max_s = np.interp(p_sigma, x, y)
            print(f"  x_max_s at p(sigma): {x_max_s:12.6g}")

        result_list.append([x_min_s, x_max_s])

        ax = axes[ig]
#        ax.tick_params(labelsize = cfg.fontsize)
        label = f'p({fit.varname[ivar]})'
        line, = ax.plot(x_list, p_list, label = label, linewidth = 1.0)
        #y=0の位置に赤い横線を描く
        ax.axhline(y = 0.0, color = 'red', linestyle = 'dashed', linewidth = 0.5)
        line_cur = ax.axvline(x = fit.pk[ivar], color = 'red', linestyle = 'dashed', linewidth = 0.5)
        ax.axhline(y = p_sigma, color = 'blue', linestyle = 'dashed', linewidth = 0.5)
        line_min = None
        line_max = None
        if x_min_s is not None:
            line_min = ax.axvline(x = x_min_s, color = 'blue', linestyle = 'dashed', linewidth = 0.5)
        if x_max_s is not None:
            line_max = ax.axvline(x = x_max_s, color = 'blue', linestyle = 'dashed', linewidth = 0.5)
        ax.set_xlabel(fit.varname[ivar])#, fontsize = cfg.fontsize)
        ax.set_ylabel('p')#, fontsize = cfg.fontsize)
        ax.set_ylim([0.0, 1.2])
        if scale == 'log':
            ax.set_xscale('log')       
#        ax.legend(fontsize = cfg.legend_fontsize)

        format = "{label}: " + fit.varname[ivar] + "={x:.4g}  p={p:.4f}"
        plot_event.annotation.add_line(label, ax, ax, x_list, p_list, line, 
                    inf_list = {"x": x_list, "p": p_list},
                    annotation_format = format, inf_format = format)

        x0 = fit.pk[ivar]
        xmin = result_list[ig][0]
        xmax = result_list[ig][1]

        format = "x0: " + fit.varname[ivar] + "={x:.4g}  p={p}"
        plot_event.annotation.add_line(label, ax, ax, x_list, p_list, line_cur, 
                    inf_list = {"x": x0, "p": 1.0},
                    annotation_format = format, inf_format = format)

        p_sigma = exp(-0.5)
        if line_min:
            format = "x0 - sigma: " + fit.varname[ivar] + "={x:.4g}  p={p}"
            plot_event.annotation.add_line(label, ax, ax, x_min_s, p_sigma, line_min, 
                    inf_list = {"x": x_min_s, "p": "exp(-1/2)"},
                    annotation_format = format, inf_format = format)

        if line_max:
            format = "x0 + sigma: " + fit.varname[ivar] + "={x:.4g}  p={p}"
            plot_event.annotation.add_line(label, ax, ax, x_max_s, p_sigma, line_max, 
                    inf_list = {"x": x_max_s, "p": "exp(-1/2)"},
                    annotation_format = format, inf_format = format)
        
        ig += 1
        tkplt.pause(0.01)

#    plt.tight_layout()
    plot_event.register_annotation_event(fig, activate = True, print_level = 0)

    ig = 0
    print()
    print("Accuracy within one sigma:")
    for ivar in range(len(fit.pk)):
        if fit.optid[ivar] == 0:
            continue

        x0 = fit.pk[ivar]
        xmin = result_list[ig][0]
        xmax = result_list[ig][1]
        if xmin is None:
            dxmin = f"{'*':^10}"
            xmin  = f"{'*':^12}"
        else:
            dxmin = f"{xmin - x0:8.4g}"
            xmin  = f"{xmin:12.6g}"
        if xmax is None:
            dxmax = f"{'*':^10}"
            xmax  = f"{'*':^12}"
        else:
            dxmax = f"{xmax - x0:8.4g}"
            xmax  = f"{xmax:12.6g}"
        print(f"{ivar:03d}: {fit.varname[ivar]:10}: {x0:12.6g} {dxmin} + {dxmax} in ({xmin} - {xmax})")
        ig += 1

    tkplt.pause(0.01)
    input("\nPress ENTER to terminate>>")

def scan_one(app, cfg, fit, target_var, itarget, xk = None, fmin = None, title = None, pause = True, print_level = 1, outlog = None):
    if xk is None:
        xk = fit.pk

    if fmin is None:
        fmin_min = 1.0e300
    else:
        fmin_min = fmin
        
    if outlog:
        outlog.write(f"  scan_one: itarget={itarget} var={fit.varname[itarget]}: fmin={fmin_min:g}\n")

    if type(cfg.x0) is str:
        if cfg.x0 == '':
            cfg.x0 = fit.kmin[itarget]
        else:
            cfg.x0 = pfloat(cfg.x0, cfg.x0)
    if type(cfg.x1) is str:
        if cfg.x1 == '':
            cfg.x1 = fit.kmax[itarget]
        else:
            cfg.x1 = pfloat(cfg.x1, cfg.x1)

    if outlog:
        outlog.write(f"  Scan var #{itarget} {cfg.target_var}: x range: {cfg.x0} - {cfg.x1}\n")
    
    if print_level:
        print(f"Scan var #{itarget} {cfg.target_var}: x range: {cfg.x0} - {cfg.x1}")

    
    x_list = []
    fmin_list = []
    ysim_all_list = []
    if fit.fplot:
        print()
        print("Plot initialize")
        tkplt, root, tkpyplot = select_plt(use_tkinter = cfg.use_tkplt, plt = plt, parent = None, 
                title = "Scan parameters")
        fig, axes = tkplt.subplots(1, 2, figsize = cfg.figsize, dpi = 100, tight_layout = False)
#        fig, axes = plt.subplots(1, 2, figsize = cfg.figsize)

        if cfg.use_tkplt:
            tkplt.add_toolbar()
            canvas = tkplt.add_canvas(fig)

        plot_event = tkPlotEvent(plt)
        plot_event.add_stop_button()
        plot_event.prepare_annotation()
        plot_event.prepare_popup_menu(fig, parent = root)

        axfmin = axes[0]
        axFit  = axes[1]
        if title:
            axfmin.set_title(title)

        axfmin.tick_params(labelsize = cfg.fontsize)
        line_fmin, = axfmin.plot(x_list, fmin_list, label = 'fmin', marker = 'o')
        axfmin.set_xlabel(cfg.target_var, fontsize = cfg.fontsize)
        axfmin.set_ylabel('fmin', fontsize = cfg.fontsize)
        if fit.pk_scale[itarget] == 'log':
            axfmin.set_xscale('log')
        axfmin.set_yscale('log')
        axfmin.legend(fontsize = cfg.legend_fontsize)

        axFit.tick_params(labelsize = cfg.fontsize)
        idx_list = range(len(fit.x_list[0]))
        line_fit, = axFit.plot(idx_list, fit.y_list[0], label = 'input', linestyle = '-', color = 'red', linewidth = 2.0)
        axFit.set_xlabel("index", fontsize = cfg.fontsize)
#        axFit.set_xlabel(fit.x_labels[0], fontsize = cfg.fontsize)
        axFit.set_ylabel(fit.y_labels[0], fontsize = cfg.fontsize)
        axFit.set_yscale('log')
#        axFit.legend(fontsize = cfg.legend_fontsize)

        format = "{label}\n line#{iline}\n data#{idata}:\n index={index:10.4g}\n mu={mu:10.4g}\n  fmin={fmin:10.4g}"
        plot_event.annotation.add_line("input", axFit, axFit, idx_list, fit.y_list[0], line_fit, 
                    inf_list = {"index": idx_list, "mu": fit.y_list[0], "fmin": 0},
                    annotation_format = format, inf_format = format)

# Don't use tight_layout() because it causes errors for zero data in log plot
#        tkplt.tight_layout()
        plot_event.stop_button.adjust_plot()
        tkplt.pause(1.0e-5)

    eps = max([1.0e-5, abs(cfg.x1 - cfg.x0) * 1.0e-5])
    x_initial = xk[itarget]
    print()
    print(f"Scanning parameter {cfg.target_var}")
    xrange_list = [x_initial] + fit.get_xrange_list(itarget, cfg.nx, print_level = 1)
    print("    xrange_list:", [f"{x:10.4g}" for x in xrange_list])

    fit.open_history()
    ix_fmin = 0
    x_min = xk[itarget]
    xk_fmin = xk.copy()
    xk = xk.copy()
    for ix, x in enumerate(xrange_list):
        if plot_event.stop_button.status == 'stop':
            plot_event.stop_button.set_text('finished')
            return False

        xk[itarget] = x

        print()
        print(f"next x({cfg.target_var})={x:12.6g}")
        pkstr = [f"{v:10.4g}".strip() for v in xk]
        print(f"  ix={ix:03d}: xk=", *pkstr)
        optpk = fit.extract_parameters(xk)
        fmin = mf.minimize_func(optpk, fit = fit, run = True, print_level = print_level)

        simlabels = [f"{s}(sim)" for s in fit.y_labels]
        labels    = [*fit.x_labels, *fit.y_labels, *simlabels]
        data_list = [*fit.x_list, *fit.y_list, *fit.yc_list]
        if print_level:
            print_data(labels = labels, data_list = data_list,
                    label_format = '{:^15}', data_format = '{:>15.4g}', header = "Simulated data:", nmax = 20, print_level = 0)
        print(f"  fmin={fmin:12.6g}")
        print(f"Save input and simulated data to [{fit.fitfile}]")
        save_data(fit.fitfile, labels, data_list)

        x_list.append(x)
        fmin_list.append(fit.fmin)
        ysim_all_list.append(fit.yc_list[0])
        if fmin < fmin_min:
            ix_fmin = ix
            x_min = x
            xk_fmin  = xk.copy()
            fmin_min = fmin

        if fit.fplot:
            if ix == 0:
                axfmin.plot(x_list, fmin_list, label = 'fmin[0]', marker = '*', markersize = 15.0)
            axfmin.lines[0].set_xdata(x_list)
            axfmin.lines[0].set_ydata(fmin_list)
            axfmin.set_xlim([min(x_list), max(x_list)])
            axfmin.set_ylim([min(fmin_list), max(fmin_list)])

#            axFit.plot(fit.x_list[0], fit.yc_list[0], 
            idx_list = range(len(fit.x_list[0]))
            label = f"cal#{ix} {cfg.target_var}=" + f"{x:10.4g}".strip()
            line_fit, = axFit.plot(idx_list, fit.yc_list[0], 
                    label = label, 
                    linewidth = 0.5, linestyle = 'dashed', marker = '^', markersize = 1.5)
#            axFit.legend(fontsize = cfg.legend_fontsize)

            format = "{label}\n line#{iline}\n data#{idata}:\n index={index:10.4g}\n T={T:8.3f} N={N:10.4g}\n mu={mu:10.4g}\n  fmin={fmin:10.4g}"
            plot_event.annotation.add_line(label, axFit, axFit, idx_list, fit.y_list[0], line_fit, 
                    inf_list = {"index": idx_list, "T": fit.x_list[0], "N": fit.x_list[1], "mu": fit.y_list[0], "fmin": fmin},
                    annotation_format = format, inf_format = format)

            tkplt.tight_layout()
            plot_event.stop_button.adjust_plot()

            tkplt.pause(1.0e-5)

        format = "data #{idata}: " + f"{cfg.target_var}" + "={x:10.4g} fmin={fmin:10.4g}"
        plot_event.annotation.add_line("fmin", axfmin, axfmin, x_list, fmin_list, line_fmin, 
                        inf_list = {"x": x_list, "fmin": fmin_list},
                    annotation_format = format, inf_format = format)

        plot_event.register_annotation_event(fig, activate = True, print_level = 0)
#        plot_event.annotation.refresh_cursors()

        popup_menu = plot_event.popup_menu.menu
        popup_menu.add_command(label="run",  command = lambda: print("runが選択されました"))
        plot_event.register_popup_menu_event()

    fit.close_history()

    print()
    print(f"Take minimum fmin={fmin_min} at iter={ix_fmin} {cfg.target_var}={x_min:10.4g}")
#    print(f"Take minimum fmin={fmin_min} at iter={ix_fmin} {cfg.target_var}={x_min:10.4g} with ", xk_fmin)
    print(f"Save parameters to [{cfg.parameterfile}]")

    fit.ix_fmin = ix_fmin
    fit.x_min   = x_min
    fit.pk      = xk_fmin
    fit.fmin    = fmin_min
    fit.print_variables(heading = "Optimum variables")
    fit.save_parameters(cfg.parameterfile, section = "Parameters", keys = fit.varname)

    if fit.fplot:
        if pause:
            app.terminate(pause = True)

    return fmin_min

def mapping(app, cfg):
    eps = 1.0e-5
    plot_fmax = 1.0e3

    print()
    print("Map sampling:")
    fit = mf.init_fit(app, cfg)

    print(f"Read input data from [{cfg.infile}]")
    fit.x_labels, fit.y_labels, fit.x_list, fit.y_list, fit.w_list \
            = mf.read_data(cfg.infile, cfg.sample, cfg.Tmin, cfg.Tmax, cfg.Nmin, cfg.Nmax, 
                           ix_sort = cfg.ix_sort)
    fit.labels = fit.x_labels + fit.y_labels
    print(f"Save input data to [{fit.datafile}]")
    save_data(fit.datafile, fit.labels, fit.x_list + fit.y_list)

    if type(cfg.target_var) is int:
        vars = [str(cfg.target_var)]
    else:
        vars = cfg.target_var.split(',')
    if len(vars) == 0 or (len(vars) == 1 and vars[0] == ''):
        vars = [0, 1]
    elif len(vars) == 1:
        itarget, varname = index2val(vars[0], fit.varname, (None, None))
        if itarget == 1:
            vars = [0, 1]
        else:
            vars = [vars[0], 1]

#    print("vars=", vars)
    itarget_list = []
    var_list = []
    for i, v in enumerate(vars):
        itarget, vars[i] = index2val(v, fit.varname, (None, None))
        itarget_list.append(itarget)
        var_list.append(vars[i])

    print()
    print(f"Target parameters:", *vars)
    itarget0 = itarget_list[0]
    var0     = var_list[0]
    xrange_list0 = [fit.pk[itarget0]] + fit.get_xrange_list(itarget0, cfg.nx, print_level = 1)

    outlog = None
#    outlog = open('out.log', 'a')
    if outlog:
        outlog.write('Started\n')

    fmin = None
    xk = fit.pk.copy()
    for ix, x in enumerate(xrange_list0):
        xk[itarget0] = x
        cfg.target_var = var_list[1]
        itarget = itarget_list[1]
        cfg.x0 = ''
        cfg.x1 = ''
        
        if outlog:
            outlog.write(f"ix={ix}: {var_list[0]}={x}\n") 
            xkstr = 'xk='
            for x in xk:
                xkstr += f"{x:g} "
            outlog.write(xkstr + "\n")

        fmin = scan_one(app, cfg, fit, cfg.target_var, itarget, xk = xk, fmin = fmin, title = f"{var_list[0]}={x:g}", 
                        pause = False, print_level = 0, outlog = outlog)
        if fmin is False:
            break

        xk = fit.pk.copy()

        if outlog:
            outlog.write(f"  Minimum at ix_min={fit.ix_fmin} with {cfg.target_var}={fit.x_min:g}: fmin={fit.fmin:g}")
            xkstr = '    xk='
            for x in xk:
                xkstr += f"{x:g} "
            outlog.write(xkstr)      
    if outlog:
        outlog.write('Finished\n') 
        outlog.close()

    app.terminate(pause = True)

def sampling(app, cfg):
    eps = 1.0e-5
    plot_fmax = 1.0e3
    
    print()
    print("Random sampling:")
    fit = mf.init_fit(app, cfg)

    print(f"Read input data from [{cfg.infile}]")
    fit.x_labels, fit.y_labels, fit.x_list, fit.y_list, fit.w_list \
            = mf.read_data(cfg.infile, cfg.sample, cfg.Tmin, cfg.Tmax, cfg.Nmin, cfg.Nmax, 
                           ix_sort = cfg.ix_sort)
    fit.labels = fit.x_labels + fit.y_labels
    print(f"Save input data to [{fit.datafile}]")
    save_data(fit.datafile, fit.labels, fit.x_list + fit.y_list)

    plot_ymin = min(fit.y_list[0])
    plot_ymax = max(fit.y_list[0])

    print()
    fit.print_variables(heading = "Fitting parameters:")
    print(f"# of sampling: {cfg.nsample}")

    fit.open_history()
    optpk = fit.extract_parameters()
    fmin_min = mf.minimize_func(optpk, fit = fit, run = True)
    i_fmin = 0
    xk_fmin = fit.pk.copy()
    plot_fmax = min([plot_fmax, fmin_min * 100.0])
    print("fmin_min=", fmin_min)

    def inf_format_func(plotevent, sel, iline, idata, label, inf, x = 0, y = 0, opt = None):
        if inf is None:
            return

        iteration = inf["iteration"][idata]
        fmin = inf["fmin"][idata]
        xk   = inf["xk"][idata]
        fit  = inf["fit"]
        print(f"iteration={iteration}  fmin={fmin:10.4g}:")
        for varname, x in zip(fit.varname, xk):
            print(f"  {varname:10}: {x:g}")

    xk_history = [fit.pk]
    x_list = [i_fmin]
    fmin_list = [fmin_min]
    ysim_all_list = []
    if fit.fplot:
        print()
        print("plot")
        tkplt, root, tkpyplot = select_plt(use_tkinter = cfg.use_tkplt, plt = plt, parent = None, 
                title = "Random sampling")

        fig, axes = tkplt.subplots(1, 2, figsize = cfg.figsize, dpi = 100, tight_layout = True)

        plot_event = tkPlotEvent(plt)
        plot_event.add_stop_button()
        plot_event.prepare_annotation()
        plot_event.register_annotation_event(fig, activate = True, print_level = 0)

        axfmin = axes[0]
        axFit  = axes[1]

        axfmin.tick_params(labelsize = cfg.fontsize)
        line_fmin, = axfmin.plot(x_list, fmin_list, label = 'fmin', marker = 'o')
        axfmin.set_xlabel('sampling index', fontsize = cfg.fontsize)
        axfmin.set_ylabel('fmin', fontsize = cfg.fontsize)
        axfmin.set_yscale('log')
        axfmin.legend(fontsize = cfg.legend_fontsize)
        
        axFit.tick_params(labelsize = cfg.fontsize)
        index_list = range(len(fit.y_list[0]))
        line_fit, = axFit.plot(index_list, fit.y_list[0], label = 'input', linestyle = '-', color = 'red', linewidth = 2.0)
        axFit.set_xlabel("index", fontsize = cfg.fontsize)
#        axFit.set_xlabel(fit.x_labels[0], fontsize = cfg.fontsize)
        axFit.set_ylabel(fit.y_labels[0], fontsize = cfg.fontsize)
        axFit.set_ylim([plot_ymin, plot_ymax])
        axFit.set_yscale('log')
        axFit.legend(fontsize = cfg.legend_fontsize)

        format = "{label}\n data#{idata}:\n iteration={iteration}\n fmin={fmin:10.4g}"
        plot_event.annotation.add_line("fmin", axfmin, axfmin, x_list, fmin_list, line_fmin, 
                    inf_list = {"iteration": x_list, "fmin": fmin_list, "fit": fit, "xk": xk_history},
                    annotation_format = format, inf_format = inf_format_func)

        format = "{label}\n data#{idata}:\n index={index:10.4g}\n y={y:10.4g}"
        plot_event.annotation.add_line("input", axFit, axFit, index_list, fit.y_list[0], line_fit, 
                    inf_list = {"index": index_list, "y": fit.y_list[0]},
                    annotation_format = format, inf_format = format)

#        tkplt.tight_layout()
        tkplt.pause(1.0e-5)

    pk_original = fit.pk.copy()
    for i in range(1, cfg.nsample + 1):
        print("status:", plot_event.stop_button.status)
        if plot_event.stop_button.status == 'stop':
            plot_event.stop_button.set_text('finished')
            break

        xk = []
        for ivar in range(len(fit.varname)):
            if fit.optid[ivar] == 0:
                xk.append(pk_original[ivar])
                continue

            rand = random.random()
            if fit.pk_scale[ivar] == 'log':
                xmin = fit.kmin[ivar]
                xmax = fit.kmax[ivar]
                if xmin < eps:
                    xmin = log(eps)
                else:
                    xmin = log(xmin)
                if xmax < eps:
                    xmax = log(eps)
                else:
                    xmax = log(xmax)
                xk.append(exp(xmin + rand * (xmax - xmin)))
            else:
                xmin = fit.kmin[ivar]
                xmax = fit.kmax[ivar]
                xk.append(xmin + rand * (xmax - xmin))

        print(f"#{i:04d}:")
        for varname, x in zip(fit.varname, xk):
            print(f"  {varname:10}: {x:12.6g}")

        optpk = fit.extract_parameters(pk = xk)
        fmin = mf.minimize_func(optpk, fit = fit, run = True, print_level = 0)
        print(f"  fmin={fmin:12.6g}")
#        print(f"Save input and simulated data to [{fit.fitfile}]")
#        save_data(fit.fitfile, labels, data_list)

        xk_history.append(xk)
        x_list.append(i)  # iteration
        fmin_list.append(fit.fmin)
        ysim_all_list.append(fit.yc_list[0])
        if fmin < fmin_min:
            i_fmin = i
            xk_fmin = xk.copy()
            fmin_min = fmin

        if fit.fplot:
            axfmin.lines[0].set_xdata(x_list)
            axfmin.lines[0].set_ydata(fmin_list)
            axfmin.set_xlim([min(x_list), max(x_list)])
            axfmin.set_ylim([min(fmin_list), plot_fmax]) #max(fmin_list)])

            label = f'#{i}'
            index_list = range(len(fit.yc_list[0]))
            line_fit, = axFit.plot(index_list, fit.yc_list[0], 
                    label = label, linewidth = 0.5, linestyle = 'dashed', marker = '^', markersize = 1.5)
#            plot_ymin = max([plot_ymin, min(fit.yc_list[0])])
#            plot_ymax = min([plot_ymax, max(fit.yc_list[0])])
            axFit.set_ylim([plot_ymin, plot_ymax])
#            axFit.legend(fontsize = cfg.legend_fontsize)

            format = "{label}\n data#{idata}:\n index={index:10.4g}\n ycal={y:10.4g}"
            plot_event.annotation.add_line(f"cal #{i}", axFit, axFit, index_list, fit.yc_list[0], line_fit, 
                    inf_list = {"index": index_list, "y": fit.yc_list[0]},
                    annotation_format = format, inf_format = format)

#            tkplt.tight_layout()
            plot_event.stop_button.adjust_plot()  
            tkplt.pause(1.0e-5)

            plot_event.annotation.refresh_cursors()

    fit.close_history()

    print()
    print(f"Take minimum fmin={fmin_min} at sampling #{i_fmin}")
#    print(f"Take minimum fmin={fmin_min} at sampling #{i_fmin} with ", xk_fmin)
    print(f"Save parameters to [{cfg.parameterfile}]")
    fit.pk = xk_fmin
    fit.print_variables(heading = "Optimum variables")
    fit.save_parameters(cfg.parameterfile, section = "Parameters", keys = fit.varname)

    if fit.fplot:
        app.terminate(pause = True, input_text = "\nPress ENTER to terminate>>\n")

def scan_all(app, cfg, fit):
    print()
    print("Scan all parameters:")

    outlog = None
#    outlog = open('out.log', 'a')
    if outlog:
        outlog.write("Start scan_all\n")

    fmin = 1.0e300
    xk = fit.pk.copy()
    for itarget in range(len(fit.varname)):
        cfg.target_var = fit.varname[itarget]
        if fit.optid[itarget] == 0:
            print(f"Var #{itarget} {cfg.target_var} is not optimized. Skip.")
            continue

#        print("Scan var #{itarget} {cfg.target_var]")
        cfg.x0 = ''
        cfg.x1 = ''
        
        if outlog:
            outlog.write(f"itarget={itarget} {cfg.target_var} from fmin={fmin:g}")
            xkstr = 'xk='
            for x in xk:
                xkstr += f"{x:g} "
            outlog.write(xkstr + "\n")

        fmin = scan_one(app, cfg, fit, cfg.target_var, itarget, xk = xk, fmin = fmin, pause = False, print_level = 0, outlog = outlog)
        if fmin is False:
            break

        xk = fit.pk.copy()

        if outlog:
            outlog.write(f"  Minimum at ix_min={fit.ix_fmin} with {cfg.target_var}={fit.x_min:g}: fmin={fit.fmin:g}")
            xkstr = '    xk='
            for x in xk:
                xkstr += f"{x:g} "
            outlog.write(xkstr)      
    if outlog:
        outlog.write('Finished\n') 
        outlog.close()

    app.terminate(pause = True)

def scan(app, cfg):
    print("")
    print("Scan parameter for TFT simulations")
    fit = mf.init_fit(app, cfg)

    print(f"Read input data from [{cfg.infile}]")
    fit.x_labels, fit.y_labels, fit.x_list, fit.y_list, fit.w_list \
            = mf.read_data(cfg.infile, cfg.sample, cfg.Tmin, cfg.Tmax, cfg.Nmin, cfg.Nmax, 
                           ix_sort = cfg.ix_sort)
    fit.labels = fit.x_labels + fit.y_labels
    print(f"Save input data to [{fit.datafile}]")
    save_data(fit.datafile, fit.labels, fit.x_list + fit.y_list)

#    print()
#    fit.print_variables(heading = "Fitting parameters:")

    if type(cfg.target_var) is str and cfg.target_var == 'all':
        itarget = None
    else:
        itarget, cfg.target_var = index2val(cfg.target_var, fit.varname, (None, None))

    if (cfg.target_var is None or cfg.target_var == '') and cfg.target_var != 'all':
        print(f"Error: target var [{cfg.target_var}] is not in the varname list ", fit.varname)
        app.terminate(pause = True)

    if cfg.target_var == 'all':
        return scan_all(app, cfg, fit)
    else:
        return scan_one(app, cfg, fit, cfg.target_var, itarget, print_level = 0)

def mlr(app, cfg):
    mlr           = tkMLR()
    mld_processed = tkMLData()
    mld_converted = tkMLData()
    mld_input_d_train_std = tkMLData()
    mld_input_d_test_std  = tkMLData()
    mld_input_o_train     = tkMLData()
    mld_input_o_test      = tkMLData()
    mld_predict_train     = tkMLData()
    mld_predict_test      = tkMLData()
    cfg.hidden_layer_sizes = (5, 5)

    print()
    print("Machine learning regression:")
    print(f"Training history file: {cfg.trainfile}")
    print(f"Test data ratio: {cfg.test_data_ratio}")
    print(f"Maximum fmin for MLR: {cfg.fmin_max_mlr}")
#    cfg.print_parameters()

    fit = mf.init_fit(app, cfg)
    print()
    print(f"Read input data from [{cfg.trainfile}]")
    mlr.read_data(cfg.trainfile)
    print("Original data:")
    print(f"  Number of columns: {mlr.ncolumns}")
    print(f"  Number of data   : {mlr.ndata}")
#    print("mlr.df:\n", mlr.df)

    mld_processed.df = mlr.extract_data(df = mlr.df, 
                            delete_na = True, limit_max_list = [['min:fmin', cfg.fmin_max_mlr]])
    print("Processed data:")
    print( "  delete non numerical data")
    print(f"  max value for [min:fmin]: {cfg.fmin_max_mlr}")
    print(f"  Number of columns: {mld_processed.ncolumns}")
    print(f"  Number of data   : {mld_processed.ndata}")
#    print(mld_processed.df)

    exclude_labels    = mlr.get_exclude_labels(keys = fit.varname, id = fit.optid, regexp_list = [r"^\-", r"^'\-"])
    objective_labels  = mlr.get_objective_labels(regexp_list = [r"^o:", r"^t:", r"^min:", r"^max:"])
    descriptor_labels = mlr.get_descriptor_labels(labels = mlr.columns, 
                            remove_keys = exclude_labels + objective_labels)
    print("labels:")
#    print("  all           :", mlr.columns)
    print("  descriptors   :", descriptor_labels)
    print("  objective vars:", objective_labels)
    print("  exclude       :", exclude_labels)

    mld_processed.descriptor_df = mld_processed.df[descriptor_labels]
    mld_processed.objective_df  = mld_processed.df[objective_labels]

    print()
    print("Scale:")
    scale_dict = {}
    for varname, scale in zip(fit.varname, fit.pk_scale):
        scale_dict[varname] = scale
    mld_converted.descriptor_df = mlr.convert_df(mld_processed.descriptor_df, scale_dict = scale_dict)

    print()
    print("descriptor data (converted)")
    print("  scale:", scale_dict)
    ndata = len(mld_converted.descriptor_df)
    print(f"  ndata: {ndata}")
    print(mld_converted.descriptor_df)
    if ndata == 0:
        print()
        print("Warning: All data have been removed. Stop")
        return

    print()
    print("Standardize:")
    mld_converted.descriptor_df_std = mlr.standardize(mld_converted.descriptor_df)
    print()
    print("descriptor data (converted+standardized)")
    print(mld_converted.descriptor_df_std)
    print()
    print("objective var data:")
    print(mld_processed.objective_df)

    d_data = mld_converted.descriptor_df_std
    o_data = mld_processed.objective_df
    d_train, d_test, o_train, o_test = mlr.split_data(d_data, o_data, test_size = cfg.test_data_ratio)
    mld_input_d_train_std.df = d_train
    mld_input_d_test_std.df  = d_test
    mld_input_o_train.df     = o_train
    mld_input_o_test.df      = o_test
    print("descriptor_data train:", mld_input_d_train_std.ndata)
    print("descriptor_data test :", mld_input_d_test_std.ndata)
    print("objective_data train :", mld_input_o_train.ndata)
    print("objective_data test  :", mld_input_o_test.ndata)

    mlr.model = mlr.fit(
                    descriptor_df = mld_input_d_train_std.df, 
                    objective_df  = mld_input_o_train.df,
                    method = cfg.method, 
                    alpha = cfg.alpha, l1_ratio = cfg.l1_ratio, use_cv = cfg.use_cv, cv = cfg.cv,
                    max_num_probes = cfg.max_num_probes, interval = cfg.interval,
                    num_rand_basis = cfg.num_rand_basis, score_mode = cfg.score_mode, seed = cfg.seed,
                    max_iter = cfg.nmaxiter, tol = cfg.tol, print_level = 1)

    mld_predict_train.df = mlr.predict(mld_input_d_train_std.df, mld_input_o_train.columns)
    mlr.score_train = None
    if mlr.std is not None and hasattr(mlr, "mean_m_sigma"):
        mlr.mean_m_sigma_train = mlr.mean_m_sigma.copy()
        mlr.mean_p_sigma_train = mlr.mean_p_sigma.copy()
        if hasattr(mlr, "score"):
            mlr.score_train = mlr.score.copy()

    mld_predict_test.df  = mlr.predict(mld_input_d_test_std.df,  mld_input_o_test.columns)
    mlr.score_test = None
    if mlr.std is not None and hasattr(mlr, "mean_m_sigma"):
        mlr.mean_m_sigma_test = mlr.mean_m_sigma.copy()
        mlr.mean_p_sigma_test = mlr.mean_p_sigma.copy()
        if hasattr(mlr, "score"):
            mlr.score_test = mlr.score.copy()

    print("predicted data (train):")
    print(mld_predict_train.df)
    print("predicted data (test):")
    print(mld_predict_test.df)
    mae_train, mse_train, r2_train = mlr.error(mld_input_o_train.df, mld_predict_train.df)
    mae_test,  mse_test,  r2_test  = mlr.error(mld_input_o_test.df,  mld_predict_test.df)
    print(f"Train data: MAE={mae_train:10.4g} MSE={mse_train:10.4g} R2={r2_train:10.4g}")
    print(f"Test data:  MAE={mae_test:10.4g} MSE={mse_test:10.4g} R2={r2_test:10.4g}")
        
    #グラフを描く
    print()
    print("plot")
    if mlr.std is not None:
        fig, axes = plt.subplots(1, 2, figsize = (12, 8))
        axes.flatten()
    else:
        fig, axes = plt.subplots(1, 2, figsize = (12, 8))
        axes.flatten()

    plot_event = tkPlotEvent(plt)
    ax = axes[0]
    ax.tick_params( labelsize = cfg.fontsize)
    mlr.plot_accuracy(ax, fig, 
            [{"label": "train", "x_df": mld_input_o_train.df, "y_df": mld_predict_train.df,
                       "descriptor_df": mld_input_d_train_std.df, "objective_df": mld_input_o_train.df, 
                       "color": "blue"}, 
             {"label": "test",  "x_df": mld_input_o_test.df,  "y_df": mld_predict_test.df,
                       "descriptor_df": mld_input_d_test_std.df,  "objective_df": mld_input_o_test.df, 
                       "color": "red"} ],
            fontsize = cfg.fontsize, legend_fontsize = cfg.legend_fontsize,
            plot_event = plot_event, register_events = "annotation|move_text"
            )

    ax  = axes[1]
    ax.tick_params( labelsize = cfg.fontsize)
    if hasattr(mlr, "mean_m_sigma_train"):
        mlr.plot_index_fmin(ax, fig,
                [{"label": "train", "input_df": mld_input_o_train.df, "predict_df": mld_predict_train.df, "color": 'blue',
                  "-sigma_df": mlr.mean_m_sigma_train, "+sigma_df": mlr.mean_p_sigma_train, "color_sigma": 'lightblue', 
                  "score_df": mlr.score_train, "color_score": 'orange',
                  "descriptor_df": mld_input_d_train_std.df, "objective_df": mld_input_o_train.df, 
                  },
                 {"label": "test", "input_df": mld_input_o_test.df,  "predict_df": mld_predict_test.df, "color": 'red',
                  "-sigma_df": mlr.mean_m_sigma_test, "+sigma_df": mlr.mean_p_sigma_test,  "color_sigma": 'pink', 
                  "score_df": mlr.score_test, "color_score": 'orange',
                  "descriptor_df":mld_input_d_test_std.df,  "objective_df": mld_input_o_test.df, 
                  }],
                objective_label = objective_labels[0],
                plot_event = None, register_events = "annotation|move_text",
                fontsize = cfg.fontsize, legend_fontsize = cfg.legend_fontsize
                )
    else:
        mlr.plot_index_fmin(ax, fig,
                [{"label": "train", "input_df": mld_input_o_train.df, "predict_df": mld_predict_train.df, "color": "blue",
                  "descriptor_df": mld_input_d_train_std.df, "objective_df": mld_input_o_train.df},
                 {"label": "test",  "input_df": mld_input_o_test.df,  "predict_df": mld_predict_test.df,  "color": "red",
                  "descriptor_df": mld_input_d_train_std.df, "objective_df": mld_input_o_train.df}],
                objective_label = objective_labels[0],
                plot_event = None, register_events = "annotation|move_text",
                fontsize = cfg.fontsize, legend_fontsize = cfg.legend_fontsize
                )

    plt.tight_layout()
    plt.pause(1.0e-5)

#    print()
#    print("Save predicted data to [predicted.xlsx]")
#    df = pd.DataFrame(ml.predicted_df, columns = ml.objective_data.labels)
#    df.to_excel("predicted.xlsx", index = False)

    app.terminate(input_text = "\nPress ENTER to terminate>>\n", pause = True)


def main():
    app     = tkApplication()

#初期値の確定順序
#intialize(): 起動時引数と初期値の設定
    cfg = mf.initialize(app)
#--infile引数をチェックし、読み込みファイル名を最初に確定する
    cfg.infile = app.check_arg('--infile', defval = None, vartype = 'str')

#cfg.infileからログファイル名を作り、console出力をredirectする
    cfg.logfile = app.replace_path(cfg.infile)
    app.redirect(heading = f"Open logfile [{cfg.logfile}]",
                targets = ["stdout", cfg.logfile], mode = 'w')

    app.print_title(f"#  {mf.ProgramName}")

#cfg.infileが入力ファイルか設定ファイルであるかによって、infileとconfig_pathを返す
#  設定ファイルである場合、infile_storedをNoneにして、後で設定ファイルからcfg.infileを読み込む
    cfg.infile, cfg.parameterfile, infile_stored \
            = app.analyze_infile(cfg.infile, infile_ext = [".xlsx"], config_ext = [".in", ".ini", ".prm"], print_level = 1)

#起動時引数で与えられたパラメータをcfgに設定
    app.update_vars(cfg,apply_default = True)
#設定ファイルを読み込み、cfgに設定。#cfg.infileも書き換えられる
    mf.read_parameters(cfg.parameterfile, cfg)
#infile_stored = Noneの場合、設定ファイルのcfg.infileをinfile_storedに再保存
#infile_storedが最終的な読み込みファイル名になる
    if infile_stored is None:
        infile_stored = cfg.infile
#起動時引数で与えられたパラメータを最優先にするため、再設定
    app.force_given_args(cfg)
#cfg.infileを infile_storedに最終設定
    cfg.infile = infile_stored
#    cfg.print_parameters()

    cfg.historyfile   = app.replace_path(cfg.infile, template = ["{dirname}", "{filebody}-history.xlsx"])
    cfg.stopfile      = app.replace_path(cfg.infile, template = ["{dirname}", "stop"])
    print(f"infile: {cfg.infile}")
    print(f"parameterfile: {cfg.parameterfile}")


# cfg.modeで与えられた関数が定義されていれば、呼び出す
    if cfg.mode in globals():
        globals()[cfg.mode](app, cfg)
    else:
        app.terminate("Error in main: Invalide mode [{}]".format(cfg.mode), usage = app.usage, pause = True)

#    app.terminate("", usage = app.usage, pause = True)


if __name__ == "__main__":
    main()
