import numpy as np
from numpy import sin, cos, tan, pi, exp, log, sqrt


from tklib.tksci.tksci import log10, e, kB
from tklib.tkutils import getarg, getintarg, getfloatarg, pconv_by_type, print_line, print_data


# 与えられた変数 params と、座標 x におけるプロファイル関数の値
def ycal(x, params):
    ret = 0.0
    for ip in range(0, len(params), 3):
        I0 = params[ip]
        x0 = params[ip+1]
        w  = params[ip+2]
        if w < 0.001:
            w = 0.001
        a = 0.832554611 / w
        X = a * (x - x0)
        ret += I0 * exp(-X * X)
    return ret

def ycal_list(xd, params):
#    print("pa=", params)
    y = []
    for i in range(len(xd)):
        y.append(ycal(xd[i], params))
    return y

# 最小化する目的関数。残差二乗和
def minimize_func(xk, fit):
    S2 = 0.0
    for i in range(fit.nx):
        yc = ycal(fit.xd[i], xk)
        d  = fit.yd[i] - yc
        S2 += d * d
    return S2

# i番目の最適化変数による一次微分。SIMPLEX法では使わない
def diff1i(i, xk, fit):
    ip  = i // 3
    idx = i % 3
    I0 = xk[ip*3]
    x0 = xk[ip*3+1]
    w  = xk[ip*3+2]

    ret = 0.0
    for id in range(fit.nx):
        if w < 0.001:
            w = 0.001

        yc = ycal(fit.xd[id], [I0, x0, w])
        dy = fit.yd[id] - yc

        a  = 0.832554611 / w
        dx = a * (fit.xd[id] - x0)
        e  = exp(-dx*dx)
        if idx == 0: # I0
            dif = e
        elif idx == 1: # x0
            dif = 2.0 * a * dx * I0 * e
        elif idx == 2: # w
            dif = 2.0 * dx * dx / w * I0 * e
        else:
            print("Error in diff1i: Invalid index (ivar={}, ipeak={}, diff_idx={})".format(i, ip, idx))
            exit()
        d = -2.0 * dif * dy
        ret += d

    return ret

# 一次微分関数配列。SIMPLEX法では使わない
def diff1(xk, fit):
    df = np.empty(len(xk))
    for i in range(len(xk)):
        df[i] = diff1i(i, xk, fit)
    return df

def plot_prepare(fit):
    if fit.fplot == 1:
        fit.figure  = fit.plt.figure(figsize = (10, 5))
        fit.ax_fit  = fit.figure.add_subplot(1, 2, 1)
        fit.rawdata = fit.ax_fit.plot(fit.xd, fit.yd, color = 'blue', linestyle = '', linewidth = 0.5,
                            fillstyle = 'full', marker = 'x', markersize = 5)
        fit.inidata = fit.ax_fit.plot(fit.xd, fit.yini, color = 'black', linestyle = 'dashed', linewidth = 0.5)
        fit.fitdata = fit.ax_fit.plot([], [], color = 'red', linestyle = '-', linewidth = 0.5)

        fit.ax_conv = fit.figure.add_subplot(1, 2, 2)
        fit.ax_conv.set_yscale('log')
        fit.convdata  = fit.ax_conv.plot([], [], color = 'black', linestyle = '-', linewidth = 0.5,
                            fillstyle = 'full', marker = 'o', markersize = 5)
        fit.plt.pause(0.001)

def plot_update(fit):
    if fit.fplot == 1:
        fit.convdata[0].set_data(fit.xiter, fit.yfmin)
        fit.ax_conv.set_xlim((0.0, max(fit.xiter) + 1.0))
        fit.ax_conv.set_ylim((min(fit.yfmin) * 0.8, max(fit.yfmin) * 1.2))

        fit.fitdata[0].set_data(fit.xd, fit.ycal)

        fit.plt.pause(0.2)

def plot_finalize(fit):
    if fit.fplot == 1:
        pass
    pass

# 最適化の各サイクルごとに呼び出されるコールバック関数
def callback(xk, fit):
    fmin = minimize_func(xk, fit)
    fit.ycal = ycal_list(fit.xd, xk)

    print("callback {}: xk=", end = '')
    for x in xk:
        print(f"{x:10.4g} ", end = '')
    print()
    print(f"   fmin={fmin:12.4g}")
    
    fit.iter += 1
    fit.xiter.append(fit.iter)
    fit.yfmin.append(fmin)

    plot_update(fit)

