import os
import sys
import shutil
import glob
import csv
import re
import numpy as np
from numpy import exp, log, sin, cos, tan, arcsin, arccos, arctan, pi
from scipy.interpolate import interp1d
from pprint import pprint
from matplotlib import pyplot as plt


from tklib.tkapplication import tkApplication
from tklib.tkfile import tkFile
from tklib.tkutils import colors
from tklib.tkutils import getarg, getintarg, getfloatarg, pint, pfloat, terminate, split_file_path, replace_path
from tklib.tkvariousdata import tkVariousData
from tklib.tkgraphic.tkplotevent import tkPlotEvent


"""
Convert any file to XLSX by auto judge
"""

usage_str = '''
"  (i) usage: python {} plugins input_path xmin xmax".format(sys.argv[0])
"         ex: python {} {} {} {} {}".format(sys.argv[0], cparams.plugins, cparams.infile, cparams.xmin, cparams.xmax)
'''[1:-1]


plugin_ver = "data:0.2"  # data_list_type = [[x, y]]


#================================
# global parameters
#================================
default_ext = ""
input_type  = "auto"
output_type = "MS-Excel .xlsx"

figsize = (8, 4)
fontsize = 12
legend_fontsize = 12


#=============================
# Treat argments
#=============================
def usage(app = None, cparams = None):
    cparams = app.get_params()
#    app.usage(infile = cparams.infile)
    for s in app.usage_str.split('\n'):
        cmd = 'print({})'.format(s.rstrip())
        eval(cmd)

def initialize(app = None, cparams = None):
    cparams.debug = 0
    cparams.findvalidstructure = True

    cparams.plugins = ''
    cparams.infile = f'*{default_ext}'

    cparams.xmin  = 20.0
    cparams.xmax  = 120.0


def update_vars(app = None, cparams = None):
    if getarg(2, None) is None:
        app.terminate(usage = usage)
    
    cparams.plugins = getarg(1, cparams.plugins)
    cparams.infile  = getarg(2, cparams.infile)
    cparams.xmin    = getfloatarg(3, cparams.xmin)
    cparams.xmax    = getfloatarg(4, cparams.xmax)


def check_file_type(infile, inf = None, app = None, cparams = None):
    if default_ext == '':
        return {"file_type": input_type}
    return None

def get_output_path(infile, inf = None, app = None, cparams = None):
    return replace_path(infile, os.path.join("{dirname}", "{filebody}-out.xlsx"))

def get_input_type(inf = None, app = None, cparams = None):
    return {"file_type": input_type}

def get_output_type(inf = None, app = None, cparams = None):
    return {"file_type": output_type}

def read_file0(module_names, modules, path, app, cparams):
        module = None
        for i in range(len(modules)):
            name = module_names[i]
            m = modules[i]

            file_type  = m.check_file_type(path, app = app, cparams = cparams)
            print(f"try [{name}] for [{path}]: file_type={file_type}")
            if file_type is not None and 'Error' not in file_type:
                module = m
                module_name = name
                break

        if module is None and 'Error':
            return None, None, None

        inf = module.read_data(path, app = app, cparams = cparams)

        return module_name, module, inf


def read_file(plugin, path, app, cparams):
    plugin_dir, plugin_filename, plugin_name, ext = split_file_path(plugin, check_dir = True)
#    print("dir=", plugin_dir, plugin_filename)

    module_names, modules = app.load_modules(plugin_dir, plugin_filename, target = "read_data", is_print = True)
    for m in modules:
        if hasattr(m, "get_input_type"):
            input_type  = m.get_input_type(app = app, cparams = cparams)
        if hasattr(m, "get_output_type"):
            output_type = m.get_output_type(app = app, cparams = cparams)
        print(f"  {m.name}: input_type={input_type}  output_type={output_type}")

    module_name, module, inf = read_file0(module_names, modules, cparams.infile, app, cparams)
#    print("module_name=", module_name)
    if module_name is None or len(module_name) == 0:
        return None, None, None

    return module_name, module, inf

def read_data(infile, app = None, cparams = None, print_level = 0):
    if print_level:
        print("")
        print(f"Read [{infile}]")

    if not os.path.isfile(infile):
        app.terminate(f"Error: Can not find [{infile}]")

    print()
    print(f"Load modules from [{cparams.plugins}] (auto judge)")
    plugin_list = cparams.plugins.split(';')
    for plugin in plugin_list:
        module_name, module, inf = read_file(os.path.join(os.environ['tkprog_X_path'], plugin), cparams.infile, app, cparams)
        if module is not None:
            return inf

    exit()
    return None

def print_data(inf, app = None, cparams = None):
    print("")
    print("filename :", inf["filename"])
    print("sample   :", inf["sample_name"])
    print("yscale   :", inf["yscale"])
    print("labels   :", inf["labels"][0])
    print("nspectrum:", inf["nspectrum"])
    print("ndata    :", inf["ndata"])
#    print(f"2Q range : {inf['xmin']} - {inf['xmax']} degrees, {{inf['xstep']}} step")

    nspectrum = inf["nspectrum"]
    data_list = inf["data_list"]

    print("\t".join(inf["labels"][0]))
    for i in range(inf["ndata"]):
        v = [f"{data_list[0][ispectrum][i]}" for ispectrum in range(nspectrum + 1)]
        print("\t".join(v))

def plot_data(inf, app = None, cparams = None):
    labels    = inf["labels"]
    data_list = inf["data_list"]
    x       = data_list[0][0]
    xlabel  = labels[0][0]
    y_list  = data_list[0][1:]
    ylabels = labels[0][1:]

    fig = plt.figure(figsize = figsize)

    ax = fig.add_subplot(111)
    ax.set_title(f"{cparams.infile}")
    ax.tick_params(labelsize = fontsize)

    ncolors = len(colors)
    for i in range(len(ylabels)):
        ax.plot(x, y_list[i], label = ylabels[i], color = colors[i % ncolors], linewidth = 0.5)

    xmin = max([min(x), cparams.xmin])
    xmax = min([max(x), cparams.xmax])
    ax.set_xlim(xmin, xmax)

    ax.set_xlabel(labels[0], fontsize = fontsize)
    ax.set_ylabel(ylabels[0], fontsize = fontsize)
    ax.legend(fontsize = legend_fontsize)

    plt.tight_layout()
    plt.pause(0.1)
    
    input("Press ENTER to terminate>>")
    

def save_data(outfiles, inf, app = None, cparams = None, print_level = 0):
    labels    = inf["labels"][0]
    data_list = inf["data_list"][0]
    outfile   = outfiles[0]

    print("")
    print(f"Save to [{outfile}]")
    tkVariousData().to_excel(outfile, labels, data_list)

    return True

def convert(inf, app = None, cparams = None, print_level = 0):
    return inf

def main():
#==================================================================
# Initialize parameters
#==================================================================
    app     = tkApplication(usage_str  = usage_str, globals = globals(), locals = locals())
    cparams = app.get_params()

    initialize(app, cparams)
    update_vars(app, cparams)

    print("")
    print( "==========================================================================")
    print(" Convert CIF file to powder XRD pattern")
    print( "==========================================================================")
    print(f"Input file: {cparams.infile}")

    logfile = app.replace_path(cparams.infile, template = ["{dirname}", "{filebody}-out.txt"])
    print(f"Open logfile [{logfile}]")
    app.redirect(targets = ["stdout", logfile], mode = 'w')

    cparams.outfile = get_output_path(cparams.infile)

    print("Output file:")
    print(f"  xlsx file: {cparams.outfile}")

    file_type = check_file_type(cparams.infile)
    if file_type is None:
        app.terminate(f"Error: [{cparams.infile}] invalid file type", usage = usage)
    if 'Error' in file_type:
        app.terminate(file_type, usage = usage)

    inf = read_data(cparams.infile, app = app, cparams = cparams)
    print_data(inf)
    inf = convert(inf, cparams = cparams)
    save_data([cparams.outfile], inf, app = app, cparams = cparams)
    plot_data(inf, app = app, cparams = cparams)

    app.terminate(usage = usage)


if __name__ == "__main__":
    main()
